
%%
%% Self-Guided Ergo 5 interaction tour, Part 4. (advanced.thy)
%%

% This is the fourth part of a self-guided introduction to the
% Ergo proof tool.
%
% At this stage, it is assumed that you are familiar with the emacs
% interface to Ergo, that you can perform some simple proofs and
% write simple tactics, and that you have a basic understanding of
% sequent calculus and window inference.
%
% If Ergo is not currently loaded (with the cpred theory), place
% the cursor at the beginning of the following line, and press F5
% (or C-c C-t).

[cpred].

%%
%% Advanced Strategies.
%%

%%
%% Contents of this file.
%%

% The lists theory.
% Window inference in Ergo.
% Advanced tactic construction.   


%%
%% The lists theory.
%%

% In this tutorial, we will develop a theory about lists. At one
% point in the tutorial changes will be saved. All theorems that
% you have proven to date (in this instance of Ergo) will be saved.
% If you do not wish to save changes, simply ignore the instruction
% to send the "save." "abandon." and "[lists]." commands to Ergo. 
% Alternatively, if you do save but wish to revert to the initial 
% state, simply delete global.erg again and delete 
% any files with the extension '.erg.qge' in your current directory.
%
% To begin, we'll delete any preexisting theories about lists.
% Before deleting a theory, we must ensure that we're not already
% in it. Provided the prompt isn't '[lists]: ', we can 
% proceed:

delete_theory_and_dependents(lists).

% Ergo may inform you that the theory doesn't exist. It is good
% practice to attempt to delete a theory before adding it, just
% in case an old version of it already exists in the system.
%
% Now we can begin a new theory.

add_theory(lists).

% We now have a new prompt. Any new inference rules we enter
% will be added to the lists theory.
%
% This theory will be about typed lists, we will make use of 
% classical predicate logic, types and equality in the statement
% of the inference rules of this theory. cpred_eq_t is a theory
% of classical predicate logic, with types, and weak equality,
% so we will make it an ancestor of the lists theory.

include_theory([cpred_eq_t],all).

% Strong equality matches our usual understanding of equality,
% and allows us to make the window inference opening rules (for
% lists) derived inference rules (ie. theorems). Strong equality
% is defined in the eq theory (weak equality is in weak_eq).

include_theory([eq],all).

% At this point, we can examine the ancestor theories;
% The following command can also be invoked, by seleting
% "Ancestors (of this theory)" in the Ergo menu (see intro.thy
% for more information).

show(ancestor).

% In our theory we will make use of quantified expressions.
% So that we do not need to escape these object variables every
% time they occur, we can pre-declare them.

declare_object_vars(x).
declare_object_vars(y).

show(object_vars).

% We will use '::' as the list constructor (ie. cons) but first
% it must be declared. It has an arity of 2, so we add it as 
% follows:

declare '::'(_,_).

% And so that we can use it infix (ie. in expressions like 'X :: XS'),
% we add it as an operator (with precedence 450, and right associativity).

add_operator(450, rightassoc, '::').

% If we list the current operators, we see the relative precedence
% of cons. The second argument of the displayed terms indicate the
% associativity (in prolog notation).
%   xfy  -> rightassoc
%   yfx  -> leftassoc
%   xfx  -> nonassoc
%    fx  -> prefix
%   xf   -> postfix
%  quant -> quant

show(operator).

help(add_operator).

% We also need to declare a constant, 'nil', to represent the
% empty list.

declare nil.

% And a "type", list.

declare list(_).

% So, at this point we have set up most of the notation and 
% operators we need to start stating properties of lists. These
% properties will be stated as inference rules, but because 
% they have no proof (ie. are axiomatic), we declare them
% as 'axiom' instead of 'theorem'.
%
% We begin by providing some typing rules about lists.
%
% nil is a list of any type;

axiom nil_list ===
      nil : list(T).

% A list is of type list(T) iff the head is of type T, and
% the tail is of type list(T);

axiom cons_list_axiom ===
      X:T and Y:list(T) <=> (X :: Y):list(T).

%  The following theorem is useful for doing proofs about types.

theorem cons_list ===
      X:T,
      Y:list(T)
      -----------------
      (X :: Y):list(T).

prove_rule(cons_list).
rule(l__iff__r).
[one,rule(cons_list_axiom)].
rule(and_intro).
save_postulate_proof(1).
quit_proof.


% Anything that is of type list(T) is nil or an instance 
% of cons;

axiom list_type ===
      X:list(T) => (not (X = nil) <=> ex x:T ex y:list(T) X = x :: y).

% Now some functions and operators along with axiomatic definitions
% of their properties;
% (These rule names need not match the function names, they are only
% chosen this way for convenience).

% Concatenation (Append);

declare '++'(_,_).
add_operator(550, rightassoc, '++').

axiom concat_1 === 
      XS:list(T) => (nil ++ XS = XS).
axiom concat_2 === 
      (X::XS):list(T) and YS:list(T) => (X::XS ++ YS = X::(XS ++ YS)).

% Head;

declare head(_).

axiom head === 
      (X::XS):list(T) => (head(X::XS) = X).

% Tail;

declare tail(_).

axiom tail ===
      (X::XS):list(T) => (tail(X::XS) = XS).

% Reverse;

declare reverse(_).

axiom reverse_1 === 
      reverse(nil) = nil.
axiom reverse_2 === 
      (X::XS):list(T) => (reverse(X::XS) = reverse(XS) ++ (X::nil)).


% Finally, we need a theorem about induction. The general case of
% induction cannot be proven (with the current set of ancestors)
% and so must be defined axiomatically;

axiom induction ===
  [nil/x]A,
  all x0:T (all x1:list(T) [x1/x]A => [(x0::x1)/x]A) 
  ----------------------------------------
  all x:list(T) A
     provided
      x0 not_free_in A, x1 not_free_in A, x0 not_free_in x1, x not_free_in x0,
      x not_free_in x1, x not_free_in T, x0 not_free_in T, x1 not_free_in T.

% Note the not_free_in constraints in the above axiom - this is quite
% typical of inference rules involving quantifiers. The intention is that
% x typically does occur free in A (otherwise we wouldn't be bothered doing
% induction), but the other object variables do not occur free in A. Further,
% none of the object variables occur free in T. We also want the object 
% variables to be mutually different and we do this by making them
% not_free_in each other.

% At this point, we may decide that we have completely specified
% the declarations and primitive inference rules of the theory, 
% and that any additions to the theory should be provable from 
% these rules.

close_theory.

% Now, we will only be able to add definitions and derived
% inference rules.
%
% Definitions are effectively rewrite rules; their definitions 
% can be folded or unfolded anywhere (in any theory that inherits 
% jprop). (The tactics found in unfold.gum can
% automatically do these rewrites, given a path to a subterm.)

define second(X) === head(tail(X)).

% list_type2 is simply a restatement of list_type (and follows
% directly from it, so is a derived inference rule - a 'theorem').

theorem list_type2 ===
      X:list(T)
      -------------------------------------------------
      not (X = nil) <=> ex x:T ex y:list(T) X = x :: y.

% An instance of cons is not nil;
% This follows from list_type (and so is a theorem).

theorem list_disjoint ===
      B:T and CS:list(T) => not (B::CS = nil).


% At this point, we'll prove the theorems in the theory;

load_tactic_file(init_proof,jprop).
prove_rule(list_type2).
init_proof.
[2].

% You will note that ... appears in two places in the display of the
% proof state. This is because the write depth is set to a smallish value.
% This can be increased by giving the write_depth parameter a new value 
% as follows.
set write_depth === 100.

rule(imp_elim).            % Add an implication and an unbound
                           % variable so we can "transform" the
                           % term.
[rule(list_type),one].
rule(assump).

{incomplete_retry_delays}. % There are delayed unification 
                           % problems. That would otherwise prevent
                           % us from saving the proof.
save_postulate_proof(1).
quit_proof.


prove_rule(list_disjoint).
rule(imp_intro).
rule(and_elim_c).

rule(r__iff__l).           % A "transformation" by unification
[one,rule(list_type2)].    % but because list_type2 has a premise
                           % we must also prove it.

[4].
rule(ex_intro).
rule(and_intro).

[9].
rule(ex_intro).
rule(and_intro).

[12].
rule(eq_reflex).

[6].
rule(cons_list).

[rule(assump(2)),rule(assump(3))].

[8,11].
[rule(assump(2)),rule(assump(3))].

{incomplete_retry_delays}.
save_postulate_proof(1).
quit_proof.


%%
%% Window inference in Ergo.
%%

% In Ergo, window inference is made possible by a consistent naming
% convention used across all theories, for the identification of
% opening rules.
%
% The name of an opening rule consists of four parts separated by
% underscores;
%   - a prefix, "open".
%   - the name of the operator/function that is to be opened into.
%   - a number, indicating the subterm that the rule opens on.
%   - the name of the window relation that the opening rule
%     applies to.
%
% For example;

theorem open_cons_1_eq ===
           A = B
    ------------------------
    (A :: C) = (B :: C).

% This rule opens on the 1st subterm of cons ('::') under the 
% window relation eq ('='). The mapping of symbols such as '::'
% to meaningful names will be explained later.
%
% The relation below the line is the window relation that the opening 
% rule can be applied to, and the relation above the line is the window 
% relation for the new window that will be created.
%
% A simple interpretation of this rule could be, "Under the window
% relation =, A :: C can be transformed to B :: C, provided that
% A can be transformed to B under the window relation =".
%
% We must add an opening rule for every potential use of the
% open tactic (op).


theorem open_cons_2_eq ===
           A = B
    ------------------------
    (C :: A) = (C :: B).

theorem open_concat_1_eq ===
           A = B
    ------------------------
    (A ++ C) = (B ++ C).

theorem open_concat_2_eq ===
           A = B
    ------------------------
    (C ++ A) = (C ++ B).

theorem open_head_1_eq ===
          A = B
    ------------------------
    head(A) = head(B).

theorem open_tail_1_eq ===
          A = B
    ------------------------
    tail(A) = tail(B).

theorem open_reverse_1_eq ===
             A = B
    ------------------------
    reverse(A) = reverse(B).

theorem open_listtype_1_eq ===
          A = B
    ------------------------
    list(A) = list(B).


% These are all theorems because they all follow from strong
% equality (strong equality is the common interpretation of 
% equality, and allows us to rewrite any subterm with an 
% equivalent subterm).
%
% We'll begin to prove these opening rules, we'll make use of
% replacement_eq

show_rule(replacement_eq).

prove_rule(open_cons_1_eq).
rule(replacement_eq(x :: C, x, A, B)).
save_postulate_proof(1).
quit_proof.

% Admittedly, this proof was rather simple. But, we can make it
% even simpler by writing a tactic.
%
% This tactic would have to handle three cases;
%    - unary operator, opening at subterm 1
%    - binary operator, opening at subterm 1
%    - binary operator, opening at subterm 2
%
% term(T) unifies T with the goal of the current open node,
% this can be used to identify which particular case we
% are proving.
%
% ie.
% tactic prove_eq_open ===
%       term(WR(Unary(A),Unary(B)));
%       ---Handle first case here---
%       '|'
%       term(WR(Binary(A,C),Binary(B,C)));
%       ---Handle second case here---
%       '|'
%       term(WR(Binary(C,A),Binary(C,B)));
%       ---Handle third case here---
%       .
%
% Above, we see another feature of Qu-Prolog that becomes
% useful in Ergo; Qu-Prolog allows the use of variable
% functors in compound terms. What this means is that
% a term like A(B,C) will unify with any expression
% involving a binary operator. In fact, it doesn't even
% matter if the operator appears to be infix; Ergo's
% internal representation is always a compound term.
%
% Ergo tactics use Prolog's execution model. This means
% that when Ergo encounters a choice point (ie. any point
% where there are multiple possible paths -- such as when
% a '|' is reached), it will take the first possibility
% it encounters. If, later on, it finds that something
% goes wrong (eg. a rule application fails), it will 
% backtrack (reversing any changes) until it reaches
% the last choice point, and then try the alternate choice.
%
% This particular property of Gumtree makes it simple
% to code tactics that perform searches for proofs.
%
% To complete the tactic, we just have to apply the rules
% correctly for each case.

tactic
  prove_eq_open
mode
  prove_eq_open
arity
  1 ---> 1.

tactic prove_eq_open ===
      term(WR(Unary(A),Unary(B)));
      rule(replacement_eq(Unary(x), x, A, B))
      '|'
      term(WR(Binary(A,C),Binary(B,C)));
      rule(replacement_eq(Binary(x,C), x, A, B))
      '|'
      term(WR(Binary(C,A),Binary(C,B)));
      rule(replacement_eq(Binary(C,x), x, A, B)).

% Now we can automatically prove each opening rule.

prove_rule(open_cons_2_eq).
prove_eq_open.
save_postulate_proof(1).
quit_proof.

prove_rule(open_concat_1_eq).
prove_eq_open.
save_postulate_proof(1).
quit_proof.

prove_rule(open_concat_2_eq).
prove_eq_open.
save_postulate_proof(1).
quit_proof.

prove_rule(open_head_1_eq).
prove_eq_open.
save_postulate_proof(1).
quit_proof.

prove_rule(open_tail_1_eq).
prove_eq_open.
save_postulate_proof(1).
quit_proof.

prove_rule(open_reverse_1_eq).
prove_eq_open.
save_postulate_proof(1).
quit_proof.

prove_rule(open_listtype_1_eq).
prove_eq_open.
save_postulate_proof(1).
quit_proof.

% A trickier solution could be achieved by making use of
% Ergo's searches and unification.
% NOTE: Separate variables Premise and G are used in this tactic and
% unified within the Prolog goal. It is not possible to use the same variable
% as in the following
%
%tactic another_prove_eq_open ===
%      premises([Premise]);
%      rule(replacement_eq);
%      term(Premise);
%      {incomplete_retry_delays}.
%
% as the type checker will reject the tactic definition. Premise has type term
% while G has type thy(term). The explicit unification within the Prolog
% goal 'tricks' the type checker into accepting the tactic.

tactic
  another_prove_eq_open
mode
  another_prove_eq_open
arity
  1 ---> 1.

tactic another_prove_eq_open ===
      premises([Premise]);
      rule(replacement_eq);
      term(G);
      {Premise = G, incomplete_retry_delays}.

% This tactic should prove almost every opening rule in
% a theory that inherits strong equality.


% Now that we've proven the opening rules, we need to 
% give the window tactic information on how to locate
% these rules. This is done by adding the following
% declarations which give names to the operators.
% to the lists directory.
%

in_table(symbols, head(_), head).
in_table(symbols, tail(_), tail).
in_table(symbols, reverse(_), reverse).
in_table(symbols, ::(_,_), cons).
in_table(symbols, ++(_,_), concat).
in_table(symbols, list(_), listtype).       

% Tables can be used to define a relationship between two terms.
help(table).

% The symbols table is used to provide a mapping between constructors
% and names for those constructors that are used to build various rule
% names used in window inference. The contents of the table can be viewed
% as follows.

show_table(symbols).

% You will note lots of other constructors are listed in this table.
% This information comes from the theories in which those constructors
% are defined.
% There is a problem with defining the symbols in the theory file. This
% information is not saved when the theory is saved and so needs to be
% loaded when the theory is loaded.
% If you look in the system theory files - e.g jpred.thy you will find that
% at the end of the file is a request to load a tactic file (if it is not 
% already loaded)
% require_tactic_file(wi_support,jpred).
% If you look at jpred/wi_support.gum you will see that this file adds
% symbol mappings to the symbols table.
%
% If you are doing this tutorial over several sessions it is suggested you
% copy the symbol mappings into the file lists/wi_support.gum and
% add 
% require_tactic_file(wi_support,lists).
% to the theory declarations.

% At this point, window inference is ready to go.
%
% We can also check that all theorems have been proven.

rerun_all_postulate_proofs.

% The command succeeded, with no output, because we have no
% unproven theorems.
% If this command encounters an unproven theorem, it will
% search the appropriate directory (lists) for a proof
% file that records the steps of a proof, and execute it
% if possible. If the file lists.erg.qge is deleted, the
% list theory must be recreated, and all theorems reproven.
% Without the ability to store a proof script in special
% files (and they're stored automatically when the
% save_postulate_proof command is entered), theories would
% require tedious maintainence whenever a tactic is 
% modified.
%
% At this point, we can save the state of Ergo, and the
% current theory. 

% WARNING: Executing "save." will save all the theorems
%          that you have added since starting Ergo.
%          If you do not wish to do this, ignore this
%          step AND the following two steps.

% BEGINNING OF "SAVE SECTION"

save.

% After saving, reentering the lists theory is simply a 
% matter of entering '[lists].'

abandon.     % This will quit Ergo
[lists].     % This will restart it, and reload the lists theory

% You will need to reload the symbols - look at the system theory
% files for an alternative - this involves putting these commands in
% a .gum file in the lists directory and adding a
% require_tactic_file into the theory file.
% This will cause these declarations to be automatically loaded
% when the theory is loaded.

in_table(symbols, head(_), head).
in_table(symbols, tail(_), tail).
in_table(symbols, reverse(_), reverse).
in_table(symbols, ::(_,_), cons).
in_table(symbols, ++(_,_), concat).
in_table(symbols, list(_), listtype).       


% END OF "SAVE SECTION"

% We can list all the inference rules of the current theory
% to verify that they're all present.
% (This can be done by selecting "Show... Theory Management... Rules" 
% from the Ergo menu)

show(rule).

% We shall now prove some more properties about lists;

theorem second_is ===
      A:T, B:T, T:list(T)
      --------------------
      second(A::B::T) = B.

require_tactic_file(init_proof,jprop).
require_tactic_file(unfold,jprop).
require_tactic_file(window_tacs, jprop).
require_global_tactic_file(tacticals).

prove_rule(second_is).
init_proof.
[6].

% At this point, we'll treat it as a window inference proof.
% Even though the 'stack' does not consist of just windows,
% the window inference tactics will still work (they apply to
% only the top node).

unfold([1]).

op([1]).


% To use the inference rule, tail, we need a way of 
% discharging the rule's antecedents. We will leave that 
% problem for later, by adding an unbound hypothesis now.

first(rule(hyp_add)).

% first is a simple, but handy tactic that
% applies a tactic to the first open node. The above is 
% equivalent to typing,
%      [rule(hyp_add),skip].
%
% Note: previously, we had only encountered one-to-one 
%   correspondences between open nodes and parallel tactic
%   application. In fact, this needn't be the case; some
%   tactics may work on more than one node, and Ergo will
%   search for some valid mapping of open nodes to tactics
%   in parallel. For example, the following code will leave
%   exactly five open nodes unchanged (and is therefore
%   equivalent to 'fix(5)');
%                   [two, one, two].
%
% Now then we can apply the rule, tail, unifying the antecedent
% with the new hypothesis, hyp 4.

trans_ante(rule(tail),rule(assump)).

% Now we can close that window, and prove the hypothesis
% that we had added.

close_win.

% Again, notice that this is a hybrid style of proof,
% we just closed a window, and now we're going to prove
% a node (node 16) that is not a window.

[16].
rule(cons_list).
[rule(assump),one].
rule(cons_list).
[rule(assump),rule(assump)].

% Now that we've discharged that subproblem, the first
% open node is once again a window (there isn't any property 
% that says we must call it a window, it is just that at 
% this point it is convenient to do so).
%
% Note that we needn't even be working on the first open
% node to use the window tactics, provided that we do
% have a window in the first node that appears in the
% prompt (ie. provided that the first node that our tactic
% can be applied to is a window).

[13].

% We'll use the same 'trick', to add a hypothesis,

rule(hyp_add).

trans_ante(rule(head),[rule(assump)]).

close_win.

rule(cons_list).

[rule(assump),rule(assump)].


% We've completed the proof at this stage
% (All the remaining open nodes were created by init_proof,
% and match the premises exactly).

save_postulate_proof(1).

quit_proof.

% Two proofs using induction;

% First, proof that if A and B are lists, then A ++ B is also
% a list.

theorem concat_type ===
      A : list(T),
      B : list(T)
      ------------------
      (A ++ B) : list(T).

prove_rule(concat_type).
init_proof.
[4].

% We will be making use of all_intro, but because this
% is untyped we need to use all_t_iff_all to type the
% bound variable (so that we can then use induction).

show_rule(all_t_iff_all).

% This rule requires an implication, so we add it as
% follows.

rule(imp_elim).
[one,rule(assump(2))].

% Then the all_elim rule;

rule(all_elim(x : list(B) => ((x ++ C) : list(B)),x,A)).

% Then 'transform' the goal;

rule(l__iff__r).
[one,rule(all_t_iff_all)].

% At this point we can use induction

rule(induction).

% First handle the nil case

[12].
start_window_proof.
op([1]).
trans_ante(rule(concat_1),rule(assump)).
close_win.
stop_window_proof.
rule(assump).

% Then the cons case

[13].
rule(all_intro).
rule(all_intro).
rule(imp_intro).

start_window_proof.
op([1]).

first(rule(hyp_add((x3 :: x4) : list(B)))).
[one,rule(cons_list),skip].
[one,rule(assump),rule(assump),skip].

trans_ante(rule(concat_2),[rule(assump),rule(assump)]).

close_win.
stop_window_proof.

rule(cons_list).
[rule(assump),rule(assump)].

save_postulate_proof(1).

quit_proof.

% Another property; the associative law for append:

theorem concat_assoc ===
      A : list(T),
      B : list(T),
      C : list(T)
      ------------------------------
      A ++ (B ++ C) = (A ++ B) ++ C.

prove_rule(concat_assoc).
init_proof.
[6].

% Like concat_type we need to transform the goal before
% we can apply induction.

rule(imp_elim).
[one,rule(assump(3))].

rule(all_elim(x : list(B) => (x ++ (C ++ D) = (x ++ C) ++ D),x,A)).

rule(l__iff__r).
[one,rule(all_t_iff_all)].

rule(induction).

% Handle the nil case;

[16].
rule(hyp_add).

% Add the typing constraints to the hypothesis:

[one,rule(concat_type)].
[one,rule(assump(2)),rule(assump(1))].

trans_ante(rule(concat_1),rule(assump)).

op([1]).

trans_ante(rule(concat_1),rule(assump)).

% Above, the concat_1 rule failed in the forward direction,
% so it was applied in reverse.

close_win.
close_win.

% And then the cons case;

[17].
rule(all_intro).
rule(all_intro).
rule(imp_intro).

% It will be easier to transform the RHS to the LHS, so
% we can use the tactic rev_rel (also in window.gum).

rev_rel.

% Resolve more typing conditions;

rule(hyp_add(x3 :: x4 : list(B))).
[one,rule(cons_list)].
[one,rule(assump),rule(assump)].

op([1]).
trans_ante(rule(concat_2),[rule(assump),rule(assump)]).
close_win.

% And more typing conditions;

rule(hyp_add(x3 :: (x4 ++ C) : list(B))).
[one,rule(cons_list)].
[one,rule(assump),rule(concat_type)].
[one,rule(assump),rule(assump)].

trans_ante(rule(concat_2),[rule(assump),rule(assump)]).

op([2]).
first(rev_rel).
first(rule(assump(7))).

% Above, we unified the first open node with hypothesis 7, which
% discharged the node, and bound M to a "transformed" term.

% Some more typing conditions;

rule(hyp_add(C ++ D : list(B))).
[one, rule(concat_type)].
[one, rule(assump), rule(assump)].

% And now we can complete the proof;

trans_ante(rule(concat_2),[rule(assump),rule(assump)]).
close_win.

save_postulate_proof(1).

quit_proof.

%
% In the above proofs we manually discharged several type conditions.
% Below we look at the horn tactic which could have been used to discharge
% these constraints automatically. After reading the section below you might 
% like to think about how using horn could have saved effort in the above
%  proofs.
% Note that the horn tactic uses forward inference in much the same way
% as the typing conditions were discharged in the above proofs.
%

%%
%% Advanced tactic construction.
%%

% The inference rules in lists are all 'guarded' by typing 
% conditions. This is necessary to ensure that the rules
% are consistent and meaningful, particularly if the lists
% theory is ever to be extended (ie. inherited from).
% This also means that before using any rule, the types
% must be (tediously) checked.
%
% We will develop a tactic to perform this automatically,
% it will have some resemblance to the horn_n tactic in horn.gum.
%
% This tactic will be called with a list. The tactic will
% (hopefully) add the hypothesis that the list is indeed of
% type list.
%
% For this example we will create a new theory that extends
% lists (so that we can have elements in the list).

delete_theory_and_dependents(binary_lists).
add_theory(binary_lists).
include_theory([lists],all).
declare on.
declare off.
declare binary.

axiom on_binary ===
      on : binary.
axiom off_binary ===
      off : binary.

close_theory.

% In this theory we could prove that;

theorem on_off_on_and_off_is_on_off_on_off ===
  on :: off :: on :: nil ++ off :: nil = on :: off :: on :: off :: nil.

% The proof of this requires us to discharge the antecedents of
% the rules concat_1 and concat_2.
%
% Our tactic would take the list as an argument and is defined as follows.

tactic
  add_list_hyp(List)
mode
  add_list_hyp(thy(term))
arity
  1 ---> 1.

tactic add_list_hyp(List) ===
      rule(hyp_add(List : list(T)));
      [one,add_list_hyp_prove].

tactic
  add_list_hyp_prove
mode
  add_list_hyp_prove
arity
  1 ---> 0.
  
tactic add_list_hyp_prove ===
      rule(cons_list);
      [
            rule(on_binary) 
            '|' 
            rule(off_binary)
      ,
            rule(nil_list)
            '|'
            add_list_hyp_prove
      ].

% In the tactic declaration for add_list_hyp above the mode part
% gives the type of the argument. The type here is thy(term) which
% says that the argument is a term of the theory. By using this
% declaration users can call the tactic with an argument that is in
% abbreviated form (as is seen in the user interface) rather than in the
% internal form. For example, the term   off :: nil : list(binary)
% has internal form 
% 'lists.:'('lists.::'('binary_lists.off', 'lists.nil'), 
%           'lists.list'('binary_lists.binary'))
% which shows that each symbol is qualified by the name of the theory
% in which it is declared.
%
% Both of the above tactics have ground arities. The first tactic
% has arity 1 ---> 1 which says that the tactic takes as input a single
% open proof node and if successful produces a single open node.
% The second tactic produces no open nodes.
% 
% The allowable types that can be used in tactic modes can be viewed
% using help.

help(modes_help_ergo5).

% We can now test this tactic;

prove_rule(on_off_on_and_off_is_on_off_on_off).

add_list_hyp(on :: off :: on :: nil).
add_list_hyp(off :: nil).
trans_ante(rule(concat_2),[rule(assump),rule(assump)]).
op([2]).
first(add_list_hyp(off :: on :: nil)).
trans_ante(rule(concat_2),[rule(assump),rule(assump)]).
op([2]).
first(add_list_hyp(on :: nil)).
trans_ante(rule(concat_2),[rule(assump),rule(assump)]).
op([2]).
trans_ante(rule(concat_1),rule(assump)).
close_win.
close_win.
close_win.

% Now we've transformed the focus, and can complete the proof;

close_win.

quit_proof.


% An obvious step now is to write a tactic that actually
% "executes" an append.

tactic
  simplify_append_wi_support
mode
  simplify_append_wi_support
arity
  _.

tactic simplify_append_wi_support ===
      focus(nil ++ B);
      first(add_list_hyp(B));
      trans_ante(rule(concat_1),[rule(assump)])
      '|'
      focus(A ++ B);
      first(add_list_hyp(A));
      trans_ante(rule(concat_2),[rule(assump),rule(assump)]);
      op([2]);
      simplify_append_wi_support;
      close_win.

tactic
  simplify_append_wi
mode
  simplify_append_wi
arity 
  _.

tactic simplify_append_wi ===
      focus(_ ++ B);
      add_list_hyp(B);
      simplify_append_wi_support.

% Now we can use this tactic to immediately simplify any append
% of concrete lists:

prove_rule(on_off_on_and_off_is_on_off_on_off).
simplify_append_wi.

% If we type retry, we notice it finds identical solutions.

retry.

quit_proof.

% The execution of the above tactic can produce a great many choice
% points (recall that a choice point is any point in execution when
% more than one path may be taken). If we know that only one solution
% will be found, or that we only want one solution, we can reduce
% non-determinism by using a 'cut'.
%
% Redefining only simplify_append_wi;

% First remove the old definition.
% Note that tactic definitions can only be removed for dynamic
% definitions. Internally tactics defined on-the-fly as above or
% loaded from source files are translated into Prolog code and
% asserted. Deleting is implemented by using Prolog retract.
% On the other hand tactics loaded from compiled files cannot be
% deleted because they are loaded as compiled Prolog code.
% In fact using compile_tactic_file, for example, on a gumtree tactic
% file first translates the definitions into Prolog code, optimises
% the code using the arity information of tactics and rules and
% then compiles the resulting Prolog clauses into compiled Prolog.

tactic delete simplify_append_wi.

tactic simplify_append_wi ===
      !(
            focus(_ ++ B);
            add_list_hyp(B);
            simplify_append_wi_support
      ).

% Ergo's cut  !(...)  behaves as Prolog's once(...). It will solve
% the body of the cut, but will only take the first solution it finds.
% We can now test the 'retry' command on the new simplify_append_wi.

prove_rule(on_off_on_and_off_is_on_off_on_off).
simplify_append_wi.

retry.

% retry fails, because we have made the tactic deterministic. All 
% the possibilities (ie. just one, in this case) are exhausted and 
% so Ergo returns to the previous state.

quit_proof.

% It turns out that a similar kind of typing mechanism already exists
% in a tactic provided with Ergo.
%
% To use it we must declare a tactic table. A tactic table is effectively a
% tactic that acts like a large, dynamically changing disjunction.

declare_tactic_table
  lists_types
doc
  'A table of tactics for deducing type facts about lists.'.

%
% We should ensure that the table is empty;

erase_tactic_table(lists_types).

% And now we add some typing rules. in_table permits
% us to add any tactic, but the tactic we are about
% to use this with requires that it be a rule application.

in_tactic_table(lists_types, rule(cons_list)).
in_tactic_table(lists_types, rule(nil_list)).
in_tactic_table(lists_types, rule(off_binary)).
in_tactic_table(lists_types, rule(on_binary)).

% We can now use this table with the horn tactic;

require_tactic_file(horn,jprop).

prove_rule(on_off_on_and_off_is_on_off_on_off).

horn(lists_types, (on :: off :: on :: nil) : list(_)).

% Notice that horn has also added intermediate hypotheses,
% if this is a problem, we can require that at most one new
% hypthesis is created;
 
undo.

 
at_most_one_new_hyp(horn(lists_types, (on :: off :: on :: nil) : list(_))).
 
% We can now redefine add_list_hyp to take advantage of horn,
 
quit_proof.

tactic delete  add_list_hyp(_).

tactic add_list_hyp(List) ===
      at_most_one_new_hyp(horn(lists_types, List : list(_))).
 
% Because simplify_append_wi will make use of this modified
% add_list_hyp tactic, the proof proceeds exactly as before;
 
prove_rule(on_off_on_and_off_is_on_off_on_off).
simplify_append_wi.

close_win. 

quit_proof.

% The above tactic could be made more efficient by generating the
% required type hyps up front. The following variant make better use
% of the hyps gererated by horn.

tactic
  simplify_append_wi_support2
mode
  simplify_append_wi_support2
arity
  _.

tactic simplify_append_wi_support2 ===
   !(
      focus(nil ++ _);
      trans_ante(rule(concat_1),rule(assump))
      '|'
      trans_ante(rule(concat_2),[rule(assump),rule(assump)]);
      op([2]);
      simplify_append_wi_support2;
      close_win
    ).

tactic
  simplify_append_wi2
mode
  simplify_append_wi2
arity 
  _.

tactic simplify_append_wi2 ===
  !(  
      focus(A ++ B);
      horn(lists_types, A : list(_));
      horn(lists_types, B : list(_));
      simplify_append_wi_support2
   ).

prove_rule(on_off_on_and_off_is_on_off_on_off).
simplify_append_wi2.

close_win.
save_postulate_proof(1).
 
quit_proof.        

% Efficiency considerations for gumtree tactics.
%
% In the example tactics above we produced on-the-fly declarations
% and definitions.
% For tactics that are designed to be generally useful, it is usually
% better to create a file with a .gum extension and define related
% tactics within the file.
% The system already contains various .gum files that can be viewed
% for inspiration.
% Tactic files that do not refer to any specific rules or theory terms
% (such as tacticals.gum) are probably best thought of as global
% tactic files. Such files should appear in the users working directory.
%
% Tactics that refer to rules or terms of a theory should appear in
% files within an appropriate theory sub-directory.
% For example, window_tacs.gum appear in the jprop directory of the
% system because some of tactics refer to specific rules and terms
% within jprop.
%
% So, for example, the tactics defined above could be put into a
% file  binary_list_tacs.gum (say) in the binary_list sub-directory.
%
% A reasonably efficient way of testing these tactics is to
% start Ergo and then load the tactic file with
% load_tactic_file(binary_list_tacs)
% If changes are required then the file should be edited and then 
% reloaded with
% reconsult_tactic_file(binary_list_tacs)
% When the tactic testing is finished and it is ready to be used
% then the tactics can be compiled with
% compile_tactic_file(binary_list_tacs)
% Ergo should then be restarted and the tactic file loaded the same way
% as above i.e. with load_tactic_file(binary_list_tacs).
% The system will respond by saying it is loading a .qo file rather
% than a .gum file.
% Note, however, that compiled tactics cannot be reconsulted.
%
% A compiled gumtree tactic is often quite a lot faster than the 
% uncompiled version.
% There are two reasons for the speed improvement:
% * The compiled tactic has been 'unfolded' to Prolog code and then compiled
%   rather than being interpreted by the gumtree tactic interpreter.
% * After unfolding to Prolog code, an optimising phase attempts partial
%   evaluations of some generated bookkeeping code, further improving
%   performance.
%
% Tactic writers can maximise the optimisations by being as specific as
% possible about the arities of tactics and their uses.
%
% For example, if tactic foo always has input arity 1 and output arity 2
% then, by declaring foo with arity 1 ---> 2, the optimiser can evaluate
% some calls on the Prolog predicates length and append (at optimiser time)
% that are used to manage lists of the open nodes known to gumtree.
%
% Sometimes, however, a tactic might not have such a specific arity because
% it can be used in more general situations and so might have a variable
% arity. Most tactics in window_tacs have variable arities.
% Even in this case the tactic writer may know that some uses of such 
% a tactic (within the definition of another tactic) have a specific input
% arity. Lets say that the definition of foo has a call to a tactic bar
% with variable arity but in this specific use has input arity 1.
% The tactic writer can provide this information to the optimiser as follows
% tactic foo ===
%      ....
%      one;bar
%      ....
% The tactic 'one' provides the required information to the optimiser.
%
% Another useful hint to the optimiser is when rules are used within
% tactics. If the tactic makes a call such as rule(imp_intro) then
% the optimiser does not know anything about the output arity (but of course
% it does know the input arity is 1).
% This is because there could be any number of rules with name imp_intro
% in different theories all with different arities.
% On the other hand if rule(imp_intro) is replaced by rule(jprop:imp_intro)
% then the optimiser will determine the output arity of that specific rule
% and use this information in the optimisation.

% Soundness of tactics - not an issue.

% Ergo has been designed in such a way that the Ergo 'kernel' (where
% the theory database and proofs are managed) can only be modified
% by 'methods' of the kernel.
% Assuming that these methods have been implemented correctly (by the
% Ergo implementors) then it is not possible for users to produce invalid
% proofs, even if their tactics are wrong, or if there is a bug in the
% gumtree tactic compiler or interpreter.
% In fact, the gumtree interpreter and compiler are also tactics (in this
% case Prolog programs) and also interact with the kernel only through
% the methods of the kernel. 

% This concludes the introduction to the Ergo.







