
%%
%% Self-Guided Ergo interaction tour, Part 1. (intro.thy)
%% 

% This is a self-guided introduction to interacting with the
% Ergo proof tool, via the Emacs interface.

% Overview
%
%   intro.thy    
%   First Contact with Ergo
%     An introduction to the Emacs interface, starting Ergo,
%     and using help
%
%   proofs.thy
%   Simple Interaction
%     An introduction to proofs with Ergo 5, a sequent calculus
%     "recap", basic Gumtree tactics and Quantifiers, substitutions
%     and unification
%
%   tactics.thy
%   Window Inference and other System Tactics
%     Window inference "recap", window inference in Ergo, and
%     other system tactics.
% 
%   advanced.thy
%   Advanced Strategies
%     Developing theories, integrating theories with the window
%     inference tactics, and advanced tactic construction.


%%
%% Contents of this file.
%%

% Introduction.
% Note on versioning.
% Invoking Ergo through Emacs.
% Choosing a theory.
% Browsing and obtaining information.
% Show.
% Help.
% Sending commands from buffers
% Exiting Ergo.                       


%%
%% First Contact with Ergo.
%%


%%
%% Introduction.
%%

% Ergo is an interactive proof tool designed and implemented at
% the University of Queensland's Sofware Verification Research Centre.
% It is available for download from the SVRC website located at 
% http://www.svrc.uq.edu.au/ and will compile under many
% unix systems.
%
% It is implemented in Qu-Prolog, which you will also need to
% download (from the SVRC website) before you can start Ergo.
%
% These tutorials provide an introduction to Ergo 5, a sequent
% calculus based prover. Some knowledge of sequent calculus is 
% assumed, as is some basic familiarity with your computer.
%
% For further information about Ergo 5, refer to the Ergo 5 
% Reference manual in doc/ergo5


%% 
%% Note on versioning.
%%

% This tutorial is packaged with the release of Ergo. It
% incorporates both window inference and sequent calculus style
% provers. For historical reasons, the window inference mode
% of Ergo is referred to as Ergo 4, and the sequent calculus 
% mode is referred to as Ergo 5. This tutorial uses only the
% Ergo 5 style interface. Window inference can still be 
% performed in Ergo 5 using special tactics. This will be
% elaborated on, in the Advanced Strategies tutorial.

%%
%% Invoking Ergo through Emacs.
%%

% Before starting, you should ensure that your shell exports
% the following variable settings (which should be the directory 
% containing the Ergo distribution) - e.g. :
%
% ERGOHOME=/home/groups/itp/ergo6
%
% If you have installed Ergo on your own system, you will need
% to correct the above paths, as appropriate. Note that daVinci
% is not a part of the standard Ergo distribution, but Ergo is
% fully functional without this component (daVinci provides 
% a graphical view of the proof tree).
%
% Your `.emacs' file, in your home directory, should contain the 
% following line (without the initial `%'):
%
% (load "$ERGOHOME/emacs/ergo.el")
%
% If you don't have a .emacs file in your home directory then copy the
% system default .emacs file into your home directory and add the above line.
% The system default file might live in /etc/skel or similar places.


% To invoke Ergo using Emacs, first invoke Emacs on some theory
% file (.thy extension). For this tutorial, use `intro.thy'.
% That is, type `emacs intro.thy'.  You should do this in a 
% directory which contains copies of all the following files:
%
%   intro.thy
%   proofs.thy
%   tactics.thy
%   advanced.thy


% Now to start up Ergo in an Emacs window, type C-c C-d.
% That is, Control-c, Control-d.
%
% This will result in the screen splitting to show two buffers, 
% with the lower half running Ergo.  You should see the Ergo copyright
% message immediately, then Ergo will load the theory graph and
% various tactic files. This should take a couple of seconds, but
% could take up to two minutes on slow machines.
%
% After this, you will get the theory management prompt (':'), 
% preceeded by 'No theory'.
%

% You should check that Ergo 5 (rather than Ergo 4) has indeed been 
% started by noting that various Ergo 5 specific 'modules', such as 
% newtypes5.qo, ops5.qo, gumtree.qo, tables.qo have been loaded.


% You can use Emacs to move around and edit both buffers.
% For example:
%
%   Left-mouse-clicking on a place moves the cursor to that place.
%
%   Arrow keys move the cursor through text.
%
%   Middle-mouse-clicking in the left-hand scroll bar moves to
%   the corresponding part of the text.
%  
%   The `Page Down' key scrolls through text a window-full at a time.
%   The `Page Up' key scrolls back a window-full at a time.
%
%   Alternatively, you can use the scrollbar on the left-hand side
%   of each buffer (drag with the middle mouse button held down, or
%   click with the left/right mouse button to move down/up half a
%   window-full at a time).
%
%   Left-mouse-click in a buffer to make that the current buffer.
%
%   Typing causes text to be inserted at cursor position.
%
%   Dragging the mouse with left button down selects text.
%
%   Clicking the middle mouse button pastes selected text where
%   the mouse cursor is. 
%
% Sometimes the output may appear below the visible part of the
% buffer:
%
%   Pressing C-End will move to the bottom
%   of the buffer, and position the cursor at the prompt.
%
%   Using the middle mouse button to click at the bottom of 
%   the left-hand scroll bar will move to the bottom of the buffer.


%%
%% Choosing a theory.
%%

% We'll begin by looking at a theory of classical predicate calculus.
% The source for this theory can be examined by viewing the file
% '$ERGOHOME/theories/cpred.thy'. 
%
% This theory draws from several other theories. You might find it 
% worthwhile at this point to take a quick look at one of these
% theory source files, even though it may make little sense. A fairly
% simple theory that is the ancestor of almost every other theory 
% in Ergo is in '$ERGOHOME/theories/jprop.thy'.


% Click on the Ergo buffer (called '*ergo*'), and then press C-End
% to place the cursor after the prompt.
%
% Now type the following command (at the prompt).
%
% Note that all Ergo input requires a terminating full stop.  

[cpred].

% When you type the "Return" (or "Enter") key at the end of the line,
% the line is sent off to Ergo to be executed.  If you use the mouse
% to paste commands into the *ergo* buffer, you still need to press
% "Return" at the end of the command before it is sent to Ergo.
%
% After sending the above command you will see Ergo loading up the
% theories used by cpred as well as the cpred theory itself.
%
% We see that cpred has been successfully loaded by noting that
% the prompt now indicates we're working with the cpred theory.

%%
%% Browsing and obtaining information.
%%

% There are a number of ways of obtaining information about
% the current Ergo environment:

%%
%% Show.
%%

% Pulling down the Ergo menu at the top of the screen you will see 
% a `show' item.  Choose it.
% A submenu appears with further submenus showing categories of topics
% on which you can show.

% Choose `Theory management', then `Ancestors'. This shows the
% `ancestor' theories of the current (cprop) theory. These are
% theories that the current theory draws on. The result shows that
% this theory has ancestors: cprop, ipred, iprop, jpred, jprop,
% stdops.
%
% cprop is a theory about classical propositional logic. The
% remaining ancestors are consecutively "simpler" theories, and 
% stdops defines basic system-wide operators.
%
% A modular approach has been taken to the design of the theory
% graph. Most theories are fairly minimal, and are built up
% out of simpler theories. This maximises flexibility - cprop 
% includes all of iprop, but adds the law of excluded middle - 
% proofs, tactics and theorems created for iprop will still 
% be accessible in cprop

% The 'Operators' and 'Symbols' items from the 'Syntax' submenu
% will show operators and symbols defined in the current theory.
% Since cpred defines no extra operators or symbols, this will
% return nothing. We will return to this later.

% The show menu can be used to request other information too.
% Many of the options are not implemented in Ergo 5, or only 
% work in certain modes.

%%
%% Help.
%%

% Ergo has some on-line help facilities. (Not the help menu at
% the top, that is the Emacs help).
%
% Choosing `help' from the Ergo menu (or entering "help." as a command)
% gives a list of topics on which help is available. 
% Topics that appear in bold can be clicked on to get further help.
% For example, click (left mouse button) on 
%
%        `Theory/Database Commands (The DBMS user interface)'
%
% then click again on
%
%        `Theory Extension Commands'
%
% and then on
%
%        `declare Symbol.'
%
% to get help on the `declare' command.
%
% Typing 'k' within the help window will `kill' the currently displayed
% help topic and return to the previous topic.
%
% The help command can also be used to search for all help pages
% that contain a given string.  For example, type:  

help "save".

%%
%% Sending commands from buffers
%% 

% You can send commands to Ergo from other Emacs buffers.
% Typing Function Key 5 (F5) sends a term to Ergo. It sends
% text from the current cursor position to the next full stop.
% Try this on the following commands in this buffer. Position
% the cursor on the first character of a command and type F5.

show(status).
help "sugar".

% Another way of sending commands is to type F8 *after* a 
% command. This sends everything between the last full stop and
% the current cursor, and starts a new line (it will also send
% a full stop if there isn't already one).
%
% Choosing `Describe: Describe Buffer Modes' under the `Help menu' at the top
% displays these and other Emacs key bindings for Ergo.
%
% Note that C-c C-t is an alternative to F5, and C-c C-p is an 
% alternative to F8.

%%
%% Exiting Ergo.
%%

% To exit, and retain changes (such as the addition of proofs
% to the database), type C-d or enter the command
% 'bye.'
%
% If you do not wish to save any changes, simply exit emacs, and when
% asked click (or type) yes, to whether you would like to kill active
% processes. Alternatively, you can type abandon at the database prompt.

abandon.

% This concludes the introduction to the emacs interface for Ergo
%
% If you wish to proceed, please open the file proofs.thy now.
% This can be done with C-x C-f proofs.thy







