;; Support for editing elves parameters.
;; $Id: ergo-elves.el,v 2.5 1996/03/08 07:39:31 nickson Exp $

(provide 'ergo-elves)
(require 'ergo-mode)

(defun ergo-edit-elves ()
  "Edit the elves list."
  (interactive)
  (ergo-send-string "emacs_edit_elves" 'echo))


(defconst ergo-elf-regexp
  (concat "\\([a-z]+\\) " ;;elfkind
          "\\([0-9]+\\) " ;;elfnumber
          "\\([^ ]+\\) "  ;;thy
          "\\([^ ]+\\)\n" ;;name
          ))

(defvar ergo-active-elves)
(defvar ergo-hinted-elves)
(defvar ergo-ignored-elves)
(defvar ergo-other-elves)
(defvar ergo-edit-elves-buffer nil)

(defun ergo-edit-elves-handler (str)
  (cond ((eq str t)
         (setq ergo-active-elves nil
               ergo-hinted-elves nil
               ergo-ignored-elves nil
               ergo-other-elves nil
               ergo-edit-elves-buffer (get-buffer-create "*Edit-Elves*"))
         (let ((pop-up-frames t)
               (default-frame-alist (ergo-frame-parameters
                                       'ergo-help-frame-alist
                                       "ergo-elves"
                                       default-frame-alist)))
           (display-buffer ergo-edit-elves-buffer)))
        ((null str)
         (ergo-with-buffer ergo-edit-elves-buffer
           (erase-buffer)
           (ergo-edit-elves-mode)
           (insert (ergo-button "Load Default"
                                "emacs_edit_elves(load_default)" 'exec)
                   "\t"
                   (ergo-button "Load Current"
                                "emacs_edit_elves(load_current)" 'exec)
                   "\t\t"
                   (ergo-button "Save Default"
                                "emacs_edit_elves(save_default)" 'exec)
                   "\t"
                   (ergo-button "Save Current"
                                "emacs_edit_elves(save_current)" 'exec)
                   "\n\n")
           (insert "\nFully active elves:\n")
           (setq ergo-active-elves
                 (sort ergo-active-elves
                       (function (lambda (x y) (< (car x) (car y)))))
                 ergo-hinted-elves
                 (sort ergo-hinted-elves
                       (function (lambda (x y) (< (car x) (car y)))))
                 ergo-ignored-elves
                 (sort ergo-ignored-elves
                       (function (lambda (x y) (< (car x) (car y))))))
           (let ((e ergo-active-elves))
             (while e
               (rplaca e (cdr (car e)))
               (insert (format "   %s:%s\n" (nth 0 (car e)) (nth 1 (car e))))
               (setq e (cdr e))))
           (insert "\nPassive hints:\n")
           (let ((e ergo-hinted-elves))
             (while e
               (rplaca e (cdr (car e)))
               (insert (format "   %s:%s\n" (nth 0 (car e)) (nth 1 (car e))))
               (setq e (cdr e))))
           (insert "\nIgnored elves:\n")
           (let ((e ergo-ignored-elves))
             (while e
               (rplaca e (cdr (car e)))
               (insert (format "   %s:%s\n" (nth 0 (car e)) (nth 1 (car e))))
               (setq e (cdr e))))
           (insert "\nUnlisted elves:\n")
           (let ((e ergo-other-elves))
             (while e
               (rplaca e (cdr (car e)))
               (insert (format "   %s:%s\n" (nth 0 (car e)) (nth 1 (car e))))
               (setq e (cdr e))))))
        ((string-match ergo-elf-regexp str)
         (let ((elf-list (intern (format "ergo-%s-elves"
                                         (ergo-matched-text str 1)))))
           (set elf-list
                (cons (list (string-to-int (ergo-matched-text str 2))
                            (ergo-matched-text str 3)
                            (ergo-matched-text str 4))
                      (symbol-value elf-list)))
           (ergo-debug "added %S to %s\n"
                       (car (symbol-value elf-list)) elf-list)
           (ergo-debug "new value of %s is %S\n"
                       elf-list
                       (symbol-value elf-list))))
        (t
         (ergo-handler-error "Bad elf list: %s" str))))

(defun ergo-save-elves-handler (str)
  (let ((elves-list 
         (mapconcat
            (function (lambda (xx) (format "'%s:%s'" (nth 0 xx) (nth 1 xx))))
            (append ergo-active-elves ergo-hinted-elves ergo-ignored-elves)
            ", "))
        (active-elves (length ergo-active-elves))
        (auto-hints (+ (length ergo-active-elves) (length ergo-hinted-elves))))
    (cond ((string= "default\n" str)
           (ergo-send-string
              (format "set default_elves_list === [%s].\n" elves-list)
              'echo)
           (ergo-send-string
              (format "set default_active_elves === %d.\n" active-elves)
              'echo)
           (ergo-send-string
              (format "set emacs_auto_hints === %d.\n" auto-hints)
              'echo))
          ((string= "current\n" str)
           (ergo-send-string
              (format "set elves_list === [%s].\n" elves-list)
              'log)
           (ergo-send-string
              (format "set active_elves === %d.\n" active-elves)
              'log)
           (ergo-send-string
              (format "set emacs_auto_hints === %d.\n" auto-hints)
              'log)))))

(defvar ergo-edit-elves-mode-map
  (cons 'keymap (cons "Ergo Edit Elves"  ergo-readonly-common-map))
  "Keymap used for Ergo Edit-Elves buffers.")
(define-key ergo-edit-elves-mode-map " " 'ergo-elves-next)
(define-key ergo-edit-elves-mode-map "\C-?" 'ergo-elves-previous)
(define-key ergo-edit-elves-mode-map "+" 'ergo-elves-promote)
(define-key ergo-edit-elves-mode-map "-" 'ergo-elves-demote)
(define-key ergo-edit-elves-mode-map "p" 'ergo-elves-promote)
(define-key ergo-edit-elves-mode-map "d" 'ergo-elves-demote)
(define-key ergo-edit-elves-mode-map "P" 'ergo-elves-promote)
(define-key ergo-edit-elves-mode-map "D" 'ergo-elves-demote)
(define-key ergo-edit-elves-mode-map "?" 'ergo-elves-document)

(defun ergo-edit-elves-mode ()
  "Major mode for editing Ergo elves.
``P'' promotes an elf, either exchanging it with the preceding elf in its
category, or promoting it to the previous (higher) category.
``D'' demotes an elf, either exchanging it with the following elf in its
category, or relegating it to the next (lower) category.
``?'' shows the documentation for an elf."
  (setq buffer-read-only t
        major-mode 'ergo-edit-elves-mode
        mode-name "Ergo-Elves")
  (use-local-map ergo-edit-elves-mode-map))

(defconst ergo-elf-display-regexp "   \\([^:]+\\):\\([^\n]+\\)\n")

(defun ergo-elves-previous ()
  (interactive)
  (beginning-of-line 0)
  (let ((done nil))
    (while (not done)
      (cond ((bobp)
             (setq done t))
            ((looking-at ergo-elf-display-regexp)
             (forward-char 3)
             (setq done t))
            (t
             (beginning-of-line 0))))))

(defun ergo-elves-next ()
  (interactive)
  (beginning-of-line 2)
  (let ((done nil))
    (while (not done)
      (cond ((eobp)
             (setq done t))
            ((looking-at ergo-elf-display-regexp)
             (forward-char 3)
             (setq done t))
            (t
             (beginning-of-line 2))))))

(defun ergo-elves-promote ()
  (interactive)
  (ergo-with-buffer ergo-edit-elves-buffer
    (let (thispos prevpos thiself prevelf ptr)
      (beginning-of-line nil)
      (or (looking-at ergo-elf-display-regexp)
          (error "No elf on this line!"))
      (setq thispos (point)
            thiself (list (ergo-matched-text nil 1) (ergo-matched-text nil 2)))
      (ergo-elves-previous)
      (beginning-of-line nil)
      (if (looking-at ergo-elf-display-regexp)
          (setq prevpos (point)
                prevelf (list (ergo-matched-text nil 1)
                              (ergo-matched-text nil 2))))
      (cond ((and (member thiself ergo-active-elves)
                  (null prevelf))
             (error "This is the highest-ranked elf!"))
            ((member thiself ergo-active-elves)
             (goto-char thispos)
             (transpose-lines 1)
             (goto-char prevpos)
             (setq ptr (member prevelf ergo-active-elves))
             (rplaca ptr thiself)
             (rplaca (cdr ptr) prevelf))
            ((equal thiself (car ergo-hinted-elves))
             (goto-char thispos)
             (beginning-of-line 2)
             (delete-region thispos (point))
             (search-backward "\n\nP")
             (forward-char 1)
             (insert (format "   %s:%s\n" (nth 0 thiself) (nth 1 thiself)))
             (setq ergo-active-elves (nconc ergo-active-elves (list thiself))
                   ergo-hinted-elves (cdr ergo-hinted-elves)))
            ((member thiself ergo-hinted-elves)
             (goto-char thispos)
             (transpose-lines 1)
             (goto-char prevpos)
             (setq ptr (member prevelf ergo-hinted-elves))
             (rplaca ptr thiself)
             (rplaca (cdr ptr) prevelf))
            ((equal thiself (car ergo-ignored-elves))
             (goto-char thispos)
             (beginning-of-line 2)
             (delete-region thispos (point))
             (search-backward "\n\nI")
             (forward-char 1)
             (insert (format "   %s:%s\n" (nth 0 thiself) (nth 1 thiself)))
             (setq ergo-hinted-elves (nconc ergo-hinted-elves (list thiself))
                   ergo-ignored-elves (cdr ergo-ignored-elves)))
            ((member thiself ergo-ignored-elves)
             (goto-char thispos)
             (transpose-lines 1)
             (goto-char prevpos)
             (setq ptr (member prevelf ergo-other-elves))
             (rplaca ptr thiself)
             (rplaca (cdr ptr) prevelf))
            ((member thiself ergo-other-elves)
             (goto-char thispos)
             (beginning-of-line 2)
             (delete-region thispos (point))
             (search-backward "\n\nU")
             (forward-char 1)
             (insert (format "   %s:%s\n" (nth 0 thiself) (nth 1 thiself)))
             (setq ergo-ignored-elves (nconc ergo-ignored-elves (list thiself))
                   ergo-other-elves (delete thiself ergo-other-elves)))))))

(defun ergo-elves-demote ()
  (interactive)
  (ergo-with-buffer ergo-edit-elves-buffer
    (let (thispos thiself nextpos nextelf ptr)
      (beginning-of-line nil)
      (or (looking-at ergo-elf-display-regexp)
          (error "No elf on this line!"))
      (setq thispos (point)
            thiself (list (ergo-matched-text nil 1) (ergo-matched-text nil 2)))
      (ergo-elves-next)
      (beginning-of-line nil)
      (setq nextpos (point)
            nextelf (list (ergo-matched-text nil 1) (ergo-matched-text nil 2)))
      (cond ((setq ptr (member thiself ergo-active-elves))
             (if (cdr ptr)
                 (progn
                   (goto-char nextpos)
                   (transpose-lines 1)
                   (backward-to-indentation 1)
                   (rplaca ptr nextelf)
                   (rplaca (cdr ptr) thiself))
               (goto-char thispos)
               (beginning-of-line 2)
               (delete-region thispos (point))
               (beginning-of-line 3)
               (insert (format "   %s:%s\n" (nth 0 thiself) (nth 1 thiself)))
               (backward-to-indentation 1)
               (setq ergo-hinted-elves (cons thiself ergo-hinted-elves)
                     ergo-active-elves (delete thiself ergo-active-elves))))
            ((setq ptr (member thiself ergo-hinted-elves))
             (if (cdr ptr)
                 (progn
                   (goto-char nextpos)
                   (transpose-lines 1)
                   (backward-to-indentation 1)
                   (rplaca ptr nextelf)
                   (rplaca (cdr ptr) thiself))
               (goto-char thispos)
               (beginning-of-line 2)
               (delete-region thispos (point))
               (beginning-of-line 3)
               (insert (format "   %s:%s\n" (nth 0 thiself) (nth 1 thiself)))
               (backward-to-indentation 1)
               (setq ergo-ignored-elves (cons thiself ergo-ignored-elves)
                     ergo-hinted-elves (delete thiself ergo-hinted-elves))))
            ((setq ptr (member thiself ergo-ignored-elves))
             (if (cdr ptr)
                 (progn
                   (goto-char nextpos)
                   (transpose-lines 1)
                   (backward-to-indentation 1)
                   (rplaca ptr nextelf)
                   (rplaca (cdr ptr) thiself))
               (goto-char thispos)
               (beginning-of-line 2)
               (delete-region thispos (point))
               (beginning-of-line 3)
               (insert (format "   %s:%s\n" (nth 0 thiself) (nth 1 thiself)))
               (backward-to-indentation 1)
               (setq ergo-other-elves (cons thiself ergo-other-elves)
                     ergo-ignored-elves (delete thiself ergo-ignored-elves))))
            ((member thiself ergo-other-elves)
             (error "This elf can be demoted no further!"))))))

(defun ergo-elves-document ()
  (interactive)
  (beginning-of-line nil)
  (or (looking-at ergo-elf-display-regexp)
      (error "No elf on this line!"))
  (ergo-send-string (format "help('elf:%s', '%s')"
                            (ergo-matched-text nil 2)
                            (ergo-matched-text nil 1))
                    'echo))

