;; Support for listing rules and building queries.
;; $Id: ergo-rules.el,v 2.22 1997/07/31 07:05:20 bof Exp $

(provide 'ergo-rules)
(require 'ergo-mode)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Listing rules
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar ergo-rules-frame nil
  "The frame used to display matching rules and construct queries.")

(defvar ergo-query-buffer nil
  "The buffer used to display the current query.")
(defvar ergo-matched-rules-buffer nil
  "The buffer used to display all matching rules.")
(defvar ergo-selected-rule-buffer nil
  "The buffer used to display selected rule.")

(defvar ergo-apprules-count 0)
(defvar ergo-matched-rule nil)

(defun ergo-rules-handler (arg)
  (cond ((eq arg t)
         (ergo-show-default-rules-frame)
         (ergo-with-buffer ergo-selected-rule-buffer
           (erase-buffer)
           (insert "[[Searching...]]"))
         (setq ergo-apprules-count 0
               ergo-matched-rule nil))
        ((null arg)
         (ergo-with-buffer ergo-selected-rule-buffer
           (erase-buffer)
           (insert (cond ((stringp ergo-apprules-count)
                          ergo-apprules-count)
                         ((> ergo-apprules-count 0)
                          "[[Search complete]]")
                         ((= ergo-apprules-count 0)
                          "[[No rules]]")))))
        ((string= arg "end\n")
         (ergo-with-buffer ergo-matched-rules-buffer
           (save-excursion
             (if (> ergo-apprules-count 0)
                 nil
               (goto-char (point-min))
               (while (memq 'ergo-hint (text-properties-at (point)))
                 (beginning-of-line 2))
               (delete-region (point) (point-max)))
             (goto-char (point-max))
             (insert (ergo-matched-rule "rule") ","
                     (ergo-matched-rule "dirn") ","
                     (ergo-matched-rule "cond") " ")
             (ergo-tab-to-column 30)
             (insert (ergo-matched-rule "ishape") ","
                     (ergo-matched-rule "oshape") " ")
             (ergo-tab-to-column 38)
             (insert (ergo-matched-rule "oterm"))
             (add-text-properties
              (progn (beginning-of-line nil) (point))
              (progn (end-of-line nil) (point))
              (list 'ergo-apprules-command (ergo-matched-rule "cmd")
                    'ergo-apprules-data ergo-matched-rule
                    'face ergo-active-face))
             (insert "\n")
             (setq ergo-apprules-count (1+ ergo-apprules-count)
                   ergo-matched-rule nil)
             (message "Got %d rule%s"
                      ergo-apprules-count
                      (if (= ergo-apprules-count 1) "" "s")))))
        ((string= arg "\n")
         nil)
        ((and (= (aref arg 0) ?\[)
              (= (aref arg 1) ?\[))
         (setq ergo-apprules-count arg))
        ((string-match "\\([a-z]+\\): \\(.*\\)$" arg)
         (let ((tag   (ergo-matched-text arg 1))
               (value (ergo-matched-text arg 2)))
           (aput 'ergo-matched-rule
                 tag
                 (cons value (cdr-safe (assoc tag ergo-matched-rule))))))
        (t
         (ergo-handler-error "apprules-handler: bad data."))))

(defun ergo-show-default-rules-frame (&optional delete-old)
  "Show the default rules frame."
;; TO DO: use X resource and ergo-make-panes in ergo-show-default-rules-frame.
  (interactive (list t))
  (if (and delete-old (frame-live-p ergo-rules-frame))
      (delete-frame ergo-rules-frame))
  (ergo-make-buffer "query")
  (ergo-make-buffer "rules")
  (ergo-make-buffer "selected-rule")
  (if (frame-live-p ergo-rules-frame)
      ;; a cut-down version of ergo-make-frame-visible
      (make-frame-visible ergo-rules-frame)
    (setq ergo-rules-frame
          (make-frame (ergo-frame-parameters
                       'ergo-rules-frame-alist
                       "ergo-rules"
                       '((name . "Ergo Rules Frame")
                         (height . 45)
                         (width . 80)))))
    (select-frame ergo-rules-frame)
    (split-window-vertically 10)
    (switch-to-buffer ergo-query-buffer)
    (let ((inhibit-read-only t))
      (erase-buffer)
      (ergo-insert-with-properties '(read-only t intangible t)
                                   ergo-initial-query))
    (other-window 1)
    (split-window-vertically -8)
    (switch-to-buffer ergo-matched-rules-buffer)
    (other-window 1)
    (switch-to-buffer ergo-selected-rule-buffer)
    ))

(defun ergo-matched-rule (tag &optional data)
  (or (car-safe (cdr-safe (assoc tag (or data ergo-matched-rule))))
      ""))

(defun ergo-list-focus-rules ()
  "List the rules that tightly match the current focus."
  (interactive)
  (ergo-send-string (format "emacs_focus_rules(%s)"
                            (if (cdr-safe ergo-last-focus-select)
                                (ergo-focus-cmd (cdr ergo-last-focus-select))
                              "[]"))
                    'exec))

(defun ergo-rules-previous (arg)
  "Back up a line in the Ergo Rules buffer."
  (interactive "p")
  (ergo-rules-move (- arg)))

(defun ergo-rules-next (arg)
  "Go forward a line in the Ergo Rules buffer."
  (interactive "p")
  (ergo-rules-move arg))

(defun ergo-rules-move (arg)
  (forward-line arg)
  (ergo-rules-select-rule))

(defun ergo-rules-apply-rule ()
  "Apply the rule on the current line of the Ergo Rules buffer.
Normally, this is the same as \\[ergo-rules-use-rule]; it differs if some
tactic has set up a list_matching_rule handler to treat rules differently."
  (interactive)
  (ergo-do-rule nil 'log 'focus 'update))

(defun ergo-rules-use-rule ()
  "Use the rule on the current line of the Ergo Rules buffer."
  (interactive)
  (ergo-do-rule "use" 'log 'focus 'update))

(defun ergo-rules-fwd-rule ()
  "Fwd the rule on the current line of the Ergo Rules buffer."
  (interactive)
  (ergo-do-rule "fwd" 'log 'focus 'update))

(defun ergo-rules-rule-rule ()
  "Print the rule on the current line of the Ergo Rules buffer."
  (interactive)
  (ergo-do-rule "rule" 'echo nil nil "rule"))

(defun ergo-rules-irule-rule ()
  "Print the instantiated rule on the current line of the Ergo Rules buffer."
  (interactive)
  (ergo-do-rule "irule" 'echo 'focus))

(defun ergo-rules-template-rule ()
  "Edit a query template for the rule on the current line
 of the Ergo Rules buffer."
  (interactive)
  (ergo-do-rule "query_template" 'exec 'focus))

(defun ergo-rules-select-rule ()
  "Select the rule on the current line of the Ergo Rules buffer."
  (interactive)
  (ergo-do-rule nil 'select 'focus)
  (let (data)
    (ergo-with-buffer ergo-matched-rules-buffer
      (setq data (or (get-text-property (1- (point))
                                        'ergo-apprules-data)
                     (get-text-property (1- (point))
                                        'ergo-hint-doc))))            
    (ergo-with-buffer ergo-selected-rule-buffer
      (erase-buffer)
      (if (stringp data)
          (insert data)
        (insert (ergo-matched-rule "kind" data)
                " "
                (ergo-matched-rule "rule" data)
                ","
                (ergo-matched-rule "dirn" data)
                ","
                (ergo-matched-rule "cond" data)
                "\n")
        (ergo-rule-write-conditions (cdr-safe (assoc "cterm" data)))
        (insert "Input:     " (ergo-matched-rule "iterm" data) "\n")
        (insert "Relation:  " (ergo-matched-rule "rreln" data) "\n")
        (insert "Output:    " (ergo-matched-rule "oterm" data) "\n")
        (ergo-rule-write-body (ergo-matched-rule "body" data))))))

(defun ergo-rule-write-conditions (conds)
  (while conds
    (insert "Condition: " (car conds) "\n")
    (setq conds (cdr conds))))

(defun ergo-rule-write-body (body)
  (if (and body (not (string= body "")))
      (insert "Body:      " body "\n")))

(defun ergo-do-rule (cmd &optional special focus update &rest components)
  "Do CMD on the rule on the current line of the Ergo Rules buffer.
CMD defaults to the command for applying the rule (generally `use').
Optional args SPECIAL, FOCUS, UPDATE are as ergo-send-string.
The arguments of the command are deduced by combining the executable elements
of the query (ergo-executable-query-constraints) and the COMPONENTS (rest of
args) -- default components are \"rule\", \"dirn\"and \"cond\"."
  (if (memq (event-basic-type last-command-event)
            '(mouse-1 mouse-2 mouse-3))
      (let ((epos (event-start last-command-event)))
        (set-buffer (window-buffer (posn-window epos)))
        (goto-char (posn-point epos))
        (beginning-of-line nil)
        (set-window-point (posn-window epos) (point))))
  (or components (setq components '("rule" "dirn" "cond")))
  (ergo-with-buffer ergo-matched-rules-buffer
    (ergo-replace-text-properties (point-min) (point-max)
                                  'face ergo-selected-face ergo-active-face)
    (beginning-of-line nil)
    (let* ((start (point))
           (defcmd (or (get-text-property (point) 'ergo-apprules-command)
                       (get-text-property (point) 'ergo-hint)))
           (query (ergo-query-current-query 'exec-only))
           (data (get-text-property (point) 'ergo-apprules-data))
           (rulespec (mapconcat (function (lambda (cpt)
                                            (nth 1 (assoc cpt data))))
                                components
                                ", ")))
      (if (string= query "")
          (setq query rulespec)
        (setq query (concat rulespec ", " query)))
      (skip-chars-forward "^ ")
      (add-text-properties start (point) (list 'face ergo-selected-face))
      (if (and defcmd (not (string= defcmd "")))
          (if (and (string-match "[a-z]+(" defcmd) cmd)
              (setq cmd (concat cmd (substring defcmd (1- (match-end 0)))))
            (setq cmd defcmd))
        (setq cmd (format "%s([%s])" (or cmd "use") query)))
      (ergo-send-string cmd special focus update))))

(defun ergo-rules-get-command-on-line (&optional noerror)
  ;; Return the command for the current line of the Ergo Rules buffer,
  ;; It is an error if there is no rule on the line, unless noerror is non-nil.
  (save-excursion
    (beginning-of-line nil)
    (let ((cmd )
          (data (get-text-property (point) 'ergo-apprules-data)))
      (cond ((and cmd (not (string= cmd "")))
             cmd)
            (data
             (format "%s,%s,%s"
                     (ergo-matched-rule "rule" data)
                     (ergo-matched-rule "dirn" data)
                     (ergo-matched-rule "cond" data)))
            (noerror
             nil)
            (t
             (error "No rule on line."))))))

(defun ergo-rules-mode ()
  "Major mode for browsing Ergo rules.
Press `\\[ergo-rules-apply-rule]' to apply rule on current line,
      `\\[ergo-rules-rule-rule]' to print it (command `rule'),
      `\\[ergo-rules-select-rule]' to select it.
`DEL' and `SPC' move back and forward through the list.

\\{ergo-rules-mode-map}"
  (interactive)
  (setq major-mode 'ergo-rules-mode
        mode-name "Ergo-Rules"
        truncate-lines t
        buffer-read-only t)
  (use-local-map ergo-rules-mode-map))

(defun ergo-selected-rule-mode ()
  "Major mode for looking at an Ergo rule.

\\{ergo-selected-rule-mode-map}"
  (interactive)
  (setq major-mode 'ergo-selected-rule-mode
        mode-name "Ergo-Selected-Rule"
        truncate-lines t
        buffer-read-only t)
  (use-local-map ergo-selected-rule-mode-map))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Building queries
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun ergo-query-handler (arg)
  (cond ((eq arg t)
         (ergo-show-default-rules-frame)
         (ergo-with-buffer ergo-query-buffer
           (delete-region (1- (length ergo-initial-query))
                          (1- (buffer-size)))))
        ((null arg)
         nil)
        ((string-match "\\([a-z]+\\): \\(.*\\)$" arg)
         (let* ((tag (ergo-matched-text arg 1))
                (val (ergo-matched-text arg 2))
                (vals (copy-sequence
                         (cdr (assoc tag ergo-query-constraints)))))
           (setq vals (cons val (delete val vals)))
           (ergo-with-buffer ergo-query-buffer
             (ergo-query-add-constraint (cons tag vals)))))))

(defconst ergo-initial-query
  (concat (ergo-button "Search" '(ergo-query-search))
          "   "
          (ergo-button "Stop" '(ergo-rules-stop-search))
          "   "
          (ergo-button "Use" '(ergo-rules-use-rule))
          "   "
          (ergo-button "Fwd" '(ergo-rules-fwd-rule))
          "   "
          (ergo-button "Default" '(ergo-list-focus-rules))
          "   "
          (ergo-button "Add Constraint ..." '(ergo-query-add-constraint))
          "\n[\n]\n"))

;; doesn't work properly, because the hook runs *before* the text is inserted!!
(defun ergo-query-insert-in-field (start end)
  (add-text-properties start (1+ end)
      (list 'query-frobbable (get-text-property (1+ end)
                                   'inheritable-query-frobbable)
            'front-sticky t
            'face ergo-active-face)))

(defun ergo-query-current-query (&optional executable-only)
  (let ((query nil)
        spos epos tag)
    (ergo-with-buffer ergo-query-buffer
      (if (= (buffer-size) 0)
          nil
        (setq spos (point-min))
        (while (/= (char-after spos) ?\[)
          (setq spos (1+ spos)))
        (setq spos (+ spos 3))
        (while (< spos (1- (point-max)))
          (setq epos spos)
          (while (/= (char-after epos) ?\()
            (setq epos (1+ epos)))
          (setq tag (format "%s" (buffer-substring spos epos)))
          (if (or (not executable-only)
                  (member tag ergo-executable-query-constraints))
              (setq query (cons (list tag) query)))
          (setq spos (1+ epos))
          (while (/= (char-after epos) ?\n)
            (setq epos (1+ epos)))
          (if (= (char-after (1- epos)) ?,)
              (setq epos (- epos 2))
            (setq epos (- epos 1)))
          (if (or (not executable-only)
                  (member tag ergo-executable-query-constraints))
              (rplacd (car query) (format "%s" (buffer-substring spos epos))))
          (setq spos (+ epos 4)))))
    (mapconcat (function (lambda (c)
                           (format "%s(%s)" (car c) (cdr c))))
               (nreverse query)
               ", ")))
  
(defun ergo-query-search ()
  (ergo-send-string (format "emacs_rules([%s])" (ergo-query-current-query))
                    'exec))

(defconst ergo-query-constraints
  '(("thys" "[curr]" "[*]" "[upto()]")
    ("name" "_")
    ("dirn" "_" "f" "r" "t" "rt")
    ("cond" "_" "c" "u")
    ("reln" "_")
    ("frozen" "false" "true")
    ("matching" "any" "tight" "loose" "frozen")
    ("imatch" "0,inf")
    ("omatch" "0,inf")
    ("i" "_")
    ("c" "_")
    ("iterm" "_" "current_focus")
    ("oterm" "_" "current_goal")
    ("cterm" "_" "hyp:_")
    ("subterm" "_")
    ("conjs" "_")))

(defconst ergo-executable-query-constraints
  '("i" "c" "oterm" "cterm" "conjs"))

(defun ergo-query-add-constraint (&optional constr)
  "Add a constraint to the current query.
CONSTR (or item in ergo-query-constraints menu) is a list of strings,
whose car is the constraint name and cdr is a list of values, the first of
which is the default value."
  (or constr
      (setq constr (ergo-easy-menu "Constraint" ergo-query-constraints)))
  (let ((inhibit-read-only t))
    (if (null constr)
        nil
      (ergo-with-buffer-tangible
         (goto-char (- (point-max) 3))
         (ergo-insert-with-properties 
            (list 'read-only t
                  'intangible t
                  'insert-in-front-hooks '(ergo-query-insert-in-field)
                  'inheritable-query-frobbable (cdr constr))
            (if (= (char-after (1- (point))) ?\)) ",\n" "\n"))
         (ergo-insert-with-properties '(read-only t intangible t)
                                      (concat " " (car constr)))
         (ergo-insert-with-properties (list 'read-only t
                                            'intangible t
                                            'rear-nonsticky t)
                                      "(")
         (ergo-insert-with-properties (list 'front-sticky t
                                            'face ergo-active-face
                                            'query-frobbable (cdr constr))
                                      (car (cdr constr)))
         (ergo-insert-with-properties 
            (list 'read-only t
                  'intangible t
                  'insert-in-front-hooks '(ergo-query-insert-in-field)
                  'inheritable-query-frobbable (cdr constr))
            ")"))))
  (goto-char (point-min)))

(defun ergo-query-template-mode ()
  "Major mode for building Ergo query templates.
You can select a constraint for editing by clicking on the constraint name.
Some constraints have a menu of typical values that can be selected by
clicking on the current value.
\\[ergo-query-delete-field] will clear out the current value;
repeating \\[ergo-query-delete-field] will delete the constraint completely.
\\{ergo-query-template-mode-map}
"
  (setq major-mode 'ergo-query-template-mode
        mode-name "Ergo-Query")
  (use-local-map ergo-query-template-mode-map))

(defun ergo-query-next-field (arg)
  "Go to the next query field."
  (interactive "p")
  (forward-line arg))

(defun ergo-query-previous-field (arg)
  "Go to the previous query field."
  (interactive "p")
  (if (eobp) (setq arg (1+ arg)))
  (forward-line (- arg)))

(defun ergo-query-delete-field (&optional nokill)
  "Delete a query field.
If the field is non-empty, clear it out (and put the old value in the kill
ring).  If the field is empty, delete the whole entry."
  ;; with optional arg nokill, leave non-empty entries alone.
  (interactive)
  (ergo-with-buffer-tangible
   (let ((inhibit-read-only t)
         start end)
     (beginning-of-line nil)
     (forward-sexp 1)
     (setq start (1+ (point)))
     (forward-sexp 1)
     (setq end (1- (point)))
     (message "text is \"%s\"" (buffer-substring start end))
     (if (or (string= (buffer-substring start end) " ")
             (string= (buffer-substring start end) ""))
         ;; delete the whole constraint
         (progn
           (beginning-of-line nil)
           (setq start (point))
           (end-of-line nil)
           (if (= (char-after (1- (point))) ?,)
               nil
             (save-excursion
               ;; delete comma from previous line
               (end-of-line 0)
               (if (= (char-after (1- (point))) ?,)
                   (progn
                     (backward-delete-char 1)
                     (setq start (1- start))))))
           (beginning-of-line 2)
           (delete-region start (point)))
       ;; erase the value
       (if nokill
           (forward-line 1)
         (goto-char start)
         (insert-and-inherit " ")
         (delete-region (1+ start) (1+ end))
         (goto-char start))))))

(defun ergo-query-send ()
  "Send the query to Ergo.
Clean out any empty fields first."
  (interactive)
  (ergo-query-clean)
  (ergo-send-string (format "%s" (buffer-string)) 'log 'focus 'update))

(defun ergo-query-clean ()
  "Clean out any empty fields from the query template."
  (interactive)
  (goto-char (point-min))
  (forward-line 1)
  (while (not (eobp))
    (ergo-query-delete-field 'nokill)))

(defun ergo-query-frob (pos)
  (let* ((vals (get-text-property pos 'query-frobbable))
         (val (and vals
                   (cdr vals)                            
                   (ergo-easy-menu "Value"
                                   (get-text-property pos 'query-frobbable)))))
    (if val
        (progn
          (goto-char (or (previous-single-property-change
                          (1+ pos) 'query-frobbable)
                         0))
          (ergo-insert-with-properties (list 'front-sticky t
                                             'face ergo-active-face
                                             'query-frobbable vals)
             val)
          (delete-region (point)
                         (next-single-property-change
                          (point) 'query-frobbable))
          (goto-char (1- (point)))))))

(defun ergo-rules-stop-search ()
  "Stop searching for rules."
  (interactive)
  (ergo-interrupt)
  (ergo-send-string "c\n" nil))

(defun ergo-query-undo ()
  (interactive)
  (let ((inhibit-read-only t))
    (ergo-with-buffer-tangible
     (undo))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Hints
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar ergo-elf-name nil)
(defvar ergo-elf-action nil)
(defvar ergo-elf-msg nil)
(defvar ergo-elf-help nil)
(defvar ergo-elf-doc-cache nil)

(defun ergo-hints-handler (str)
  (cond ((eq str t)
         (ergo-show-default-rules-frame)
         (ergo-with-buffer ergo-matched-rules-buffer
           (goto-char (point-min))
           (while (memq 'ergo-hint (text-properties-at (point)))
             (delete-region (point) (progn (beginning-of-line 2) (point))))))
        ((null str)
         nil)
        ((string-match "^elf: \\(.*\\)$" str)
         (setq ergo-elf-name (ergo-matched-text str 1)
               ergo-elf-action nil
               ergo-elf-msg nil
               ergo-elf-help nil))
        ((string-match "^action: \\(.*\\)$" str)
         (setq ergo-elf-action (ergo-matched-text str 1)))
        ((string-match "^msg: \\(.*\\)$" str)
         (setq ergo-elf-msg (ergo-matched-text str 1)))
        ((string-match "^end: $" str)
         (ergo-with-buffer ergo-matched-rules-buffer
           (ergo-insert-with-properties (list 'ergo-hint ergo-elf-action
                                              'ergo-hint-doc ergo-elf-help
                                              'face ergo-active-face)
               ergo-elf-name
               (make-string (max 1 (- 20 (length ergo-elf-name))) ?\ )
               ergo-elf-msg
               "\n"))
         (aput 'ergo-elf-doc-cache ergo-elf-name ergo-elf-help))
        ((string-match "^{{{cached}}}$" str)
         (setq ergo-elf-help (cdr (assoc ergo-elf-name ergo-elf-doc-cache))))
        (t
         (setq ergo-elf-help
               (concat ergo-elf-help str)))))

