%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: assert.ql,v 1.53 2001/06/04 01:24:07 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file contains all the procedures for adding information
%% to the Ergo database.  Together with dbms/query.ql and
%% dbms/fact_types.ql, it hides the internal naming system used
%% for database predicates.
%%
%% NOTE: this file should not contain any command/doc declarations.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% We print the banner and declare global dynamics here, as this
%% is the first file to be linked, and the dynamics must exist
%% before we can process any command declarations.
%% 
%% Actually, assert.ql is a sensible place to declare things as dynamic
%% anyway...  :-)

?- banner,
   create_global_database.


%% assert_theory( Fact ).    %% goes into current theory.
%% assert_theory( Fact, Thy ).
%% retract_theory( Fact, Thy ).
%% retractall_theory( Fact, Thy ).
%%
%% assert_global( Fact ).
%% retract_global( Fact ).
%% retractall_global( Fact ).
%%
%%  assert_theory/2 asserts Fact into theory Thy in the Ergo database.
%%  assert_theory/1 is the same, but uses current_dir to determine
%%      which theory Fact should be asserted into.
%%  retract_theory/2 retracts the first matching instance of Fact
%%      from theory Thy.
%%  retractall_theory/2 retracts all matching instances of Fact
%%      from theory Thy.
%%
%%      The *_theory procedures generally require the outermost
%%      functor of Fact to satisfy theory_fact/2.  However, if
%%      Thy = 'global', then they call the *_global procedures
%%      and in this case, the outermost functor of Fact must
%%      satisfy global_fact/2.
%%
%%      This ability of the *_theory procedures to perform operations
%%      on the global part of the database gives a convenient way of
%%      manipulating facts like tactic/3 that may be either global or
%%      theory-specific.
%%
%%
%%  assert_global/1 asserts Fact into the global part of the database.
%%  retract_global/1 retracts the first matching instance of Fact
%%      from the global part of the database.
%%  retractall_global/2 retracts all matching instances of Fact
%%      from the global part of the database.
%%
%%      The *_global procedures require the outermost functor of Fact
%%      to satisfy global_fact/2.
%%
%%  NB. There is no support for asserting clauses with Prolog bodies,
%%      since all Prolog code (eg. axiom side-conditions) should be 
%%      stored as data so that the prover can manipulate the code
%%      in a flexible fashion, such as printing it, delaying its 
%%      execution etc.
%%
%%  mode assert_theory( @compound ).
%%  mode assert_theory( @compound, @atom ).
%%  mode retract_theory( +compound, @atom ).
%%  mode retractall_theory( +compound, @atom ).
%%
%%  mode assert_global( @compound ).
%%  mode retract_global( +compound ).
%%  mode retractall_global( +compound ).
%%

assert_theory(Fact) :-
    global_state_lookup(current_dir,Thy),
    assert_theory(Fact, Thy).
 
assert_theory(Fact, global) :-
    !,
    assert_global(Fact).
assert_theory(Fact, Thy) :-
    atom(Thy),
    functor(Fact, F, N),
    PRE(assert_theory, loaded_theory(Thy,_)),
    touch_theory_fact(Thy, F, N, AZ, _),
    dbms_asserted_form(Thy, Fact, RealFact),
    assertaz(AZ, RealFact),
    announce_event(assert(Thy,Fact)).


assert_theory(Fact, Code, Thy) :-
    atom(Thy),
    functor(Fact, F, N),
    PRE(assert_theory, loaded_theory(Thy,_)),
    touch_theory_fact(Thy, F, N, AZ, _),
    dbms_asserted_form(Thy, Fact, RealFact),
    assertaz(AZ, RealFact, Code),
    announce_event(assert(Thy,Fact)). %% Could pass Code as an extra argument?

assertaz(a, RealFact, Code) :- asserta((RealFact :- Code)).
assertaz(z, RealFact, Code) :- assertz((RealFact :- Code)).


%% assert_axthm(Name, Kind, Ass, Stmt, Body, Thy).
%%
%%  All axioms/theorems/postulates should be asserted using this,
%%  because it may preprocess the rule in some way before asserting it.
%%
assert_axthm(Name, Kind, [], Stmt, Body, Thy) :-
    !,
    rule_shape(Stmt, Shape),
    assert_theory(rule(Name, Kind, Shape, Stmt, Body), Thy).
assert_axthm(_, _, _, _, _, _) :-
    handle_error(['assert_axthm: assumptions are not supported.']).

%% private rule_shape(Stmt, Shape).
%% private rule_shape_arg(Stmt, Val0, Val).
%% private rule_shape_arg(N, Stmt, Val0, Val).
%%
%%  Build a Shape skeleton describing the shape of Stmt.
%%  A skeleton can be:
%%   Num
%%   b(Num, Num)
%%   b(Num, b(Num, Num)).
%%  'b' indicates any binary functor.
%%  Num indicates the number of atoms occurring in the term at that position.
%%
rule_shape(A, 0) :-
    var(A),
    !.
rule_shape(_(A, B), b(AV, 0)) :-
    var(B),
    !,
    rule_shape_arg(A, 0, AV).
rule_shape(_(A, _(B, C)), b(AV, b(BV, CV))) :-
    !,
    rule_shape_arg(A, 0, AV),
    rule_shape_arg(B, 0, BV),
    rule_shape_arg(C, 0, CV).
rule_shape(_(A, B), b(AV, BV)) :-
    !,
    rule_shape_arg(A, 0, AV),
    rule_shape_arg(B, 0, BV).
rule_shape(A, AV) :-
    rule_shape_arg(A, 0, AV).

rule_shape_arg(T, V0, V) :-
    ( any_variable(T) ->
        V = V0
    ; sub(T) ->
        V = V0
    ; atomic(T) ->
        V is V0+1
    ; quant(T) ->
        quantify(T, Q, L, B),
        rule_shape_arg(Q, V0, V1),
        rule_shape_arg(L, V1, V2),
        rule_shape_arg(B, V2, V)
    ; compound(T) ->
        functor(T, F, N),
        rule_shape_arg(F, V0, V1),
        rule_shape_arg(N, T, V1, V)
    ).
rule_shape_arg(0, _, V, V) :-
    !.
rule_shape_arg(N, T, V0, V) :-
    PRE('rule_shape_arg', N > 0),
    arg(N, T, A),
    rule_shape_arg(A, V0, V1),
    N1 is N - 1,
    rule_shape_arg(N1, T, V1, V).

assert_global(Fact) :-
    functor(Fact, F, N),
    touch_global_fact(F, N, AZ, _),
    dbms_asserted_form(global, Fact, RealFact),
    assertaz(AZ, RealFact),
    announce_event(assert(global,Fact)).

assertaz(a, Fact) :- asserta(Fact).
assertaz(z, Fact) :- assertz(Fact).

retract_theory(Fact, global) :-
    !,
    retract_global(Fact).
retract_theory(Fact, Thy) :-
    atom(Thy),
    PRE(retract_theory, loaded_theory(Thy,_)),
    functor(Fact, F, N),
    touch_theory_fact(Thy, F, N, AZ, Retract),
    check_is_retractable(Retract, F, N),
    (
        AZ = z,
        %% loaded clauses come before newly asserted clauses,
        %% so we `retract' the loaded clauses first.
        query_loaded_theory_fact(Thy, Fact),
        dbms_retracted_form(Thy, Fact, RetractedFact),
        assert(RetractedFact)
    ;
        dbms_asserted_form(Thy, Fact, RealFact),
        retract((RealFact :- true))
    ;
        AZ = a,
        %% loaded clauses come after newly asserted clauses,
        %% so we `retract' the loaded clauses last.
        query_loaded_theory_fact(Thy, Fact),
        dbms_retracted_form(Thy, Fact, RetractedFact),
        assert(RetractedFact)
    ).


check_is_retractable(true, _, _).
check_is_retractable(false, F, N) :-
    handle_error([q(F/N),' facts cannot be retracted.']).


query_loaded_theory_fact(Thy, Fact) :-
    dbms_loaded_form(Thy, Fact, LoadedFact),
    functor(LoadedFact, LF, N),
    current_predicate(LF/N),
    dbms_retracted_form(Thy, Fact, RetractedFact),
    call_predicate(LoadedFact),
    \+ call_predicate(RetractedFact).


retractall_theory(Fact, global) :-
    !,
    retractall_global(Fact).
retractall_theory(Fact, Thy) :-
    atom(Thy),
    PRE(retractall_theory, loaded_theory(Thy,_)),
    functor(Fact, F, N),
    touch_theory_fact(Thy,F,N,_,Retract),
    check_is_retractable(Retract, F, N),
    %% Retract all matching loaded facts.
    ( \+ query_loaded_theory_fact(Thy, Fact) ->
        true
    ;
        dbms_retracted_form(Thy, Fact, RetractedFact),
        assert(RetractedFact)
    ),
    %% Retract all matching asserted facts.
    dbms_asserted_form(Thy, Fact, AssertedFact),
    retractall(AssertedFact),
    !.


retract_global(Fact) :-
    functor(Fact, F, N),
    touch_global_fact(F, N, _, _),
    dbms_asserted_form(global, Fact, RealFact),
    retract((RealFact :- true)),
    !.


retractall_global(Fact) :-
    functor(Fact, F, N),
    touch_global_fact(F, N, _, _),
    dbms_asserted_form(global, Fact, RealFact),
    retractall(RealFact),
    !.



%% tactic_assert_theory( Fact ).    %% goes into current theory.
%% tactic_assert_theory( Fact, Thy ).
%% tactic_retract_theory( Fact, Thy ).
%% tactic_retractall_theory( Fact, Thy ).
%%
%% tactic_assert_global( Fact ).
%% tactic_retract_global( Fact ).
%% tactic_retractall_global( Fact ).
%%
%%  These are the tactic versions of assert_theory/global.
%%  They are restricted to 'tactic' fact types (ie. facts
%%  that have the core(false) property) so
%%  that tactics cannot corrupt the core Ergo database.
%%
%%  mode tactic_assert_theory( @compound ).
%%  mode tactic_assert_theory( @compound, @atom ).
%%  mode tactic_retract_theory( +compound, @atom ).
%%  mode tactic_retractall_theory( +compound, @atom ).
%%
%%  mode tactic_assert_global( @compound ).
%%  mode tactic_retract_global( +compound ).
%%  mode tactic_retractall_global( +compound ).
%%

tactic_assert_theory(Fact) :-
    global_state_lookup(current_dir,Thy),
    tactic_assert_theory(Fact, Thy).

tactic_assert_theory(Fact, Thy) :-
    check_tactic_fact(Thy, Fact),
    assert_theory(Fact, Thy).

tactic_retract_theory(Fact, Thy) :-
    check_tactic_fact(Thy, Fact),
    retract_theory(Fact, Thy).

tactic_retractall_theory(Fact, Thy) :-
    check_tactic_fact(Thy, Fact),
    retractall_theory(Fact, Thy).

tactic_assert_global(Fact) :-
    check_tactic_fact(global, Fact),
    assert_global(Fact).

tactic_retract_global(Fact) :-
    check_tactic_fact(global, Fact),
    retract_global(Fact).

tactic_retractall_global(Fact) :-
    check_tactic_fact(global, Fact),
    retractall_global(Fact).


%% private check_tactic_fact(Thy,Fact).
%%
%%  Succeeds if:
%%    (a) Thy=global and Fact is a declared global fact, or
%%    (b) Thy is some other atom and Fact is a declared theory fact.
%%
%%  Generates an error otherwise.
%%
%%  mode check_tactic_fact(@atom,@compound).
%%
check_tactic_fact(Thy,Fact) :-
    atom(Thy),
    compound(Fact),
    functor(Fact,F,N),
    atom(F),
    is_tactic_fact(Thy,F,N),
    !.
check_theory_tactic_fact(Thy,Fact) :-
    functor(Fact,F,N),
    handle_error(['tactics cannot change ',q(Thy),' ',q(F/N),' facts.']).

is_tactic_fact(global,F,N) :-
    !,
    global_fact(F,N,false,_,_,_,_).
is_tactic_fact(_Thy,F,N) :-
    %% \+ _Thy=global.
    theory_fact(F,N,false,_,_,_,_).


%% touch_theory_fact(TheoryName, Functor, Arity, AZ, Retract).
%% touch_global_fact(Functor, Arity, AZ, Retract).
%% touch_database(TheoryName).
%%
%%  Marks part of the Ergo database as having been modified,
%%  so that it will be saved by the next 'save' command.
%%  TheoryName must be the name of an existing theory, or the
%%  atom 'global'.  Therefore, these procedures should not normally
%%  be called directly by the user, because they do not include
%%  error checking.
%%
%%  AZ is 'a' if new clauses should be added before all existing
%%  clauses, or 'z' if they should be added after all existing clauses.
%%
%%  Retract is true if the clauses are allowed to be retracted,
%%  otherwise it is false.
%%
%%  mode touch_theory_fact(@theory, @atom, @integer, ?atom).
%%  mode touch_global_fact(@atom, @integer, ?atom).
%%  mode touch_database(@atom).
%%
touch_theory_fact(Thy, F, N, AZ, Retract) :-
    theory_fact(F,N,_,true,AZ1,Retract1,_),
    !,
    touch_database(Thy),
    AZ = AZ1,
    Retract = Retract1.
touch_theory_fact(_, F, N, AZ, Retract) :-
    theory_fact(F,N,_,false,AZ1,Retract1,_),
    !,
    AZ = AZ1,
    Retract = Retract1.
touch_theory_fact(_, F, N, _, _) :-
    handle_error([w(F), '/', w(N), ' is not a known theory fact.']).

touch_global_fact(F, N, AZ, Retract) :-
    global_fact(F,N,_,true,AZ1,Retract1,_),
    !,
    touch_database(global),
    AZ = AZ1,
    Retract = Retract1.
touch_global_fact(F, N, AZ, Retract) :-
    global_fact(F,N,_,false,AZ1,Retract1,_),
    !,
    AZ = AZ1,
    Retract = Retract1.
touch_global_fact(F, N, _, _) :-
    handle_error([w(F), '/', w(N), ' is not a known global fact.']).

touch_database(Thy) :- query_global(needs_saving(Thy)), !.
touch_database(Thy) :- assert_global(needs_saving(Thy)).

%% dbms_asserted_form(Thy, Fact, AssertedFact).
%% dbms_retracted_form(Thy, Fact, RetractedFact).
%% dbms_loaded_form(Thy, Fact, LoadedFact).
%%
%%  For the theory Thy return the internal Ergo database fact
%%  that is the asserted/retracted equivalent of Fact.
%%  Note that Thy may be 'global'.
%%
%%  (The need for storing retracted facts arises because we allow
%%  theories to be compiled, but still allow the facts in those
%%  theories to be retracted.  This is implemented by asserting
%%  a copy of each retracted fact, then each query of a
%%  compiled fact checks that that compiled fact is not retracted).
%%
%% mode dbms_asserted_form(@theory, @term, ?term).
%% mode dbms_retracted_form(@theory, @term, ?term).
%% mode dbms_loaded_form(@theory, @term, ?term).

dbms_asserted_form(global, Fact, AFact) :-
    !,
    functor(Fact, F, N),
    concat_atom(['/A/',F], AF),
    functor(AFact, AF, N),
    same_args(Fact, AFact, 1, N).
dbms_asserted_form(Thy, Fact, AFact) :-
    functor(Fact, F, N),
    concat_atom([Thy,'/A/',F], AF),
    functor(AFact, AF, N),
    same_args(Fact, AFact, 1, N).

dbms_retracted_form(Thy, Fact, AFact) :-
    functor(Fact, F, N),
    concat_atom([Thy,'/R/',F], AF),
    ( theory_fact(F, N, _, _, _, _, Props),
      member(key(KeyField), Props) ->
          %% One field (KeyField) is enough to uniquely identify a fact.
          functor(AFact, AF, 1),
          arg(KeyField, Fact, Arg),
          arg(1, AFact, Arg)
    ;
        functor(AFact, AF, N),
        same_args(Fact, AFact, 1, N)
    ).

dbms_loaded_form(Thy, Fact, AFact) :-
    functor(Fact, F, N),
    concat_atom([Thy,'/L/',F], AF),
    functor(AFact, AF, N),
    same_args(Fact, AFact, 1, N).


%% dbms_asserted_name(Thy, Name, AssertedName).
%% dbms_loaded_name(Thy, Name, LoadedName).
%% dbms_new_searcher_name(Thy, SearcherName).
%% dbms_loaded_searcher_name(Thy, SearcherName).
%%
%%  Similar to dbms_*_form, but only transforms the name.
%%
%%  dbms_new_searcher_name determines the name of the `searcher'
%%  predicate for theories that have been created during the current
%%  session, while dbms_loaded_searcher_name determines the name of
%%  the searcher predicate for theories that have been loaded/consulted.
%%
%%  There is no dbms_retracted_name, because we always go through
%%  dbms_retracted_form to allow it to change the arity.
%%
%%  mode dbms_asserted_name(@theory, @atom, ?atom).
%%  mode dbms_loaded_name(@theory, @atom, ?atom).
%%  mode dbms_new_searcher_name(@theory, ?atom).
%%  mode dbms_loaded_searcher_name(@theory, ?atom).
%%
dbms_asserted_name(global, F, AF) :- !,atom_concat2('/A/', F, AF).
dbms_asserted_name(Thy, F, AF)    :- concat_atom([Thy,'/A/',F], AF).

dbms_loaded_name(Thy, F, AF)      :- concat_atom([Thy,'/L/',F], AF).

dbms_new_searcher_name(Thy, S)    :- atom_concat2(Thy, '/Asearcher', S).

dbms_loaded_searcher_name(Thy, S) :- atom_concat2(Thy, '/Lsearcher', S).


%% create_theory_searcher(Thy).
%%
%%  Creates a simple `searcher' predicate for Thy, and asserts
%%  the loaded_theory/2 fact.
%%  Thy must be a newly created theory with no compiled/loaded facts.
%%
%%  mode create_theory_searcher(@theory).
%%
create_theory_searcher(Thy) :-
    dbms_new_searcher_name(Thy, SName),
    %% This will remove any previously created searcher for Thy.
    dynamic(SName/1),
    theory_fact(F,N),
    functor(Fact, F, N),
    dbms_asserted_form(Thy,Fact,AFact),
    assert((SName(Fact) :- !,AFact)),
    fail.
create_theory_searcher(Thy) :-
    dbms_new_searcher_name(Thy, SName),
    assert((SName(FF) :- dbms_asserted_form(Thy,FF,AFF), call_predicate(AFF))),
    asserta(loaded_theory(Thy, SName)).

