%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##
%% 
%% CVS $Id: expansions.ql,v 1.26 1995/05/02 22:51:42 marku Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% WARNING!
%% Some expansions should be able to run OUTSIDE Ergo.
%% Beware of using Ergo-defined procedures in this file!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% This file contains term_expansion procedures for source and tactic files.
%% Four kinds of expansions are defined.
%%
%% 1. DECLARATION EXPANSIONS (can be outside Ergo, unless wrappers are used).
%%     Declarations look like:
%%      Table command
%%         Cmd, ...
%%      doc
%%         DocString, ...
%%      extends
%%         Node, ...
%%      mode
%%         ModeSpec, ...
%%      wrapper
%%         TacticWrapper ---> SubProofWrappers
%%      implementation
%%         Impl.
%%     All clauses but the first command are optional; not all clauses are
%%      meaningful for all command Tables.
%%
%%     All doc clauses are combined (in order) with the descriptions of the
%%      command clauses to define a single doc node.  Wrapper and
%%      implementation clauses pertain to the latest command clause.
%%      A command clause MUST appear first.
%%
%%     Example 1:
%%       dbms command
%%         mycmd(A, B)
%%       doc
%%         'Do my command on A and B',nl,
%%         'no more and no less!'
%%       extends
%%         dbms
%%       command
%%         mycmd(A)
%%       doc
%%         'B defaults to []'
%%       implementation
%%         mycmd(A, [])
%%       mode
%%         mycmd(+integer),
%%         mycmd(+integer, ?list(atom)),
%%         mycmd(?integer, +list(atom)).
%% 
%%     This generates two commands mycmd/1 and mycmd/2 in the dbms table.  Only
%%     mycmd/2 needs to be impld directly in Prolog, since the `implementation'
%%     clause furnishes an implementation for mycmd/1.  A doc node
%%     (cross-linked to both dbms commands) is created from the `command' and
%%     `doc' clauses.
%%
%%   Example 2:
%%     TO DO: Something with wrappers.
%% 
%% 2. MACRO EXPANSIONS (can be outside Ergo).
%%     The following appear only as `facts':
%%       @include(File).       -- include File.
%%       @define(Mac, Def).    -- define macro
%%       @define(Mac).         -- define macro
%%       @undef(Mac).          -- remove definition
%%       @if(Goal).            -- start conditional clauses
%%       @ifdef(Mac).          -- same as @if(is_defined_macro(Mac)).
%%       @ifndef(Mac).         -- same as @if(\+ is_defined_macro(Mac)).
%%       @else.                -- reverse conditional clauses
%%       @endif.               -- end conditional clauses
%%
%%     The following appear _within_ clauses:
%%       @macro(Mac)           -- replaced by the expansion of Mac
%%       @if Goal then Term1 else Term2 endif
%%                             -- replaced by Term1 or Term2 according to Goal
%%
%% 3. THEORY EXPANSIONS (must be in Ergo, with a current theory).
%%     Within clauses, terms of the following forms are converted to theory
%%     terms by reading from Chars using the operator and object_var tables
%%     and abbreviations of the current theory.
%%       @thy(Chars)           -- use current theory and [] arity.
%%       @thy(Thy,Chars)       -- use specified theory and [] arity.
%%       @thy(Thy,Chars,Arity) -- use specified theory and arity.
%%     NB: currently, Thy must be the current theory (or a variable).
%%     The term @current_thy is converted to the name of the current theory.
%%
%% 4. TACTIC PREFIXING (can be outside Ergo, but normally will not be).
%%     Every definition and call on a procedure defined in the file is prefixed
%%     with a `module' name: the current theory.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% consult_with_expansions(Expansions, File).
%% reconsult_with_expansions(Expansions, File).
%% compile_with_expansions(Expansions, File).
%%
%%  Consult/reconsult/compile File with Expansions turned on.
%%   Expansions can include: decl(ThyOrGlobal), macro, thy(Thy), prefix(Thy).
%%
%%  mode consult_with_expansions(@list(nonvar), @atom).
%%  mode reconsult_with_expansions(@list(nonvar), @atom).
%%  mode compile_with_expansions(@list(nonvar), @atom).
%%
consult_with_expansions(Exps, File) :-
    unwind_protect((setup_expansions(Exps, File),
                    consult(File)),
                   remove_expansions(Exps)).
reconsult_with_expansions(Exps, File) :-
    unwind_protect((setup_expansions(Exps, File),
                    reconsult(File)),
                   remove_expansions(Exps)).
compile_with_expansions(Exps, File) :-
    unwind_protect((setup_expansions(Exps, File),
                    fcompile(File)),
                   remove_expansions(Exps)).

%% private setup_expansions(Expansions, File).
%%
%%  Set up Expansions for consulting or compiling File.
%%
%%  mode setup_expansions(@list(nonvar), @atom).
%%
setup_expansions([], _).
setup_expansions([Exp|Exps], File) :-
    ( Exp = decl(Thy) ->
        add_multi_expansion_vars(expand_decls(Thy))
    ; Exp = macro ->
        add_multi_expansion(mac_expand_term_1),
        add_subterm_expansion(mac_expand_subterm_1)
    ; Exp = thy(Thy) ->
        add_subterm_expansion_vars(expand_thy_term(Thy, _))
    ; Exp = prefix(Thy) ->
        build_tac_prefix_procs(Thy, File),
        add_multi_expansion(tac_add_prefixes)
    ;
        exp_program_error(unrecoverable,
                          setup_expansions([Exp|Exps], File),
                          ['bad expansion ', w(Exp)])
    ),
    setup_expansions(Exps, File).

%% private remove_expansions(Exps).
%%
%%  Remove Ergo expansions Exps.
%%
%%  mode remove_expansions(@list(atom)).
%%
remove_expansions([]).
remove_expansions([Exp|Exps]) :-
    ( Exp = decl(_) ->
        del_multi_expansion_vars(expand_decls(_))
    ; Exp = macro ->
        del_multi_expansion(mac_expand_term_1),
        del_subterm_expansion(mac_expand_subterm_1)
    ; Exp = thy(_) ->
        del_subterm_expansion_vars(expand_thy_term(_,_))
    ; Exp = prefix(_) ->
        retractall(tac_prefix_procs(_, _)),
        del_multi_expansion(tac_add_prefixes)
    ;
        exp_program_error(unrecoverable,
                          remove_expansions([Exp|Exps]),
                          ['remove_expansions: bad expansion ', w(Exp)])
    ),
    remove_expansions(Exps).

%% 
%% 1. DECLARATION EXPANSIONS.
%%

?- op(1150, xfy, command),
   op(1150, xfy, doc),
   op(1150, xfy, doc_fragment),
   op(1150, xfy, extends),
   op(1150, xfy, mode),
   op(1150, xfy, wrapper),
   op(990,  xfy, '--->'),
   op(1150, xfy, implementation).

%% expand_decls(ThyOrGlobal, Decls, Queries, Vars).
%%
%%  Expand the declarations that preceed Ergo procedure definitions into
%%  queries that build command entries, doc nodes etc.
%%
%%  mode expand_decls(@theory, @compound, -list(term), @list(compound)).
%%
expand_decls(Thy, command(Table, Rest), Qs, Vars) :-
    exp_debug(['trying decl expansion on ', w(command(Table, Rest))]),
    ( expand_decls(Rest, command, [], Qs0, Vars, Table, Thy,
                   [], Doc, _, LC, _, LCD),
      exp_debug(['Parsed declarations; now completing queries.']),
      expand_complete_queries(Qs0, Qs1, Thy, DocName, Qs0),
      exp_debug(['Completed queries ; now building doc.']),
      expand_create_doc_query(Doc, Thy, DocName, Table, LC, Qs1, Qs),
      exp_debug(['Built doc; expansion succeeded.']) ->
          true
    ;
        exp_data_error(warning,
                       declaration,
                       ['Unacceptable declaration ignored:', nl,
                        tab, w(command(Table, Rest))]),
        Qs = []
    ),
    exp_debug(['done decl expansion on ', w(command(Table,Rest)), nl,
               '  resulting queries:', nl,tab,tab, w(Qs)]).

%%% Caution: this clause is called from the dbms `doc' command as well as from
%%% expansions.
%%% To do: Abstract out the appropriate functionality.
expand_decls(Thy, doc(Name, Rest), Qs, _Vars) :-
    ( expand_decls(Rest, doc, [], Qs0, [], concept, Thy,
                   [], Doc1, _, _, _, _) ->
        reverse(Doc1, Doc),
        make_query(doc(Thy, Name, Doc, []), Q),
        Qs1 = [Q | Qs0],
        expand_complete_queries(Qs1, Qs, Thy, Name, Qs1)
    ;
        exp_data_error(warning,
                       declaration,
                       ['Unacceptable declaration ignored:', nl,
                        tab, w(doc(Name, Rest))]),
        Qs = []
    ),
    exp_debug(['done decl expansion on ', w(doc(Name,Rest))]).

expand_decls(Thy, doc_fragment(Head, Rest), Qs, _Vars) :-
    ( expand_decls(Rest, doc, [], Qs0, [], fragment, Thy,
                   [], Doc1, _, _, _, _) ->
        reverse(Doc1, Doc),
        make_query(frag(Thy, Head, Doc), Q),
        Qs1 = [Q | Qs0],
        expand_complete_queries(Qs1, Qs, Thy, Head, Qs1)
    ;
        exp_data_error(warning,
                       declaration,
                       ['Unacceptable declaration ignored:', nl,
                        tab, w(doc_fragment(Head, Rest))]),
        Qs = []
    ),
    exp_debug(['done decl expansion on ', w(doc_fragment(Head,Rest))]).

%% private expand_decls(Decls, Command, Qs0, Qs, Vars, Table, Thy,
%%                      Doc0, Doc, LastCmd0, LastCmd,
%%                      LastCmdDesc0, LastCmdDesc).
%%
%%  A Decls has a Command, a Clause and (maybe) some more Decls.
%%
%%  mode expand_decls(@compound, @atom, +list(compound), -list(compound),
%%            @list(compound), @cmdtab, @thyname,
%%            +writelist, -writelist, +term, -term, +atom, -atom).
%%
expand_decls(command(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, LCD0, LCD) :-
    !,
    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
                        H0, H1, LC0, LC1, LCD0, LCD1),
    expand_decls(Rest, command, Qs1, Qs, Vs, T, Thy,
                 H1, H, LC1, LC, LCD1, LCD).
expand_decls(doc(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, LCD0, LCD) :-
    !,
    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
                        H0, H1, LC0, LC1, LCD0, LCD1),
    expand_decls(Rest, doc, Qs1, Qs, Vs, T, Thy,
                 H1, H, LC1, LC, LCD1, LCD).
expand_decls(extends(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, LCD0, LCD) :-
    !,
    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
                        H0, H1, LC0, LC1, LCD0, LCD1),
    expand_decls(Rest, extends, Qs1, Qs, Vs, T, Thy,
                 H1, H, LC1, LC, LCD1, LCD).
expand_decls(mode(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, LCD0, LCD) :-
    !,
    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
                        H0, H1, LC0, LC1, LCD0, LCD1),
    expand_decls(Rest, mode, Qs1, Qs, Vs, T, Thy,
                 H1, H, LC1, LC, LCD1, LCD).
expand_decls(wrapper(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, LCD0, LCD) :-
    !,
    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
                        H0, H1, LC0, LC1, LCD0, LCD1),
    expand_decls(Rest, wrapper, Qs1, Qs, Vs, T, Thy,
                 H1, H, LC1, LC, LCD1, LCD).
%expand_decls(subwrappers(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
%             H0, H, LC0, LC, LCD0, LCD) :-
%    !,
%    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
%                        H0, H1, LC0, LC1, LCD0, LCD1),
%    expand_decls(Rest, subwrappers, Qs1, Qs, Vs, T, Thy,
%                 H1, H, LC1, LC, LCD1, LCD).
expand_decls(implementation(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, LCD0, LCD) :-
    !,
    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
                        H0, H1, LC0, LC1, LCD0, LCD1),
    expand_decls(Rest, implementation, Qs1, Qs, Vs, T, Thy,
                 H1, H, LC1, LC, LCD1, LCD).
expand_decls(Clause, Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, LCD0, LCD) :-
    %% now we've processed all the keywords out of Clause, so it should
    %% be a comma-separated list of `lines', or a single `line'.
    expand_decls_clause(Clause, Cmd, Qs0, Qs, Vs, T, Thy,
                        H0, H, LC0, LC, LCD0, LCD).

%% private expand_decls_clause(Clause, Cmd, Qs0, Qs, Vars, Table, Thy,
%%                             Doc0, Doc, LastCmd0, LastCmd,
%%                             LastCmdDesc0, LastCmdDesc).
%%
%%  A Clause is a comma-separated list of Lines.
%%
%%  mode expand_decls_clause(@compound, @atom,
%%                           +list(compound), -list(compound),
%%                           @list(compound), @cmdtab, @thyname,
%%                           +writelist, -writelist, +term, -term,
%%                           +atom, -atom).
expand_decls_clause((L1,Ls), Cmd, Qs0, Qs, Vs, T, Thy,
                    H0, H, LC0, LC, LCD0, LCD) :-
    !,
    expand_decls_line(Cmd, L1, Qs0, Qs1, Vs, T, Thy,
                      H0, H1, LC0, LC1, LCD0, LCD1),
    expand_decls_clause(Ls, Cmd, Qs1, Qs, Vs, T, Thy,
                        H1, H, LC1, LC, LCD1, LCD).
expand_decls_clause(L, Cmd, Qs0, Qs, Vs, T, Thy,
                    H0, H, LC0, LC, LCD0, LCD) :-
    expand_decls_line(Cmd, L, Qs0, Qs, Vs, T, Thy,
                      H0, H, LC0, LC, LCD0, LCD).

%% private expand_decls_line(Cmd, Line, Qs0, Qs, Vars, Table, Thy,
%%                           Doc0, Doc, LastCmd0, LastCmd,
%%                           LastCmdDesc0, LastCmdDesc).
%% 
%%  A Line is processed depending on the Cmd it pertains to.
%%  Cmd is the keyword immediately preceeding the group of
%%  lines in which this Line appears.
%%
%%  mode expand_decls_line(@atom, @compound,
%%                    +list(compound), -list(compound),
%%                    @list(compound), @cmdtab, @thyname,
%%                    +writelist, -writelist, +term, -term,
%%                    +atom, -atom).
%%
%%% command: Create the query, leaving CallModes and Impl to be filled in.
%%% Also add a command-description entry to the doc and update LastCmd.
expand_decls_line(command, _, _, _, _, concept, _, _, _, _, _, _, _) :-
    !,
    exp_data_error(unrecoverable,
                   declaration,
                   ['Command clause not permitted in doc declaration.']).
expand_decls_line(Type, Line, _Qs0, _Qs, _Vs, _Tab, Thy,
                  _H0, _H, _LC0, _LC, _LCD0, _LCD) :-
    exp_debug(['Expanding ', w(Type), ' line: ', q(Line), ' in ', w(Thy)]),
    fail.
expand_decls_line(command, Call, Qs, [Q | Qs], Vs, Tab, Thy, H,
                  % NB: the following gets reversed!!!
                  [nl, ' command.',  w(Tab), ' ', w(Thy), ' -- a ', CallA | H],
                  _, Call, _, CallA) :-
    make_query(command(Thy, Tab, Call, _Modes, _Impl, _HN), Q),
    copy_term((Call,Vs),  (CallC,VsC)),
    exp_bind_vars_names(VsC),
    open_string(write, Strm),
    write_term(Strm, CallC, [op_table(Thy)]),
    stream_to_atom(Strm, CallA),
    close(Strm),
    exp_debug(['Created command description ', q(CallA)]).

%%% doc: Add a line to the doc; the query will be created at the end.
expand_decls_line(doc, Line, Qs, Qs, _, _, _, H, [Line|H], LC, LC, LCD, LCD) :-
    ( true /* prefer: is_legal_writelist_elem(Line) */ ->
        true
    ;
        exp_data_error(unrecoverable,
                       declaration,
                       ['Bad doc entry: ', q(Line)])
    ).

%%% extends: Add a doc extension.
expand_decls_line(extends, ExtNode, Qs, [Q|Qs], _Vs, _Tab, Thy,
                  Doc, Doc, LC, LC, LCD, LCD) :-
    make_query(ext(Thy, ExtNode, XRefTag, ThisNode), Q),
    ( var(LCD) ->
        XRefTag = [ThisNode]
    ;
        XRefTag = [LCD]
    ).

%%% mode: Add an entry to the CallModes in the appropriate query.
expand_decls_line(mode, Mode, Qs, Qs, _, Tab, Thy,
                  Doc, Doc, LC, LC, LCD, LCD) :-
    functor(Mode, F, N),
    functor(Call, F, N),
    make_query(command(Thy, Tab, Call, Modes, _Impl, _Doc), Q),
    ( member(Q, Qs) ->
        open_append(Modes, [Mode])
    ;
        exp_data_error(unrecoverable,
                       declaration,
                       ['Mode declaration ', w(Mode), nl,
                        tab, 'has no corresponding command declaration.'])
    ).

%%% wrapper: create a wrapper query for the last command
expand_decls_line(wrapper, Line, Qs, [Q|Qs], _Vs, Tab, Thy,
                  Doc, Doc, Call, Call, LCD, LCD) :-
    !,
    expand_wrapper(Line, Call, Wrapper, SubWrappers),
    make_query(wrap(Thy, Call, Wrapper, SubWrappers), Q),
    ( Tab \== tactic ->
        exp_data_error(unrecoverable,
                       declaration,
                       [w(Call), ': only tactics can have wrappers!'])
    ; make_query(wrap(_, Call, _, _), Q1),
      member(Q1, Qs) ->
        exp_data_error(unrecoverable,
                       declaration,
                       [w(Call), ': multiple wrappers!'])
    ;
        true
    ).

%%% implementation: add to the Impl for the last command.
expand_decls_line(implementation, Impl, Qs, Qs, _Vs, Tab, Thy,
                  H, H, LC, LC, LCD, LCD) :-
    make_query(command(Thy, Tab, LC, _, RC, _Doc), Q),
    ( member(Q, Qs) ->
        open_append_commas(RC, (Impl,_))
    ;
        exp_data_error(unrecoverable,
                       declaration,
                       ['Implementation declaration ', w(Impl), nl,
                        tab, 'has no corresponding command declaration.'])
    ).


%% private expand_wrapper(+Line, +Cmd, -Wrapper, -SubWrappers).
%%
expand_wrapper('--->'(Wrapper,SubWrappers), _, Wrapper, SubWrappers) :-
    list(Wrapper),  %% simple version of check_is_legal_writelist(Wrapper)
    list(SubWrappers),  %% may be open or closed.
    !.
expand_wrapper(Wrapper, _, Wrapper, []) :-
    %% The tactic has no subproofs.
    list(Wrapper),  %% simple version of check_is_legal_writelist(Wrapper),
    !.
expand_wrapper(Wrapper, Call, _, _) :-
    exp_data_error(unrecoverable,
                   declaration,
                   [w(Call), ': bad wrapper!']).


%% private expand_complete_queries(Queries0,Queries, Thy, DocName, AllQueries).
%%
%%  Complete a list of queries: bind Thy and DocName, close each modes list,
%%  and bind Impl if not already provided.  Note that AllQueries must
%%  be the complete Queries0 list.
%%
%%  mode expand_complete_queries(+list(compound), ?list(compound),
%%                               @theory, @atom, +list(compound)).
%%
expand_complete_queries([], [], _, _, _).
expand_complete_queries([Q|_], _, _, _, _) :-
    exp_debug(['About to complete query: ', w(Q)]),
    fail.
expand_complete_queries([Q|Qs0], [Q2|Qs], Thy, DocName, AllQs) :-
    make_query(command(Thy, Tab, Call, Modes, Impl0, DocName), Q),
    !,
    expand_complete_queries_modes(Modes, Call),
    expand_complete_queries_impl(Call, AllQs, Impl0, Impl),
    make_query(command(Thy, Tab, Call, Modes, Impl, DocName), Q2),
    exp_debug(['Completed command query: ',
               'Call: ', q(Call),nl,
               'Modes: ', q(Modes),nl,
               'Impl: ', q(Impl),nl,
               'DocName: ', q(DocName)]),
    expand_complete_queries(Qs0, Qs, Thy, DocName, AllQs).
expand_complete_queries([Q|Qs0], [Q|Qs], Thy, DocName, AllQs) :-
    make_query(ext(Thy, ExtNode, XRefTag, DocName), Q),
    !,
    exp_debug(['Completed query: ', q(extqy(Thy,ExtNode,XRefTag,DocName))]),
    expand_complete_queries(Qs0, Qs, Thy, DocName, AllQs).
expand_complete_queries([Q|Qs0], [Q|Qs], Thy, DocName, AllQs) :-
    expand_complete_queries(Qs0, Qs, Thy, DocName, AllQs).

%% private expand_complete_queries_modes(Modes, Call).
%%
%%  Complete the modes by closing the list and issuing a warning if no modes
%%  were given.
%%
%%  mode expand_complete_queries_modes(?mode, @cmd).
%% 
expand_complete_queries_modes(Modes, Call) :-
    open_to_closed(Modes),
    ( Modes = [] ->
        exp_data_error(warning, declaration, ['no modes for ', w(Call)])
    ;
        true
    ).

%% private expand_complete_queries_impl(Call, Queries, Impl0, Impl).
%%
%%  Complete the implementation: add the subproofs (if any).
%%
%%  mode expand_complete_queries_impl(@cmd, @list(compound), ?code, -code).
%% 
expand_complete_queries_impl(Call, Queries, Impl0, Impl) :-
    make_query(wrap(_, Call, _, SubMs), Q),
    member(Q, Queries),
    !,
    %% The tactic has a wrapper, so we must build the correct
    %% implementation clause so that call_tac/wrap_tac will handle
    %% any subproofs correctly.
    ( SubMs == [] ->
        Impl = Call
    ;
        Impl = '--->'(Call,_)
    ),
    ( var(Impl0) ->
        true
    ;
        ergo_data_error(warning, var(Impl0),
                        ['Tactic ',q(Call),' implementation was ignored.'])
    ).
expand_complete_queries_impl(_, _, Impl0, Impl) :-
    nonvar(Impl0),
    !,
    open_append_commas(Impl0, true),
    strip_trailing_true(Impl0, Impl).
expand_complete_queries_impl(Call, _, _, Call).


strip_trailing_true((A,true), A) :- !.
strip_trailing_true((A,B0), (A,B)) :- !, strip_trailing_true(B0,B).
strip_trailing_true(A, A).


%% private expand_create_doc_query(Doc, Thy, Name, Table, LastCmd,
%%                                  Queries0, Queries).
%%
%%  Build the doc node and add a query to install it.
%%  Table and LastCmd are used to create the Name.
%%
%%  mode expand_create_doc_query(+writelist, @theory,
%%                                @atom, @cmdtab, +compound,
%%                                +list(compound), -list(compound)).
%%
expand_create_doc_query([], _, _, _, _, Qs, Qs) :- !.
expand_create_doc_query(Doc, Thy, Name, Tab, LC, Qs, [Q|Qs]) :-
    make_query(doc(Thy, Name, List, Cmds), Q),
    functor(LC, F, N),
    make_doc_name(Tab, F, N, Name),
    reverse(Doc, List),
    expand_decls_get_commands(Qs, Cmds).

%% private make_query(Struct, Query).
%%
%%  Query is a query that will create a fact corresponding to Struct.
%%  Eventually, this procedure should generate *clauses* (of multifile
%%  procedures) rather than queries.
%%  
%%  mode make_query(@compund, -compound).
%%
make_query(command(Thy, Tab, Call, Modes, Impl, Doc),
   (?- add_command(Thy, Tab, Call, Modes, Impl, Doc))).
make_query(doc(Thy, Name, Doc, Cmds),
   (?- add_doc(Thy, Name, Doc, Cmds))).
make_query(ext(ThisThy, ExtNode, XRefTag, ThisNode),
   (?- add_doc_extension(ThisThy, ExtNode,
                         [xref(XRefTag, ThisThy, ThisNode)]))).
make_query(frag(Thy, Head, Doc),
   (?- add_doc_fragment(Thy, Head, Doc))).
make_query(wrap(Thy, Call, Msg, Subs),
   (?- add_tactic_wrap_info(Thy, Call, Msg, Subs))).

%% private make_doc_name(Table, Functor, Arity, Name).
%%
%%  Name is the name of the doc node for Functor/Arity in Table.
%%
%%  mode make_doc_name(@cmdtab, @atom, @integer, ?atom).
%% 
make_doc_name(Tab, F, N, Name) :-
    concat_atom([Tab, :, F, /, N], Name).

%% private expand_decls_get_commands(Queries, Commands).
%%
%%  Create the command cross-references for the doc node.
%%
%%  mode expand_decls_get_commands(+list(compound), -list(compound)).
%%
expand_decls_get_commands([], []).
expand_decls_get_commands([Q|Qs], [cmd(Thy, Table, Fun, Arity)|Cs]) :-
    make_query(command(Thy, Table, Call, _, _, _), Q),
    !,
    functor(Call, Fun, Arity),
    expand_decls_get_commands(Qs, Cs).
expand_decls_get_commands([_Q|Qs], Cs) :-
    expand_decls_get_commands(Qs, Cs).

%% 
%% 2. MACRO EXPANSIONS.
%%
%% mac_expand_term(Term0, Terms).
%%
%%  Expand a top-level clause.
%%  Macro-related `facts' will expand to nothing (after setting the appropriate
%%  environment); ordinary clauses expand to nothing (if the current @if
%%  condition is false) or to themselves (if it is true).
%%
mac_expand_term(Term0, Terms) :-
    ( mac_expand_term_1(Term0, Terms) ->
        true
    ; Terms = [Term0]
    ).

%% define_macro(Mac).
%% define_macro(Mac, Def).
%%
%%  Define macro Mac to expand to Def (default itself).
%%  Issue a warning if there is already a defined macro that unifies with Mac.
%%
%%  mode define_macro(@nonvar).
%%  mode define_macro(@nonvar, @term).
%%
define_macro(Mac) :-
    define_macro(Mac, Mac).

define_macro(Mac, Def) :-
    ( is_defined_macro(Mac) ->
        exp_data_error(warning,
                       define_macro(Mac, Def),
                       ['Macro ', Mac, ' is already defined.'])
    ; assert(current_macro(Mac, Def))
    ).

%% undefine_macro(Mac).
%%
%%  Remove Mac's definition, if it has one.  Note that this will also remove a
%%  definition of a more general or more specific macro.
%%
%%  mode undefine_macro(@nonvar).
%%
undefine_macro(Mac) :-
    retractall(current_macro(Mac)).
            
%% is_defined_macro(Mac).
%% is_defined_macro(Mac, Def).
%%
%%  Is Mac currently defined (with definition Def)?
%%
%%  mode is_defined_macro(?nonvar).
%%  mode is_defined_macro(?nonvar, ?term).
%%
is_defined_macro(Mac) :-
    clause(current_macro(Mac, _), _).
is_defined_macro(Mac, Def) :-
    clause(current_macro(Mac, Def), _).

%% mac_include_file(File, Clauses).
%%
%%  Clauses are the clauses (and facts and queries) in File.
%%
%%  mode mac_include_file(@filename, -list(nonvar)).
%%
mac_include_file(File, Clauses) :-
    open(File, read, Stream),
    mac_include_file(Stream, [], Clauses),
    close(Stream).

mac_include_file(Stream, Clauses0, Clauses) :-
    read_term(Stream, Term, [variable_names(Vars)]),
    expand_term(Term, Terms, Vars),
    ( Terms == end_of_file ->
        Clauses = Clauses0
    ; list(Terms) ->
        append(Terms, Clauses0, Clauses1),
        mac_include_file(Stream, Clauses1, Clauses)
    ;
        Clauses1 = [Terms | Clauses0],
        mac_include_file(Stream, Clauses1, Clauses)
    ).
    
%% mac_expand_subterm(Subterm, Exp).
%%
%% Macroexpand subterms -- @if/then/else/endif and @macro.
%%
mac_expand_subterm(Sub, Exp) :-
    ( mac_expand_subterm_1(Sub, Exp) ->
        true
    ; Exp = Sub
    ).

?- op(990, yf,  endif).
?- op(989, xfx, else).
?- op(988, xfx, then).
?- op(200, fx,  @).

?- dynamic(current_macro/2).
?- dynamic(ifdef_flags/1).


%DBGmac_expand_term_1(Term, _) :-
%DBG    write_term_list(['trying macro expansion on ', w(Term), nl]),
%DBG    fail.
mac_expand_term_1(H, _) :-
    compound(H),
    functor(H, V, _),
    var(V),
    !,
    fail.
mac_expand_term_1(@include(File), Clauses) :-
    mac_include_file(File, Clauses).
mac_expand_term_1(@define(Mac), []) :-
    define_macro(Mac).
mac_expand_term_1(@define(Mac, Def), []) :-
    define_macro(Mac, Def).
mac_expand_term_1(@undef(Mac), []) :-
    undefine_macro(Mac).
mac_expand_term_1(@if(Goal), []) :-
    ( call(Goal) ->
        Def = yes
    ; Def = no
    ),
    asserta(ifdef_flags(Def)).
mac_expand_term_1(@ifdef(Mac), []) :-
    ( is_defined_macro(Mac) ->
        Def = yes
    ; Def = no
    ),
    asserta(ifdef_flags(Def)).
mac_expand_term_1(@ifndef(Mac), []) :-
    ( is_defined_macro(Mac) ->
        Def = no
    ; Def = yes
    ),
    asserta(ifdef_flags(Def)).
mac_expand_term_1(@else, []) :-
    ( retract(ifdef_flags(Def)) ->
        true
    ; exp_data_error(warning,
                     @else,
                     ['@else has no matching @ifdef'])
    ),
    ( Def = yes ->
        NDef = no
    ; NDef = yes
    ),
    asserta(ifdef_flags(NDef)).
mac_expand_term_1(@endif, []) :-
    ( retract(ifdef_flags(_)) ->
        true
    ; exp_data_error(warning,
                     @endif,
                     ['@endif has no matching @ifdef'])
    ).
mac_expand_term_1(end_of_file, end_of_file) :-
    clause(ifdef_flags(_), _),
    !,
    retractall(ifdef_flags(_)),
    exp_data_error(warning,
                   end_of_file,
                   ['@ifdef has no matching @endif']).
mac_expand_term_1(_Clause, []) :-
    clause(ifdef_flags(Def), _),
    !,
    Def = no.

mac_expand_subterm_1(H, _) :-
    compound(H),
    functor(H, V, _),
    var(V),
    !,
    fail.
mac_expand_subterm_1(@macro(Mac), Def) :-
    is_defined_macro(Mac, Def).
mac_expand_subterm_1(endif(else(then(@if(G), T1), T2)), T) :-
    ( call(G) ->
        mac_expand_subterm(T1, T)
    ; mac_expand_subterm(T2, T)
    ).
mac_expand_subterm_1(endif(else(then(@ifdef(Mac), T1), T2)), T) :-
    ( is_defined_macro(Mac) ->
        mac_expand_subterm(T1, T)
    ; mac_expand_subterm(T2, T)
    ).


%% 
%% 3. THEORY EXPANSIONS.
%% expand_thy_term(Thy, Vars, ThyTerm, ExpTerm, CVars).
%%
%%  Expand the theory term @thy(...) or @current_thy.
%%  Accumulate variables in the open list Vars, and put the result in Term.
%%  CVars is the (closed) list of variables coming from the context, so we
%%  initialize Vars from it on the first call.
%%
%%  mode expand_thy_term(@atom, ?openlist(compound), @nonvar, -term,
%%                       +list(compound)).
%%
expand_thy_term(_, _, H, _, _) :-
    compound(H),
    functor(H, V, _),
    var(V),
    !,
    fail.
expand_thy_term(Thy, _, @current_thy, Thy, _).
expand_thy_term(Thy, Vars, @thy(Chars), Term, CVars) :-
    %% NB. we don't expand @thy(Atom) or @thy(Compound), because
    %%     they are mode declarations.
    list(Chars),
    expand_thy_chars(Chars, Thy, [], Vars, Term, CVars).
expand_thy_term(Thy, Vars, @thy(XThy, Chars), Term, CVars) :-
    ( XThy = @current_thy ->
        YThy = Thy
    ;
        YThy = XThy
    ),
    check_or_error(Thy=XThy,
                   ['Sorry, @thy terms may only refer to current theory.']),
    expand_thy_chars(Chars, Thy, [], Vars, Term, CVars).
expand_thy_term(Thy, Vars, @thy(XThy, Chars, Arity), Term, CVars) :-
    check_or_error(Thy=XThy,
                   ['Sorry, @thy terms may only refer to current theory.']),
    expand_thy_chars(Chars, Thy, Arity, Vars, Term, CVars).

%% expand_thy_chars(Chars, Thy, Arity, Vars, Term, CVars).
%%
%%  Does the actual expansion.
%%
%%  mode expand_thy_chars(@list(char), @theory, @arity, ?openlist(compound),
%%                        -term, +list(compound)).
%%
expand_thy_chars(Chars, _, _, _, _, CVs) :-
    exp_debug(['trying theory expansion on ', w(@thy(Term)), nl,
                       'vars are: ', w(CVs)]),
    fail.
expand_thy_chars(Chars, Thy, Arity, Vars, Term, CVars) :-
    ( is_chars(Chars) ->
        true
    ;
        exp_data_error(unrecoverable, @thy(Chars),
                       ['argument of @thy must be a chars.'])
    ),
    ( var(Vars) ->
        closed_to_open(CVars, Vars)
    ;
        true
    ),
    append(Chars, " . ", Ch),
    open_string(read(Ch), Strm),
    unwind_protect(
        (
            exp_debug(['about to read from string ', chars(Ch),
                       ' in theory ', w(Thy)]),
            read_term(Strm, Term0, [variable_names(TVars), op_table(Thy)])
        ),
        close(Strm)
                  ),
    % currently works only for current theory:
    deabbreviate(Term0, Thy, Arity, Term),
    exp_merge_varslists(TVars, Vars),
    !,
    exp_debug(['theory expansion succeeded: got ', q(Term)]).
expand_thy_chars(Chars, Thy, Arity, _, _, _) :-
    exp_data_error(warning, @thy(Chars),
                   [@thy(Thy, Chars, Arity), ': theory expansion failed.']),
    fail.

%% 
%% 4. THEORY PREFIXING.
%%

?- dynamic(tac_prefix_procs/2).

%% tac_add_prefixes(Term0, Term).
%%
%%  Add a module prefix to all instances of procedures defined in the file.
%%
%%  mode tac_add_prefixes(@nonvar, -nonvar).
%%
tac_add_prefixes(Term0, Term) :-
    tac_add_prefixes(Term0, Term, Flag),
    !,
    nonvar(Flag).

tac_add_prefixes(Term, Term, _) :-
    var(Term),
    !.
tac_add_prefixes(Higher, Higher, _) :-
    functor(Higher, Var, _),
    var(Var),
    !.
tac_add_prefixes(--->(Tac0,CodeList0), --->(Tac,CodeList), Flag) :-
    !,
    functor(Tac0, Fun, Ar0),
    nonvar(Fun),
    Ar1 is Ar0 + 1,
    functor(RTac, Fun, Ar1),
    ( tac_prefix_procs(RTac, PTac) ->
        Flag = true,
        functor(PTac, PFun, Ar1),
        functor(Tac, PFun, Ar0),
        same_args(Tac0, Tac, 1, Ar0)
    ;
        Tac = Tac0
    ),
    tac_add_prefixes_codelist(CodeList0, CodeList, Flag).
tac_add_prefixes(Term0, Term, Flag) :-
    tac_prefix_procs(Term0, Term1),
    !,
    Flag = true,
    exp_debug(['prefixed ', q(Term0), ' giving ', q(Term1), nl]),
    tac_add_prefixes(Term1, Term, Flag).
tac_add_prefixes(Term0, Term, Flag) :-
    meta_predicate(Term0, GoalArgs),
    !,
    functor(Term0, Fun, Ar),
    functor(Term,  Fun, Ar),
    tac_add_prefixes(1, Ar, GoalArgs, Term0, Term, Flag).
tac_add_prefixes(Term, Term, _).

tac_add_prefixes(N, Ar, _, _, _, _) :-
    N > Ar,
    !.
tac_add_prefixes(N, Ar, GA, T0, T, F) :-
    % N =< Ar,
    arg(N, T0, A0),
    arg(N, T, A),
    ( GA = [N|RGA] ->
        tac_add_prefixes(A0, A, F)
    ;
        A0 = A,
        GA = RGA
    ),
    N1 is N + 1,
    tac_add_prefixes(N1, Ar, RGA, T0, T, F).

tac_add_prefixes_codelist(V, V, _) :-
    var(V),
    !.
tac_add_prefixes_codelist([], [], _).
tac_add_prefixes_codelist([Code0 | Codes0], [Code | Codes], Flag) :-
    tac_add_prefixes(Code0, Code, Flag),
    tac_add_prefixes(Codes0, Codes, Flag).

%% QP built-ins
meta_predicate((?- _),                   [1]).
meta_predicate((:- _),                   [1]).
meta_predicate((_ :- _),                 [1,2]).
meta_predicate((_ , _),                  [1,2]).
meta_predicate((_ ; _),                  [1,2]).
meta_predicate((_ -> _),                 [1,2]).
meta_predicate((_ ^ _),                  [2]).
meta_predicate((\+ _),                   [1]).
meta_predicate(bagof(_,_,_),             [2]).
meta_predicate(call(_),                  [1]).
meta_predicate(catch(_,_,_),             [1,3]).
meta_predicate(findall(_,_,_),           [2]).
meta_predicate(once(_),                  [1]).
meta_predicate(setof(_,_,_),             [2]).
meta_predicate(unwind_protect(_,_),      [1,2]).
meta_predicate(delay(_,_),               [2]).

%% Ergo declarations
meta_predicate(add_command(_,_,_,_,_,_), [5]).
meta_predicate(check_or_error(_,_),      [1]).    
meta_predicate(add_listener(_,_),        [2]).
meta_predicate(add_handler(_,_),         [2]).
meta_predicate(less_verbose(_,_),        [2]).
meta_predicate(save_hyps(_),             [1]).
meta_predicate(open_pos(_,_),            [2]).
meta_predicate(open_pos(_,_,_),          [3]).
meta_predicate(open_hyp(_,_),            [2]).
meta_predicate(open_hyp(_,_,_),          [2]).  %% TODO: make tactic arg last?

%% Other
% meta_predicate(Goal, GoalArgs) :-
%    look at mode decls.

%% private build_tac_prefix_procs(Module, File).
%% private build_tac_prefix_proc(Term, Module).
%%
%%  Record the need to prefix the procedures in File with Module.
%%
%%  mode build_tac_prefix_procs(@atom, @atom).
%%  mode build_tac_prefix_proc(@term, @atom).
%%
build_tac_prefix_procs(Mod, File) :-
    open(File, read, Strm),
    repeat,
      %% do_consult will have set up op and obvar tables
      read(Strm, Term),  
      build_tac_prefix_proc(Term, Mod),
      Term == end_of_file,
    !,
    close(Strm).

build_tac_prefix_proc([], _) :-
    !.
build_tac_prefix_proc([T|Ts], Mod) :-
    !,
    build_tac_prefix_proc(T, Mod),
    build_tac_prefix_proc(Ts, Mod).
build_tac_prefix_proc(end_of_file, _) :-
    !.
build_tac_prefix_proc((?- _Query), _) :-
    !.
build_tac_prefix_proc((:- _Decl), _) :-
    !.
build_tac_prefix_proc((@ _Decl), _) :-
    !.
build_tac_prefix_proc((command(_,_)), _) :-
    !.
build_tac_prefix_proc((doc(_,_)), _) :-
    !.
build_tac_prefix_proc((Head :- _Body), Mod) :-
    !,
    build_tac_prefix_proc(Head, Mod).
build_tac_prefix_proc(Head, Mod) :-
    (
        atom(Head)
    ;
        compound(Head),
        functor(Head, F, _),
        atom(F)
    ),
    !,
    ( tac_prefix_procs(Head, _) ->
        true
    ;
        functor(Head, Fun, Ar),
        functor(UnPref, Fun, Ar),
        concat_atom([Mod, ':', Fun], PreFun),
        functor(Pref, PreFun, Ar),
        same_args(UnPref, Pref, 1, Ar),
        assert(tac_prefix_procs(UnPref, Pref))
    ).

%% 
%% UTILITIES
%% private exp_merge_varslists(Vars1, Vars2).
%%
%%  Vars1 is a closed list of Var=Name pairs; Vars2 is an open list of such
%%  pairs.  For every Name in Vars1, if that name also occurs in Vars2, unify
%%  the corresponding Var; otherwise, add the pair to Vars2.
%%
%%  mode exp_merge_varslists(+list(compound), ?openlist(compound)).
%%
exp_merge_varslists([], _).
exp_merge_varslists([V=N | Vs1], Vs2) :-
    ( open_member(V2=N, Vs2) ->
        V = V2
    ;
        open_append(Vs2, [V=N])
    ),
    exp_merge_varslists(Vs1, Vs2).

%% private exp_bind_vars_names(VarsAndNames).
%%
%%  Takes a list of Variable=Name pairs and binds the Variables to their Names.
%%
%%  mode exp_bind_vars_names(+list(compound)).
%%
exp_bind_vars_names([]).
exp_bind_vars_names([N=N|Ns]) :-
    exp_bind_vars_names(Ns).

%% private open_append_commas(Commas, Tail).
%%
%%  Binds the tail of Commas (a comma-separated sequence with an uninstantiated
%%  tail) to Tail.
%%
%%  mode open_append_commas(+term, +term).
%%
open_append_commas(Var, Tail) :-
    var(Var),
    !,
    Var = Tail.
open_append_commas((_,Rest), Tail) :-
    open_append_commas(Rest, Tail).

%% 
%% TEMPORARY DEFINITIONS:
%% Ergo should catch these exceptions and use its handle_error mechanism.
exp_program_error(Sev, Goal, Msg) :-
    exception(ergo_program(Sev, Goal, Msg)).

exp_data_error(Sev, Goal, Msg) :-
    append(Msg, [nl], Msg1),
    exception(ergo_data(Sev, Goal, Msg1)).

%% We need this here in case we run outside Ergo.
exp_debug(WriteList) :-
    ( clause(dbg(expansions), _) ->
        debug(expansions, WriteList)
    ;
        true
    ).
