%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: interpret.ql,v 1.40 2003/12/19 01:32:49 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Support for theory INTERPRETATION.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../lib/ergo.h"

%% TO DO:
%% 3. check 1-1 property for defined symbols (all symbols for instantiate)
%%     (and add symbol facts for all such names)
%% 4. special kind of deabbreviation for RHS of 1-1 maplets
%%     (must map between subtheories in the same way as Subint)


database command
  interpret(Abs,Sub,Inst,Repar,Subint)
extends
  theory_construction_cmds
mode
  interpret(@theory,@theory,@list(compound),@list(compound),@list(compound))
implementation
  global_state_lookup(current_dir,CD),
  %% ensure that the Abs theory is loaded (before deabbreviating!)
  cd(Abs),
  cd(CD),
  deabbreviate_interpretation(Inst, Abs, Inst1),
  catch(interpret(Abs,Sub,Inst1,Repar,Subint), interp_error, delete_subtheory(Sub, CD))
doc
  'Interpret abstract theory Abs into the current theory.',nl,
  'The interpreted rules go into a new subtheory called Sub.',nl,
  'Inst is a list of AbsTerm--->ConTerm instantiations.',nl,
  'Repar is a list of Theory--->SubTheory reparentings --',nl,
  'each Theory must have already been interpreted into SubTheory.',nl,
  'Subint is a list of AbsSub--->ConSub subinterpretations, describing how',nl,
  'subtheories of Abs map to new subtheories of the current theory.'.

database command
  instantiate(Abs,Sub,Inst,Repar,Subint)
extends
  theory_construction_cmds
mode
  instantiate(@theory,@theory,@list(compound),@list(compound),@list(compound))
implementation
  global_state_lookup(current_dir,CD),
  %% ensure that the Abs theory is loaded  (before deabbreviating!)
  cd(Abs),
  cd(CD),
  deabbreviate_instantiation(Inst, Abs, Inst1),
  catch(instantiate(Abs,Sub,Inst1,Repar,Subint), interp_error, delete_subtheory(Sub, CD))
doc
  'Instantiate abstract theory Abs into the current theory.',nl,
  'The instantiated rules go into a new subtheory called Sub.',nl,
  'Inst is a list of AbsTerm--->ConTerm instantiations.',nl,
  'Repar is a list of Theory--->SubTheory reparentings --',nl,
  'each Theory must have already been interpreted into SubTheory.',nl,
  'Subint is a list of AbsSub--->ConSub subinterpretations, describing how',nl,
  'subtheories of Abs map to new subtheories of the current theory.'.

%% interpret(AbsThy, SubThy, Instantiations, Reparentings, Subinterpretations).
%%
%%  AbsThy is the name of the abstract theory to be interpreted.
%%  The interpretation goes into the current theory, in SubThy which is used
%%   to identify the interpretation in future.
%%  Instantiations is a list of AbsTerm--->ConTerm maplets for instantiating
%%   the declared and defined constants and functions of the theory; the
%%   constraints on each AbsTerm and ConTerm are:
%%     1. AbsTerm must be either an atom or a compound term with all arguments
%%        distinct metavariables.
%%     2. If AbsTerm is a defined symbol of AbsThy, ConTerm must be a term with
%%        the same structure as AbsTerm, whose principal symbol is fresh (as
%%        the abstract definition will become a concrete definition).
%%     3. If AbsTerm is a declared symbol of AbsThy, ConTerm must be a legal
%%        term of the concrete (current) theory, which: 
%%        a) contains no free object variables, and
%%        b) contains no metavariables other than those in AbsTerm.
%%  If some abstract symbol is not explicitly instantiated, a default
%%    instantiation is constructed with the same name (modulo theory prefix) as
%%    the abstract.  As with explicit instantiation, the resulting concrete
%%    term must have a fresh functor for defined abstract terms, and must be
%%    already declared in the current theory for declared abstract terms.
%%  Reparentings is a list of Theory--->SubTheory maplets that explains how
%%   each ancestor of AbsThy is to be mapped to an existing interpreted SubThy
%%   of the current theory.  If no symbols in some ancestor are to be
%%   instantiated, that ancestor need not be named here, and it becomes an
%%   ancestor of the current theory (if it is not one already).
%%  Subinterpretations is a list of AbsSubThy--->ConSubThy maplets describing
%%   how each subtheory AbsSubThy of AbsThy is to be interpreted into a
%%   different subtheory of the target theory.
%%
%%  mode interpret(+atom, +atom, @list(term), @list(ground), @list(ground)).
%%
interpret(Abs, Sub, Inst0, Repar, Subint) :-
    %% ensure that the Abs theory is loaded.
    global_state_lookup(current_dir,CD),
    cd(Abs),
    cd(CD),
    debug(interpretation, ['Checking supplied instantiations']),
    check_interpretation_map(Inst0, Abs),
    debug(interpretation, ['Processing reparentings']),
    process_reparentings(Repar, CD, Inst0, Inst),
    check_inherited_ancestors(Abs, CD, Repar),
    (
      debug(interpretation, ['Recording supplied interpretation']),
      record_interpretation(Abs, Sub, Subint, Inst),
      debug(interpretation, ['Recording default instantiations']),
      record_default_instantiations(Abs, Sub, Subint),
      interpret_theory(Abs, Sub, Inst, interpret),
      interpret_subtheories(Subint, Sub, interpret),
      cd(Sub),
      include_theory(CD, all),
      cd(CD),
      close_theory(Sub),
      !
    ;
      throw(interp_error)
    ).

instantiate(Abs, Sub, Inst0, Repar, Subint) :-
    global_state_lookup(current_dir,CD),
    %% note: current theory must be open for instantiate but not for interpret
    check_open(CD),
    %% ensure that the Abs theory is loaded.
    cd(Abs),
    cd(CD),
    debug(interpretation, ['Checking supplied instantiations']),
    check_instantiation_map(Inst0, Abs),
    debug(interpretation, ['Processing reparentings']),
    process_reparentings(Repar, CD, Inst0, Inst),
    check_inherited_ancestors(Abs, CD, Repar),
    (
      debug(interpretation, ['Recording supplied interpretation']),
      record_interpretation(Abs, Sub, Subint, Inst),
      debug(interpretation, ['Recording default instantiations']),
      record_default_instantiations(Abs, Sub, Subint),
      interpret_theory(Abs, Sub, Inst, instantiate),
      interpret_subtheories(Subint, Sub, instantiate),
      cd(Sub),
      include_theory(CD, all),
      cd(CD),
      close_theory(Sub),
      !
    ;
      throw(interp_error)
    ).

%% interpret_subtheories(Subint, Sub, Kind).
%%
%%  Interpret abstract subtheories into subtheories of the current theory
%%  according to Subint and the interpretation recorded in Sub.
%%  Kind is `interpret' (translate axioms to postulates) or `instantiate'
%%  (leave axioms alone).
%%
%%  mode interpret_subtheories(@list(ground), @theory, @atom).
%%
interpret_subtheories([], _, _).
interpret_subtheories(['--->'(ASub,CSub) | SubInt], Int, Kind) :-
    interpret_theory(ASub, CSub, Int, Kind),
    interpret_subtheories(SubInt, Int, Kind).

%% interpret_theory(Abs, Sub, Int, Kind).
%%
%%  Interpret abstract theory Abs into concrete subtheory Sub, using Int's
%%  recorded interpretation.
%%  Kind is `interpret' (translate axioms to postulates) or `instantiate'
%%  (leave axioms alone).
%%
%%  mode interpret_theory(@theory, @theory, @theory, @atom).
%% 
interpret_theory(Abs, Sub, Int, Kind) :-
    ( Kind = interpret ->
        msg(interpret_theory, ['Interpreting ', Abs, ' into ', Sub])
    ; Kind = instantiate ->
        msg(interpret_theory, ['Instantiating ', Abs, ' into ', Sub])
    ;
        handle_error(['interpret_theory: unrecognized kind.'])
    ),
    (
        theory_fact(Name, Arity),
        Name \= infrule_apply,
        functor(Fact, Name, Arity),
        query_theory(Fact, Abs),
        debug(interpretation, ['Interpreting fact ', w(Fact)]),
	ergo_version(ERGO_VER),
        interpret_fact(ERGO_VER, Name, Fact, Abs, Sub, Int, Kind),
        fail
    ;
        close_theory(Sub)
    ),
    ( ergo_version(ergo4)
      ->
         add_rulegroups(Abs,Sub)
      ;
         true
    ).
    

%% interpret_fact(ErgoVersion, FactName, Fact, AbsThy, SubThy, Int, Kind).
%%
%%  Try to interpret a theory fact Fact (with principal functor FactName) into
%%  SubThy using the interpretation recorded in Int.
%%
%%  mode interpret_fact(+atom, +atom, @term, +theory, +theory, +theory, +atom).
%%
% Facts to look for (Ergo4):
% abbrev/4, compiled_rule/8, default_prover/2, equivalence/1, interpretation/2,
% interpreted_theory/2, object_vars/1, openrule/8, preorder/1,
% reverse_of/2, rulegroup/7, save_clauses/3, stronger_than/2,
% symbol/3, tactic_file/1
% Currently only interprets rules.
%
% Facts to look for (Ergo5):
% default_prover/2, interpretation/2,
% interpreted_theory/2, object_vars/1, 
% Currently only interprets rules.
%

interpret_fact(ergo4, rule, rule(Name, Type, _, Stmt, Body), Abs, Sub, Int, Kind) :-
    !,
    interpret_rule(Type, Name, [], Stmt, Body, Abs, Sub, Int, Kind).
interpret_fact(ergo4, tactic_file, tactic_file(File), Abs, Sub, Int, _) :-
    !,
    interpret_tactic_file(File, Abs, Sub, Int),
    add_tactic_file(Sub, File).
interpret_fact(ergo, infrule, INFRULE, _Abs, Sub, _Int, Kind) :-
    !,
    interpret_infrule(INFRULE, Sub, Kind).
interpret_fact(ergo, rule, rule(Name, _Type, _, 'ergo.='(D,T), Body), _, Sub, Int, Kind) :-
    !,
    ( member('--->'(D,_), Int)
      ->
          ( ancestor(weak_eq) ->
              true
          ;
              (
                 handle_error(['Theory must have weak_eq as an ancestor in order to instantiate/interpret defined symbol.'])
               ; 
                 throw(interp_error)
               )
          ),
          concat_atom([Kind, Name], '_', RuleName),
          ( Kind = instantiate
            ->
             assert_theory(infrule_status(RuleName, fundamental), Sub)
           ;
             true
          ),
          Concl = 'weak_eq.='(D, T),
          interpret_infrule(infrule(RuleName, RuleName, Concl, _ConclID, [], Body, true),
                            Sub, Int, Kind)
      ;
          interpret_terms([D,T], Sub, [ID,IT]),
          interpret_nfis(Body, Sub, BodyI),
          assert_axthm(Name, defn, [], 'ergo.='(ID,IT), BodyI, Sub)
    ).
interpret_fact(ergo, infrule_apply, _INFRULE, _Abs, _Sub, _Int, _Kind) :- !.
interpret_fact(ergo, infrule_status, infrule_status(Name, Status), _, Sub, _, Kind) :-
    !,
    ( Status = fundamental
      ->
         ( Kind = instantiate
           ->
	     assert_theory(infrule_status(Name, fundamental), Sub)
           ;
             true
         )
      ;
         assert_theory(infrule_status(Name, Status), Sub)
    ).
interpret_fact(_, Name, Fact, _, _, _, _) :-
    functor(Fact,Name,Arity),
    ignorable_fact(Name,Arity),
    !.
interpret_fact(_, Name, _, _, _, _, _) :-
    handle_error(['Failed to interpret ', q(Name), '.']).
interpret_fact(_, _, _, _, _, _, _) :-
    throw(interp_error).

%% private ignorable_fact(Name,Arity).
%%
%%  True of all facts that can safely be ignored during interpretation.
%%
%%  mode ignorable_fact(?atom, ?integer).
%%
ignorable_fact(interpretation,2).
ignorable_fact(interpreted_theory,2).
ignorable_fact(object_vars,1).
ignorable_fact(status,1).
ignorable_fact(symbol,3).
ignorable_fact(abbrev, 4).
ignorable_fact(tactic_file, 1).
ignorable_fact(Name,Arity) :-
    theory_tactic_fact(Name,Arity).



%% interpret_rule(Type, Name, AssumpList, Statement, Body,
%%                AbsThy, SubThy, Int, Kind).
%%
%%  Interpret a rule (axiom, postulate or theorem) into SubThy using the
%%  instantiations recorded in Int.
%%                  interpret     instantiate
%%   Axiom     ---> postulate     axiom
%%   Theorem   ---> theorem       theorem
%%   Postulate ---> postulate     postulate
%%   Defn      ---> defn          defn
%%
%%  mode interpret_rule(@term, +atom, @term, @term, @term,
%%                      +theory, +theory, +theory, +atom).
%%
interpret_rule(Type, Name, Ass, Stmt, Body, _Abs, Sub, Int, Kind) :-
    interpret_terms([Stmt|Ass], Int, [StmtI|AssI]),
    % check it's a legal term of the current theory (may be redundant as it is
    % done in check_interpretation??)
    check_term(StmtI),
    check_terms(AssI),
    ( Type = axiom, Kind = interpret ->
        %% interpreted axiom: interpret the body and add a postulate
        interpret_nfis(Body, Int, BodyI),
        assert_axthm(Name, postulate, AssI, StmtI, BodyI, Sub)
    ; Type = axiom, Kind = instantiate ->
        %% instantiated axiom: leave as axiom
        interpret_nfis(Body, Int, BodyI),
        assert_axthm(Name, axiom, AssI, StmtI, BodyI, Sub)
    ; Type = axiom(_) ->
        %% param_axiom: create a param_axiom (should really be a
        %% `param_postulate', but there is no way to discharge it).
        interpret_tactic_body(Body, Int, BodyI),
        assert_axthm(Name, Type, AssI, StmtI, BodyI, Sub)
    ; Type = theorem(Deps) ->
        %% theorem: add a theorem.
        interpret_nfis(Body, Int, BodyI),
        translate_dependencies(Deps, Int, TDeps),
        assert_axthm(Name, theorem(TDeps), AssI, StmtI, BodyI, Sub)
    ; Type = postulate ->
        interpret_nfis(Body, Int, BodyI),
        assert_axthm(Name, postulate, AssI, StmtI, BodyI, Sub)
    ; Type = defn ->
        %% definition: add a definition.
        interpret_nfis(Body, Int, BodyI),
        assert_axthm(Name, defn, AssI, StmtI, BodyI, Sub)
    ;
        handle_error(['Unknown rule type ''', w(Type), '''.', nl])
    ),
    !.
interpret_rule(_, Name, _, _, _, _, _, _, Kind) :-
    handle_error(['Failed to ', w(Kind), ' rule ', w(Name)]).

interpret_infrule(infrule(NameF, Name, Concl, ConclID, Seqs, Before, After), 
                  Sub, _Kind) :-
    After = true,   % No oracles !!
    interpret_terms([Concl|Seqs], Sub, [ConclI|SeqsI]),
    interpret_constraints(Before, Sub, BeforeI),
    Info = info(_Thy:NameF,_,_,ConclID,NodeIds),
    rule_check_theory_term(ConclI, Info),
    rule_check_seqs(SeqsI, Info, Seqs1, SeqIds),
    NodeIds = [ConclID|SeqIds],
    rule_compile(ConclID, (rule_unify(ConclI, BConcl), BeforeI),
                 After, Seqs1, SeqIds, Code),
    assert_theory(infrule(NameF,Name,ConclI,ConclID,SeqsI,BeforeI,After), Sub),
    mode_from_rule(Name, RName, _),
    assert_theory(infrule_apply(NameF,RName,BConcl,ConclID,SeqIds), Code, Sub),
    !.
interpret_infrule(InfRule, _Sub, Kind) :-
    handle_error(['Failed to ', w(Kind), ' inference rule ', q(InfRule)]).
interpret_infrule(_, _, _) :-
    throw(interp_error).


%% interpret_constraints(Constraints, Int, IConstraints).
%%
%%  Apply Int's instantiation function recursively to the terms inside
%%  context_search constraints.
%%
%%  mode interpret_terms(@term, +list(term), -term).
%%

interpret_constraints((A,B), Int, (AI,BI)) :-
    interpret_constraints(A, Int, AI),
    interpret_constraints(B, Int, BI),
    !.
interpret_constraints(context_search(H,CID,HID,Term), Int, 
                      context_search(H,CID,HID,TermI)) :-
    interpret_term(Term, Int, TermI),
    !.
interpret_constraints(C, _, C).

%% interpret_terms(Terms0, Int, Terms).
%%
%%  Apply Int's instantiation function recursively to Terms0 giving Terms.
%%
%%  mode interpret_terms(@list(term), +list(term), -list(term)).
%%
interpret_terms([], _, []).
interpret_terms([T0|Ts0], Int, [T|Ts]) :-
    interpret_term(T0, Int, T),
    interpret_terms(Ts0, Int, Ts).


%% interpret_term(Term0, Int, Term).
%%
%%  Apply Int's instantiation function recursively to Term0 giving Term.
%%
%%  mode interpret_term(@term, +atom, -term).
%%
interpret_term(T0, Int, T) :-
    transform_subterms(interpret_subterm(Int), T0, T).


%% interpret_subterm(Int, Term0, Term).
%%
%%  Apply Int's instantiation function once to Term0 giving Term.
%%
%%  mode interpret_subterm(+atom, @term, -term).
%%
interpret_subterm(Int, T0, T) :-
    ( var(T0) ->
        T = T0
    ; ergo_version(ergo4), T0 = param_const('schema.state_var',ZVar0) ->
        %% This is a Cogito-specific hack, because it relies on knowledge
        %% about the axiomatization of the schema theory.
        %% This interpretation of parameterized constants is sound for
        %% that axiomatization, even though it is not sound in general.
        interpret_cogito_state_var(Int, ZVar0, ZVar),
        T = param_const('schema.state_var',ZVar)
    ; query_theory(interpretation(T0, T1), Int) ->
        simplify_term(T1, T)
    ;
        T = T0
    ).


%% record_interpretation(AbsThy, SubThy, SubInterpretation, Instantiation).
%%
%%  Create each target subtheory and record the details of the interpretation
%%  in SubThy.  We record the whole interpretation (including all
%%  instantiations and subinterpretations) in SubThy (which is the subtheory
%%  resulting from the main interpretation); each `younger brother' subtheory
%%  Sub2 (from SubInterpretation) contains interpreted_theory(Sub2, SubThy).
%%
%%  mode record_interpretation(+theory, +theory,
%%                             +list(compound), +list(compound)).
%%
record_interpretation(Abs, Sub, SubInt, Inst) :-
    add_subtheory(Sub),
    assert_theory(interpreted_theory(Abs, Sub), Sub),
    record_instantiations(Inst, Sub),
    (
        member('--->'(A1, S1), SubInt),
        add_subtheory(S1),
        assert_theory(interpreted_subtheory(A1, S1), Sub),
        assert_theory(interpreted_theory(A1, Sub), S1),
        fail
    ;
        true
    ).

%% record_instantiations(Inst, Sub).
%%
%%  Record instantiations Inst in subtheory Sub.
%%
%%  mode record_instantiations(@list(compound), @theory).
%%
record_instantiations([], _).
record_instantiations(['--->'(AT,CT)|Inst], Sub) :-
    record_instantiation(AT, CT, Sub),
    record_instantiations(Inst, Sub).

%% record_instantiation(AbsTerm, ConTerm, Sub).
%%
%%  Record the instantiation AbsTerm--->ConTerm in subtheory Sub.
%%
%%  mode record_instantiation(@thy(nonvar), @thy(nonvar), @theory).
%%
record_instantiation(Abs, Con, Sub) :-
    debug(interpretation, ['Recording instantiation ', w(Abs),'--->',w(Con)]),
    ( query_theory(interpretation(Abs, _), Sub) ->
        handle_error(['Error: repeated instantiation of ', thy(Abs)])
    ;
        assert_theory(interpretation(Abs, Con), Sub)
    ).

%% interpreted_symbol(Name, Arity, Sub).
%% interpreted_symbol(Name, Arity, Sub, SubName).
%%
%%  True iff Sub has an interpretation for Name/Arity.
%%  SubName is the declared name of the resulting term (if meaningful).
%%
%%  mode interpreted_symbol(@atom, @arity, @theory).
%%  mode interpreted_symbol(@atom, @arity, @theory, @atom).
%%
interpreted_symbol(Name, [], Sub) :-
    query_theory(interpretation(Name, _), Sub).
interpreted_symbol(Name, [N], Sub) :-
    integer(N),
    !,
    functor(Term, Name, N),
    query_theory(interpretation(Term, _), Sub).
interpreted_symbol(Name, [quant(BArity)], Sub) :-
    query_theory(interpretation(Term, _), Sub),
    quant(Term),
    quantify(Term, Name, Vars, _),
    bound_vars_to_arity(Vars, BArity).

interpreted_symbol(Name, [], Sub, SubName) :-
    query_theory(interpretation(Name, SubName), Sub).
interpreted_symbol(Name, [N], Sub, SubName) :-
    integer(N),
    !,
    functor(Term, Name, N),
    query_theory(interpretation(Term, SubTerm), Sub),
    functor(SubTerm, SubName, N).
interpreted_symbol(Name, [quant(BArity)], Sub, SubName) :-
    query_theory(interpretation(Term, SubTerm), Sub),
    quant(Term),
    quant(SubTerm),
    quantify(Term, Name, Vars, _),
    quantifier(SubTerm, SubName),
    bound_vars_to_arity(Vars, BArity).

%% record_default_instantiations(AbsTheory, SubTheory, SubInterpretation).
%%
%%  Augment the instantiations for SubTheory so that it interprets every symbol
%%  of each abstract theory (including abstract subtheories).
%%
%%  mode record_default_instantiations(+theory, +theory, @list(compound)).
%%
record_default_instantiations(Abs, Sub, SubInt) :-
    record_default_instantiations1(Abs, Sub, Sub),
    (
        member('--->'(A1,S1), SubInt),
        %% NB -- record insts in Sub, *not* S1
        record_default_instantiations1(A1, Sub, S1),
        fail
    ;
        true
    ).
    
record_default_instantiations1(Abs, Int, Sub) :-
    (
        query_theory(symbol(Name, Arity, SKind), Abs),
        debug(interpretation,
              ['Checking symbol ', w(Name/Arity), '(', w(SKind), ')']),
        record_default_instantiation(SKind, Name, Arity, Abs, Int, Sub),
        fail
    ;
        true
    ).

%% record instantiation in Int, declare defined symbols in Sub.
record_default_instantiation(declared, Name, Arity, Abs, Int, Sub) :-
    ( interpreted_symbol(Name, Arity, Int) ->
        debug(interpretation,
              [w(Name/Arity), ' does not need a default instantiation.'])
    ; Name = param_const(_, _) ->
        true
    ; Arity = [] ->
        debug(interpretation,
              [' recording default instantiation for ', w(Name/Arity)]),
        internal_name(Name, Abs, Name1),
        internal_name(SubName, Sub, Name1),
        declare_name(Name1, Arity, Sub, declared),
        record_instantiation(Name, SubName, Int)
    ; Arity = [N], integer(N) ->
        debug(interpretation,
              [' recording default instantiation for ', w(Name/Arity)]),
        internal_name(Name, Abs, Name1),
        internal_name(SubName, Sub, Name1),
        declare_name(Name1, Arity, Sub, declared),
        length(Args, N),
        AbsTerm =.. [Name|Args],
        SubTerm =.. [SubName|Args],
        record_instantiation(AbsTerm, SubTerm, Int)
    ;
        handle_error(['Cannot interpret declared symbol ', w(Name/Arity), '.'])
    ).
record_default_instantiation(defined, Name, Arity, Abs, Int, Sub) :-
    ( interpreted_symbol(Name, Arity, Int, SubName) ->
        debug(interpretation,
              [w(Name/Arity), ' does not need a default instantiation.']),
        internal_name(SubName, Sub, Name1),
        declare_name(Name1, Arity, Sub, defined)
    ; Arity = [] ->
        debug(interpretation,
              [' recording default instantiation for ', w(Name/Arity)]),
        internal_name(Name, Abs, Name1),
        internal_name(SubName, Sub, Name1),
        debug(interpretation,
              [' generating new defined symbol for ', w(Name/Arity)]),
        declare_name(Name1, Arity, Sub, defined),
        record_instantiation(Name, SubName, Int)
    ; Arity = [N], integer(N) ->
        debug(interpretation,
              [' recording default instantiation for ', w(Name/Arity)]),
        internal_name(Name, Abs, Name1),
        internal_name(SubName, Sub, Name1),
        declare_name(Name1, Arity, Sub, defined),
        length(Args, N),
        AbsTerm =.. [Name|Args],
        SubTerm =.. [SubName|Args],
        record_instantiation(AbsTerm, SubTerm, Int)
    ; Arity = [quant([untyped])] ->
        debug(interpretation,
              [' recording default instantiation for ', w(Name/Arity)]),
        internal_name(Name, Abs, Name1),
        internal_name(SubName, Sub, Name1),
        declare_name(Name1, Arity, Sub, defined),
        quantify(AbsTerm, Name, [!x], Body),
        quantify(SubTerm, SubName, [!x], Body),
        record_instantiation(AbsTerm, SubTerm, Int)
    ; Arity = [quant([typed])] ->
        debug(interpretation,
              [' recording default instantiation for ', w(Name/Arity)]),
        internal_name(Name, Abs, Name1),
        internal_name(SubName, Sub, Name1),
        declare_name(Name1, Arity, Sub, defined),
        quantify(AbsTerm, Name, [!x:T], Body),
        quantify(SubTerm, SubName, [!x:T], Body),
        record_instantiation(AbsTerm, SubTerm, Int)
    ;
        handle_error(['Cannot interpret defined symbol ', w(Name/Arity), '.'])
    ).


%% private check_interpretation_map(Inst, AbsThy).
%%
%%  Check that Inst is a list of instantiations AbsTerm--->ConTerm,
%%  and for each instantiation, either:
%%   1. AbsTerm is an unconstrained skeleton from a declared symbol of AbsThy,
%%      and ConTerm is a legal term of the current (target) theory; or
%%   2. AbsTerm is an unconstrained skeleton from a defined symbol of AbsThy,
%%      and ConTerm is a similar skeleton (with an unprefixed name).
%%
check_interpretation_map([], _).
check_interpretation_map(['--->'(Abs,Con) | Map], AbsThy) :-
    check_interpretation_map_aux(Abs, Con, AbsThy),
    check_interpretation_map(Map, AbsThy).
check_interpretation_map(['--->'(Abs,_Con) | _Map], AbsThy) :-
    handle_error([q(Abs), ' is not a symbol of theory ', q(AbsThy)]).

check_interpretation_map_aux(param_const(_Name, Val), Con, _AbsThy) :-
    !,
    atomic(Val),
    check_or_error(
         atom(Con), 
         [thy(Con), ' must be an atom for interpretation map.',nl]).
check_interpretation_map_aux(Abs, Con, AbsThy) :-
    debug(interpretation, ['checking ',q(Abs),' ---> ',q(Con)]),
    defn_lhs(Abs, Name, Arity, AMVs, AOVs, AQB),
    ( AT = AbsThy
    ; query_global(subthy_owner(AT, AbsThy))
    ),
    query_theory(symbol(Name, Arity, Kind), AT),
    !,
    debug(interpretation, [q(Kind),' in theory ',q(AT)]),
    ( Kind = declared ->
        check_definition(Abs, Con, Name, Arity, Constraints),
        check_or_error(
            Constraints = true,
            [thy(Abs),' ---> ',thy(Con),' requires constraints.',nl,
             'Not implemented yet.'])
    ; Kind = defined,
      defn_lhs(Con, _CName, Arity, CMVs, COVs, CQB),
      a(AMVs,AOVs,AQB)==a(CMVs,COVs,CQB) ->
        true
    ;
        handle_error([q(Abs), ' is a defined symbol of ', q(AbsThy), nl,
                      ' but ', q(Con), ' is not a valid renaming of it.'])
    ).


%% private check_instantiation_map(Inst, AbsThy).
%%
%%  Check that Inst is a list of instantiations AbsTerm--->ConTerm, where
%%  AbsTerm is an unconstrained skeleton from a symbol of AbsThy,
%%  and ConTerm is a similar skeleton (with an unprefixed name).
%%
check_instantiation_map([], _).
check_instantiation_map(['--->'(Abs,Con) | Map], AbsThy) :-
    check_instantiation_map_aux(Abs, Con, AbsThy),
    check_instantiation_map(Map, AbsThy).
check_instantiation_map(['--->'(Abs,_Con) | _Map], AbsThy) :-
    handle_error([q(Abs), ' is not a symbol of theory ', q(AbsThy)]).


check_instantiation_map_aux(param_const(_Name, Val), Con, _AbsThy) :-
    !,
    atomic(Val),
    check_or_error(
         atom(Con),
         [thy(Con), ' must be an atom for instantiation map.',nl]).
check_instantiation_map_aux(Abs, Con, AbsThy) :-
    debug(interpretation, ['checking ',q(Abs),' ---> ',q(Con)]),
    defn_lhs(Abs, Name, Arity, AMVs, AOVs, AQB),
    ( AT = AbsThy
    ; query_global(subthy_owner(AT, AbsThy))
    ),
    query_theory(symbol(Name, Arity, Kind), AT),
    !,
    debug(interpretation, [q(Kind),' in theory ',q(AT)]),
    ( defn_lhs(Con, _CName, Arity, CMVs, COVs, CQB),
      a(AMVs,AOVs,AQB)==a(CMVs,COVs,CQB) ->
        true
    ;
        handle_error([q(Abs), ' is a defined symbol of ', q(AbsThy), nl,
                      ' but ', q(Con), ' is not a valid renaming of it.'])
    ).


%% process_reparentings(Reparentings, CD, Inst0, Inst).
%%
%%  Process Reparentings into CD, adding the inherited instantiations to Inst0
%%  giving Inst.
%%
%%  mode process_reparentings(+list(term), +atom, @list(term), -list(term)).
%%
process_reparentings([], _, Inst, Inst).
process_reparentings(['--->'(Abs,Sub) | Repar], CD, Inst0, Inst) :-
    check_or_error(
        query_global(subthy_owner(Sub, CD)),
        [q(Sub), ' is not a subtheory of ', q(CD), '.']),
    check_or_error(
        query_theory(interpreted_theory(Abs, Sub), Sub),
        [q(Sub), ' was not created by interpreting ', q(Abs), '.']),
    findall('--->'(A,C), query_theory(interpretation(A,C), Sub), Inst1),
    append(Inst1, Inst0, Inst2),
    process_reparentings(Repar, CD, Inst2, Inst).


%% check_inherited_ancestors(Abs, Con, Repar).
%%
%%  Check that all ancestors of Abs are either already inherited by Con, or are
%%  the subject of a reparenting in Repar.
%%
%%  mode check_inherited_ancestors(@atom, @atom, @list(ground)).
%%
check_inherited_ancestors(Abs, Con, Repar) :-
    \+ (real_strict_ancestor(Abs, Anc),
        \+ check_or_error(
            ( member('--->'(Anc,_), Repar)
            ; strict_ancestor(Con, Anc)
            ),
            [q(Anc), ' is an ancestor of ', q(Abs), nl,
             ' but is not interpreted or inherited by ', q(Con), '.'])
       ).


%% translate_dependencies(Deps, Sub, TDeps).
%%
%%  Take a list of theorem dependencies and translate it according to the
%%  interpretation for Sub.
%%  NOT YET IMPLEMENTED.
%%
%%  mode translate_dependencies(+list(term), @atom, -list(term)).
%%
translate_dependencies(Deps, _, Deps).

%% interpret_tactic_file(File, Abs, Sub, Int).
%%
%%  Interpret all theory terms in Abs/File and produce Sub/File.
%%  NOT YET IMPLEMENTED.
%%
%%  mode interpret_tactic_file(@filename, @theory, @theory, @list(ground)).
%% 
interpret_tactic_file(File, Abs, Sub, _Int) :-
    writelist(['Copying tactic file.',nl]),
    mkdir(Sub, 8'777, _),
    concat_atom(['cp ', Abs, /, File, '.ql ', Sub, /, File, '.ql'], Cmd),
    os(system(Cmd)).

%% interpret_nfis(Body0, Int, Body).
%%
%%  Interpret the nfi list Body0 using Into to give Body.
%%
%%  mode interpret_nfis(@code, @theory, -code).
%% 
interpret_nfis(true, _, B) :-
    !,
    B = true.
interpret_nfis((A0,B0), I, (A,B)) :-
    !,
    interpret_nfis(A0, I, A),
    interpret_nfis(B0, I, B).
interpret_nfis(not_free_in(X,T0), I, not_free_in(X,T)) :-
    !,
    interpret_term(T0, I, T).
interpret_nfis(distinct(T0), _, distinct(T)) :-
    !,
    %% T0 should be all object variables, which don't need interpretation.
    T = T0.
interpret_nfis(T, _, _) :-
    handle_error(['Cannot interpret axiom body ', q(T)]).

%% interpret_tactic_body(Code0, Int, Code).
%%
%%  Interpret all interp(V=T) goals in Code0 using Int to give Code.
%%  This version inpects only top-level conjuncts of Code0.
interpret_tactic_body(Code, _, Code) :-
    var(Code),
    !.
interpret_tactic_body((A0,B0), Int, (A,B)) :-
    !,
    interpret_tactic_body(A0, Int, A),
    interpret_tactic_body(B0, Int, B).
interpret_tactic_body(interp(V=T0), Int, interpretable(V=T)) :-
    !,
    interpret_term(T0, Int, T).
interpret_tactic_body(Code, _, Code).

%% interp(V=T).
%%
%%  Unify V and T.
%%  Used to wrap unifications whose right-hand-sides should be subject to
%%  interpretation.
%% 
%%  mode interp(?term, ?term).
%%
interp(T=T).


%% private interpret_cogito_state_var(Int, StateVar0, StateVar).
%%
%%  Translates the Cogito state var whose name is StateVar0.
%%
interpret_cogito_state_var(Int, Z0, Z) :-
    check_or_error(
        split_atom(Z0, Mod0, '#', Name),
        ['Illegal Cogito zvar: ',q(Z0)]),
    (
        query_theory(interpreted_theory(Mod0, Mod), Int)
    ;
        query_theory(interpreted_subtheory(Mod0, Mod), Int)
    ),
    !,
    concat_atom([Mod,'#',Name], Z).
interpret_cogito_state_var(_, Z0, Z0).



%% private add_rulegroups(Abs,Conc).
%%
%% adds the rulegroups from theory Abs to theory Conc.
%% note no translation/interpretation of names is neccessary.
add_rulegroups(Abs,Conc):-
    global_state_lookup(current_dir,CD),
    add_rulegroups1(Abs,Conc),
    cd(CD).

add_rulegroups1(Abs,Conc):-
    cd(Abs),
    setof((Group,Name,Dir),
          (Thy1, In, Conds, Reln, Out, ThmThm) ^
           ((Thy1 = Abs; query_global(subthy_owner(Thy1, Abs))),
            query_theory(rulegroup(Group, In, Conds, Reln, Out, ThmThm:Name, Dir))),
          Groups),
    cd(Conc),
    member((Group1,Name1,Dir1), Groups),
    rulegroup(Group1,Name1,Dir1),
    fail.

add_rulegroups1(_,_).
