%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: main.ql,v 1.60 2006/02/13 04:17:02 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file defines the initial behaviour of the prover.
%%
%% main/1 is the first procedure called by Qu-Prolog when the
%% prover is started.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


?- %% A minimal operator table for saving proofs and *.erg files.
   %% For backwards compatibility, it is a subset of the 'user' table.
   % ',' is in there automatically.
   op(proof,50,xfx,':'),
   op(proof,400,yfx,'/'),
   op(proof,600, xfx, not_free_in).

?- add_global_exception_handler(signal(Sev,_Goal,_Msg,Sig),
                                handle_signal(Sig, Sev)),
   add_global_exception_handler(sigint_exception(Sev,_,_), handle_sigint_exception(Sev)),
   add_global_exception_handler(
       syntax_error(Severity, Goal, Message, Stream, StreamPos0,
                    StreamPosN, LineNo0, LineNoN, Tokens, TokensPos),
       handle_syntax_error(
           syntax_error(Severity, Goal, Message, Stream, StreamPos0,
                        StreamPosN, LineNo0, LineNoN, Tokens, TokensPos))).

handle_syntax_error(Data) :-
    get_exception_message(Data, Msg),
    write_term_list(stderr, Msg).

handle_signal(Signal, Severity) :-
    handle_event(signal(Signal, Severity)),
    !.

handle_signal('SIGINT', Sev) :-
    !,
    thread_atomic_goal(
       thread_push_goal(thread0, handle_sigint(Sev))).

handle_signal(Sig, _) :- 
    default_signal_handler(Sig).

handle_sigint(Sev) :-
    exception(sigint_exception(Sev,_,'unhandled SIGINT')).

handle_sigint_exception(Sev):-
    ip_lookup(in_tactic, X), X == true,
    !,
    writelist(stderr,
              [nl,
               '[c]ontinue/[a]bort(Tactic)/',
               '[q]uit(Proof)/[abandon](Ergo)/[h]elp: ']),
%    flush_output(stderr),
    get_ergo_line(stdin, Line),
    ( answer_matches(Line, "help", 0) ->
        writelist(stderr,
         [nl,
          '  Continue   - resume current execution.',nl,
          '  Abort_tactic - fail the current tactic.',nl,
          '  Quit Proof - fail out of the current proof and return to',nl,
          '               the theory-level user interface.',nl,
          '  Abandon    - exit from Ergo (WITHOUT SAVING CHANGES!!)',nl,
          '  Help       - display this message.',nl]),
        handle_sigint(Sev)
    ; answer_matches(Line, "continue", 1) ->
        true
    ; answer_matches(Line, "quit", 1) ->
        announce_event(proof(Thy, Name, aborted)),
        throw(abort_proof(Thy:Name))
    ; answer_matches(Line, "abort_tactic", 1) ->
%        announce_event(proof(Thy, Name, aborted)),
        throw(abort_tactic(Thy:Name))
    ; answer_matches(Line, "abandon", 7) ->
        announce_event(proof(Thy, Name, aborted)),
        abandon
    ;
        writelist(stderr,['Error: Illegal/incomplete answer.',nl]),
        handle_sigint(Sev)
    ).
handle_sigint_exception(Sev):-
    in_proof(Thy:Name),
    !,
    writelist(stderr,
              [nl,
               '[c]ontinue/',
               '[q]uit(Proof)/[abandon](Ergo)/[h]elp: ']),
%    flush_output(stderr),
    get_ergo_line(stdin, Line),
    ( answer_matches(Line, "help", 0) ->
        writelist(stderr,
         [nl,
          '  Continue   - resume current execution.',nl,
          '  Quit Proof - fail out of the current proof and return to',nl,
          '               the theory-level user interface.',nl,
          '  Abandon    - exit from Ergo (WITHOUT SAVING CHANGES!!)',nl,
          '  Help       - display this message.',nl]),
        handle_sigint(Sev)
    ; answer_matches(Line, "continue", 1) ->
        true
    ; answer_matches(Line, "quit", 1) ->
        announce_event(proof(Thy, Name, aborted)),
        throw(abort_proof(Thy:Name))
    ; answer_matches(Line, "abandon", 7) ->
        announce_event(proof(Thy, Name, aborted)),
        abandon
    ;
        writelist(stderr,['Error: Illegal/incomplete answer.',nl]),
        handle_sigint(Sev)
    ).
handle_sigint_exception(Sev) :-
    /* not in proof */
    writelist(stderr, [nl, '[c]ontinue/',
                       '[abandon](Ergo)/[h]elp: ']),
%    flush_output(stderr),
    get_ergo_line(stdin,Line),
    ( answer_matches(Line, "help", 0) ->
        writelist(stderr,
         [nl,
          '  continue     - resume current execution.',nl,
          '  abandon      - exit from Ergo (WITHOUT SAVING CHANGES!!).',nl,
          '  help         - display this message.',nl]),
        handle_sigint(Sev)
    ; answer_matches(Line, "continue", 1) ->
        true
    ; answer_matches(Line, "abandon", 7) ->
        abandon
    ;
        writelist(stderr,['Error: Illegal/incomplete answer.',nl]),
        handle_sigint(Sev)
    ).

%% main(Args).
%%
%%  Startup procedure for Ergo.
%%  Args is a list of all the command line arguments after
%%  the first '--' argument (all previous arguments are handled
%%  by Qu-Prolog.
%%
%%  mode main(@list(atom)).
%%
main(Args) :-
    ( member(ergo4, Args) ->
        assert(ergo_version(ergo4))
    ;
        assert(ergo_version(ergo))
    ),
    with_local_exception_handler(initialize_prover(Args),
                                   sigint_exception(_,_,_),
%                                  signal(_,_,_,'SIGINT'),
                                 (write(' You''re too impatient!'),nl,halt)),
    user_interface,
    bye.

initialize_prover(Args) :-
    ip_set(less_verbose, 0),
    cmdline_args(Args),
    do_load_foreign,
    initialize_macros,
    global_state_set(current_dir, none),
    consult_global,
    announce_event(prover(startup)),
    load_global_tactics,
    ctr_set(dbms_interfaces, 0),
    ( ergo_version(ergo4) ->
        initialize_proof_structure
    ;
        true
    ).


%% bye.
%%
%%  The bye/0 command saves all modified data (except for any incomplete
%%  proofs, which are just discarded with a warning), and generates a
%%  prover(exit) announcement before exiting.  Rather than being
%%  called directly, bye/0 is usually invoked automatically when
%%  the theory-level user interface procedure returns successfully.
%%  When invoked this way, there cannot be any incomplete proofs.
%%
%%  mode bye.
%%
bye :-
    ctr_get(dbms_interfaces, Num),
    Num > 0,
    !,
    handle_error(['cannot exit a nested theory-level interface via bye/0.']).
bye :-
    get_current_proofs(Names),
    ( Names == [] ->
        true
    ;
        writelist(['Warning: discarded incomplete proofs: ',q(Names),nl])
    ),
    save,
    announce_event(prover(exit)),
    get_error_count(Errs),
    ( Errs > 0 ->
        halt(1)
    ;
        halt(0)
    ).


general command
  abandon
extends
  save_exit_cmds
mode
  abandon
doc
  'Exit the prover immediately without saving anything!',nl,
  'WARNING: this command discards all theory changes and proofs',nl,
  'done since the last save command.'.

abandon :-
    announce_event(prover(abandon)),
    halt(0).


%% set_dbms_interface(Tactic).
%%
%%  Sets the dbms interface to be Tactic/0.
%%  This will take effect next time Ergo is started,
%%  (or the next time user_interface/0 is called, eg. when
%%  a new theory interface is started up).
%%
%%  mode set_dbms_interface(@atom).
%%
set_dbms_interface(T) :-
    retractall_global(dbms_interface(_)),
    assert_global(dbms_interface(T)).


user_interface :-
    (query_global(dbms_interface(T)) ->
        ctr_get(dbms_interfaces, Num0),
        Num is Num0 + 1,
        ctr_set(dbms_interfaces, Num),
        unwind_protect(
            call_predicate(T),
            ctr_set(dbms_interfaces, Num0))
    ;
        write('No user interface has been defined.'), nl,
        repeat,
        get_command('Ergo Prolog Command ?- ', user, Cmd),
        ui_do_term(Cmd)
    ),
    !.

ui_do_term(end_of_file) :-
    !.
ui_do_term(Cmd) :-
    (call_predicate(Cmd) ->
        true
    ;
        write(failed),
        nl
    ),
    fail.



error_count parameter
  [set_error_count, get_error_count]
doc
  'This parameter records the number of errors that have occurred in',nl,
  'the current Ergo session.  It can be incremented by calling the',nl,
  'increment_error_count/0 procedure.  For example, user interfaces',nl,
  'may choose to do this whenever a command fails.',
  para,
  'This parameter is used to determine the exit status of Ergo for',nl,
  'normal exits (eg. via the ''bye'' command).  If the error count is',nl,
  'zero, then the exit status is zero, otherwise it is one.'.

get_error_count(Val) :-
    global_state_lookup('Error@Count', Val).

set_error_count(Val) :-
    PRE('set_error_count', integer(Val)),
    0 =< Val,
    global_state_set('Error@Count', Val).

increment_error_count :-
    global_state_increment('Error@Count', _).


?- set_error_count(0).

