%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: parameters.ql,v 1.6 2002/06/27 05:10:19 itp Exp $
%%
%%  This file contains commands and documentation nodes for 'parameters'.
%%  Parameters are (name,value) pairs whose value can be set by the
%%  user via a standard interface (the `set Param === Value' command).


parameters header
  'Parameters'
doc
  'Ergo has several global parameters that control various aspects of',nl,
  'its behaviour, such as how terms are printed, how verbose the output',nl,
  'should be etc.',
  para,
  'Each parameter has a name and a value.',nl,
  'The ',xref(['show(ParamName)'], global, 'general:show/1'),
  ' command can be used to display',nl,
  'the current value of a parameter.',nl,
  'The ',xref(['set ParamName === Value'], global, 'general:set/1'),
  ' command can be used to set a',nl,
  'parameter to a new value.',
  para,
  'The parameters are:',nl
extends
  master_help_node.


general command
  set('==='(Parameter, Value))
header
  'set Parameter === Value'
extends
  general_cmds
doc
  'Set one of the ', xref([parameters], global, parameters), '.'
implementation
  set_parameter(Parameter, Value)
mode
  set(@compound).

%% command set_parameter(Name, Value).
%% command get_parameter(Name, Value).
%%
%%  Set/get Parameter=Value, in the way specified by Parameter's declaration.
%%
%%  mode set_parameter(@atom, @term).
%%  mode get_parameter(@atom, ?term).
%%
set_parameter(Name, Value) :-
    query_global_or_ancestors(parameter(Name, _, SetCmd, _)),
    !,
    call_predicate(SetCmd,Value),
    announce_event(param_changed(Name, Value)).
set_parameter(Name, _) :-    
    handle_error(['parameter ', w(Name), ' has not been declared.']).
    
get_parameter(Name, Value) :-
    query_global_or_ancestors(parameter(Name, _, _, GetCmd)),
    !,
    call_predicate(GetCmd,Value).
get_parameter(Name, _) :-    
    handle_error(['parameter ', w(Name), ' has not been declared.']).


