%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% CVS $Id: types.ql,v 1.6 2000/11/15 06:29:44 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file contains procedures for declaring and storing types which 
%% are used in mode declarations. 'type' is a global_fact in the database.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% builtin_type(Name,Type,DocNode,PrologPred).
%%
%%  These are predefined entries in the global part of the
%%  database.  Such entries can never be deleted.
%%
%%  mode builtin_type(@nonvar, @list(nonvar), @docnode, @code).
%%
type_term header
  'term'
doc
  'Any Qu-Prolog term.'
extends
  modes_help_ergo4,
  modes_help_ergo5.

builtin_type(term, [], type_term, notypecheck).

type_anyvar header
  'anyvar'
doc
  'Any (meta or object) variable.'
extends
  modes_help_ergo4,
  modes_help_ergo5.

builtin_type(anyvar,  [term],  type_anyvar, notypecheck).

type_var header
  'var'
doc
  'Any Qu-Prolog meta variable.'
extends
  modes_help_ergo4,
  modes_help_ergo5.

builtin_type(var,  [anyvar],  type_var, var).

type_obvar header
  'obvar'
doc
  'Any Qu-Prolog object variable.'
extends
  modes_help_ergo4,
  modes_help_ergo5.

builtin_type(obvar, [anyvar], type_obvar, obvar).

type_nonvar header
  'nonvar'
doc
  'Any term other than variables.'
extends
  modes_help_ergo5,
  modes_help_ergo4.

builtin_type(nonvar, [term], type_nonvar, nonvar).

type_ground header
  'ground'
doc
  'Any term not containing any variables.'
extends
  modes_help_ergo5,
  modes_help_ergo4.

builtin_type(ground, [nonvar], type_ground, ground).

type_atomic header
  'atomic'
doc
  'Any atom or number.'
extends
  modes_help_ergo5,
  modes_help_ergo4.

builtin_type(atomic, [ground], type_atomic, atomic).

type_number header
  'number'
doc
  'Any number.'
extends
  modes_help_ergo5,
  modes_help_ergo4.

builtin_type(number, [atomic], type_number, number).

type_real header
  'real'
doc
  'Any real number.'
extends
  modes_help_ergo5,
  modes_help_ergo4.

builtin_type(real, [number], type_real, notypecheck).

type_integer header
  'integer'
doc
  'Any integer.'
extends
  modes_help_ergo5,
  modes_help_ergo4.

builtin_type(integer, [number], type_integer, integer).

type_atom header
  'atom'
doc
  'Any Qu-Prolog name.  Syntax is [a-z][A-Za-z0-9_]*, or ''...''.'
extends
  modes_help_ergo5,
  modes_help_ergo4.

builtin_type(atom, [atomic], type_atom, atom).

type_emptylist header
  'emptylist'
doc
  'The empty list ([]).'
extends
  modes_help_ergo5,
  modes_help_ergo4.

builtin_type(emptylist, [openlist(_),list(_)], type_emptylist, is_emptylist).

type_compound header
  'compound'
doc
  'Any structured term + [].'
extends
  modes_help_ergo5,
  modes_help_ergo4.

builtin_type(compound, [nonvar], type_compound, compound).

type_gcomp header
  'gcomp'
doc
  'Any ground compound.'
extends
  modes_help_ergo5,
  modes_help_ergo4.

builtin_type(gcomp, [compound,ground], type_gcomp, notypecheck).

type_quant header
  'quant'
doc
  'Any quantified term.'
extends
  modes_help_ergo5,
  modes_help_ergo4.

builtin_type(quant, [compound], type_quant, notypecheck).

type_list_Type header
  'list(Type), openlist(Type)'
doc
  'Any list whose elements are of type Type.'
extends
  modes_help_ergo5,
  modes_help_ergo4.

builtin_type(list(Type), [nonvar], type_list_Type, satisfies_list_type(Type)).
builtin_type(openlist(Type), [term], type_list_Type, notypecheck).

%% The internal types are not linked to the master_help_node.
%% These types are used only for modes in the trusted core prover.
modes_help_internal doc
    'Modes and types for Ergo internal data structures.'
header
    'Internal Modes and Types'.

%% Ergo internal data structures
type_arity header
  'arity'
doc
  'A fully-instantiated arity.'
extends
  modes_help_internal.

builtin_type(arity, [list(ground)], type_arity, notypecheck).

type_hrec header
  'hrec'
doc
  'hrec is the structure used to represent each hypothesis.'
extends
  modes_help_internal.

builtin_type(hrec, [compound], type_hrec, notypecheck).

type_proof header
  'proof'
doc
  'proof is the (FocusProofList,HypList) structure used to represent',nl,
  'the complete proof of a window.'
extends
  modes_help_internal.

builtin_type(proof, [compound], type_proof, notypecheck).

type_cdep_node header
  'cdep_node'
doc
  xinc(global,internal_types)
extends
  modes_help_internal.

builtin_type(cdep_node, [compound], type_cdep_node, notypecheck).

type_cdep_dms header
  'cdep_dms'
doc
  xinc(global,internal_types)
extends
  modes_help_internal.

builtin_type(cdep_dms, [compound], type_cdep_dms, notypecheck).

type_udep_node header
  'udep_node'
doc
  xinc(global,internal_types)
extends
  modes_help_internal.

builtin_type(udep_node, [compound], type_udep_node, notypecheck).

type_udep_dms header
  'udep_dms'
doc
  xinc(global,internal_types)
extends
  modes_help_internal.

builtin_type(udep_dms, [compound], type_udep_dms, notypecheck).

%% private to treeprover/context.ql
type_cref header
  'cref'
doc
  xinc(global,internal_types), ' A pointer into ''Context@@Forest''.'
extends
  modes_help_internal.

builtin_type(cref, [integer], type_cref, notypecheck).

%% private to treeprover/context.ql
type_cstruct header
  'cstruct'
doc
  xinc(global,internal_types), ' A context structure within ''Context@@Forest''.'
extends
  modes_help_internal.

builtin_type(cstruct, [compound], type_cstruct, notypecheck).

internal_types doc_fragment
  'For internal use only.'.

%% add_type(Name, Type, DocNode). 
%% add_type(Name, Type, DocNode, Pred).
%% 
%% Declares a new type as a synonym of existing type Type or list of types.
%% The meaning of the new type is given by doc node DocNode.
%% DocNode should be called `type_Name' which is of the form:
%%
%%      type_Name header
%%        'Name'
%%      doc
%%        Doc
%%      extends
%%        modes_help_ergoX.
%%
%% X indicates which to version of ergo the type is specific.
%% Type checking of commands involves checking that each parameter 
%% satisfies the specified type in the mode declaration of that command. 
%% Pred is a current_predicate which serves as an additional type check.
%% 
%% mode add_type(@nonvar, @nonvar, @docnode).
%% mode add_type(@nonvar, @list(nonvar), @docnode).
%% mode add_type(@nonvar, @nonvar, @docnode, @code).
%% mode add_type(@nonvar, @list(nonvar), @docnode, @code).
%%
add_type(Name, Type, DocNode) :-
   add_type(Name, Type, DocNode, []). 

add_type(Name, Type, DocNode, Pred) :-
   check_or_error(
        ( Name =.. [F|Args],
          atom(F),
          check_distinct_metavars(Args)
        ),
        ['illegal type name: ', q(Name)]
   ),
   nonvar(Type),
   check_atom(DocNode),
   ( list(Type) ->
         check_nonempty_type(Type, Name),
         map(check_defined_type(Name), [Type]),
         TypeList = Type
   ;
         check_defined_type(Name, Type),
         TypeList = [Type]
   ),
   check_pred(Name, Pred),
   check_or_error(
         \+ query_global(type(Name,_,_,_)),
         ['type ', q(Name),' is already defined.']
   ),
   ( query_theory(doc_node(DocNode, _, _, _), global) ->
         true
   ;
         writelist(['Warning: doc node ',q(DocNode),' not found for type ',q(Name),'.',nl])
   ),
   assert_global(type(Name, TypeList, DocNode, Pred)).

check_defined_type(Name, Type) :-
   nonvar(Type),
   ( query_global(type(Type,_,_,_)) ->
         true
   ;
         writelist(['Warning: type ', q(Type),' is not defined for type ',q(Name),'.',nl])
   ).

check_nonempty_type(Type, _) :-
    \+ Type = [],
    !.
check_nonempty_type(_, Name) :-
    handle_error(['cannot have an empty list of type equivalences for type ',q(Name),'.']). 

check_pred(_, notypecheck).
check_pred(_, Pred) :-
    nonvar(Pred),
    Pred =.. [F|_],
    current_predicate(F/_).

check_pred(Name, Pred) :-
    writelist(['Warning: ',q(Pred),' is not a valid predicate list for type ',q(Name),'.',nl]).

%% is_emptylist(L).
%%
%%  Succeeds if L is the emptylist 
%%
%%  mode is_emptylist(@list(_)).
%%
is_emptylist(L) :-
   L == [].

%% satisfies_list_type(Type, List).
%%
%%  Checks that each element in list List satisfies_type Type.
%%
%%  mode satisfies_list_type(@ground, @list(_)).
%%
satisfies_list_type(Type, List) :-
   list(List),
   map(satisfies_type(Type), [List]).

