%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: macro_expansions.ql,v 1.5 2001/06/04 01:24:09 itp Exp $
%%
%% See expansions.ql for how to use this file.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% WARNING!
%% This file should be able to run OUTSIDE Ergo.
%% Beware of using Ergo-defined procedures in this file!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% MACRO EXPANSIONS (can be outside Ergo).
%%     The following appear only as `facts':
%%       @include(File).       -- include File.
%%       @define(Mac, Def).    -- define macro
%%       @define(Mac).         -- define macro
%%       @undef(Mac).          -- remove definition
%%       @if(Goal).            -- start conditional clauses
%%       @ifdef(Mac).          -- same as @if(is_defined_macro(Mac)).
%%       @ifndef(Mac).         -- same as @if(\+ is_defined_macro(Mac)).
%%       @else.                -- reverse conditional clauses
%%       @endif.               -- end conditional clauses
%%
%%     The following appear _within_ clauses:
%%       @@Mac                 -- replaced by the expansion of Mac
%%       @if Goal then Term1 else Term2 endif
%%                             -- replaced by Term1 or Term2 according to Goal
%%

%% command enable_macro_expansions.
%% command disable_macro_expansions.
%%
%%  Enable/disable macro expansions.
%%
enable_macro_expansions :-
    add_ergo_expansion(mac_expand_term),
    add_ergo_expansion(mac_expand_subterms).

disable_macro_expansions :-
    remove_ergo_expansion(mac_expand_term),
    remove_ergo_expansion(mac_expand_subterms).

?- op(990, yf,  endif).
?- op(989, xfx, else).
?- op(988, xfx, then).
?- op(200, fx,  @).
?- op(200, fx,  @@).

%% mac_expand_term(Term0, Terms, Vars).
%%
%%  Expand a top-level clause.
%%  Macro-related `facts' will expand to nothing (after setting the appropriate
%%  environment); ordinary clauses expand to nothing (if the current @if
%%  condition is false) or to themselves (if it is true).
%%
mac_expand_term(Term, Exp, _Vars) :-
    nonvar(Term),
    nonvar_functor(Term),
    exp_debug(['trying macro expansion on ', w(Term)]),
    mac_expand_term_2(Term, Exp).

%% mac_expand_subterms(Term0, Term, Vars).
%%
%%  Expand @@, @macro and @if... forms in subterms.
%%
mac_expand_subterms(Term, Exp, _Vars) :-
    nonvar(Term),
    nonvar_functor(Term),
    exp_debug(['trying macro expansion on ', w(Term)]),
    mac_expand_subterms_2(Term, Exp, Flag),
    nonvar(Flag).

%% define_macro(Mac).
%% define_macro(Mac, Def).
%%
%%  Define macro Mac to expand to Def (default itself).
%%  Issue a warning if there is already a defined macro that unifies with Mac.
%%
%%  mode define_macro(@nonvar).
%%  mode define_macro(@nonvar, @term).
%%
define_macro(Mac) :-
    define_macro(Mac, Mac).

define_macro(Mac, Def) :-
    ( is_defined_macro(Mac) ->
        exp_data_error(warning,
                       define_macro(Mac, Def),
                       ['Macro ', Mac, ' is already defined.'])
    ; assert(current_macro(Mac, Def))
    ).

%% undefine_macro(Mac).
%%
%%  Remove Mac's definition, if it has one.  Note that this will also remove a
%%  definition of a more general or more specific macro.
%%
%%  mode undefine_macro(@nonvar).
%%
undefine_macro(Mac) :-
    retractall(current_macro(Mac, _)).
            
%% is_defined_macro(Mac).
%% is_defined_macro(Mac, Def).
%%
%%  Is Mac currently defined (with definition Def)?
%%
%%  mode is_defined_macro(?nonvar).
%%  mode is_defined_macro(?nonvar, ?term).
%%
is_defined_macro(Mac) :-
    clause(current_macro(Mac, _), _).
is_defined_macro(Mac, Def) :-
    clause(current_macro(Mac, Def), _).

%% mac_include_file(File, Clauses).
%%
%%  Clauses are the clauses (and facts and queries) in File.
%%
%%  mode mac_include_file(@filename, -list(nonvar)).
%%
mac_include_file(File, Clauses) :-
    open(File, read, Stream),
    mac_include_file(Stream, [], Clauses),
    close(Stream).

mac_include_file(Stream, Clauses0, Clauses) :-
    read_term(Stream, Term, [variable_names(Vars)]),
    expand_term(Term, Terms, Vars),
    ( Terms == end_of_file ->
        Clauses = Clauses0
    ; list(Terms) ->
        append(Terms, Clauses0, Clauses1),
        mac_include_file(Stream, Clauses1, Clauses)
    ;
        Clauses1 = [Terms | Clauses0],
        mac_include_file(Stream, Clauses1, Clauses)
    ).

%% current_macro(Macro, Expansion).
%%
%%  The table of currently-defined macros and theoir expansions.
%% 
?- dynamic(current_macro/2).

%% ifdef_flags(Value).
%%
%%  The stack of nested ifdef flags.
%%  Each Value can be:
%%   yes  -- all current conditions are true.
%%   no   -- latest condition is false but all earlier conditions are true.
%%   skip -- some earlier condition is false.
%% 
?- dynamic(ifdef_flags/1).

%% mac_ifdef_flag.
%%
%%  True iff the current ifdef flag is true (including an empty stack).
%%
mac_ifdef_flag :-
    ( clause(ifdef_flags(Flag), _) ->
        Flag = yes
    ;
        true
    ).

%% private mac_expand_term_2(Term, Clauses).
%%
%%  Expander for top-level terms.
%%  Called only when term is compound with nonvar functor.
%% 
mac_expand_term_2(@include(File), Clauses) :-
    mac_ifdef_flag,
    !,
    mac_include_file(File, Clauses).
mac_expand_term_2(@define(Mac), []) :-
    mac_ifdef_flag,
    !,
    define_macro(Mac).
mac_expand_term_2(@define(Mac, Def), []) :-
    mac_ifdef_flag,
    !,
    mac_expand_subterms_2(Def, XDef, _),
    define_macro(Mac, XDef).
mac_expand_term_2(@undef(Mac), []) :-
    mac_ifdef_flag,
    !,
    undefine_macro(Mac).
mac_expand_term_2(@if(Goal), []) :-
    mac_expand_subterms_2(Goal, XGoal, _),
    ( \+ mac_ifdef_flag ->
        Push = skip
    ; call_predicate(XGoal) ->
        Push = yes
    ;
        Push = no
    ),
    asserta(ifdef_flags(Push)).
mac_expand_term_2(@ifdef(Mac), []) :-
    ( \+ mac_ifdef_flag ->
        Push = skip
    ; is_defined_macro(Mac) ->
        Push = yes
    ;
        Push = no
    ),
    asserta(ifdef_flags(Push)).
mac_expand_term_2(@ifndef(Mac), []) :-
    ( \+ mac_ifdef_flag ->
        Push = skip
    ; is_defined_macro(Mac) ->
        Push = no
    ;
        Push = yes
    ),
    asserta(ifdef_flags(Push)).
mac_expand_term_2(@else, []) :-
    ( retract(ifdef_flags(Pop)) ->
        true
    ;
        exp_data_error(warning,
                       @else,
                       ['@else has no matching @ifdef'])
    ),
    ( Pop = yes ->
        Push = no
    ; Pop = no ->
        Push = yes
    ;
        Push = skip
    ),
    asserta(ifdef_flags(Push)).
mac_expand_term_2(@endif, []) :-
    ( retract(ifdef_flags(_)) ->
        true
    ;
        exp_data_error(warning,
                       @endif,
                       ['@endif has no matching @ifdef'])
    ).
mac_expand_term_2(end_of_file, end_of_file) :-
    clause(ifdef_flags(_), _),
    !,
    retractall(ifdef_flags(_)),
    exp_data_error(warning,
                   end_of_file,
                   ['@ifdef has no matching @endif']).
mac_expand_term_2(_Clause, []) :-
    \+ mac_ifdef_flag,
    !.

%% private mac_expand_subterms_2(Term, Exp, Flag).
%%
%%  The expander for subterms.
%%  Always succeeds, sets Flag if any work done.
%% 
mac_expand_subterms_2(Term, Term, _Flag) :-
    any_variable(Term),
    !.
mac_expand_subterms_2(Term0, Term, Flag) :-
    sub(Term),
    !,
    substitute(Term0, Subs, Body0),
    substitute(Term, Subs, Body),
    mac_expand_subterms_2(Body0, Body, Flag).
mac_expand_subterms_2(Term0, Term, Flag) :-
    quant(Term0),
    !,
    quantify(Term0, Quant, Vars, Body0),
    quantify(Term, Quant, Vars, Body),
    mac_expand_subterms_2(Body0, Body, Flag).
mac_expand_subterms_2(Term0, Term, _Flag) :-
    atomic(Term0),
    !,
    Term = Term0.
mac_expand_subterms_2(Term0, Term, Flag) :-
    compound(Term0),
    nonvar_functor(Term0),
    mac_expand_subterm(Term0, Term),
    !,
    Flag = yes.
mac_expand_subterms_2(Term0, Term, Flag) :-
    compound(Term0),
    functor(Term0, Fun, Arity),
    functor(Term, Fun, Arity),
    mac_expand_subterms_2(Arity, Term0, Term, Flag).

mac_expand_subterms_2(0, _, _, _) :-
    !.
mac_expand_subterms_2(N, Term0, Term, Flag) :-
    % N > 0,
    arg(N, Term0, Arg0),
    arg(N, Term, Arg),
    mac_expand_subterms_2(Arg0, Arg, Flag),
    N1 is N - 1,
    mac_expand_subterms_2(N1, Term0, Term, Flag).

%% private mac_expand_subterm(Term0, Term).
%%
%%  Term0 is compound with nonvar functor.
%%  Succeeds only if some work is done.
%% 
mac_expand_subterm(@macro(Mac), Def) :-
    is_defined_macro(Mac, Def).
mac_expand_subterm(@@(Mac), Def) :-
    is_defined_macro(Mac, Def).
mac_expand_subterm(endif(else(then(@if(G), T1), T2)), T) :-
    mac_expand_subterms_2(G, XG, _),
    ( call_predicate(XG) ->
        mac_expand_subterms_2(T1, T, _)
    ;
        mac_expand_subterms_2(T2, T, _)
    ).
mac_expand_subterm(endif(else(then(@ifdef(Mac), T1), T2)), T) :-
    ( is_defined_macro(Mac) ->
        mac_expand_subterms_2(T1, T, _)
    ;
        mac_expand_subterms_2(T2, T, _)
    ).
