%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: macros.ql,v 1.20 2002/06/27 05:10:21 itp Exp $

#include "../lib/ergo.h"

%% initialize_macros.
%%
%%  Initializes the stack of input files to contain just the
%%  current input stream.
%%
%%  mode initialize_macros.
%%
initialize_macros :-
    current_input(S),
    assert(macro_stack([S])).


general command
  macro_and_delete(File),
  macro(File)
extends
  general_cmds
doc
  'Sets the input stream to be the file called File.',nl,
  'When the end of that file is reached, the original input stream',nl,
  '(eg. the terminal) is restored and, for macro_and_delete only,',nl,
  'File is deleted.',
  para,
  'Macro files can be nested, to an arbitrary depth.',
  para,
  'The theory/tactic ',xref(load_path,global,'parameter:load_path'),
  ' is ',emph(not),' used to find File.'
mode
  macro_and_delete(@filename),
  macro(@filename).


macro(File) :-
    error('Opening Macrofile: '),
    errornl(File),
    open(File, read, Stream, [eof_action(eof_code)]),
    retract(macro_stack(Others)),
    assert(macro_stack([Stream|Others])),
    !.


macro_and_delete(File) :-
    error('Opening and deleting Macrofile: '),
    errornl(File),
    open(File, read, Stream),
    retract(macro_stack(Others)),
    assert(macro_stack([Stream|Others])),
    delete_file(File),
    !.


%% exit_macro.
%%
%%  Closes the current input stream and pops it off the input stack.
%%  This must only be called when 'in_macro' is true.
%%
%%  mode exit_macro.
%%
exit_macro :-
    retract(macro_stack([Curr|Others])),
    assert(macro_stack(Others)),
    close(Curr),
    errornl('Exiting Macrofile.'),
    !.


%% in_macro.
%%
%%  Succeeds if the input stack contains at least two streams.
%%
%%  mode in_macro.
%%
in_macro :- macro_stack([_,_|_]).


%% current_ergo_input(Stream).
%%
%%  Returns the current input stream.
%%
%%  mode current_ergo_input.
%%
current_ergo_input(S) :- 
    macro_stack([S|_]).


