%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% CVS $Id: pretty.ql,v 1.14 2000/11/15 06:29:46 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% A simple pretty-printer for Qu-Prolog terms.
%%
%%   Warning: the algorithm is (exponentially?) slow for large terms.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% pretty_print(Term, Width, IndentList).
%% pretty_print(Term, Width, IndentList, Prec).
%%
%%  The four argument version adds brackets if necessary,
%%  where Prec is the precedence of the surrounding operator.
%%
%%  mode pretty_print(@term, @integer, @writelist).
%%  mode pretty_print(@term, @integer, @writelist, @integer).
%%
pretty_print(T, W, I, Prec) :-
    (term_precedence(T, TPrec), TPrec >= Prec ->
        write('('),
        pretty_print(T,W,I),
        write(')')
    ;
        pretty_print(T,W,I)
    ).


pretty_print(T, W, I) :-
    (pp_small_term(T,W) ->
        writelist([q(T)])
    ; sub(T) ->
        substitute(T, Subs, Term),
        pp_indent(W,I, W2,I2),
        pp_subs(Subs, W, I),
        writelist_indent([nl], I2),
        current_op(SubsPrec, xfy, '*'),
        pretty_print(Term, W2, I2, SubsPrec)
    ; list(T) ->
        pp_indent(W,I, W2,I2),
        writelist_indent(['[',nl], I2),
        current_op(CommaPrec, xfy, ','),
        pp_list_elems(T, W2, I2, CommaPrec),
        writelist_indent([nl,']'], I)
    ; compound(T) ->
        functor(T,F,Arity),
        pp_indent(W,I, W2,I2),
        writelist_indent([q(F),'(',nl], I2),
        current_op(CommaPrec, xfy, ','),
        pp_args(T, 1, Arity, W2, I2, CommaPrec),
        writelist_indent([')'], I2)
    ;
        quantify(T, Q, Vars, Body),
        %% and not preceeded by a substitution.
        pp_indent(W,I, W2,I2),
        current_op(QuantPrec, quant, Q),
        writelist_indent(['(',q(Q),' ',q(Vars),' ',nl], I2),
        pretty_print(Body, W2, I2, QuantPrec),
        writelist_indent([')'], I2)
    ).


%% term_precedence(+Term, -Precedence)
%%
%%   NB. since we always bracket quantifiers, they have precedence 0.
%%
term_precedence(T, P) :-
    substitute(T, [_|_], _),
    current_op(P, xfy, '*').
term_precedence(T, 0) :-
    nonvar(T),
    T = [_|_],
    !.
term_precedence(T, P) :-
    compound(T),
    functor(T, F, _Arity),
    current_op(P, _, F).
%% TO DO: Extend term_precedence to handle prefix/infix overloads.
term_precedence(T, 0).


%% pp_subs prints a list of substitutions.
%%
pp_subs([], _, _).
pp_subs([S|SS], W, I) :-
    pp_subs(SS, W, I),
    writelist([q(S)]).


%% private pp_args(+Term, +Num, +Arity, +Width, +Indent, +Prec).
%%
%%   Writes the arguments of T (from Num to Arity, inclusive),
%%   separated by a comma, newline and Indent.
%%   Prec is the precedence of ','.
%%
pp_args(T, N, Arity, W, I, Prec) :-
    N < Arity,
    !,
    arg(N, T, Arg),
    pretty_print(Arg, W, I, Prec),
    writelist_indent([',',nl], I),
    N1 is N + 1,
    pp_args(T, N1, Arity, W, I, Prec).
pp_args(T, N, N, W, I, Prec) :-
    arg(N, T, Arg),
    pretty_print(Arg, W, I, Prec).


%% private pp_list_elems(+List, +Width, +Indent, +Prec)
%%
%%   Pretty prints a list, starting each element on a new line,
%%   indented by the list Indent.
%%   Prec is the precedence of ','.
%%
pp_list_elems([E], W2, I2, Prec) :-
    !,
    pretty_print(E, W2, I2, Prec).
pp_list_elems([E|ES], W2, I2, Prec) :-
    pretty_print(E, W2, I2, Prec),
    writelist_indent([',',nl], I2),
    pp_list_elems(ES, W2, I2, Prec).


%% private pp_indent(+Width0,+Indent0, -Width,-Indent).
%%
%%   Indent is an indentation list that is one level deeper than Indent0
%%   and Width is a correspondingly decremented Width0.
%%
pp_indent(W0,I, W,['  '|I]) :-
    W1 is W0 - 2,
    (W1 < 10 -> W=10 ; W=W1).


%% private pp_small_term(+Term,+Width)
%%
%%   Succeeds if Term should be printed on one line
%%   (or couldn't be printed reasonably on two).
%%
pp_small_term(T,_) :-
    var(T).
pp_small_term(T,_) :-
    obvar(T).
pp_small_term(T,_) :-
    atomic(T).
pp_small_term(T,W) :-
    term_width(T,TW),
    CHECK('pp_small_term: term_width', (integer(TW), integer(W))),
    TW =< W.


%% term_width(+Term, -Width)
%%
%%   Width is the approximate number of characters that will
%%   be required to write Term out on a single line.
%%
term_width(In, Width) :-
    (substitute(In, Subs, Term), \+ Subs = [] ->     
        term_width(Term, W),
        subslists_width(Subs, 0, WS),
        Width is W + WS
    ; %% named meta/object variables.
      get_var_name(In, InName) ->
          name(InName,Str),
          length(Str,Width)
    ; %% unnamed meta-variables.
      var(In) ->
          Width = 2
    ; %% unnamed object-variables.
      obvar(In) ->
          Width = 3
    ; integer(In) ->
        name(In,Str),
        length(Str,Width)
    ; atom(In) ->
        atom_length(In, Width)
    ; compound(In) ->
        functor(In, F, Arity),
        atom_length(F, W1),
        ( current_op(_,_,F) ->
            W2 is W1 + Arity
        ;
            W2 is W1 + Arity * 2 + 3
        ),
        termargs_width(In, Arity, W2, Width)
    ; quant(In) ->
        quantify(In, Q, Vars, Body),
        atom_length(Q,WQ),
        length(Vars, NVars),
        term_width(Body, WBody),
        Width is WQ + WBody + (3 + 2) * NVars  % 3 for Var + 2 spaces
    ;
        handle_error(['unknown term type in pretty_print'])
    ).


%% private termargs_width(...).
%%
%%  This sums the widths of the arguments of functor In
%%
termargs_width(In, N, Sum, Total) :-
    (N =:= 0 ->
        Total = Sum
    ;
        arg(N, In, Arg),
        N1 is N - 1,
        term_width(Arg, W),
        NewSum is Sum + W,
        termargs_width(In, N1, NewSum, Total)
    ).



subslists_width([], Total, Total).

subslists_width([S|Subs], Sum, Total) :-
    subs_width(S, -2, W),
    NewSum is Sum + W + 3,  % 3 for '[]*'
    subslists_width(Subs, NewSum, Total).


subs_width([], Total, Total).

subs_width([T/_|Subs], Sum, Total) :-
    term_width(T, W),
    NewSum is Sum + W + 3 + 2,    % 3 for Var, 2 for comma
    subs_width(Subs, NewSum, Total).

