%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: thy_expansions.ql,v 1.7 2000/11/15 06:29:46 itp Exp $
%%
%% See expansions.ql for how to use this file.

%% THEORY EXPANSIONS (must be in Ergo, with a current theory).
%%     Within clauses, terms of the following forms are converted to theory
%%     terms by reading from Chars using the operator and object_var tables
%%     and abbreviations of the current theory.
%%       @thy(Chars)           -- use current theory and [] arity.
%%       @thy(Thy,Chars)       -- use specified theory and [] arity.
%%       @thy(Thy,Chars,Arity) -- use specified theory and arity.
%%     NB: currently, Thy must be the current theory (or a variable).
%%     The term @current_thy is converted to the name of the current theory.
%%

%% command enable_theory_expansions(Thy).
%% command disable_theory_expansions.
%%
%%  Enable/disable @thy expansions for Thy.
%%
%%  mode enable_theory_expansions(@theory).
%%  mode disable_theory_expansions.
%%
enable_theory_expansions(Thy) :-
    announce_event(need_operators(Thy)),
    add_ergo_expansion(expand_thy_term(Thy, _)).
disable_theory_expansions :-
    remove_ergo_expansion(expand_thy_term(_,_)).

?- op(200, fx, @).

%% expand_thy_term(Thy, Vars, ThyTerm, ExpTerm, CVars).
%%
%%  Expand the theory term @thy(...) or @current_thy.
%%  Accumulate variables in the open list Vars, and put the result in Term.
%%  CVars is the (closed) list of variables coming from the context, so we
%%  initialize Vars from it on the first call.
%%
%%  mode expand_thy_term(@atom, ?openlist(compound), @nonvar, -term,
%%                       +list(compound)).
%%
expand_thy_term(Thy, Vars, Term, Exp, CVars) :-
    nonvar(Term),
    nonvar_functor(Term),
    exp_thy_debug(['trying theory subterm expansion on ', w(Term)]),
    expand_thy_subterms(Term, Exp, Thy, Vars, CVars, Flag),
    nonvar(Flag).

%% private expand_thy_subterms(Term, Exp, Thy, Vars, CVars, Flag).
%%
%%  The expander for subterms.
%%  Always succeeds, sets Flag if any work done.
%% 
expand_thy_subterms(Term, Term, _, _, _, _Flag) :-
    any_variable(Term),
    !.
expand_thy_subterms(Term0, Term, Thy, Vars, CVars, Flag) :-
    sub(Term),
    !,
    substitute(Term0, Subs, Body0),
    substitute(Term, Subs, Body),
    expand_thy_subterms(Body0, Body, Thy, Vars, CVars, Flag).
expand_thy_subterms(Term0, Term, Thy, Vars, CVars, Flag) :-
    quant(Term0),
    %% \+ sub(Term0)
    !,
    quantify(Term0, Quant, BVars, Body0),
    quantify(Term, Quant, BVars, Body),
    expand_thy_subterms(Body0, Body, Thy, Vars, CVars, Flag).
expand_thy_subterms(Term0, Term, _, _, _, _Flag) :-
    atomic(Term0),
    !,
    Term = Term0.
expand_thy_subterms(Term0, Term, Thy, Vars, CVars, Flag) :-
    compound(Term0),
    nonvar_functor(Term0),
    expand_thy_subterm(Term0, Term, Thy, Vars, CVars),
    !,
    Flag = yes.
expand_thy_subterms(Term0, Term, Thy, Vars, CVars, Flag) :-
    compound(Term0),
    functor(Term0, Fun, Arity),
    functor(Term, Fun, Arity),
    expand_thy_subterms(Arity, Term0, Term, Thy, Vars, CVars, Flag).

expand_thy_subterms(0, _, _, _, _, _, _) :-
    !.
expand_thy_subterms(N, Term0, Term, Thy, Vars, CVars, Flag) :-
    % N > 0,
    arg(N, Term0, Arg0),
    arg(N, Term, Arg),
    expand_thy_subterms(Arg0, Arg, Thy, Vars, CVars, Flag),
    N1 is N - 1,
    expand_thy_subterms(N1, Term0, Term, Thy, Vars, CVars, Flag).

%% private expand_thy_subterm(Term0, Term, Thy, Vars, CVars).
%%
%%  Term0 is compound with nonvar functor.
%%  Succeeds only if some work is done.
%% 
expand_thy_subterm(@current_thy, Thy, Thy, _, _).
expand_thy_subterm(@thy(Chars), Term, Thy, Vars, CVars) :-
    %% NB. we don't expand @thy(Atom) or @thy(Compound), because
    %%     they are mode declarations.
    list(Chars),
    expand_thy_chars(Chars, Thy, [], Vars, Term, CVars).
expand_thy_subterm(@thy(XThy, Chars), Term, Thy, Vars, CVars) :-
    ( XThy = @current_thy ->
        true
    ;
        XThy = Thy ->
            true
    ;
        exp_data_error(unrecoverable,
                       ['Sorry, @thy terms may only refer to current theory.'])
    ),
    expand_thy_chars(Chars, Thy, [], Vars, Term, CVars).
expand_thy_subterm(@thy(XThy, Chars, Arity), Term, Thy, Vars, CVars) :-
    ( XThy = @current_thy ->
        true
    ;
        XThy = Thy ->
            true
    ;
        exp_data_error(unrecoverable,
                       ['Sorry, @thy terms may only refer to current theory.'])
    ),
    expand_thy_chars(Chars, Thy, Arity, Vars, Term, CVars).

%% expand_thy_chars(Chars, Thy, Arity, Vars, Term, CVars).
%%
%%  Does the actual expansion.
%%
%%  mode expand_thy_chars(@list(char), @theory, @arity, ?openlist(compound),
%%                        -term, +list(compound)).
%%
expand_thy_chars(Chars, _, _, _, _, CVs) :-
    exp_thy_debug(['trying theory expansion on ', w(@thy(Chars)), nl,
                       'vars are: ', w(CVs)]),
    fail.
expand_thy_chars(Chars, Thy, Arity, Vars, Term, CVars) :-
    ( is_chars(Chars) ->
        true
    ;
        exp_data_error(unrecoverable, @thy(Chars),
                       ['argument of @thy must be a chars.'])
    ),
    ( var(Vars) ->
        closed_to_open(CVars, Vars)
    ;
        true
    ),
    append(Chars, " . ", Ch),
    open_string(read(Ch), Strm),
    unwind_protect(
        (
            exp_thy_debug(['about to read from string ', chars(Ch),
                       ' in theory ', w(Thy)]),
            read_term(Strm, Term0, [variable_names(TVars),
                                    op_table(Thy),
                                    obvar_prefix_table(Thy)])
        ),
        close(Strm)
                  ),
    exp_thy_debug(['read term: ', q(Term0)]),
    %%%% DANGER! DANGER!  We are about to call out of the prover.
    %% Either this code should itself move out of the prover,
    %% or deabbreviation should be done via the hook mechanism.
    deabbreviate(Term0, Thy, Arity, Term), % currently works for current thy
    exp_thy_debug(['deabbreviated term is: ', q(Term)]),
    exp_thy_debug(['this terms var list is: ', q(TVars)]),
    exp_thy_debug(['previous vars list is: ', q(Vars)]),
    exp_merge_varslists(TVars, Vars),
    exp_thy_debug(['merged vars list is: ', q(Vars)]),
    !,
    exp_thy_debug(['theory expansion succeeded: got ', q(Term)]).
expand_thy_chars(Chars, Thy, [], _, _, _) :-
    !,
    %% We put quotes around it and turn it into an atom, because
    %% the QP write_term_list doesn't support chars/1 arguments yet.
    append([34|Chars],[34],Chars2),
    atom_codes(AChars, Chars2),
    exp_data_error(warning, @thy(Thy,Chars,[]),
                   [w(@thy(AChars)), ': theory expansion failed.']),
    fail.
expand_thy_chars(Chars, Thy, Arity, _, _, _) :-
    append([34|Chars],[34],Chars2),
    atom_codes(AChars, Chars2),
    exp_data_error(warning, @thy(Thy,Chars,Arity),
                   [w(@thy(Thy,AChars,Arity)), ': theory expansion failed.']),
    fail.

%% private exp_merge_varslists(Vars1, Vars2).
%%
%%  Vars1 is a closed list of Var=Name pairs; Vars2 is an open list of such
%%  pairs.  For every Name in Vars1, if that name also occurs in Vars2, unify
%%  the corresponding Var; otherwise, add the pair to Vars2.
%%
%%  mode exp_merge_varslists(+list(compound), ?openlist(compound)).
%%
exp_merge_varslists([], _).
exp_merge_varslists([V=N | Vs1], Vs2) :-
    ( open_member(V2=N, Vs2) ->
        V = V2
    ;
        open_append(Vs2, [V=N])
    ),
    exp_merge_varslists(Vs1, Vs2).


%% Theory expansions must always run inside Ergo, so we
%% call the standard debug/2 procedure, but with a special
%% first argument to distinguish the debug messages in this
%% file from other `expansions' debug messages.
%%
exp_thy_debug(Msg) :- debug(thy_expansions, Msg).
