%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% CVS $Id: hyps.ql,v 1.55 2000/11/15 06:29:47 itp Exp $

#include "../lib/ergo.h"


%% Access procedures for hypothesis records.
hrec_num(            hrec(N,_,_,_,_,_), N).
hrec_term(           hrec(_,T,_,_,_,_), T).
hrec_numterm(        hrec(N,T,_,_,_,_), N, T).
hrec_just(           hrec(_,_,_,J,_,_), J).
hrec_conj_just(      hrec(_,_,_,cj(J),_,_), J).
hrec_cdep(           hrec(_,_,_,_,C,_), C).
hrec_udep(           hrec(_,_,_,_,_,U), U).
hrec_deleted(        hrec(_,_,D,_,_,_)) :- D == deleted.
hrec_not_deleted(    hrec(_,_,D,_,_,_)) :- var(D).
hrec_not_justified(  hrec(_,_,_,cj(J),_,_)) :- var(J).

justify_hrec(hrec(No,Term,_,cj(J),CNode,UNode), Just, CDep, UDep) :-
    var(J),
    thaw_var(J),
    J = Just,
    cdep_discharge(CNode, CDep),
    udep_discharge(UNode, UDep),
    is_subproof_id(SPId),
    announce_event(hyp(discharge,SPId,No,Term)).

delete_hrec(hrec(_,_,deleted,_,_,_)).


%% add_hyp(Term, Number, Just, ConjDMS, UseDMS).
%%
%%  Internal procedure to add a single (justified) hypothesis to
%%  the current hypothesis list.  It assumes that syntax and
%%  new-var checking are either already done, or are unnecessary.
%%  Use add_conj/4 to add conjectures, not this procedure.
%%
%%  If Number is specified, it is used, otherwise a new number is generated.
%%
%%  mode add_hyp(@thy(term), ?integer, +nonvar, +cdep_dms, +udep_dms).
%%
add_hyp(Term, Number, Just, CDep, UDep) :-
    var(Number),
    !,
    ip_lookup(hypnr, OldNumber),
    Number is OldNumber + 1,
    ip_set(hypnr, Number),
    add_hyp(Term, Number, Just, CDep, UDep).
add_hyp(Term, Number, Just, CDep, UDep) :-
    PRE('add_hyp', integer(Number)),
    cdep_create_discharged(CNode,CDep),
    udep_create_discharged(UNode,UDep),
    ip_lookup(hyplist, Hyps0),
    insert_hyp(Hyps0, Number, hrec(Number,Term,_,Just,CNode,UNode), Hyps),
    ip_set(hyplist, Hyps),
    is_subproof_id(SPId),
    ( add_hyp_kind(Just, Kind) ->
        announce_event(hyp(add(Kind),SPId,Number,Term))
    ;
        true
    ).

%% private insert_hyp(Hyps0, Number, HRec, Hyps).
%%
%%  Maintain the hyps in reverse-numeric order.
%%  This will be efficient if add_hyp is called with smaller numbers first.
%%
insert_hyp([], _, HRec, [HRec]) :-
    !.
insert_hyp(Hyps, Num, HRec, [HRec|Hyps]) :-
    Hyps = [hrec(Num1,_,_,_,_,_)|_],
    Num1 < Num,
    !.
insert_hyp([Hyp|Hyps0], Num, HRec, [Hyp|Hyps]) :-
    PRE('insert_hyp', Num1 >= Num),
    %% The number of Hyp is greater than that of HRec.
    insert_hyp(Hyps0, Num, HRec, Hyps).


%% Inherited hyps are not announced.  We hope/expect that future versions of
%% Ergo will *share* the parent hyps.  Conceptually, inherited hyps will remain
%% at the parent level, but be accessible from the child window.

add_hyp_kind(i,            _         ) :- !,fail.
add_hyp_kind(tr(_),        translated) :- !.
add_hyp_kind(new,          new       ) :- !.
add_hyp_kind(sl(_),        proved    ) :- !.
add_hyp_kind(sr(_),        proved    ) :- !.
add_hyp_kind(oh(_,_),      proved    ) :- !.
add_hyp_kind(ht(_,_,_,_),  proved    ) :- !.
add_hyp_kind(J, _) :-
    handle_error(program, ['bad hyp just: ',q(J)]).


%% add_conj(Term, Number, CDep, UDep).
%%
%%  An internal predicate which adds a conjecture stating 'Term'
%%  to the current hypothesis list.  It assumes that syntax and
%%  new-var checking are either already done, or are unnecessary.
%%
%%  mode add_conj(@thy(term), -integer, -cdep_dms, -udep_dms).
%%
add_conj(Term, Number, CDep, UDep) :-
    freeze_var(Just),
    cdep_create(CDep),
    udep_create(UDep),
    ip_lookup(hyplist, Hyps),
    ip_lookup(hypnr, No),
    Number is No + 1,
    ip_set(hyplist, [hrec(Number,Term,_,cj(Just),CDep,UDep)|Hyps]),
    ip_set(hypnr, Number),
    is_subproof_id(SPId),
    announce_event(hyp(add(conjectured),SPId,Number,Term)).


proof command
  split_hyp(HypNr)
extends
  hyp_proofcmds
doc
  'Split a conjunct hypothesis.  This is equivalent to two',nl,
  'h_trans operations, using the following theorems:',nl,
  nl,
  'ergo:and_elim_l === (A and B) => A',nl,
  nl,
  'ergo:and_elim_r === (A and B) => B',nl,
  nl,
  'The original hypothesis is not deleted.'
mode
  split_hyp(?hypname).

split_hyp(Hyp) :-
    get_hyp(Hyp, hrec(N, ERGO_AND(A,B), _, _, CDep, UDep)), 
    add_hyp(A, _, sl(N), CDep, UDep),
    add_hyp(B, _, sr(N), CDep, UDep).


%% command split_left(Hyp, Term, NewHypNr).
%% command split_right(Hyp, Term, NewHypNr).
%%
%%  Perform each side of the split in a separate operation.  This is
%%  most useful when rerunning a proof, and it is not simple to reconstruct
%%  the original call to split_hyp.
%%
%%  mode split_left(?hypname, ?thy(term), -integer).
%%  mode split_right(?hypname, ?thy(term), -integer).
%%
split_left(Hyp, A, No) :-
    get_hyp(Hyp, hrec(N, ERGO_AND(A,_), _, _, CDep, UDep)), 
    add_hyp(A, No, sl(N), CDep, UDep).

split_right(Hyp, B, No) :-
    get_hyp(Hyp, hrec(N, ERGO_AND(_,B), _, _, CDep, UDep)), 
    add_hyp(B, No, sr(N), CDep, UDep).



%% decode_hyp(HypDescriptor, HypRecord).
%%
%%  Produce a partially instantiated HypRecord which contains all of the
%%  information present in the HypDescriptor.
%%
%%  mode decode_hyp(?hypname, ?hrec).
%%
decode_hyp(N,                  hrec(N2,_,_,_,_,_)) :- var(N),     !, N=N2.
decode_hyp(N,                  hrec(N2,_,_,_,_,_)) :- integer(N), !, N=N2.
decode_hyp(hyp:N,              hrec(N2,_,_,_,_,_)) :- integer(N), !, N=N2.
decode_hyp(hyp:N,              hrec(N2,_,_,_,_,_)) :- var(N),
                                                      \+ get_var_name(N,_),
                                                      !,
                                                      N=N2.
decode_hyp(hyp:T,              hrec(_,T2,_,_,_,_)) :- !, T=T2.
%% Next one is for internal use only.
decode_hyp(hrec(N,T,D,J,C,U),  hrec(N,T,D,J,C,U)).


%% private get_hyp(HypDescriptor, HypRecord).
%% private get_hyp(HypDescriptor, HypRecord, HypList).
%%
%%  Find an undeleted hypothesis record HypRecord which is compatible with
%%  the descriptor HypDescriptor (see decode_hyp/2) in the current, or
%%  supplied, hypothesis list.
%%
%%  mode get_hyp(?hypname, -hrec).
%%  mode get_hyp(?hypname, -hrec, +list(hrec)).
%%
get_hyp(Name, Hyp) :-
    ip_lookup(in_proof, true),
    ip_lookup(hyplist, HL),
    get_hyp(Name, Hyp, HL).

get_hyp(Name, HRec, HL) :-
    decode_hyp(Name, HRec),
    member(HRec, HL),
    hrec_not_deleted(HRec).



proof command
  delete_hyp(HypName)
extends
  hyp_proofcmds
doc
  'Deletes the hypothesis specified by HypName.',nl,
  'It fails if HypName specifies an undischarged conjecture.',nl,
  nl,
  'HypName may be an integer, or a term of the form:  hyp:T,',nl,
  'where T is an integer, or a pattern that matches one of the',nl,
  'hypotheses.'
mode
  delete_hyp(?hypname).

%%  Design Notes: (1) In the past, this could sometimes succeed *without*
%%  deleting the specified hypothesis, either because it was a conjecture
%%  or because the save_hyps flag was set.  This turned out to be a messy
%%  approach, so all future implementations of delete_hyp should succeed
%%  only when the hypothesis is deleted.  If you want an optional form of
%%  deletion, write another procedure which calls this one first then
%%  takes various other actions.
%%
%%  (2) Only the single argument form should be called from tactics,
%%  because the two argument form exposes implementation details,
%%  especially the dependency management system data structures,
%%  which may allow unsound proofs if they are instantiated incorrectly.
%%
delete_hyp(Name) :-
    ip_lookup(hyplist, HL),
    delete_hyp(Name, _, HL).

delete_hyp(Name, Hyp) :-
    ip_lookup(hyplist, HL),
    delete_hyp(Name, Hyp, HL).

delete_hyp(Name, HRec, HL) :- 
    get_hyp(Name, HRec, HL),
    incomplete_retry_delays,
    !,
    \+ hrec_not_justified(HRec),
    delete_hrec(HRec),
    hrec_numterm(HRec, No, Term),
    is_subproof_id(SPId),
    announce_event(hyp(delete,SPId,No,Term)).


check_delete_hyp(Hyp) :-
    delete_hyp(Hyp),
    !.
check_delete_hyp(Hyp) :-
    is_conjecture(Hyp, _),
    !,
    handle_error(['attempt to delete conjecture ',q(Hyp)]).
check_delete_hyp(Hyp) :-
    handle_error(['attempt to delete non-existent hypothesis ',q(Hyp)]).


%% number_of_hyps(?No).
%%
%%  What is the number of the most recent hypothesis?
%%
%%  mode number_of_hyps(?integer).
%%
number_of_hyps(No) :- ip_lookup(hypnr, No).


%% is_hypname(Name).
%%
%%  Succeeds if Name is of the correct form to be a hypothesis
%%  name.  That is, it succeeds if Name is an integer, or a term
%%  of the form hyp:H, where H is an integer or a theory term
%%  to match against some hypothesis.
%%
%%  mode is_hypname(@hypname).
%%
is_hypname(N) :-
    integer(N),
    !.
is_hypname(N) :-
    compound(N),
    functor(N, ':', 2),
    arg(1, N, Hyp),
    Hyp == hyp.


%% is_hyp(HypDescriptor, T).
%% is_hyp(HypDescriptor, No, T)
%%
%%  Search the hypothesis list of the current window.  This predicate
%%  used to perform an 'incomplete_retry_delays'.  This seemed unhelpful
%%  and has been removed.
%%
%%  mode is_hyp(?hypname, ?thy(term)).
%%  mode is_hyp(?hypname, ?integer, ?thy(term)).
%%
is_hyp(Hyp, T)     :- get_hyp(Hyp, hrec(_,T,_,_,_,_)).
is_hyp(Hyp, No, T) :- get_hyp(Hyp, hrec(No,T,_,_,_,_)).


%% is_justified_hyp(HypDescriptor, T).
%% is_justified_hyp(HypDescriptor, No, T)
%%
%%  Search the hypothesis list of the current window for justified
%%  hypotheses (nonconjectures).
%%
%%  mode is_justified_hyp(?hypname, ?thy(term)).
%%  mode is_justified_hyp(?hypname, ?integer, ?thy(term)).
%%
is_justified_hyp(Hyp, T)      :-
    is_justified_hyp(Hyp, _, T).

is_justified_hyp(Hyp, No, T) :-
    get_hyp(Hyp, hrec(No,T,_,Just,_,_)),
    (Just = cj(J)
      ->
        nonvar(J)
    ;
        true
    ).


%% is_conjecture(HypDescriptor, T).
%% is_conjecture(HypDescriptor, No, T)
%%
%%  Search the hypothesis list of the current window for undischarged
%%  conjectures.
%%
%%  mode is_conjecture(?hypname, ?thy(term)).
%%  mode is_conjecture(?hypname, ?integer, ?thy(term)).
%%
is_conjecture(Hyp, T)     :- get_hyp(Hyp, hrec(_,T,_,cj(J),_,_)), var(J).
is_conjecture(Hyp, No, T) :- get_hyp(Hyp, hrec(No,T,_,cj(J),_,_)), var(J).


%% is_hyp_or_theorem(RuleName0, RuleName, Stmt).
%%
%%  A predicate to find and return information on any of the
%%  'beliefs' (axioms through to conjectures) at the current
%%  point of a proof.
%%
%%  mode is_hyp_or_theorem(?rulename, -fullrulename, ?thy(term)).
%%
is_hyp_or_theorem(Name0, Name, Stmt) :-
    get_th_or_hyp(Name0, Name, Stmt, _, _, _).
                  

%% private get_th_or_hyp(Rule0, Rule, Form, Body, CDep, UDep).
%%
%%  A predicate to find and return information on any of the
%%  'beliefs' (axioms through to conjectures) at a particular point 
%%  during a proof.
%%
%%  This predicate is for internal use only, because it is likely to
%%  disappear or change its interface/behaviour.  In particular,
%%  a nasty current feature is that it returns the body of an ordinary
%%  theorem, but executes the body of a compiled rule.
%%
%%  mode get_th_or_hyp(?rulename, ?fullrulename, ?thy(term), ?code,
%%                      -cdep_dms, -udep_dms).
%%
get_th_or_hyp(Hyp, hyp:No, T, true, CDep, UDep) :- 
    get_hyp(Hyp, hrec(No,T,_,_,CDep,UDep)).
get_th_or_hyp(Rule, ProofRule, T, Body, CDep, UDep) :- 
    decode_rule(Rule, CD, RuleName),
    (
        ProofRule = proof:RuleName,
        CD = proof,
        cdep_empty(CDep),
        Body = true,
        is_ass(RuleName, T, UDep)
    ;
        look_for_theorem(CD:RuleName, T),
        (
            ProofRule = CD:RuleName,
            cdep_empty(CDep),
            udep_empty(UDep),
            rule_at(CD, RuleName, T, [], Body)
        ;
            decode_comp_rule(RuleName, CRuleName, Trace),
            once(eval_compiled_rule(CD,CRuleName,_,_,T,CDep,UDep,Trace)),
            ProofRule = CD:cr(CRuleName, Trace),
            Body = true
        )
    ).



%% save_hyps(Tactic).
%% save_hyps(Tactic, Flag).
%%
%%  Now obsolete.
%%
%%  mode save_hyps(?tactic).
%%  mode save_hyps(?tactic,@truefalse).
%%
save_hyps(Code) :-
    writelist(['WARNING: save_hyps/1 is obsolete.',nl]),
    call_tac(Code).
save_hyps(Code,_) :-
    writelist(['WARNING: save_hyps/2 is obsolete.',nl]),
    call_tac(Code).


proof command
  conjecture(Term)
extends
  hyp_proofcmds
doc
  'Adds Term as a conjecture (an unjustified hypothesis).',nl,
  'Term is checked for legal syntax and new variables.'
mode
  conjecture(@thy(term)).

conjecture(Term) :-
    conjecture(Term, _).

conjecture(Term, N) :- 
    check_term(Term),
    assimilate_new_vars([Term]),
    add_conj(Term, N, _, _),
    !.



%% used_in(HypA, 'focus').
%% used_in(HypA, HypB).
%%
%%  Was hypothesis HypA used in the production of HypB, where HypB 
%%  is either the number of another hypothesis, or 'focus'?
%%
%%  Optimised towards finding all places where 'HypA' was used.
%%
%%  Note that hypotheses are always referred to by number here.
%% 
%%  mode used_in(?integer, ?atom).
%%  mode used_in(?integer, ?integer).
%%
used_in(A, B) :-
    get_hyp(A, hrec(_,_,_,_,_,UDepA)),
    (
        B = focus,
        ip_lookup(focususe, UDepB)
    ;
        B \== focus,
        ip_lookup(hyplist, HL),
        get_hyp(B, hrec(_,_,_,_,_,UDepB), HL)
    ),
    udep_depends_on(UDepB, UDepA).



%% is_ass(Letter, Term).
%% is_ass(Letter, Term, UDMSTag).
%%
%%  Is there a proof assumption named Letter which states Term?
%%
%%  mode is_ass(?atom, ?thy(term)).
%%  mode is_ass(?atom, ?thy(term), -udep_node).
%%
is_ass(L, T) :- is_ass(L, T, _).
is_ass(L, T, U) :-
    ip_lookup(assumptions, Ass),
    member(ass(L,T,U), Ass).


proof command
  conjecture_subterm(Path,Relation,Term)
extends
  hyp_proofcmds
doc
  'If Subterm is the subterm of the current focus',nl,
  'specified by the subterm path Path, Term is an',nl,
  'ergo term, and Relation is a binary relation, then the',nl,
  'conjecture Relation(Subterm,Term) is made.'
mode
  conjecture_subterm(@subterm_path,@relation,@thy(term)).

conjecture_subterm(Path,Relation,Term) :-
   focus(Path,empty,['is_focus'(F)]),
   conjecture(Relation(F,Term)).
