%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% embryonic support for POSTULATES

%% A postulate is a rule that is asserted without proof (like an axiom), but
%% for which a proof can later be furnished (like a theorem).  This file
%% defines commands to add and to prove postulates.  Later, it should be
%% possible to enable and disable the use of postulates, both individually and
%% _en masse_.  When a postulate is disabled, neither that postulate nor any
%% theorem that depends on it should be usable.  Disabling postulates is
%% essential to the soundness of prove_postulate, otherwise unsound circular
%% inference is possible.

% postulate(Name, AssumpList, Stmt, Body).
% mode postulate(@atom, @list(term), @term, @nonvar).

% from (prover/dbms/theory_data)add_axiom/4
add_postulate(Name, Ass, Stmt, Body) :-
    	check_current_theory(Thy),
	check_atom(Name),
	check_no_name_clash_rule(Thy, Name),
	check_or_writelist(
	    (
		nonvar(Stmt)
	    ;
	        Ass=[_|_]
	    ),
	    ['Error: postulate ''',Name,''' is too general.',nl]),
	check_term(Stmt),
	check_terms(Ass),
	collect_vars((Ass,Stmt), Vars),
	check_axiom_body(Body, Vars),
	assert_theory(rule(Name, postulate, Ass, Stmt, Body), Thy).

% prove_postulate(Name, Tact)
% mode prove_postulate(@atom, @term).
prove_postulate(Name, Tact) :-
    check_current_theory(Thy),
    check_atom(Name),
    ( retract_theory(rule(Name, postulate, Ass, Stmt, Body), Thy) ->
        true
    ; writelist(['Error: ', Name, ' is not a postulate in this theory.', nl]),
      fail
    ),
    writelist(['Warning: avoid circular reasoning!', nl]),
    ( with_disabled_postulates(
            [Thy:Name],
            add_theorem(Name, Ass, Stmt, ask, (Tact, Body))),
      query_theory(rule(Name, theorem(Deps), _, _, _), Thy),
      ( member(Thy:Name, Deps) ->
          writelist(['Error: you used circular reasoning!',nl]),
          retract_theory(rule(Name, _,_,_,_), Thy),
          fail
      ; true
      ) ->
          writelist(['Proved postulate ', Name, nl])
    ; add_postulate(Name, Ass, Stmt, Body) ->
        writelist(['Failed to prove postulate ', Name, nl]),
        fail
    ; writelist(['Oops, your postulate ', Name, ' got lost.  Sorry!', nl]),
      fail
    ).

%%%%% BUG ALERT! %%%%%%
%% NB: Current implementation of dependency lists is INADEQUATE.
%% Example:
%%   define postulates P1 and P2.
%%   prove theorem T1 using P1: deps(T1)=[P1].
%%   discharge P1 using P2: deps(P1)=[P2], but deps(T1) *is not updated*.
%%   discharge P2 *erroneously* using T1.
%% Suggested fix:
%%    - don't keep transitive closure of deps, calculate it as needed.
%%    - keep separate deps facts
%%    - check deps for current theory only.
%%%%% BUG ALERT! %%%%%%

% Enabling and disabling postulates.  In all cases, Posts may be a list of
% Thy:Name, or the atom 'all'.
% disabled_postulates(Posts) -- return currently disabled postulates.
% set_disabled_postulates(Posts) -- set currently disabled postulates.
% disable_postulates(Posts) -- add to currently disabled postulates.
% enable_postulates(Posts) -- remove from currently disabled postulates.
% with_disabled_postulates(Posts, Goal) -- call Goal with Posts disabled.
%%disabled_postulates(Posts) :-
%%    query_global(disabled_postulates(Posts)).

%%set_disabled_postulates(Posts) :-
%%    retractall_global(disabled_postulates(_)),
%%    assert_global(disabled_postulates(Posts)).

%%disable_postulates(Posts) :-
%%    disabled_postulates(Old),
%%    ( Old = all ->
%%        true
%%    ; Posts = all ->
%%        set_disabled_postulates(all)
%%    ; union_list(Old, Posts, New),
%%      set_disabled_postulates(New)
%%    ).

%%enable_postulates(Posts) :-
%%    disabled_postulates(Old),
%%    ( Old = all ->
%%        check_or_writelist(
%%            Posts = all,
%%            ['Error: cannot selectively enable ',
%%             'when all postulates are disabled.', nl]),
%%        set_disabled_postulates([])
%%    ; Posts = all ->
%%        set_disabled_postulates([])
%%    ; diff_list(Old, Posts, New),
%%        set_disabled_postulates(New)
%%    ).
    
%%with_disabled_postulates(Posts, Goal) :-
%%    disabled_postulates(DPs),
%%    unwind_protect((disable_postulates(Posts), call(Goal)),
%%                   set_disabled_postulates(DPs)).
with_disabled_postulates(Posts, Goal) :-
    call(Goal).

