%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: prove.ql,v 1.82 2001/06/13 03:21:20 itp Exp $

#include "../lib/ergo.h"



%% add_theorem(Name, Ass, Term, Save, Code).
%%
%%  Add a theorem to the current theory or a subtheory thereof.
%%  Succeeds if executing Code proves the theorem Ass |- Term.
%%
%%  Name must be an atom or a Sub:Name pair, where Sub is a subtheory of the
%%  current theory.  The Name must not already in use for any rule in the
%%  current theory.
%%
%%  Ass must be a closed list of terms that are legal in the current 
%%  theory.  Currently, only the empty list is supported.
%%
%%  Term must be a term that is legal in the current theory.
%%
%%  Save must be one of the atoms: yes, no, or ask.
%%  After the theorem is proved successfully, 'yes' means it will be
%%  added to the current theory, 'no' means it will not be added and
%%  'ask' means that the user will be asked whether or not it should
%%  be added.
%%
%%  Code must be a sequence of ',' separated tactic calls,
%%  proof commands, and variables.  Variables cause the current
%%  prover to be called (see set_prover/2).
%%
%%  mode add_theorem(@nonvar, @list(term), @term, @atom, ?term).
%%
add_theorem(QName, Ass, Term, Ask, Tact) :-
    check_or_error(
        member_var(Ask, [yes,no,ask]),
        [q(Ask),' must be yes, no, or ask.']),
    decode_subthy_rule(QName, Dest, Name),
    check_legal_rule_name(Name),
    check_no_name_clash_rule(Dest, Name),
    \+ \+ (proof_init(Dest, Name, Ass, Term),
           announce_event(proof(Dest,Name,started)),
           add_theorem2(Dest, Name, Ass, Term, Ask, Tact)).


add_theorem2(Dest, Name, Ass, Term, Ask, Tact) :-
    debug(add_theorem,['about to call ', w(Tact), ' to prove ', w(Dest:Name)]),
    once(call_tac(Tact)),
    proof_finish(Term, Proof, Body, FU),
    debug(add_theorem,['proof_finish gives proof: ',q(Proof),nl,
                       'body: ',q(Body),nl,
                       'and focus usage: ',q(FU)]),
    !,
    (confirm_add_theorem(Ask, Dest:Name, Term, Body) ->
        debug(add_theorem,['starting to save proof record...']),
        save_proof_record(Dest, Name, theorem, Term, Body, Proof, FU),
        debug(add_theorem,['calculating dependencies...']),
        get_theorem_dependencies(Proof, Deps),
        debug(add_theorem,['dependencies are:',q(Deps)]),
        assert_axthm(Name, theorem(Deps), Ass, Term, Body, Dest),
        announce_event(proof(Dest,Name,finished(true)))
    ;
        msg(ergo, 1, ['Theorem ',q(Dest:Name),' was not added.']),
        announce_event(proof(Dest,Name,finished(false)))
    ).
add_theorem2(Dest, Name, _, _, _, _) :-
    announce_event(proof(Dest,Name,failed)),
    fail.
                   


proof_init(Thy, Name, Ass, Term) :-
    check_terms(Ass),
    check_term(Term),
    name_assumptions(Ass, Ass1),
    add_proof_name(Thy:Name),
    % ip_set(save_proof_record, true), %% set save_proof default
    ip_set(in_proof,     true),
    ip_set(newvars,      []),   %% Some of the variables introduced 
                                %% during proof.  Always frozen.
    ip_set(newvar_flag,  true), %% True if newvars contains all 
                                %% the new variables (and therefore
                                %% they are all frozen).
    ip_set(fixedvars,    []),
    internal_promote_var(Term), 
    accept_constraints,         %% collects known proofvars constraints.
    simplify_relation(rev(ERGO_IMPLIES), Reln),
    ip_set(assumptions, Ass1),  %% LHS of the /- operator
    ip_set(wrapnr,      0),     %% Number of most recent wrapper
    ctr_set(maxproofid,  0),     %% Max subproof id. so far in this proof.
    ctr_set(maxfocusid,  0),     %% Max pstep/3 id. so far in this proof.
    ip_set(subproofstack, []),
    push_subproof(Term,Reln,ERGO_TRUE),
    debug(subproof, ['done first push']),
    itrans_abort,
    !.


%% proof_finish(Stmt, Proof, Body, FocusUsage).
%%
%%  Succeeds if the current proof window (which must be the outermost
%%  window of the proof) is complete and does not depend upon unjustified
%%  conjectures, and the current proof constraints (not_free_in, 
%%  object variable distinctness etc.) have not overspecialised the proof.
%%
%%  mode proof_finish(@thy(term), -compound, -list(compound), -list(compound)).
%%
proof_finish(Stmt, Proof, Body, FU) :-
    pop_subproof(ERGO_TRUE, _, FU, Proof),
    ip_set(finished_proof, Proof),
    %% Promote all newvars that appear in Stmt.
    retry_delays,
    internal_promote_var(Stmt),
    ensure_constraints_accepted(
        ['All recent constraints must be accepted for',nl,
         'this proof to be successfully completed.',nl]),
    %% determine which nfi constaints are part of the theorem.
    ip_lookup(fixedvars, Vars),
    collect_constraints(Vars, ObvarNFIs, NFIs, Funnies),
    check_or_error(
        Funnies = [],
        ['unknown constraints: ',q(Funnies)]),
    append(ObvarNFIs, NFIs, AllNFIs),
    list_to_comma(AllNFIs, Body1),
    prolog_code_diff(Body1, true, Body).




%% confirm_add_theorem(Ask, FullName, Term, Body).
%%
%%  Succeeds iff the theorem should be added to the current theory.
%%  Ask must be 'yes', 'no' or 'ask'.
%%
%%  mode confirm_add_theorem(@atom, @fullrulename, @term, @compound).
%%
confirm_add_theorem(yes, _, _, _) :-
    !.
confirm_add_theorem(ask, Name, Term, Body) :-
    question(['The resultant theorem ',q(Name),nl,
              '=== ',thy(Term),nl,
              '    provided ', code(Body),nl,nl,
              'Do you wish to keep this result? '],
             yes).



%% add_proof_name(Thy:Name)
%%
%%  Consult the list of proofs in progress to ensure that we don't
%%  attempt to prove two theorems with the same name.  Add the given
%%  name to the list if it does not clash.
%% 
%%  mode add_proof_name(@gcomp).
%%
add_proof_name(Name) :-
    Name = Thy:Thm,
    check_atom(Thy),
    check_atom(Thm),
    ip_lookup(proofnames, NameList),
    check_or_error(
        \+  member(Name, NameList), 
        ['there is a proof in progress with this name.']),
    ip_set(proofnames, [Name|NameList]).



%% private name_assumptions(AssumptionTerms, Assumptions).
%%
%%  Name the assumptions (terms from the LHS of '/-') with
%%  lowercase letters. 
%%
%%  mode name_assumptions(@list(term), -list(compound)).
%%
name_assumptions(A0, A) :-
    name_assumptions(A0, A, 97).

name_assumptions([], [], _).
name_assumptions([A|B], [ass(Name, A, UNode)|NB], N) :-
    name(Name, [N]),
    Next is N + 1,
    check_or_error(
        Next < 97+26,
        ['too many assumptions (max. is 26).']),
    udep_create_empty(UNode),
    name_assumptions(B, NB, Next).


%% is_goal(Goal).
%% is_reln(Reln).
%% is_focus(Focus).
%% is_focus_id(Id).
%% is_subproof_depth(Depth).
%% is_subproof_id(Id).
%%
%%  Get various information about the current subproof.
%%
%%  The focus id. is a number that uniquely identifies the most
%%  recent line of the current subproof (ie. the line whose term
%%  is returned by is_focus/1).
%%
%%  The subproof id. is a number that uniquely identifies the
%%  current subproof.
%%
%%  mode is_goal(?term).
%%  mode is_reln(?nonvar).
%%  mode is_focus(?term).
%%  mode is_focus_id(?integer).
%%  mode is_subproof_depth(?integer).
%%  mode is_subproof_id(?integer).
%%
is_goal(Goal)      :- ip_lookup(goal,      Goal).
is_reln(R)         :- ip_lookup(relation,  R).
is_focus(Focus)    :- ip_lookup(proof,     Prf),
                      Prf = [pstep(_,Focus,_)|_].
is_focus_id(Id)    :- ip_lookup(proof,     Prf),
                      Prf = [pstep(Id,_,_)|_].
is_subproof_depth(D) :- ip_lookup(proofdepth, D).
is_subproof_id(Id)   :- ip_lookup(proofid, Id).

proof command
  nfi(Var,Term)
extends
  nfi_proofcmds
doc
  'Adds the constraint Var not_free_in Term to the current proof.'
mode
  nfi(@obvar,@thy(term)).

nfi(Var, Term) :-
    check_or_error(
        obvar(Var),
        [q(Var),' is not an object variable.']),
    check_or_error(
        constraints_accepted,
        ['new constraints cannot be added until other recently generated',nl,
         'constraints have been accepted or satisfied in some way.']),
    Var not_free_in Term,
    accept_constraints.


%% initialize_proof_structure.
%% reinitialize_proof_structure.
%%
%%  initialize_proof_structure/0 must be called once when the
%%  system is started, before invoking the (typically 
%%  repeat-fail) command line interpreter.  The values set here 
%%  are thus the default values when outside a proof.
%%
%%  reinitialize_proof_structure/0 is similar, but can be
%%  called at any time.
%%
%%  mode initialize_proof_structure.
%%
initialize_proof_structure :-
    ip_set(proofnames, []),
    reinitialize_proof_structure.

reinitialize_proof_structure :-
    ip_set(proof, fail),
    ip_set(proofid, 0),
    ip_set(subproofstack, []),
    ctr_set(maxproofid, 0),
    ctr_set(maxfocusid, 0),
    ip_set(proofdepth, 0),
    ip_set(goal, fail),
    ip_set(relation,  fail),
    ip_set(hypnr, 0),
    ip_set(wrapnr, 0),
    ip_set(assumptions, []),
    ip_set(in_proof, false),
    ip_set(fixedvars, []),
    ip_set(newvars, []),
    ip_set(newvar_flag, true),
    dynamic(hfn/2).



%% private proof_suspend(Code).
%%
%%  Suspend the current proof, execute Code, then return
%%  to the proof.
%%
%%  mode proof_suspend(+code).
%%
proof_suspend(Code) :-
    in_proof(Thy:Name),
    ctr_get(maxproofid, MaxProofId),
    ctr_get(maxfocusid, MaxFocusId),
    announce_event(proof(Thy,Name,suspended)),
    unwind_protect(
        \+ \+ proof_suspend2(Code),
        ( ctr_set(maxproofid, MaxProofId),
          ctr_set(maxfocusid, MaxFocusId),
          announce_event(proof(Thy,Name,resumed))
       )).

proof_suspend2(Code) :-
    reinitialize_proof_structure,
    call_predicate(Code).


proof command
  suspend
doc
  'Suspend the current proof and start up a new theory-level user',nl,
  'interface.  After that interface is exited, the current theory will',nl,
  'be reselected and the suspended proof will be resumed.'
mode
  suspend
implementation
  proof_suspend.

proof_suspend :-
    in_proof(Name),
    writelist(['Starting up another theory management user interface.',nl,
               'Proof of ',q(Name),' will resume when you exit.',nl]),
    proof_suspend(visit_theory_interface),
    writelist(['Resuming proof of ',q(Name),'.',nl,nl]).


%% private visit_theory_interface.
%%
visit_theory_interface :-
    global_state_lookup(current_dir,OrigThy), OrigThy \= none,
    user_interface,
    %% We cd back to OrigThy in case the user has changed theories.
    cd(OrigThy).


%% get_theorem_dependencies(Proof, Deps).
%%
%%  Build a list of the axioms, theorems and postulates upon 
%%  which this proof depends.
%%
%%  mode get_theorem_dependencies(@term, -list(term)).
%%
get_theorem_dependencies(Proof, Deps) :-
    ip_set(thm_deps, []),
    apply_proof_rules(Proof, add_thm_deps),
    ip_lookup(thm_deps, Deps1),
    sort(Deps1, Deps),
    msg(add_theorem, 2, ['Dependencies: ', q(Deps)]).


apply_proof_rules(subprf(_,HL,FL,_), Cmd) :-
    !,
    apply_focuslist_rules(FL, Cmd),
    apply_hyplist_rules(HL, Cmd).
apply_proof_rules(Proof, _) :-
    handle_error(program, ['unknown proof record: ',q(Proof)]).

apply_focuslist_rules([], _).
apply_focuslist_rules([pstep(_,_,J)|FL], Cmd) :-
    apply_focus_rules(J, Cmd),
    !,
    apply_focuslist_rules(FL, Cmd).

apply_focus_rules(new, Cmd).
apply_focus_rules(tr(Rule, _, _, Ants), Cmd) :-
    Cmd(Rule),
    map(Cmd, [Ants]).
apply_focus_rules(op(_, Proof), Cmd) :-
    apply_proof_rules(Proof, Cmd).
apply_focus_rules(wr(_,_,_,subprf(_,_,FL,_)), Cmd) :-
    apply_focuslist_rules(FL, Cmd).
apply_focus_rules(F, _) :-
    handle_error(program, ['unknown focus proof step: ',q(F)]).

apply_hyplist_rules([], _).
apply_hyplist_rules([H|HL], Cmd) :-
    hrec_just(H, J),
    apply_hypjust_rules(J, Cmd),
    !,
    apply_hyplist_rules(HL, Cmd).

apply_hypjust_rules(cj(tr(Rule,_,_,Ants)), Cmd) :-
    Cmd(Rule),
    map(Cmd, [Ants]).
apply_hypjust_rules(cj(Proof), Cmd) :-
    apply_proof_rules(Proof, Cmd).
apply_hypjust_rules(oh(_,Proof), Cmd) :-
    apply_proof_rules(Proof, Cmd).
apply_hypjust_rules(ht(Input,Rule,_,Ants), Cmd) :-
    Cmd(Input),
    Cmd(Rule),
    map(Cmd, [Ants]).
apply_hypjust_rules(i, _).
apply_hypjust_rules(new, _).
apply_hypjust_rules(sl(_), _).
apply_hypjust_rules(sr(_), _).
apply_hypjust_rules(tr(_), _).
apply_hypjust_rules(Just, _) :-
    handle_error(program, ['unknown hyp justification: ',q(Just)]).


add_thm_deps(hyp:_) :-
    !.
add_thm_deps(proof:_) :-
    !.
add_thm_deps(true)  :-
    !.
add_thm_deps(Thy:Name) :-
    !,
    ( dependency_theory(Thy) ->
        ip_lookup(thm_deps, L),
        ip_set(thm_deps, [Thy:Name | L])
    ;
        true
    ).
add_thm_deps(Rule) :-
    writelist(['WARNING: proof contains strange rule name: ',q(Rule),nl]).


%% private dependency_theory(Thy).
%%
%%  True iff Thy is a theory or subtheory for which we want to
%%  record proof dependency information.  Currently, this is
%%  the current theory and all its subtheories.  We do not need
%%  to record dependencies upon theorems in ancestor theories,
%%  because the theory graph is acyclic, so no circularities 
%%  can arise from the use of such theorems.
%%
%%  mode dependency_theory(@theory).
%%
dependency_theory(Thy) :-
    % This is a special case of the second clause.
    global_state_lookup(current_dir,Thy),
    Thy \= none,
    !.
dependency_theory(Thy) :-
    global_state_lookup(current_dir,Curr),
    is_subtheory_of(Thy, Parent),
    is_subtheory_of(Curr, Parent),
    !.

is_subtheory_of(Sub, Parent) :-
    query_global(subthy_owner(Sub, Parent)).
is_subtheory_of(Parent, Parent).





