%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: trans.ql,v 1.47 2000/11/15 06:29:47 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file defines procedures that perform a single proof step
%% on the current proof.
%%
%%   trans/4 transforms the focus.
%%   h_trans/6 adds a new hypothesis.
%%   d_trans/5 discharges a conjecture.
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Implementation.
%%
%% All three commands, trans, h_trans and d_trans, have a common structure:
%%
%%      - Get information on target term (focus/hyp) including
%%              - The term
%%              - The appropriate proof repository
%%              - The CDep and UDep information
%%              - The transformation relation
%%
%%      - Call trans/11 within a thawing_new_vars wrapper
%%
%%      - Update all of the above


obsolete(Cmd) :-
    handle_error([q(Cmd),' is obsolete.']).

obsolete(Cmd,Cmd2) :-
    writelist(['WARNING: ',q(Cmd),' is obsolete.  Using ',
               q(Cmd2),' instead.',nl]),
    call_tac(Cmd2).


proof command
  trans(Rule),
  trans(Rule, Direction),
  trans(Rule, Direction, Antecedents),
  trans(Rule, Direction, Antecedents, Goal)
extends
  focus_proofcmds
doc
  'Transform the current focus using Rule, which may be a theorem',nl,
  'or axiom from some theory, or a hypothesis.',nl,
  nl,
  'Direction determines how the rule is used to transform the focus.',nl,
  'It must be one of: ''f'' (forward), ''r'' (reverse), ''t'' (true) or',nl,
  '''rt'' (reverse true) and it defaults to ''f''.',nl,
  nl,
  'If Antecedents is an empty list (the default), then Rule will be',nl,
  'treated as an unconditional rule.  Otherwise, Antecedents must be',nl,
  'a non-empty list of rule names which will be used to discharge the',nl,
  'conditions of Rule.',nl,
  nl,
  'Goal is unified with the result of the transformation (the new focus).'
mode
  trans(+rulename),
  trans(+rulename, +direction),
  trans(+rulename, +direction, +list(rulename)),
  trans(+rulename, +direction, +list(rulename), ?thy(term)).

trans(Rule) :-
    trans(Rule, f, [], _).

trans(Rule, Ant) :-
    list(Ant),
    !,
    obsolete(trans(Rule, Ant),
             trans(Rule, f, Ant)).
trans(Rule, Dir) :-
    trans(Rule, Dir, [], _).

trans(Rule, Dir, Ant) :-
    trans(Rule, Dir, Ant, _).

trans(Rule, Dir, G, Ant) :-
    %% Check for old-style trans/4 calls.
    list(Ant),
    \+ list(G),
    !,
    writelist(['WARNING: trans/4 arguments were in the old order.',nl,
               '  They should be: trans(Rule,Dir,AntRules,Goal).',nl,
               '  I will swap AntRules<-->Goal and continue...',nl]),
    trans(Rule, Dir, Ant, G).
trans(Rule, Dir, Ant, G) :-
    is_focus(F),
    ip_lookup(relation,  Rel),
    thawing_new_vars( 
            trans(Rule, Rule2, Dir, Rel, F, true, G, Ant, Ant2, CD, UD)
                    ),
    add_proof_step(G, tr(Rule2, Dir, G, Ant2), CD, UD).


proof command
  h_trans(Name, Rule),
  h_trans(Name, Rule, Direction),
  h_trans(Name, Rule, Direction, Antecedents),
  h_trans(Name, Rule, Direction, Antecedents, Goal),
  h_trans(Name, Rule, Direction, Antecedents, Goal, NewHypNum)
extends
  hyp_proofcmds
doc
  'Uses Rule to prove a new hypothesis whose statement is Goal',nl,
  'and whose number is NewHypNum.',nl,
  nl,
  'Direction and Antecedents specify how Rule should be used.',nl,
  'Direction defaults to ''f'' (forward).',nl,
  'Antecedents defaults to [] (unconditional).',nl,
  nl,
  'Name specifies where the input of Rule comes from (typically',nl,
  'another hypothesis or ''true'') and Goal is unified with the output',nl,
  'of Rule (which becomes the new hypothesis).'
mode
  h_trans(+rulename, +rulename),
  h_trans(+rulename, +rulename, +direction),
  h_trans(+rulename, +rulename, +direction, ?list(rulename)),
  h_trans(+rulename, +rulename, +direction, ?list(rulename), ?thy(term)),
  h_trans(+rulename, +rulename, +direction, ?list(rulename), ?thy(term), -integer).

h_trans(Name, Rule) :-
    h_trans(Name, Rule, f, [], _, _).

h_trans(Name, Rule, Ant) :-
    list(Ant),
    !,
    obsolete(h_trans(Name, Rule, Ant),
             h_trans(Name, Rule, f, Ant)).
h_trans(Name, Rule, Dir) :-
    h_trans(Name, Rule, Dir, [], _, _).

h_trans(Name, Rule, Dir, Ant) :-
    h_trans(Name, Rule, Dir, Ant, _, _).

h_trans(Name, Rule, Dir, Ant, Goal) :-
    h_trans(Name, Rule, Dir, Ant, Goal, _).

h_trans(Name, Rule, Dir, G, Ant, New) :-
    %% Check for old-style h_trans/6 calls.
    list(Ant),
    \+ list(G),
    !,
    writelist(['WARNING: h_trans/6 arguments were in the old order.',nl,
               '  They should be: h_trans(Name,Rule,Dir,AntRules,Goal,_).',nl,
               '  I will swap AntRules<-->Goal and continue...',nl]),
    h_trans(Name, Rule, Dir, Ant, G, New).
h_trans(Name, Rule, Dir, Ant, G, New) :-
    (Name == true, Dir = rt ->
        Name2 = ergo:true,
        F  = ERGO_TRUE,
        FBody = true,
        cdep_empty(CDep0),
        udep_empty(UDep0)
    ; decode_hyp(Name, HREC) ->
        get_hyp(Name, HREC),
        FBody = true,
        HREC = hrec(Num, F, _, _, CDep0, UDep0),
        Name2 = hyp:Num
    ;
        get_th_or_hyp(Name, Name2, F, FBody, CDep0, UDep0)
    ),
    !,
    thawing_new_vars(
            trans(Rule, Rule2, Dir, ERGO_IMPLIES, 
                  F, FBody, G, Ant, Ant2, CD, UD)
                    ),
    cdep_union(CD, CDep0, CDep),
    add_hyp(G, New, ht(Name2, Rule2, Dir, Ant2), CDep, [UDep0|UD]).


proof command
  take_theorem(Name),
  take_theorem(Name, Form),
  take_theorem(Name, Form, NewHypNumber)
extends
  hyp_proofcmds
doc
  'Adds a copy of the rule called Name as a new hypothesis',nl,
  '(NewHypNumber) of the current window.  The new hypothesis is',nl,
  'unified with Form.  This allows rule Name to be instantiated.',nl,
  nl,
  'take_theorem was a separate proof command in earlier versions of',nl,
  'Ergo, but is now an abbreviation for h_trans(true,Name,rt,[],...).'
mode
  take_theorem(+rulename),
  take_theorem(+rulename, ?thy(term)),
  take_theorem(+rulename, ?thy(term), -integer).

take_theorem(Name)                    :- h_trans(true, Name, rt, [], _, _).
take_theorem(Name, Th)                :- h_trans(true, Name, rt, [], Th, _).
take_theorem(Name, Th, No)            :- h_trans(true, Name, rt, [], Th, No).


proof command
  d_trans(Name, Rule),
  d_trans(Name, Rule, Direction),
  d_trans(Name, Rule, Direction, Antecedents)
extends
  hyp_proofcmds
doc
  'Discharge the conjecture Name with a single application of Rule.',nl,
  'The intention when discharging is to show (HypTerm <= true).',nl,
  'This proves that the conjecture is justified within the current',nl,
  'window.  Thus the effect of d_trans is to convert the conjecture',nl,
  'Name into a normal hypothesis.',nl,
  nl,
  'Direction and Antecedents specify how Rule should be used.',nl,
  'Direction defaults to ''f'' (forward).',nl,
  'Antecedents defaults to [] (unconditional).'
mode
  d_trans(+hypname, +rulename),
  d_trans(+hypname, +rulename, +direction),
  d_trans(+hypname, +rulename, +direction, +list(rulename)).

d_trans(Name, Rule) :-
    d_trans(Name, Rule, f, []).

d_trans(Name, Rule, Ant) :-
    list(Ant),
    !,
    obsolete(d_trans(Name, Rule, Ant),
             d_trans(Name, Rule, f, Ant)).
d_trans(Name, Rule, Dir) :-
    d_trans(Name, Rule, Dir, []).

d_trans(Name, Rule, Dir, true, Ant) :-
    %% Check for old-style d_trans/5 calls.
    list(Ant),
    !,
    writelist(['WARNING: obsolete call to d_trans/5.',nl,
               '  The call should be d_trans(Name,Rule,Dir,AntRules).',nl,
               '  I will remove redundant goal and continue...',nl]),
    d_trans(Name, Rule, Dir, Ant).
d_trans(Name, Rule, Dir, Ant) :-
    get_hyp(Name, HREC),
    hrec_not_justified(HREC),
    hrec_term(HREC, F),
    thawing_new_vars(
            trans(Rule,Rule2,Dir,rev(ERGO_IMPLIES),
                  F,true,ERGO_TRUE,Ant,Ant2,CD,UD)
                    ),
    justify_hrec(HREC, tr(Rule2, Dir, ERGO_TRUE, Ant2), CD, UD).



%% trans(+Rule0, -Rule, ?Dir, +Rel, +F, +FBody, ?G, +Ant0, -Ant, -CDep, -UDep).
%%
%%   The common portion of all the trans'.  This isn't quite true, as they
%%   also all wrap the call to this predicate inside 'thawing_new_vars'.
%%
%%   The call to 'thawing_new_vars' is pushed into each of the calling 
%%   predicates to force the compiler to compile this code segment, which 
%%   would otherwise remain interpreted.
%%
%%   It is also important to realize that the code here relies heavily upon
%%   knowledge of the internal workings of the newvar mechanism.  
%%
%%  mode trans(+rulename, -fullrulename, ?direction, +nonvar, ?thy(term),+code,
%%     ?thy(term), +list(rulename), -list(fullrulename), -cdep_dms,-udep_dms).
%%
trans(Rule, Rule2, Direction, Rel, F, FBody, G, Ant, Ant2, CDep, UDep) :-
    \+ in_itrans,
    list(Ant),
    nonvar(Rule),
    choose_conditional(Ant, RuleForm, Conds, Concl),
    choose_direction(Direction, Concl, F, RuleRel, GForm),
    debug(trans, ['rulename=',q(Rule),',  skeleton=',q(RuleForm)]),
    get_th_or_hyp(Rule, Rule2, RuleForm, Body, D1, U1),
    debug(trans, ['get_th_or_hyp returns:',q(Rule2),' === ',thy(RuleForm),nl,
                  'body=',code(Body),nl,
                  'cdep=',q(D1),', udep=',q(U1)]),
    debug(trans, ['checking that antecedents: ',thy(Conds),nl,
                  'are discharged by rulelist: ',q(Ant)]),
    check_ant(Ant, Ant2, AntBody, Conds, D2, U2),
    debug(trans, ['check_ant returns cdep=',q(D2),',  udep=',q(U2)]),
    debug(trans, ['checking that ',q(RuleRel),' is stronger than ',q(Rel)]),
    is_ground_relation(RuleRel),
    is_stronger_than(RuleRel, Rel), 
    cdep_union(D1,D2,CDep),
    udep_union(U1,U2,UDep),
    debug(trans, ['attempting to unify output (',thy(GForm),
                  ') with goal (',thy(G),').']),
    collect_vars(GForm, GFVars),
    thawed_sublist(GFVars, [], NewVars),
    ip_set(newvars, NewVars),
    G = GForm,
    debug(trans, ['executing side conditions']),
    once((Body,FBody,AntBody)),
    debug(trans, ['resolving any delayed unifications']),
    incomplete_retry_delays,
    debug(trans, ['trans/11 is complete!']),
    !.

is_ground_relation(rev(R)) :-
    atom(R).
is_ground_relation(R) :-
    atom(R).


%% choose_conditional(Ant, Form, Conds, Concl).
%%
%%  This procedure splits up the entire statement of a rule (Form)
%%  into the conditions (Conds) and the conclusion (Concl),
%%  according to whether or not Ant is an empty or non-empty list.
%%  It can also be used to construct Form from Conds and Concl.
%%
%%  mode choose_conditional(?list(term), ?term, ?term, ?term).
%%
choose_conditional([],     Form,              ERGO_TRUE,  Form ).
choose_conditional([_|_],  ERGO_IMPLIES(A,B), A,          B    ).


%% choose_direction(?Direction, ?Concl, ?Input, ?RuleRel, ?Output).
%%
%%  This procedure splits up Concl into Input, RuleRel and Output,
%%  according to Direction.  It can also construct Concl from its parts.
%%  Direction will be one of: 'f', 'r', 't' or 'rt'.
%%
%%  mode choose_direction(?direction, ?term, ?term, ?term, ?term).
%%
choose_direction(f,  Form, F,         R,           G) :- Form =.. [R,F,G].
choose_direction(r,  Form, F,         rev(R),      G) :- Form =.. [R,G,F].
choose_direction(t,  F,    F,         ERGO_EQUALS, ERGO_TRUE).
choose_direction(rt, F,    ERGO_TRUE, ERGO_EQUALS, F).


%% check_ant(AntNames0, AntNames, Body, AntPreds, CDep, UDep).
%%
%%  Discharge the antecedent term AntPreds against the list of rules
%%  AntNames0.  Dependancy information on conjectures and hypotheses 
%%  referenced by AntNames is returned in CDep and UDep.  Body is the
%%  composite QuProlog guard of the rules in AntNames.
%%
%%  mode check_ant(+list(rulename), -list(rulename), -nonvar,
%%                  +nonvar, -cdep_dms, -udep_dms).
%%
check_ant([], [], true, ERGO_TRUE, CDep, UDep) :-
    cdep_empty(CDep),
    udep_empty(UDep),
    !.
check_ant([A], [Ac], Body, Form, CDep, UDep) :-
    !,
    get_th_or_hyp(A, Ac, Form, Body, CDep, UDep).
check_ant([A,B|C], [Ac,Bc|Cc], (Body,Body1), ERGO_AND(F,F1), CDep, UDep) :-
    get_th_or_hyp(A, Ac, F, Body, CD, UD),
    !,
    check_ant([B|C], [Bc|Cc], Body1, F1, CD1, UD1),
    cdep_union(CD, CD1, CDep),
    udep_union(UD, UD1, UDep).


