%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% CVS $Id: window.ql,v 1.77 2001/06/13 03:21:20 itp Exp $


#include "../lib/ergo.h"


%% command open_pos(Path, Tactic).
%% command open_pos(Path, LeftOverPath, Tactic).
%%
%%   This predicate opens a window on the focus according to Path.  At most
%%   one window rule is applied, so it is possible that the window position
%%   is described by some prefix of Path.  The unrealized portion
%%   of Path0 is returned in LeftOverPath (open_pos/2 forces this to be []).
%%
%%   Once the window has been established, Tactic is executed.  When Tactic
%%   succeeds, the window is closed.  If Tactic fails, so does open_pos.
%%
%%  mode open_pos(+subterm_path, ?tactic).
%%  mode open_pos(+subterm_path, -subterm_path, ?tactic).
%%
open_pos(A, B) :-
    open_pos(A, [], B).

open_pos(Path0, Path, Tactic) :-
    \+ in_itrans,
    is_focus(F0),
    \+ is_param_const(F0),
    is_reln(Reln0),
    (
        query_ancestors(openrule(F0,Pos,SF0,NewH,HFn,RelFn,GFn,'--->'(SF,F))),
        debug(open_pos, ['Trying openrule(',thy(F0),',',q(Pos),
                                          ',..,',q(RelFn),').'])
    ;
        ancestor(eq),
	%% TO DO: Warning about built-in knowledge of theory graph.
        %% WARNING: built-in knowledge about current theory graph here!
        %%
        %% Use a default equality opening rule, whose soundness follows
        %% from the eq.std_eq axiom.
        Pos = [_],
        subterm_change_path(F0,Pos,SF0,SF,F),
        HFn   = ['--->'(X,X)],
        RelFn = ['--->'(ERGO_EQUALS,ERGO_EQUALS)],
        GFn   = [],  %% unknown
        NewH  = [],
        debug(open_pos, ['Default eq openrule: ',thy(F0),',',q(Pos),
                         ',',thy(SF0),'...,',thy(SF),'--->',thy(F)])
    ),        
    match_paths(Pos, Path0, Path, Obvars),
    translate_relation(RelFn, Reln0, SReln),
    !,
    is_goal(G0),
    ( member('--->'(G0,SG), GFn) ->
        true
    ;
        (SG = unknown)
    ),
    assimilate_new_vars(Obvars),
    %%
    %% Generate hypotheses of new window.
    ip_lookup(hyplist, HL),
    ip_lookup(hypnr, NH),
    inherit_hyps(HFn, HL, HL1, SHL0, Obvars),
    translate_hyps(HFn, F0, NH, HL1, SHL0, SHL, NH2, Obvars),
    %%
    %% Now set up the new subproof.
    ip_lookup(focusdep, CDep0),
    push_subproof(SF0, SReln, SG, CDep0),
    ip_set(hyplist, SHL),
    ip_set(hypnr,   NH2),
    add_new_hyps(NewH),
    incomplete_retry_delays,
    !,
    call_tac(Tactic),
    !,
    pop_subproof(SF, CDep, SUDep, Proof),
    add_proof_step(F,op(Pos,Proof),CDep,SUDep).


%% translate_relation(+Func, +Current, -New).
%%
%%   Find the new window relation New according to the relation function
%%   Func, and the current relation Current.
%%
translate_relation(Func, Current, New) :-
    member('--->'(From,To), Func),
    (
        is_stronger_than(From, Current),
        simplify_relation(To, New)
    ;
        is_stronger_than(rev(From), Current),
        simplify_relation(rev(To), New)
    ),
    !.



%% match_paths(+WinRulePath, +UserPath0, -RemPath, -Obvars).
%%
%%   Matches the head of the user specified path with a window rule path,
%%   returning the leftover user path and the obvars from any quantifiers
%%   lying on the path.
%%
match_paths([],        UP,     UP,  []).
match_paths([X|WP],    [X|UP], REM, OB) :-
    integer(X),
    !,
    match_paths(WP,UP,REM,OB).
match_paths([b(V)|WP], [U|UP], REM, OB) :-
    !,
    (
        U = 2
    ;
        U = b(V)
    ),
    !,
    append(V,VS,OB),
    match_paths(WP,UP,REM,VS).
match_paths([t(N)|WP], UP0,  REM, OB) :-
    (
        UP0 = [t(N)|UP]
    ;
        UP0 = [1,N,2|UP]
    ),
    !,
    match_paths(WP,UP,REM,OB).


%% inherit_hyps(+HypTransFunc, +ParentHyps0,
%%	        -ParentHyps, -InheritedHyps, +WinObvars).
%%
%%   Apply the given hypothesis translation function to the
%%   hypothesis list of the parent window, producing that of the child
%%   window.
%%
%%   ParentHyps0 is the original list of parent hypotheses.
%%   Its deleted hypotheses are removed, the directly-inherited
%%   ones are returned in InheritedHyps and the remaining ones
%%   are returned in ParentHyps as inh(Num,Term,CDep,UDep) terms
%%   (which are then passed to translate_hyps).
%%
%%   This is a convenient time to assert the not_free constraints generated
%%   by opening windows on quantified terms.
%%
inherit_hyps(_,  [], [], [], _).
inherit_hyps(HF, [hrec(_,_,D,_,_,_)|HL], HL2, New, Obv) :-
    D == deleted,
    !,
    inherit_hyps(HF, HL, HL2, New, Obv).
inherit_hyps(HF,[hrec(Nr,T,_,_,C0,U0)|HL], HL2,
	     [hrec(Nr,T,_,i,C0,U)|New], Obv) :-
    \+ \+ member('--->'(T,T), HF),
    gen_nfi(Obv, T),
    udep_create_discharged(U, U0),
    inherit_hyps(HF, HL, HL2, New, Obv).
inherit_hyps(HF, [hrec(Nr,T,_,_,Dep,U)|HL], [inh(Nr,T,Dep,U)|HL2], New, Obv) :-
    inherit_hyps(HF, HL, HL2, New, Obv).




%% translate_hyps(+HypTransFunc, +Focus, +HypNr0, +ParentList, 
%%                +HypList0, -HypList, -HypNr, +Obvars)
%%
%%  Uses HypTransFunc to translate hypotheses in ParentList.
%%  The resulting list of hypotheses is returned in HypList,
%%  where HypList0 is the tail of HypList.
%%
%%  Focus is here so its variables can be matched and frozen.
%%
translate_hyps(_, _, NH, [], A, A, NH, _).
translate_hyps(Hfn, Focus, NH, HL2, NHL, NNHL, NNH, Obv) :-
    abolish(hfn/2),
    \+ \+ (thaw_vars(Focus), assert(hfn(Hfn,Focus))),
    do_translate_hyps(HL2, NH, Focus, NHL, NNHL, NNH, Obv).

do_translate_hyps([], NH, _, A, A, NH, _).
do_translate_hyps([inh(Nr,T,D,U0)|HL2],NH,F,NHL,
                  [hrec(Next,NT,_,tr(Nr),D,U)|NNHL],NNH,Obv) :-
    hfn(HFn,F),
    member('--->'(T,NT), HFn),
    !,
    Next is NH + 1,
    gen_nfi(Obv, T),
    udep_create_discharged(U, U0),
    do_translate_hyps(HL2, Next, F, NHL, NNHL, NNH, Obv).
do_translate_hyps([_|HL2], NH, F, NHL, NNHL, NNH, Obv) :-
    do_translate_hyps(HL2, NH, F, NHL, NNHL, NNH, Obv).


%% gen_nfi(+ObvarList, +Term)
%%
%%  Generate multiple nfi constraints.
%%
gen_nfi([], _).
gen_nfi([A|B], T) :-
    A not_free_in T,
    gen_nfi(B, T).



%% add_new_hyps(+NewHypTerms)
%%
%%   This predicate adds introduced hypotheses to the current hypothesis
%%   list, with terms specified by the list NewHypTerms.
%%
add_new_hyps([]).
add_new_hyps([H|Rest]) :-
    cdep_empty(CDep),
    udep_empty(UDep),
    add_hyp(H,_,new,CDep,UDep),
    add_new_hyps(Rest).


%% command open_hyp(Hyp, Tactic).
%% command open_hyp(Hyp, Tactic, HypNum).
%% command open_hyp(Hyp, Tactic, HypNum, HypTerm).
%%
%%  Opens a window on the hypothesis specified by Hyp, then
%%  executes Tactic in that window.  If Tactic succeeds, the
%%  focus of the resulting window (HypTerm) is added as a new
%%  hypothesis of the original window (whose number is HypNum).
%%  The original hypothesis, Hyp, is not deleted.
%%
%%  Restrictions include: only the first hypothesis that matches
%%  Hyp is opened; once/1 is wrapped around Tactic.
%%  
%%  mode open_hyp(+hypname, ?tactic).
%%  mode open_hyp(+hypname, ?tactic, -integer).
%%  mode open_hyp(+hypspec, ?code, -integer, -term).
%%
open_hyp(Hyp, Tactic) :-
    open_hyp(Hyp, Tactic, _, _).

open_hyp(Hyp, Tactic, Next) :-
    open_hyp(Hyp, Tactic, Next, _).

open_hyp(Hyp, Tactic, NewNum, NewTerm) :-
    nonvar(Hyp),
    \+ in_itrans,
    ip_lookup(hyplist, HL),
    ip_lookup(hypnr, HypNr),
    get_hyp(Hyp, hrec(Num, Term, _, _, CDep0, UDep0), HL),
    inherit_for_openhyp(HL, NewHL),
    choose_hyp_goal(Goal),
    push_subproof(Term,ERGO_IMPLIES,Goal,CDep0,UDep0),
    ip_set(hyplist, NewHL),
    ip_set(hypnr,   HypNr),
%% TO DO: Fix the lazy setting of hypnr in open_hyp.
%% We could get max number in NewHL.
    !,
    call_tac(Tactic),
    !,
    pop_subproof(NewTerm, CDep1, UDep1, Proof),
    add_hyp(NewTerm,NewNum,oh(Num,Proof),CDep1,UDep1),
    !.

%% TO DO: Remove the following hack in choose_hyp_goal.
%% A hack, because it assumes knowledge about the iprop theory.
%% However, the choice of goal does not affect soundness...
choose_hyp_goal('iprop.false') :-
    ancestor(iprop),
    !.
choose_hyp_goal(unknown).


%% inherit_for_openhyp(+HypList0, -HypList).
%%
%%   Inherit all the non-deleted hyps in HypList0
%%
inherit_for_openhyp([], []).
inherit_for_openhyp([hrec(Nr,T,Del,_,C0,U0)|HL], [hrec(Nr,T,_,i,C,U)|New]) :-
    var(Del),
    !,
    cdep_create_discharged(C,C0),
    udep_create_discharged(U,U0),
    inherit_for_openhyp(HL, New).
inherit_for_openhyp([_|HL], New) :-
    inherit_for_openhyp(HL, New).


proof command
  discharge_conjecture(Hyp)
extends
  hyp_proofcmds
doc
  'This predicate open a window for the discharge of conjecture ''Hyp''.',nl,
  'The discharge must be complete - to ''true''.  The standard rules for',nl,
  'conjectures apply - they may be introduced and used, but those which',nl,
  'have been used must be discharged before Tactic completes.'
implementation
  discharge_conjecture(Hyp) ---> [_]
mode
  discharge_conjecture(?hypname).

discharge_conjecture(Hyp, [Tactic]) :-
    \+ in_itrans,
    ip_lookup(hyplist, HL),
    ip_lookup(hypnr, HypNr),
    get_hyp(Hyp, HREC, HL),
    hrec_not_justified(HREC),
    !,
    hrec_term(HREC, Term),
    hrec_cdep(HREC, CNode),
    inherit_for_discharge(CNode, HL, SHL),
    push_subproof(Term, rev(ERGO_IMPLIES), ERGO_TRUE),
    ip_set(hyplist, SHL),
    ip_set(hypnr,   HypNr),
%% TO DO: Fix the lazy approach to setting hypnr in discharge_conjecture.
%% We could find max number in SHL.
    !,
    call_tac(Tactic),
    !,
    pop_subproof(ERGO_TRUE, CDep, UDep, Just),
    justify_hrec(HREC, Just, CDep, UDep).


%% inherit_for_discharge(+TargetCDep, +HypList0, -Hyplist).
%%
%%  Produce the list of inherited hypotheses for the discharge operation on
%%  the conjecture whose CDep dependencies are given in TargetCDep.
%%  All hypotheses are inherited except those which
%%  depend upon the target conjecture (which includes the target itself).
%%
%%  mode inherit_for_discharge(@cdep_node, @list(hrec), -list(hrec)).
%%
inherit_for_discharge(_, [], []).
inherit_for_discharge(ID,
                      [hrec(N,T,Del,_,C0,U0)|HL],
                      [hrec(N,T,_,i,C0,U)|New]) :-
    var(Del),
    \+ cdep_depends_on(C0, ID),
    !,
    udep_create_discharged(U,U0),
    inherit_for_discharge(ID, HL, New).
inherit_for_discharge(ID, [_|HL], New) :-
    inherit_for_discharge(ID, HL, New).


%% private push_subproof(Focus, Reln, Goal).
%% private push_subproof(Focus, Reln, Goal, CDep).
%% private push_subproof(Focus, Reln, Goal, CDep, UDep).
%%
%%  Saves the current subproof and creates a new one with
%%  the given Focus, Reln, Goal, CDep and UDep.  It sets the hyplist
%%  of the new subproof to [], but callers can override this
%%  afterwards.
%%
push_subproof(Focus,Reln,Goal) :-
    cdep_empty(CDep),
    push_subproof(Focus,Reln,Goal,CDep,[]).

push_subproof(Focus,Reln,Goal,CDep) :-
    push_subproof(Focus,Reln,Goal,CDep,[]).

push_subproof(Focus,Reln,Goal,CDep,UDep) :-
    push_simple_subproof(Focus, UDep),
    ip_set(relation,   Reln),      %% Required relation between successive foci
    ip_set(goal,       Goal),      %% Current goal
    ip_set(focusdep,   CDep),      %% DMS data - undischarged conjectures 
    ip_set(hyplist,    []),        %% List of hypothesis records
    ip_set(hypnr,      0),         %% Because hyplist = [].
    debug(subproof, ['new: ', nl,
                     '  Focus: ',q(Focus),nl,
                     '  Reln:  ',q(Reln),nl,
                     '  Goal:  ',q(Goal),nl,
                     '  CDep:  ',q(CDep),nl,
                     '  UDep:  ',q(UDep)]).

%% push_simple_subproof(Focus, UDep).
%%
%%  Saves the current subproof and creates a new one with
%%  the same reln, goal, hyps and cdep, and with Focus and UDep.
%%
push_simple_subproof(Focus, UDep) :-
    %% Save previous subproof components onto stack.
    ip_lookup(proofdepth, Depth0),
    ip_lookup(proofid,    Id0),
    ip_lookup(proof,      Proof0),
    ip_lookup(relation,   Reln0),
    ip_lookup(goal,       Goal0),
    ip_lookup(hyplist,    HypList0),
    ip_lookup(hypnr,      HypNr0),
    ip_lookup(focusdep,   CDep0),
    ip_lookup(focususe,   UDep0),
    %%
    ip_lookup(subproofstack, Stack),
    SubProof = subproof(Depth0,Id0,Proof0,Reln0,Goal0,
                        HypList0,HypNr0,CDep0,UDep0),
    ip_set(subproofstack, [SubProof|Stack]),
    debug(subproof, ['pushed: ',q(SubProof)]),
    %%
    %% Allocate new subproof id and focus id.
    ctr_inc(maxproofid, MaxId),
    ctr_inc(maxfocusid, FID),
    %%
    %% Set all the implicit parameters for this new subproof.
    Depth is Depth0 + 1,
    ip_set(proofdepth, Depth),     %% Level number of the current subproof.
    ip_set(proofid,    MaxId),     %% Unique Id. of the this new subproof.
    ip_set(proof, [pstep(FID,Focus,new)]), %% The proof steps of this subproof.
    ip_set(focususe,   [UDep]),    %% DMS data - the hyps used by each step.
    debug(subproof, ['new: Depth=',q(Depth),' Id=',q(MaxId)]).

%%  
%% private pop_subproof(Focus, CDep, UDep, Proof).
%% private pop_subproof(Focus, CDep, UDep, Proof, Justify).
%%
%%  Returns the key facts about the current subproof (these can
%%  also be looked up directly before calling this procedure),
%%  and restores the previous subproof off the stack.
%%  If Justify=true (the default), also checks that the final focus of the
%%  current subproof is fully justified (ie. does not depend upon any
%%  conjectures made at that level).
%%  Currently Proof = subprf(Id,HypList,PStepList,Reln).
%%
%%  mode pop_subproof(?thy(term), -cdep_dms, -list(udep_dms), ?proof).
%%  mode pop_subproof(?thy(term), -cdep_dms, -list(udep_dms), ?proof, @atom).
%%
pop_subproof(Focus,CDep,UDep,Proof) :-
    pop_subproof(Focus,CDep,UDep,Proof,true).
    
pop_subproof(Focus,CDep,UDep,SubProof,Just) :-
    ip_lookup(proofid,    Id),
    ip_lookup(proof,      PSteps),
    ip_lookup(relation,   Reln),
    ip_lookup(focusdep,   CD),
    ip_lookup(focususe,   UDep),
    ip_lookup(hyplist,    HypList),
    PSteps = [pstep(_,Focus,_)|_],
    cdep_simplify(CD, CDep),  %% an occasional simplify helps keep it small.
    ( Just == true ->
        is_justified_focus(CDep, HypList)
    ;
        true
    ),
    SubProof = subprf(Id,HypList,PSteps,Reln),
    %%
    %% Get previous subproof components off the stack.
    ip_lookup(subproofstack, Stack0),
    Stack0 = [SubProof0|Stack],
    debug(subproof, ['pop restores: ',q(SubProof0)]),
    SubProof0 = subproof(Depth0,Id0,PSteps0,Reln0,Goal0,
                         HypList0,HypNr0,CDep0,UDep0),
    %% Restore them to be the current subproof.
    ip_set(subproofstack, Stack),
    ip_set(proofdepth, Depth0),
    ip_set(proofid,    Id0),
    ip_set(proof,      PSteps0),
    ip_set(relation,   Reln0),
    ip_set(goal,       Goal0),
    ip_set(focusdep,   CDep0),
    ip_set(focususe,   UDep0),
    ip_set(hyplist,    HypList0),
    ip_set(hypnr,      HypNr0).


%% private add_proof_step(Focus, Just, CDep, UDep).
%%
%%  Adds one focus-transformation step to the current subproof.
%%
add_proof_step(Focus, Just, CDep1, UDep1) :-
    debug(add_proof_step, ['about to add proof step',nl,
                           '  focus: ', w(Focus), nl,
                           '  just:  ', w(Just), nl,
                           '  cdep:  ', w(CDep1), nl,
                           '  udep:  ', w(UDep1)]),
    ip_lookup(proof,    Proof0),
    ip_lookup(focusdep, CDep0),
    ip_lookup(focususe, UDep0),
    ctr_inc(maxfocusid, N1),
    cdep_union(CDep1,CDep0,CDep),
    UDep = [UDep1|UDep0],
    ip_set(proof, [pstep(N1,Focus,Just)|Proof0]),
    ip_set(focusdep, CDep),
    ip_set(focususe, UDep),
    debug(subproof, ['added: ',q(pstep(N1,Focus,Just)),nl,
                     '  cdep: ',q(CDep),nl,
                     '  udep: ',q(UDep)]).

%% find_subproof(Id, Ancestors, Steps, Hyps, Reln).
%%
%%  Subproof Id has proof steps Steps, hypothesis list Hyps and relation Reln.
%%  ProofSteps is a list of pstep(Number,Focus,Justification) terms, in
%%    decreasing order (latest-first).
%%  Hyps is a list of hrec(Number,Term,DelFlag,Justification,CDep,UDep) terms,
%%    in decreasing order.
%%  Ancestors is a list of subproofs giving the path to subproof Id in
%%    the subproof hierarchy; the last element is the one matching Id.
%%  The terms of Ancestors are subpf(Id,HypList,PStepList,Reln,Step,Reason)
%%    Step gives the step in the parent proof where this subproof began:
%%      new   -- this is the root of the main proof
%%      hyp:N -- this is a forwards step on hyp:N
%%      N     -- the actual step number
%%      unknown -- TEMPORARY.
%%    Reason gives the reason this subproof was started:
%%      new -- the root of the main proof
%%      discharge_conjecture -- conjecture discharged by tactic
%%      open_hyp(HypNum)
%%      open_pos(Pos)
%%      wrapped_tactic(Annotation)
%%      unknown -- TEMPORARY.
%%  You can specify partially-instantiated ancestors to constrain the search.
%%
%%  mode find_subproof(?integer, ?list(compound),
%%                     ?list(compound), ?list(compound), ?nonvar).
%%
find_subproof(Id, Proofs, Steps, Hyps, Reln) :-
    in_proof,
    ip_lookup(subproofstack, Stack),
    Stack = [_|_],
    !,
    ip_lookup(proofid, CurId),
    ip_lookup(proof, CurPS),
    ip_lookup(hyplist, CurHL),
    ip_lookup(relation, CurRel),
    reverse([subproof(_, CurId, CurPS, CurRel, _, CurHL, _, _, _)|Stack],
            TmpStack),
    TmpStack = [_Dummy|StackR],
    %% StackR has subproof 1 at the head, and the current subproof at the tail.
    find_subproof(StackR, Id, Proofs, Steps, Hyps, Reln).
find_subproof(Id, Proofs, Steps, Hyps, Reln) :-
    % \+ in_proof,
    ip_lookup(finished_proof, Proof),
    nonvar(Proof),
    !,
    Proof = subprf(CurId, CurHL, CurPS, CurRel),
    find_subproof([subproof(_, CurId, CurPS, CurRel, _, CurHL, _, _, _)],
                  Id, Proofs, Steps, Hyps, Reln).

%% find_subproof(Stack, Id, Proofs, Steps, Hyps, Reln).
%% 
%%  Nondeterministically checks elements of stack, which includes the current
%%  subproof.
%%  N.B. we cannot get justifications for these.
%% 
find_subproof([subproof(_, HId, HSteps, HReln, _, HHyps, _, _, _)|_Stack],
              Id, [subpf(HId,HHyps,HSteps,HReln,new,new)|Proofs],
              Steps, Hyps, Reln) :-
    find_subproof_subprf(subprf(HId, HHyps, HSteps, HReln),
                         Id, Proofs, Steps, Hyps, Reln).
find_subproof([subproof(_, HId, HSteps, HReln, _, HHyps, _, _, _)|Stack],
              Id, [subpf(HId,HHyps,HSteps,HReln,new,new)|Proofs],
              Steps, Hyps, Reln) :-
    find_subproof_tail(Stack, Id, Proofs, Steps, Hyps, Reln).

find_subproof_tail([subproof(_, HId, HSteps, HReln, _, HHyps, _, _, _)|_Stack],
                   Id, [subpf(HId,HHyps,HSteps,HReln,unknown,unknown)|Proofs],
                   Steps, Hyps, Reln) :-
    find_subproof_subprf(subprf(HId, HHyps, HSteps, HReln),
                         Id, Proofs, Steps, Hyps, Reln).
find_subproof_tail([subproof(_, HId, HSteps, HReln, _, HHyps, _, _, _)|Stack],
                   Id, [subpf(HId,HHyps,HSteps,HReln,unknown,unknown)|Proofs],
                   Steps, Hyps, Reln) :-
    find_subproof_tail(Stack, Id, Proofs, Steps, Hyps, Reln).

%% find_subprf_subprf(SubPrf, Id, Proofs, Steps, Hyps, Reln).
%% 
%%  Nondeterministically checks SubPrf, its hyps and its steps.
%% 
find_subproof_subprf(subprf(Id,Hyps,Steps,Reln), Id, [], Steps, Hyps, Reln).
find_subproof_subprf(subprf(_,SHyps,_,_), Id, Proofs, Steps, Hyps, Reln) :-
    member(hrec(HNum,_,_,Just,_,_), SHyps),
    find_subproof_just(Just, Id, hyp:HNum, Proofs, Steps, Hyps, Reln).
find_subproof_subprf(subprf(_,_,SSteps,_), Id, Proofs, Steps, Hyps, Reln) :-
    member(pstep(SNum,_,Just), SSteps),
    find_subproof_just(Just, Id, SNum, Proofs, Steps, Hyps, Reln).

%% find_subproof_just(JustTerm, Id, StepNum, Proofs, Steps, Hyps, Reln).
%% 
%%  Case-analyses forms of justifications that have subproofs.
%% 
%% Case 1: a conjecture discharged by a tactic.
find_subproof_just(cj(SubPrf), Id, Num,
                   [SubPf|Proofs], Steps, Hyps, Reln) :-
    subprf_to_subpf(SubPrf, Num, discharge_conjecture, SubPf),
    find_subproof_subprf(SubPrf, Id, Proofs, Steps, Hyps, Reln).
%% Case 2: an open_hyp subproof.
find_subproof_just(oh(Pos,SubPrf), Id, Num,
                   [SubPf|Proofs], Steps, Hyps, Reln) :-
    subprf_to_subpf(SubPrf, Num, open_hyp(Pos), SubPf),
    find_subproof_subprf(SubPrf, Id, Proofs, Steps, Hyps, Reln).
%% Case 3: an open_pos subproof.
find_subproof_just(op(Pos,SubPrf), Id, Num,
                   [SubPf|Proofs], Steps, Hyps, Reln) :-
    subprf_to_subpf(SubPrf, Num, open_pos(Pos), SubPf),
    find_subproof_subprf(SubPrf, Id, Proofs, Steps, Hyps, Reln).
%% Case 4: wrapped tactic.
find_subproof_just(wr(Anno,_,_,SubPrf), Id, Num,
                   [SubPf|Proofs], Steps, Hyps, Reln) :-
    subprf_to_subpf(SubPrf, Num, wrapped_tactic(Anno), SubPf),
    find_subproof_subprf(SubPrf, Id, Proofs, Steps, Hyps, Reln).

%% subprf_to_subpf(SubPrf, StepNum, Reason, SubPf).
subprf_to_subpf(subprf(Id,Hyps,Steps,Reln), Num, Reason,
                subpf(Id,Hyps,Steps,Reln,Num,Reason)).

    
%% is_justified_focus.
%% private is_justified_focus(CDep, HypList).
%%
%%  is_justified_focus/0 succeeds when the focus of the current window
%%  does not depend on any undischarged conjectures of the current window.
%%
%%  is_justified_focus/2 succeeds when CDep does not depend on any of the 
%%  hypotheses in HypList.
%%  
%%  mode is_justified_focus.
%%  mode is_justified_focus(@term, @list(compound)).
%%
is_justified_focus :-
    ip_lookup(focusdep, FD),
    cdep_simplify(FD, FD2),
    ip_lookup(hyplist, HL),
    is_justified_focus(FD2, HL).

is_justified_focus(FD, _) :-
    cdep_empty(FD),
    !.
is_justified_focus(FD, HL) :-
    \+ (hrec_just(HREC,cj(Just)),
	hrec_cdep(HREC,CDep),
	member(HREC, HL),
	var(Just),
        cdep_depends_on(FD, CDep)).


%% check_is_justified_focus.
%%
%%  Succeeds if focus does not depend on any undischarged conjectures
%%  of the current window.  Otherwise it generates an error.
%%
%%  mode check_is_justified_focus.
check_is_justified_focus :-
    check_or_error(
        is_justified_focus,
        ['focus depends upon unproven conjectures.']).

%% TO DO: Instead of the proof command set_goal being an extension of
%% focus_proofcmds, it should be decided where it *should* go.

proof command
  set_goal,
  set_goal(Term)
extends
  focus_proofcmds
doc
  'Set the goal of the current window to Term.',nl,
  'If Term is not specified, set it to be the same as the current focus.'
mode
  set_goal,
  set_goal(@thy(term)).

set_goal :-
    is_focus(F),
    ip_set(goal, F).

set_goal(G) :-
    check_term(G),
    ip_set(goal, G).

