%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: ac.ql,v 1.8 2001/06/04 01:24:10 itp Exp $

#include "../prover/lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  This file defines various useful tactics for using associative and
%%  commutative laws.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

proof command
  ac_set_focus(Goal)
doc
  'Use associative and commutative laws to rearrange the focus',nl,
  'to match Goal.'
mode
  ac_set_focus(@thy(term))
extends
  focus_proofcmds.

proof command
  ac_trans(Rule, Dirn, Ants, Goal)
doc
  'Use associative and commutative laws to rearrange the focus',nl,
  'to match the input of the trans command, then apply the trans command.'
mode
  ac_trans(+rulename, +direction, +list(rulename), ?thy(term))
extends
  focus_proofcmds.

proof command
  ac_flatten_focus
doc
  'Use associative laws to flatten (i.e. remove brackets) the focus.'
mode
  ac_flatten_focus
extends
  focus_proofcmds.

ac_set_focus(Goal) :-
    is_focus(F),
    is_reln(Reln),
    comm_code(F, Reln, CommCode),
    ass_code(F, Reln, ALCode, ARCode),
    ( acunify(F, Goal, CommCode, ALCode, ARCode, Code) ->
        call_predicate(Code)
    ;
        handle_error(['ac: cannot unify ', thy(F), ' with ', thy(Goal)])
    ).

ac_trans(Rule, Dirn, Ants, Goal) :-
    itrans_start,
    itrans_input_name(null),
    decode_rule(Rule, Thy, Name),
    itrans_rule_name(Thy, Name),
    itrans_direction(Dirn),
    ( Ants = [] ->
        itrans_conditional(u)
    ;
        itrans_conditional(c)
    ),
    itrans_output(Goal),
    itrans_input(Inp),
    itrans_finish(null),
    ac_set_focus(Inp),
    trans(Rule, Dirn, Ants, Goal).

comm_code(Op(_,_), Reln, CommCode) :-
    ( is_theorem(Comm,  RReln(Op(X,Y), Op(Y,X))),
      var(X),
      var(Y),
      X \== Y,
      is_stronger_than(RReln, Reln) ->
        CommCode = trans(Comm,f,[])
    ;
        handle_error(['ac: no usable commutative law for ', q(Op)])
    ).

ass_code(Op(_,_), Reln, ALCode, ARCode) :-
    ( is_theorem(Assoc, RReln(Op(X,Op(Y,Z)), Op(Op(X,Y),Z))),
      var(X),
      var(Y),
      var(Z),
      X \== Y,
      X \== Z,
      Y \== Z,
      is_stronger_than(RReln, Reln) ->
        ALCode = trans(Assoc,f,[]),
        ARCode = trans(Assoc,r,[])
    ; is_theorem(Assoc, RReln(Op(Op(X,Y),Z), Op(X,Op(Y,Z)))) ->
      var(X),
      var(Y),
      var(Z),
      X \== Y,
      X \== Z,
      Y \== Z,
      is_stronger_than(RReln, Reln) ->
        ALCode = trans(Assoc,r,[]),
        ARCode = trans(Assoc,f,[])
    ;
        handle_error(['ac: no usable associative law for ', q(Op)])
    ).

acunify(Term, Goal, CC, LC, RC, Code) :-
    Term = Op(_,_),
    acatoms(Term, Goal, Op),
    !,
    acsort(Term, TermSort, Goal, Op, CC, LC, RC, Sort),
    acregroup(TermSort, Goal, Op, LC, Regroup),
    acoptimize((Sort,Regroup), Code).

ac_flatten_focus :-
    is_focus(F),
    is_reln(R),
    ass_code(F, R, _, RC),
    acflatten(F, _, RC, Code),
    call_predicate(Code).

acflatten(Op(Op(A,B),C), D, RC, Code) :-
    !,
    acflatten(Op(A,Op(B,C)), D, RC, Code1),
    Code = (RC,Code1).
acflatten(Op(A,Op(B,C)), Op(A,D), RC, Code) :-
    !,
    acflatten(Op(B,C), D, RC, Code1),
    Code = open_pos([2],Code1).
acflatten(A, A, _, true).

acsort(Op(Op(A,B),C), D, Goal, Op, CC, LC, RC, Code) :-
    !,
    acsort(Op(A,Op(B,C)), D, Goal, Op, CC, LC, RC, Code1),
    Code = (RC, Code1).
acsort(Op(A,B), C, Goal, Op, CC, LC, RC, Code) :-
    PRE('acsort', atom(A)),
    !,
    acsort(B, B1, Goal, Op, CC, LC, RC, Code1),
    acfilter(Op(A,B1), C, Goal, CC, LC, RC, Code2),
    Code = (open_pos([2],Code1),Code2).
acsort(A, A, _, _, _, _, _, true).

%% acfilter(Op(A,B)): A always atomic, B sorted.
acfilter(Op(A,Op(B,C)), D, Goal, CC, LC, RC, Code) :-
    !,
    ( accompare(A, B, Goal) ->
        acfilter(Op(A,C), E, Goal, CC, LC, RC, Code1),
        D = Op(B,E)
    ;
        D = Op(A,Op(B,C)),
        Code1 = fail
    ),
    ( Code1 = fail ->
        Code = true
    ;
        Code = (LC, open_pos([1],CC), RC, open_pos([2],Code1))
    ).
acfilter(Op(A,B), C, Goal, CC, _LC, _RC, Code) :-
    PRE('acfilter', atom(B)),
    !,
    ( accompare(A, B, Goal) ->
        Code = CC,
        C = Op(B,A)
    ;
        Code = true,
        C = Op(A,B)
    ).

%% accompare(A, B, Goal) -- A follows B in Goal.
accompare(A, B, G) :-
    accompare(A, B, G, Afound, Bfound),
    !,
    nonvar(Bfound),
    var(Afound).

accompare(A, _B, A, true, _).
accompare(_A, B, B, _, true).
accompare(A, B, _Op(X,_Y), AF, BF) :-
    accompare(A, B, X, AF, BF),
    !.
accompare(A, B, _Op(_X,Y), AF, BF) :-
    accompare(A, B, Y, AF, BF).

acregroup(Term, Goal, _, _, Code) :-
    Term = Goal,
    !,
    Code = true.
acregroup(Op(A,B), Op(X,Y), Op, LC, Code) :-
    acmax(X, Op, M),
    acsplit(Op(A,B), M, Op(U,V), LC, Split),
    acregroup(U, X, Op, LC, LHS),
    acregroup(V, Y, Op, LC, RHS),
    Code = (Split, open_pos([1],LHS),open_pos([2],RHS)).

acmax(V, _, V) :-
    var(V),
    !.
acmax(Op(_A,B), Op, X) :-
    !,
    acmax(B, Op, X).
acmax(X, _, X).

acsplit(Op(A,B), M, Op(U,V), _, Code) :-
    A = M,
    !,
    U = A,
    V = B,
    Code = true.
acsplit(Op(A,B), M, Op(Op(A,U),V), LC, Code) :-
    acsplit(B, M, Op(U,V), LC, Code1),
    !,
    Code = (open_pos([2], Code1), LC).

%% acatoms(Term, Goal, Op) ensures that Term and Goal have exactly the same
%% atoms, by binding variables in either.
acatoms(Term, Goal, Op) :-
    ac_get_atoms(Term, Op, [], TAtoms, [], TVars),
    ac_get_atoms(Goal, Op, [], GAtoms, [], GVars),
    ac_merge_atoms(TAtoms, GAtoms, TRest, GRest),
    ac_dist_vars(GVars, TRest, Op),
    ac_dist_vars(TVars, GRest, Op).

ac_get_atoms(X, _, A0, A, V0, V) :-
    thawed_var(X),
    !,
    A = A0,
    V = [X|V0].
ac_get_atoms(Op(X,Y), Op, A0, A, V0, V) :-
    !,
    ac_get_atoms(X, Op, A0, A1, V0, V1),
    ac_get_atoms(Y, Op, A1, A, V1, V).
ac_get_atoms(X, _, A, [X|A], V, V).

ac_merge_atoms([], G, [], G).
ac_merge_atoms([T|Ts0], Gs0, Ts, Gs) :-
    delete(T, Gs0, Gs1),
    !,
    ac_merge_atoms(Ts0, Gs1, Ts, Gs).
ac_merge_atoms([T|Ts0], Gs0, Ts, Gs) :-
    Ts = [T|Ts1],
    ac_merge_atoms(Ts0, Gs0, Ts1, Gs).

ac_dist_vars([], [], _).
ac_dist_vars([V], Ts, Op) :-
    ac_build(Ts, Op, V).
ac_dist_vars([T,V|Vs], [T|Ts], Op) :-
    ac_dist_vars([V|Vs], Ts, Op).

ac_build([X], _, X).
ac_build([X,Y|Z], Op, Op(X,T)) :-
    ac_build([Y|Z], Op, T).

acoptimize(Code0, Code) :-
    code_flatten(Code0, _, Code1),
    ( var(Code1) ->
        Code = true
    ;
        code_peephole(Code1, Code)
    ).

code_flatten((A,B), C0, C) :-
    !,
    code_flatten(A, C0, C1),
    code_flatten(B, C1, C).
code_flatten(open_pos(P,A), C0, C) :-
    !,
    code_flatten(A, _, B),
    code_emit(open_pos(P,B), C0, C).
code_flatten(Trans, C0, C) :-
    code_emit(Trans, C0, C).

code_emit(true, Code0, Code) :-
    !,
    Code = Code0.
code_emit(open_pos(_,true), Code0, Code) :-
    !,
    Code = Code0.
code_emit(Cmd, Code0, Code) :-
    var(Code0),
    !,
    Code = Cmd.
code_emit(Cmd, Code0, Code) :-
    Code = (Code0,Cmd).

code_peephole((trans(R,_,_),trans(R,_,_)), true) :-
    !.
code_peephole(((C0,trans(R,_,_)),trans(R,_,_)), C) :-
    !,
    code_peephole(C0, C).
code_peephole((C0,C1), C) :-
    !,
    code_peephole(C0,C2),
    code_peephole(C1,C3),
    C = (C2,C3).
code_peephole(open_pos(P,C0), open_pos(P,C)) :-
    !,
    code_peephole(C0, C).
code_peephole(C, C).
