%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##


%% $Id: daVinci.ql,v 1.13 2001/06/04 01:24:10 itp Exp $
%% Generate daVinci format graphs of theory structure.

database command
  daVinci(File),
  daVinci(File, Thys),
  daVinci(File, Thys, Roots),
  daVinci_nontrans(File),
  daVinci_nontrans(File, Thys),
  daVinci_nontrans(File, Thys, Roots)
extends
  misc_cmds
doc
  'Write theory graph into <File>.daVinci.',nl,
  'With second arg Thys, draw the forest with just those theories.',nl,
  'With third arg Roots, include Roots and all their descendents too.',nl,
  'daVinci_nontrans will eliminate all transitively-deducable links.'
mode
  daVinci(@filename),
  daVinci(@filename, @list(thyname)),
  daVinci(@filename, @list(thyname), @list(thyname)),
  daVinci_nontrans(@filename),
  daVinci_nontrans(@filename, @list(thyname)),
  daVinci_nontrans(@filename, @list(thyname), @list(thyname)).

daVinci(File) :-
    bagof(Thy, Ancs^query_global(ancestor(Thy, Ancs)), Thys),
    daVinci(File, [], Thys, trans).

daVinci(File, Thys) :-
    daVinci(File, Thys, [], trans).

daVinci(File, Thys, Roots) :-
    daVinci(File, Thys, Roots, trans).

daVinci_nontrans(File) :-
    bagof(Thy, Ancs^query_global(ancestor(Thy, Ancs)), Thys),
    daVinci(File, [], Thys, nontrans).

daVinci_nontrans(File, Thys) :-
    daVinci(File, Thys, [], nontrans).

daVinci_nontrans(File, Thys, Roots) :-
    daVinci(File, Thys, Roots, nontrans).

daVinci(File, Thys, Roots, Trans) :-
    enable_signals,
    atom_concat2(File, '.daVinci', Full),
    open(Full, write, Stream),
    gsetof(Thy, Root^
           (
               member(Root, Roots),
               ancestor_of(Thy, Root),
               \+ member(Thy, Thys)
           ),
           XThys),
    append(Thys, XThys, AllThys),
    daVinci_thys(AllThys, Thys, Trans, Stream),
    close(Stream),
    disable_signals.

%% daVinci_thys(Thys, LThys, Trans, Stream).
%%
%%  Draw a graph whose nodes are Thys and XThys, and with two kinds of edges:
%%    - solid edges are direct father->son links.
%%    - dotted edges are ancestor->descendent links that cannot be inferred
%%      from the above.
%%  Indirect links can occur only under LThys, which is a subset of Thys.
%% 
%%  mode daVinci_thys(@list(thyname), @atom, @stream).
%%
daVinci_thys(Thys, LThys, Trans, Strm) :-
    gbagof(Thy-Son,
           (
               member(Thy, Thys),
               query_global(parent(Son, Thy)),
               member(Son, Thys)
           ),
           DirectEdges),
    gbagof(Thy-Dec,
           (
               member(Thy, LThys),
               real_strict_ancestor(Dec, Thy),
               member(Dec, Thys),
               \+ member(Thy-Dec, DirectEdges)
           ),
           IndirectEdges),
    ( Trans = trans ->
        MinDirectEdges = DirectEdges
    ;
        remove_transitive_edges(DirectEdges, DirectEdges, IndirectEdges,
                                MinDirectEdges)
    ),
    remove_transitive_edges(IndirectEdges, DirectEdges, IndirectEdges,
                            MinIndirectEdges),
    writelist(Strm, ['[',nl]),
    daVinci_nodes(Thys, MinDirectEdges, MinIndirectEdges, Strm),
    writelist(Strm, [']',nl]).

daVinci_nodes([], _, _, _).
daVinci_nodes([Thy], Dir, Ind, Strm) :-
    daVinci_node(Thy, Dir, Ind, Strm),
    writelist(Strm, [nl]).
daVinci_nodes([Thy1,Thy2|Thys], Dir, Ind, Strm) :-
    daVinci_node(Thy1, Dir, Ind, Strm),
    writelist(Strm, [',',nl]),
    daVinci_nodes([Thy2|Thys], Dir, Ind, Strm).

daVinci_node(Thy, Dir, Ind, Strm) :-
    daVinci_start_node(Thy, Strm),
    gbagof(Son, member(Thy-Son, Dir), DirSons),
    gbagof(Dec, member(Thy-Dec, Ind), IndSons),
    daVinci_edges(DirSons, direct, Strm),
    ( DirSons = [] ->
        daVinci_edges(IndSons, indirect, Strm)
    ; IndSons = [] ->
        true
    ;
        writelist(Strm, [',',nl]),
        daVinci_edges(IndSons, indirect, Strm)
    ),
    daVinci_finish_node(Strm).

daVinci_start_node(Node, Strm) :-
    writelist(Strm, [' l ("', w(Node),
                     '", n ("", [a("OBJECT","', w(Node), '")],',nl,
                     '   [',nl]).

daVinci_finish_node(Strm) :-
    writelist(Strm, [nl, '   ]))']).

daVinci_edges([], _, Strm).
daVinci_edges([Son], Kind, Strm) :-
    daVinci_edge(Son, Kind, Strm).
daVinci_edges([Son1,Son2|Sons], Kind, Strm) :-
    daVinci_edge(Son1, Kind, Strm),
    writelist(Strm, [',',nl]),
    daVinci_edges([Son2|Sons], Kind, Strm).

daVinci_edge(Son, direct, Strm) :-
    writelist(Strm, ['     e ("", [], r ("', w(Son), '"))']).
daVinci_edge(Son, indirect, Strm) :-
    writelist(Strm, ['     e ("", [a("EDGEPATTERN","dotted")], r ("',
                     w(Son), '"))']).

%% gsetof(Template, Goal, Answers).
%%
%%  Like setof, but can return an empty set of results, and will not backtrack
%%  on unbound variables.
%%
%%  mode gsetof(@term, +goal, -list(term)).
%%
gsetof(T, G, A) :-
    setof(T, G, A),
    !.
gsetof(T, G, []).

%% gbagof(Template, Goal, Answers).
%%
%%  Like bagof, but can return an empty bag of results, and will not backtrack
%%  on unbound variables.
%%
%%  mode gbagof(@term, +goal, -list(term)).
%%
gbagof(T, G, A) :-
    bagof(T, G, A),
    !.
gbagof(T, G, []).

%% remove_transitive_edges(Edges0, Edges1, Edges2, Edges).
%% 
%% Edges is a subset of Edges0.  Any edges that are transitively deducable from
%% Edges1+Edges2 are removed.
%%
remove_transitive_edges([], _, _, []).
remove_transitive_edges([Thy-Son|Edges0], Edges1, Edges2, Edges) :-
    ( non_unit_path(Thy, Son, Edges1, Edges2) ->
        writelist(['removed ', w(Thy-Son), nl]),
        Edges = EdgesR
    ;
        writelist(['retained ', w(Thy-Son), nl]),
        Edges = [Thy-Son | EdgesR]
    ),
    remove_transitive_edges(Edges0, Edges1, Edges2, EdgesR).

%% non_unit_path(Node1, NodeN, Edges1, Edges2).
%% non_empty_path(Node1, NodeN, Edges1, Edges2).
%%
%% There is a non-unit path from Node1 to NodeN via Edges1 and Edges2.
%%
non_unit_path(Node1, NodeN, Edges1, Edges2) :-
    member(Node1-Node2, Edges1),
    non_empty_path(Node2, NodeN, Edges1, Edges2).
non_unit_path(Node1, NodeN, Edges1, Edges2) :-
    member(Node1-Node2, Edges2),
    non_empty_path(Node2, NodeN, Edges1, Edges2).

non_empty_path(Node1, NodeN, Edges1, _Edges2) :-
    member(Node1-NodeN, Edges1).

non_empty_path(Node1, NodeN, _Edges1, Edges2) :-
    member(Node1-NodeN, Edges2).

database command
  daVinci_all(File)
extends
  misc_cmds
doc
  'Write the entire theory graph into <File>.daVinci, eliminating all', nl,
  'transitively-deducable links and subtheories.'
mode
  daVinci_all(@filename).

daVinci_all(File) :-
    findall(Thy,
            (query_global(ancestor(Thy, _)),
              \+ query_global(subthy_owner(Thy, _))
            ),
            Thys),
    daVinci_nontrans(File, Thys).
