%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##


%% $Id: dv5.ql,v 1.30 2001/06/04 01:24:10 itp Exp $
%% Interface between Ergo 5 and daVinciV2.0

#include "../prover/lib/ergo.h"

?- require_global_tactic_file(dv).

dv5_node_display parameter
  [set_dv5_node_display, get_dv5_node_display]
doc
  'This parameter specifies what to display in each daVinci proof node.',nl,
  'Value is a procedure which will be called with two arguments:',nl,
  '  Proc(@nodeid, -writelist).',nl,
  'The default value is ''dv5_node_display_term''.',nl,
  nl,
  begin(list),
  item('dv5_node_display_term'),' displays the term',nl,
  item('dv5_node_display_id_term'),' displays the id and term',nl,
  item('dv5_node_display_id_context_term'),' displays the id, context and term',nl,
  end(list)
extends
  parameters.

dv5_context_display parameter
  [set_dv5_context_display, get_dv5_context_display]
doc
  'This parameter specifies how to display context in a daVinci proof.',nl,
  'Possible values are:',
  begin(list),
  item([]),'  ''none''           -- don''t display context',
  item([]),'  ''proof_node''     -- display context changes in the proof nodes',
  item([]),'  ''c_node''         -- display context nodes without use-edges ',
  item([]),'  ''c_node_edges''   -- display context nodes with use-edges ',nl,
  '                       (not currently implemented).',
  end(list)
extends
  parameters.

set_dv5_node_display(Proc) :-
    retractall_global(dv5_node_display(_)),
    assert_global(dv5_node_display(Proc)).

get_dv5_node_display(Proc) :-
    query_global(dv5_node_display(Proc)).

set_dv5_context_display(Val) :-
    member(Val, [none,proof_node,c_node,c_node_edges]),
    !,
    retractall_global(dv5_context_display(_)),
    assert_global(dv5_context_display(Val)).
set_dv5_context_display(_) :-
    handle_error(['dv5_context_display: legal values are ',
                  w([none,proof_node,c_node,c_node_edges])]).

get_dv5_context_display(Val) :-
    query_global(dv5_context_display(Val)).

database command
  dv5_enable,
  dv5_disable
doc
  'Enable/disable the daVinci proof construction interface.'
extends
  general_cmds
mode
  dv5_enable,
  dv5_disable.

dv5_enable :-
    daVinci_init,
    dynamic(dv5_old_closed/2),
    dynamic(dv5_old_context/2),
    dynamic(dv5_old_context_use/2),
    add_listener(proof(_Thy,_Name,Action),
                 dv5_proof(Action)),
    add_listener(daVinci(State), dv5_state(State)),
    ( global_fact(dv5_node_display,1) ->
        true
    ;
        new_global_fact(dv5_node_display, 1, [save(true)]),
        new_global_fact(dv5_context_display, 1, [save(true)]),
        assert_global(dv5_node_display(dv5_node_display_term)),
        assert_global(dv5_context_display(proof_node))
    ).

dv5_disable :-
    delete_listeners(proof(_,_,_), dv5_proof(_)),
    delete_listeners(daVinci(_), dv5_state(_)).
    

%%dv5_proof(Action) :-
%%    writelist(['dv5 proof handler got: ', w(Action), nl]),
%%    fail.
dv5_proof(start([Id])) :-
    !,
    erase_all(dv5_old_open_list),
    erase_all(dv5_old_current_list),
    retractall(dv5_old_closed(_,_)),
    retractall(dv5_old_context(_,_)),
    retractall(dv5_old_context_use(_,_)),
    !,
    recorda(dv5_old_open_list, [Id], _),
    recorda(dv5_old_current_list, [Id], _),
    daVinci_start.
dv5_proof(exited) :-
    daVinci_active,
    !,
    daVinci_command(quit).
dv5_proof(gumtree_display(Open,Cur)) :-
    daVinci_active,
    !,
    dv5_redisplay(Open,Cur).
dv5_proof(_).

dv5_state(connected) :-
    daVinci_command(app_menu(create_menus(
        [menu_entry_mne(goto, 'Go to', g, control, 'G'),
         menu_entry_mne(display, 'Display', d, control, 'D'),
         menu_entry_mne(annotations, 'Annotations', a, control, 'A'),
         menu_entry_mne(redraw, 'Redraw', r, control, 'R')
        ]))),
    daVinci_command(menu(layout(orientation(bottom_up)))),
    daVinci_command(graph(new([]))),
    recorded(dv5_old_open_list, [Id], _),
    dv5_label(Id, Label),
    dv5_status_attrs(current, Attrs),
    daVinci_command(graph(update([new_node(Id, Label, Attrs)], []))),
    dv5_activate_menus([Id]).
dv5_state(selected(Nodes)) :-
    dv5_activate_menus(Nodes).
dv5_state(menu(Menu, Selected)) :-
    dv5_menu_selection(Menu, Selected).
dv5_state(disabled) :-
    dv5_disable.
dv5_state(closed).

general command
  dv5_redraw(Roots)
doc
  'Erase and redraw the subtrees rooted at Roots.'
extends
  prooftree_info_cmds
mode
  dv5_redraw(@list(nodeid)).

dv5_redraw(Nodes) :-
    dv5_delete_subtrees(Nodes, _, _, [], DelNodes),
    dv5_redraw(Nodes, [], NewNodes, [], NewEdges),
    daVinci_command(graph(update(DelNodes, []))),
    daVinci_command(graph(update(NewNodes, NewEdges))),
    recorded(dv5_old_current_list, Cur, CurRef),
    recorded(dv5_old_open_list, Open, OpRef),
    erase(CurRef),
    erase(OpRef),
    recorda(dv5_old_current_list, [], _),
    recorda(dv5_old_open_list, [], _),
    dv5_redisplay(Open,Cur).

dv5_redraw([], NU, NU, EU, EU) :-
    !.
dv5_redraw([Node|Nodes], NU0, NU, EU0, EU) :-
    !,
    dv5_redraw(Node, NU0, NU1, EU0, EU1),
    dv5_redraw(Nodes, NU1, NU, EU1, EU).
dv5_redraw(Node, NU0, NU, EU0, EU) :-
    dv5_add_node(Node, _, _, NU0, NU, EU0, EU1),
    ( proof_node_parent(Node, Parent) ->
        EU = [new_edge(Node, Parent, Node, [])|EU1]
    ;
        EU1 = EU
    ).

%% dv5_redisplay(Open, Cur).
%%
%%  Open are the currenly open nodes.
%%  Bring the da Vinci display up to date to reflect this.
%%
%%  Each node in the union of the old and new trees falls into exactly one of
%%  the following classes, according to its previous and present state:
%%
%%    p            present
%%    r          Absent  Open  Closed
%%    e Absent    1       2      3
%%    v
%%    i Open      4       5      6
%%    o
%%    u Closed    7       8      9
%%    s
%%
%%  (1): trivial; not detected.
%%  (2): a new open node.
%%   Detected by an element of Open that was not previously recorded.  
%%   Its parent in the current tree is (3), (6), or (9). 
%%  (3): A new closed node.
%%   (a) Detected as a parent of (2) or (3); -- its parent in the current tree
%%       is (3), (6) or (9).
%%   (b) Detected a descendant of (6) or (9).
%%   (c) NOT DETECTED: SEE BELOW.
%%  (4): A deleted open node.
%%   Detected as a node in the previous fringe that no longer exists. 
%%   Its parent in the old tree is (7), (8), or (9). 
%%  (5): trivial; easily detected.
%%  (6): A node that just became closed as the result of a forwards step.
%%   Detected as a node in the previous fringe that is now closed. 
%%   Its children in the current tree are (2) or (3); handle it by building a
%%   fresh subtree.
%%  (7): A deleted closed node.
%%   Detected as the parent of (4) or (7). 
%%   Its parent in the old tree is (7), (8), or (9). 
%%  (8): A closed node that was just opened by a backwards step.
%%   Detected as a new node in the fringe. 
%%   Its children in the old tree were (4) or (7); handle it by deleting all
%%   subtrees.
%%  (9): A closed node that is still closed.
%%   (a) Trivial and not detected.
%%   (b) Detected as the parent of (2) or (3); handle by adding new subtrees.
%%   (c) Detected as the parent of (4) or (7); handle by deleting old subtrees.
%%   (d) NOT DETECTED: SEE BELOW.
%%
%%  ********************
%%  *** NOT DETECTED ***
%%  ********************
%%  One scenario will NOT BE DETECTED by this code, because it involves no open
%%  nodes.
%%  A node is closed and has only closed descendants.  A failure removes those
%%  descendents, but then a forwards step adds new ones and closes them all.
%%  E.g. (rule(assump) | (rule(implies_intro); rule(assump))).
%% 
dv5_redisplay(Open, Curr) :-
    recorded(dv5_old_open_list, OldOpen, OpRef),
    recorded(dv5_old_current_list, OldCurr, CurrRef),
    dv5_update_nodes(OldOpen, Open, OldOpen, [], NU, [], EU, [], AU1),
    ( Curr == OldCurr ->
        AU = AU1
    ;
        erase(CurrRef),
        recorda(dv5_old_current_list, Curr, _),
        dv5_set_current(OldCurr, Curr, Open, AU1, AU)
    ),
    ( NU == [], EU == [] ->
        true
    ;
        daVinci_command(graph(update(NU, EU)))
    ),
    ( AU == [] ->
        true
    ;
        dv5_change_attrs(AU)
    ),
    daVinci_command(special(select_nodes(Curr))),
    ( Open == OldOpen ->
        true
    ;
        erase(OpRef),
        recorda(dv5_old_open_list, Open, _)
    ).

dv5_set_current([], [], _, AU, AU).
dv5_set_current([], [New|News], _, AU0, AU) :-
    dv5_set_status(current, New, AU0, AU1),
    dv5_set_current([], News, _, AU1, AU).
dv5_set_current([Old|Olds], News, Open, AU0, AU) :-
    delete(Old, News, News1),
    !,
    dv5_set_current(Olds, News1, Open, AU0, AU).
dv5_set_current([Old|Olds], News, Open, AU0, AU) :-
    PRE('dv5_set_current', (\+ member(Old, News))),
    member(Old, Open),
    !,
    dv5_set_status(open, Old, AU0, AU1),
    dv5_set_current(Olds, News, Open, AU1, AU).
dv5_set_current([Old|Olds], News, Open, AU0, AU) :-
    PRE(
        'dv5_set_current: \+ member(Old, Open)',
        (\+ member(Old, News), \+ member(Old, Open))
    ),
    dv5_set_current(Olds, News, Open, AU0, AU).

%% dv5_update_nodes(OldNodes, NewNodes, OldOpen, NU0, NU, EU0, EU, AU0, AU).
%% dv5_update_nodes(NewNodes, OldOpen, NU0, NU, EU0, EU, AU0, AU).
%%
dv5_update_nodes([], New, OldOpen, NU0, NU, EU0, EU, AU0, AU) :-
    dv5_update_nodes(New, OldOpen, NU0, NU, EU0, EU, AU0, AU).
dv5_update_nodes([Node|Old0], New0, OldOpen, NU0, NU, EU0, EU, AU0, AU) :-
    dv5_find_old_ancestor(Node, Anc),
    dv5_update_node(Anc, Old0, Old, New0, New, NU0, NU1, EU0, EU1, AU0, AU1),
    dv5_update_nodes(Old, New, OldOpen, NU1, NU, EU1, EU, AU1, AU).

dv5_update_nodes([], _, NU, NU, EU, EU, AU, AU).
dv5_update_nodes([Node|New0], OldOpen, NU0, NU, EU0, EU, AU0, AU) :-
    dv5_find_new_ancestor(Node, Anc, OldOpen),
    dv5_update_node(Anc, _, _, [Node|New0], New, NU0, NU1, EU0, EU1, AU0, AU1),
    dv5_update_nodes(New, OldOpen, NU1, NU, EU1, EU, AU1, AU).

%% dv5_update_node(Node, Old0, Old, New0, New, NU0, NU, EU0, EU, AU0, AU).
%%
%%  Node is in both trees.  Update it and all its subtrees.
%% 
dv5_update_node(Node, Old0, Old, New0, New, NU0, NU, EU0, EU, AU0, AU) :-
    ( retract(dv5_old_closed(Node, OldChildren)) ->
        OldStatus = closed,
        dv5_delete_subtrees(OldChildren, Old0, Old, NU0, NU1)
    ;
        Old0 = Old,
        OldStatus = open,
        NU1 = NU0
    ),
    ( proof_node_children(Node, NewChildren) ->
        NewStatus = closed,
        assert(dv5_old_closed(Node, NewChildren)),
        dv5_context_use(Node, EU0, EU1),
        dv5_add_subtrees(NewChildren, Node, New0, New, NU1, NU, EU1, EU)
    ;
        delete_all(Node, New0, New),
        !,
        NewStatus = open,
        NU = NU1,
        dv5_context_use(Node, EU0, EU)
    ),
    ( OldStatus == NewStatus ->
        AU = AU0
    ;
        dv5_set_status(NewStatus, Node, AU0, AU)
    ).

%% dv5_delete_subtrees(Nodes, Old0, Old, NU0, NU).
%%
dv5_delete_subtrees([], Old, Old, NU, NU).
dv5_delete_subtrees([Node|Nodes], Old0, Old, NU0, NU) :-
    dv5_delete_node(Node, Old0, Old1, NU0, NU1),
    dv5_delete_subtrees(Nodes, Old1, Old, NU1, NU).

dv5_delete_node(Node, Old0, Old, NU0, NU) :-
    ( retract(dv5_old_closed(Node, Children)) ->
        dv5_delete_subtrees(Children, Old0, Old1, NU0, NU1)
    ;
        Old1 = Old0,
        NU1 = NU0
    ),
    ( retract(dv5_old_context(NNode, Node)) ->
        NU2 = [delete_node(NNode)|NU1]
    ;
        NU2 = NU1
    ),
    delete_all(Node, Old1, Old),
    !,
    NU = [delete_node(Node) | NU2].

%% dv5_add_subtrees(Nodes, Parent, New0, New, NU0, NU, EU0, EU).
%%
dv5_add_subtrees([], _, New, New, NU0, NU, EU0, EU) :-
    NU = NU0,
    reverse(EU0, EU).
dv5_add_subtrees([Node|Nodes], Parent, New0, New, NU0, NU, EU0, EU) :-
    dv5_add_node(Node, New0, New1, NU0, NU1, EU0, EU1),
    EU2 = [new_edge(Node, Parent, Node, [])|EU1],
    dv5_add_subtrees(Nodes, Parent, New1, New, NU1, NU, EU2, EU).

dv5_add_node(Node, New0, New, NU0, NU, EU0, EU) :-
    ( proof_node_children(Node, Children) ->
        assert(dv5_old_closed(Node, Children)),
        dv5_context_use(Node, EU0, EUA),
        dv5_add_subtrees(Children, Node, New0, New1, NU0, NU1, EUA, EU1),
        Status = closed
    ;
        New1 = New0,
        NU1 = NU0,
        EU1 = EU0,
        Status = open
    ),
    delete_all(Node, New1, New),
    !,
    dv5_status_attrs(Status, Attrs),
    dv5_label(Node, Label0),
    dv5_context_intro(Node, Label0, Label, NU1, NU2, EU1, EU),
    NU = [new_node(Node,Label,Attrs)|NU2].

%% dv5_set_status(Status, Node, AU0, AU).
%% 
dv5_set_status(Status, Node, AU0, AU) :-
    dv5_status_attrs(Status, Attrs),
    AU = [node(Node, Attrs)|AU0].

%% dv5_status_attrs(Status, Attrs).
%%
dv5_status_attrs(current, [shape(ellipse),border(double)]).
dv5_status_attrs(open, [shape(ellipse),border(single)]).
dv5_status_attrs(closed, [shape(box),border(single)]).
dv5_status_attrs(context, [shape(box),border(double)]).
%%dv5_status_attrs(current, [shape(box),border(double)]).
%%dv5_status_attrs(open, [shape(box),border(double)]).
%%dv5_status_attrs(closed, [shape(box),border(single)]).
%%dv5_status_attrs(context, [shape(ellipse),border(single)]).

%% dv5_find_old_ancestor(Node, Ancestor).
%% 
%%  Search the old tree for an ancestor that is shared by the new tree.
%% 
dv5_find_old_ancestor(Node, Anc) :-
    ( proof_node_term(Node, _) ->
        Anc = Node
    ; dv5_old_closed(Parent, Nodes),
      member(Node, Nodes) ->
        dv5_find_old_ancestor(Parent, Anc)
    ;
        fail
    ).

%% dv5_find_new_ancestor(Node, Ancestor, OldOpen).
%%
%%  Search the new tree for an ancestor that is shared by the old tree.
%% 
dv5_find_new_ancestor(Node, Anc, OldOpen) :-
    ( dv5_old_closed(Node,_) ->
        Anc = Node
    ; member(Node, OldOpen) ->
        Anc = Node
    ; proof_node_parent(Node, Parent) ->
        dv5_find_new_ancestor(Parent, Anc, OldOpen)
    ;
        fail
    ).

dv5_change_attrs(Attrs) :-
    %% need to do just the first (latest) entry in the list for each node.
    dv5_change_attrs(Attrs, [], Attrs2),
    daVinci_command(graph(change_attr(Attrs2))).

dv5_change_attrs([], As, As).
dv5_change_attrs([node(N,A)|Ns], As0, As) :-
    ( member(node(N,_), As0) ->
        dv5_change_attrs(Ns, As0, As)
    ;
        dv5_change_attrs(Ns, [node(N,A)|As0], As)
    ).

%% dv5_label(Node, WriteList).
%%
%%  WriteList is the label to use for Node.
%% 
dv5_label(Id, Label) :-
    get_dv5_node_display(Proc),
    call_predicate(Proc(Id, Label)).

%% TO DO: The definition of dv5_default_node_display is left here only
%% for backwards compatibility. It can be removed as soon as all references
%% to it have been removed (including, for example, global.erg files).
%% MB 5/8/1997.

dv5_default_node_display(Id, Wl) :-
    dv5_node_display_term(Id, Wl).

dv5_node_display_term(Id, [thy(Term)]) :-
    proof_node_term(Id, Term).

dv5_node_display_id_term(Id, [w(Id), ' ::: ', thy(Term)]) :-
    proof_node_term(Id, Term).

dv5_node_display_id_context_term(Id, [w(Id), ' ::: ', wl(Wc), ' ---> ', thy(Term)]) :-
    proof_node_term(Id, Term),
    proof_node_context(Id, Context),
    cspec_to_writelist(Context, Wc).

%% dv5_context_intro(Node, Label0, Label, NU0, NU, EU0, EU).
%%
%%  Introduce new context at Node -- how this is portrayed depends on the
%%  dv5_context_display parameter.
%% 
dv5_context_intro(Node, Label0, Label, NU0, NU, EU0, EU) :-
    get_dv5_context_display(Val),
    ( Val == none ->
        Label = Label0,
        NU = NU0,
        EU = EU0
    ; Val == proof_node ->
        dv5_proof_node_contexts(Node, Label0, Label),
        NU = NU0,
        EU = EU0
    ; member(Val, [c_node,c_node_edges]) ->
        Label = Label0,
        dv5_c_node_contexts(Node, NU0, NU, EU0, EU)
    ).

dv5_proof_node_contexts(Node, Label0, Label) :-
    proof_node_parent(Node, Parent),
    context_names(CNames),
    !,
    map(dv5_context_diff(Node, Parent), [CNames,Diffs]),
    dv5_label_contexts(Diffs, Label, Label0).

dv5_c_node_contexts(Node, NU0, NU, EU0, EU) :-
    proof_node_parent(Node, Parent),
    context_names(CNames),
    !,
    map(dv5_context_diff(Node, Parent), [CNames,Diffs]),
    dv5_add_context_nodes(Diffs, Node, NU0, NU, EU0, EU).

dv5_label_contexts([]) -->
    !,
    [].
dv5_label_contexts([Diff|Diffs]) -->
    ( {Diff = ===(CN,CN)} ->
        []
    ;
        cspec_to_writelist(Diff),
        ['\\n']
    ),
    dv5_label_contexts(Diffs).

dv5_add_context_nodes(Diffs, Node, NU0, NU, EU0, EU) :-
    dv5_label_contexts(Diffs, Label, []),
    append(Label1, [_], Label),
    !,
    concat_atom([c,Node], NName),
    assert(dv5_old_context(NName, Node)),
    dv5_status_attrs(context, Attrs),
    NU = [new_node(NName, Label1, Attrs) | NU0],
    EU = [new_edge(NName, NName, Node, []) | EU0].
dv5_add_context_nodes(_, _, NU, NU, EU, EU).

dv5_context_diff(Node, Parent, CName, Diff) :-
    context_diff(Node, Parent, CName, Diff1),
    abbrev_rule_cspec(Parent, =(CName,Diff1), Diff).

%% dv5_context_use(Node, EU0, EU).
%% 
%% Update context-use edges for Node.

%% TO DO: Work out if ``Vile hack!!!'' refers to still-active code or to
%% the mostly commented out dv5_context_use.
%% Ray's last comment: ``No'' regarding if, underlined the second disjunct.

%%dv5_context_use(Id, EU0, EU) :-
%%    get_dv5_context_display(Val),
%%    Val == c_node_edges,
%%    !,
%%    ( retract(dv5_old_context_use(Id, OCId)) ->
%%        concat_atom([Id,hyp,OCId], OEdge),
%%        EU1 = [delete_edge(OEdge)|EU0]
%%    ;
%%        EU1 = EU0
%%    ),
%%    ( proof_node_rule(Id, natded_iprop, assump) ->
%%        proof_node_term(Id, Term),
%%        context_search(hyp, Id, CId, Term),
%%        concat_atom([hyp,CId], NName),
%%        concat_atom([Id,hyp,CId], NEdge),
%%        assert(dv5_old_context_use(Id, CId)),
%%        EU = [new_edge(NEdge, NName, Id, [edgepattern(dotted)])|EU1]
%%    ;
%%        EU = EU1
%%    ).
dv5_context_use(_, EU, EU).

%% dv5_activate_menus(Nodes).
%%
%%  Nodes have been selected.
%%  Activate the appropriate menus.
%%
dv5_activate_menus(Nodes) :-
    %% `goto' active if only open nodes are selected
    %% `display', `annotate', `redraw' always active
    recorded(dv5_old_open_list, Open, _),
    ( diff_list(Nodes, Open, []) ->
        Active = [goto,display,annotations,redraw]
    ; 
        Active = [display,annotations,redraw]
    ),
    daVinci_command(app_menu(activate_menus(Active))).

dv5_menu_selection(goto, Nodes) :-
    emacs_send_string([proof_nodes(Nodes)]).
dv5_menu_selection(display, Nodes) :-
    emacs_send_string(['show_nodes_and_context(', proof_nodes(Nodes), ')']).
dv5_menu_selection(annotations, Nodes) :-
    emacs_send_string(['show_annotations(', proof_nodes(Nodes), ')']).
dv5_menu_selection(redraw, Nodes) :-
    dv5_redraw(Nodes).

?- dv5_disable. %% don't use dv5 otherwise add this - , dv5_enable.
