%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: emacs.ql,v 1.108 2007/03/27 23:40:57 itp Exp $
%%
%% Support for Emacs interface.

emacs_magic parameter
  [set_emacs_magic, current_emacs_magic]
doc
  'Control the generation of emacs magic commands.',nl,
  'Value is a list saying what kinds of magic to enable:',
  begin(list),
  item('  help:'),' help goes in own window, with clickable xrefs.',
  item('  xref:'),' help goes in transcript window, with clickable xrefs.',
  item('  proof:'),' proof frame handling (Ergo4 only).',
  item('  dialog:'),' questions are asked in dialog boxes (Ergo4 only).',
  end(list),
  nl,
  'Alternatively, Value is the atom ''default'', in which case the',nl,
  'value is obtained from the parameter ',
  xref([default_emacs_magic],global,'parameter:default_emacs_magic'), '.', nl,
  '`set emacs_magic === default'' is generated automatically whenever the',nl,
  'Emacs interface starts Ergo.'
extends
  parameters.

default_emacs_magic parameter
  [set_default_emacs_magic, get_default_emacs_magic]
doc
  'Default value for ',
  xref([emacs_magic],global,'parameter:emacs_magic'), '.'
extends
  parameters.

emacs_proof_panes parameter
  [set_emacs_proof_panes, get_emacs_proof_panes]
doc
  'List of panes to be included in the proof frame.',nl,
  'Value is a list of atoms from the set',nl,
  '   ''[subproof,hypotheses,proof_control,subproof_stack]''.',nl,
  'This parameter is set automatically by the emacs interface,',nl,
  'based on the panes named in ergo-proof-frame-panes.'
extends
  parameters.

emacs_transcript_info parameter
  [set_emacs_transcript_info, get_emacs_transcript_info]
doc
  'List of names of data that should be displayed in the transcript',nl,
  'window before every prompt.',nl,
  'Can include: [window_level, window_id, focus, hyps, newvars, reln, goal].'
extends
  parameters.

?- dynamic(current_emacs_magic/1),
   dynamic(current_emacs_proof_panes/1),
   dynamic(current_emacs_transcript_info/1),
   ( global_fact(default_emacs_magic,1) ->
       true
   ;
       new_global_fact(default_emacs_magic,1,[save(true)]),
       assert_global(default_emacs_magic([help,proof]))
   ).

set_emacs_magic(Default) :-
    Default == default,
    !,
    get_default_emacs_magic(Magic),
    set_emacs_magic(Magic).
set_emacs_magic(Kinds) :-
    diff_list(Kinds, [help,xref,proof,dialog], []),
    !,
    retractall(current_emacs_magic(_)),
    assert(current_emacs_magic(Kinds)),
    delete_listeners(proof(_,_,_), emacs_proof(_,_,_)),
    delete_handlers(proof_output(_,_), emacs_proof_output(_,_,_)),
    delete_handlers(showing_help(_,_), emacs_showing_help(_,_)),
    delete_handlers(write(_, _, _), emacs_write_xref(_,_,_,_)),
    delete_handlers(write(_, _, _), emacs_write_emph(_,_)),
    delete_handlers(question(_, _, _), emacs_dialog(_,_,_,_)),
    delete_listeners(param_changed(_,_), emacs_proof_output(_,_,_)),
    ( member(proof, Kinds) ->
        add_listener(proof(Thy,Name,Action),
                     emacs_proof(Action, Thy, Name)),
        get_emacs_proof_panes(Panes),
        add_handler(proof_output(Kind,Term),
                    emacs_proof_output(Kind, Term, Panes)),
        add_listener(param_changed(Name,Val),
                     emacs_proof_output(param(Name), Val, Panes))
    ;
        true
    ),
    ( member(help, Kinds) ->
        add_handler(write(Strm, _, xref(WL, Thy, Name)),
                    emacs_write_xref(Strm, WL, Thy, Name)),
        add_handler(write(Strm, _, emph(WL)),
                    emacs_write_emph(Strm, WL)),
        add_handler(showing_help(Action,Node),
                    emacs_showing_help(Action,Node))
    ; member(xref, Kinds) ->
        add_handler(write(Strm, _, xref(WL, Thy, Name)),
                    emacs_write_xref(Strm, WL, Thy, Name))
    ;
        true
    ),
    ( member(dialog, Kinds) ->
        add_handler(question(Prompt,Default,Answer),
                    emacs_dialog(Prompt, [yes,no], Default, Answer))
    ;
        true
    ).
set_emacs_magic(Kinds) :-
    handle_error(['Bad value ''', w(Kinds), '''',
                  ' for parameter ''emacs_magic''.']).

set_default_emacs_magic(Magic) :-
    retractall_global(default_emacs_magic(_)),
    assert_global(default_emacs_magic(Magic)).

get_default_emacs_magic(Magic) :-
    query_global(default_emacs_magic(Magic)).

set_emacs_proof_panes(Panes) :-
    diff_list(Panes, [subproof_stack,hypotheses,proof_control,subproof], []),
    !,
    retractall(current_emacs_proof_panes(_)),
    assert(current_emacs_proof_panes(Panes)),
    ( current_emacs_magic(Kinds), member(proof, Kinds) ->
        %% reset emacs_magic to get hooks updated
        set_emacs_magic(Kinds)
    ;
        true
    ).
set_emacs_proof_panes(Panes) :-
    handle_error(['Bad value ''', w(Panes), '''',
                  ' for parameter ''emacs_proof_panes''.']).

get_emacs_proof_panes(Panes) :-
    current_emacs_proof_panes(Panes),
    !.
get_emacs_proof_panes([subproof_stack,hypotheses,proof_control,subproof]).

set_emacs_transcript_info(Info) :-
    diff_list(Info,
              [window_level, window_id, focus, hyps, newvars, reln, goal], []),
    !,
    retractall(current_emacs_transcript_info(_)),
    assert(current_emacs_transcript_info(Info)).
set_emacs_transcript_info(Info) :-
    handle_error(['Bad value ''', w(Info), '''',
                  ' for parameter ''emacs_transcript_info(_)),
                 ''.']).

get_emacs_transcript_info(Info) :-
    current_emacs_transcript_info(Info),
    !.
get_emacs_transcript_info([focus]).


%% emacs_magic(Command).
%%
%%  Send MAGIC COMMAND to emacs.
%%
%%  mode emacs_magic(@atom).
%% 
emacs_magic(Command) :-
    write('MAGIC '), write(Command), nl.


%%general command
%%  send_theories
%%doc
%%  'Send the names of all theories to emacs.'
%%mode
%%  send_theories.

send_theories :-
    emacs_magic('THEORIES'),
    ( query_global(ancestor(Thy,_)),
      writeq(Thy),
      nl,
      fail
    ;
      true
    ),
    write(global), nl,
    emacs_magic('NORMAL').

%% emacs_showing_help(Action, Doc).
%%
%%  Start/end output to help buffer with heading Doc.
%%
emacs_showing_help(start, Doc) :-
    emacs_magic('HELP'),
    writelist(['%%%%node: ' | Doc]),
    nl.
emacs_showing_help(end, _) :-
    emacs_magic('NORMAL').
    
%% emacs_write_xref(WriteList, Thy, Name).
%% emacs_write_xref(Stream, WriteList, Thy, Name).
%%
%%  Write and mouse-sensitize a help cross-reference.
%%  If Stream is specified, fail unless Strem is the standard output.
%%
%%  mode emacs_write_xref(@writelist, @thyname, @term).
%%  mode emacs_write_xref(@stream, @writelist, @thyname, @term).
%%
emacs_write_xref(Strm, WL, Thy, Name) :-
    stream_property(Strm, alias(user_output)),
    emacs_write_xref(WL, Thy, Name).

emacs_write_xref(WL, Thy, Name) :-
    emacs_magic('XREF'),
    writelist(WL), nl,
    write('COMMAND'), nl,
    writelist(['help(', q(Name), ', ', q(Thy), ')']),
    emacs_magic('END').

%% emacs_write_emph(Stream, WriteList).
%%
%%  Write and emphasize some text.
%%
%%  mode emacs_write_emph(@stream, @writelist).
%%
emacs_write_emph(Strm, WriteList) :-
    stream_property(Strm, alias(user_output)),
    !,
    emacs_magic('EMPH'),
    writelist(WriteList),
    emacs_magic('END').

general command
  synchronize_emacs(Tag)
doc
  'Used by the emacs interface to synchronize the processes.'
mode
  synchronize_emacs(@integer).

synchronize_emacs(Tag) :-
    emacs_magic('SYNC'),
    write(Tag),
    emacs_magic('NORMAL').

%% emacs_echo_input(Caller, Input).
%%
%%  Caller has just read Input.
%%  Echo it.
%%
%%  mode emacs_echo_input(@term, @term).
%%
emacs_echo_input(get_command(_Prompt, OpTable, _Term), Term) :-
    write_term(user_output, Term,
               [remember_name(true),
                op_table(OpTable),
                obvar_prefix_table(OpTable)]),
    write(user_output, '.'),
    nl.
emacs_echo_input(get_ergo_line(_,_), Line) :-
    write_chars(Line, user_output, noquotes),
    nl.

%% emacs_reader_special_execute_hook(Cmd).
%%
%%  Handler for reader special execute commands.
%%  These commands will be generated by the Emacs interface.
%%
emacs_reader_special_execute_hook(send_theories) :-
    send_theories.
emacs_reader_special_execute_hook(emacs_focus_rules(Path)) :-
    emacs_focus_rules(Path).
emacs_reader_special_execute_hook(emacs_rules(Query)) :-
    deabbreviate_cmd([emacs_rules(@query)], emacs_rules(Query), Cmd),
    emacs_hints,
    call_predicate(Cmd).
emacs_reader_special_execute_hook(proof_info) :-
    in_proof(Thy:Name),
    emacs_proof_frame(Thy:Name).
emacs_reader_special_execute_hook(help(T)) :-
    do_help(T,_).
emacs_reader_special_execute_hook(help(T,Thy)) :-
    do_help(T,Thy).
emacs_reader_special_execute_hook(expand_subproof_line(L)) :-
    expand_subproof_line(L).
emacs_reader_special_execute_hook(expand_subproof_stack_line(L)) :-
    expand_subproof_stack_line(L).
emacs_reader_special_execute_hook(expand_hypotheses_line(L)) :-
    expand_hypotheses_line(L).
emacs_reader_special_execute_hook(set_parameter(P,V)) :-
    set_parameter(P, V).
emacs_reader_special_execute_hook(emacs_browse_subproof(N)) :-
    emacs_browse_subproof(N).
emacs_reader_special_execute_hook(emacs_quit_browser) :-
    emacs_quit_browser.
emacs_reader_special_execute_hook(summarize_subproof) :-
    summarize_subproof.
emacs_reader_special_execute_hook(synchronize_emacs(Tag)) :-
    synchronize_emacs(Tag).
emacs_reader_special_execute_hook(emacs_refresh) :-
    emacs_refresh.
emacs_reader_special_execute_hook(emacs_edit_elves(Cmd)) :-
    emacs_edit_elves(Cmd).
emacs_reader_special_execute_hook(emacs_do_deferred_output) :-
    emacs_do_deferred_output.

expand_subproof_line(*) :-
%% TO DO: Understand and rewrite the comment ``just one line'' that is applied
%% to expand_subproof_line, to expand_subproof_stack_line and
%% expand_hypotheses_line.
%% Ray's last comment: These commands are supposed to operate on just a single
%% entry, but instead they force expansion of all entries. (*)
  local_write_depth(0,
     (
         emacs_magic('SUBPROOF'),
         write(clear),nl,
         is_subproof_id(PrfId),
         show_subproof(PrfId, 0, 0, FinalNum),
         emacs_magic('NORMAL'),
         is_focus_id(FinalId),
         ctr_set(emacs_proofstep_id, FinalId),
         ctr_set(emacs_proofstep_num, FinalNum)
    )).

expand_subproof_stack_line(*) :-
  local_write_depth(0,
     (
         is_subproof_id(SID),
         emacs_magic('CONTEXT'),
         show_subproof_stack,
         emacs_magic('NORMAL'),
         ctr_set(emacs_subproof_id, SID)
    )).

expand_hypotheses_line(*) :-
  local_write_depth(0,
     (
         emacs_magic('HYPS'),
         write(clear), nl,
         show_hyps_ascending(0),
         emacs_magic('NORMAL')
    )).

%% emacs_proof(Action, Thy, Name).
%%
%%  Notice that a proof has started or finished.
%%  Prepare to display an appropriate proof frame next time some proof data are
%%  available.
%%
%%  mode emacs_proof(@nonvar, @theory, @rulename).
%%
emacs_proof(started, Thy, Name) :-
    emacs_proof_frame(Thy:Name).
emacs_proof(failed, _, _) :-
    emacs_proof_frame(none).
emacs_proof(finished(_), _, _) :-
    emacs_proof_frame(none).
emacs_proof(aborted, _, _) :-
    emacs_proof_frame(none).
%% The next two are not yet generated by Ergo.
emacs_proof(suspended, _, _) :-
    emacs_proof_frame(none).
emacs_proof(resumed, Thy, Name) :-
    emacs_proof_frame(Thy:Name),
    show_window.

%% emacs_proof_frame(Proof).
%%
%%  Bring up, pop down or reinitialize the proof frame.
%%
%%  mode emacs_proof_frame(@nonvar).
%%
emacs_proof_frame(Thy:Name) :-
    emacs_magic('STARTPROOF'),
    writelist(['theory: ', w(Thy), nl, 'name: ', w(Name), nl]),
    ( get_current_proofs([_,_|_]) ->
        writelist(['extra:   (there are suspended proofs)', nl])
    ;
        true
    ),
    emacs_magic('NORMAL'),
    ctr_set(displayedsubproof, 0),
    ctr_set(emacs_subproof_id, 0),
    ctr_set(emacs_proofstep_id, 0),
    ctr_set(emacs_proofstep_num, 0).
emacs_proof_frame(none) :-
    emacs_magic('ENDPROOF'),
    emacs_magic('NORMAL').

%% emacs_proof_output(Kind, Term, Panes).
%%
%%  Dispose of a line of proof output.
%%
%%  mode emacs_proof_output(@nonvar, @thy(term), @list(atom)).
%%
emacs_proof_output(window, [Thy:Name, WinLev, WinId], _) :-
    emacs_transcript(window_level, WinLev),
    emacs_transcript(window_id, WinId),
    emacs_magic('WINDOW'),
    writelist(['thy: ',   w(Thy), nl,
               'name: ',  w(Name), nl,
               'id: ',    q(WinId), nl,
               'level: ', q(WinLev), nl
              ]),
    emacs_magic('NORMAL').
emacs_proof_output(focus, Focus, Panes) :-
    emacs_transcript(focus, Focus),
    emacs_defer_output(focus, Panes).
emacs_proof_output(allhyps, _, Panes) :-
    emacs_transcript(hyps, 0),
    member(hypotheses, Panes),
    !,
    emacs_defer_output(allhyps, _).
emacs_proof_output(changedhyps, Change, Panes) :-
    emacs_transcript(hyps, Change),
    member(hypotheses, Panes),
    !,
    emacs_defer_output(changedhyps, Change).
emacs_proof_output(newvars, Vs, Panes) :-
    emacs_transcript(newvars, Vs),
    member(proof_control, Panes),
    !,
    emacs_defer_output(newvars, Vs).
emacs_proof_output(reln, R, Panes) :-
    emacs_transcript(reln, R),
    member(proof_control, Panes),
    !,
    emacs_defer_output(reln, R).
emacs_proof_output(goal, G, Panes) :-
    emacs_transcript(goal, G),
    member(proof_control, Panes),
    !,
    emacs_defer_output(goal, G).
emacs_proof_output(param(Name), Value, Panes) :-
    member(proof_control, Panes),
    !,
    query_global(emacs_param(Name, Label, Kind)),
    emacs_update_param(Name, Label, Kind, w(Value)).
emacs_proof_output(proof_step(N,Id,Just,Reln), Focus, _) :-
    get_write_depth(D),
    ( Just == new ->
        true
    ;
        emacs_pstep_msg(Just, JMsg),
        writelist([max_depth(D,[tab(4),q(Reln),tab,tab|JMsg])]),
        nl
    ),
    writelist([q(N),':',tab]),
    ( is_focus_id(Id) ->
        indprint(Focus)
    ;
        writelist([max_depth(D, [thy(Focus)]), nl])
    ).
emacs_proof_output(subproof_stack(_Level,ID,F,_Path), Explanation, _) :-
    get_write_depth(D),
    writelist(['{{{', q(ID), '}}}:',tab,max_depth(D,[thy(F)]),nl,
               tab,tab,tab|Explanation]),
    nl.
emacs_proof_output(hyp(Kind,Num), Hyp, _) :-
    ip_lookup(emacs_in_transcript, Flag),
    nonvar(Flag),
    !,
    get_write_depth(D),
    hyp_marker(Kind, Marker),
    ( Kind = discharged ->
        Lbl = '  hyp  '
    ;
        Lbl = '  conj '
    ),
    writelist([left(11, [w(Lbl), q(Num), chars(Marker)]),
               max_depth(D,[thy(Hyp)]),nl]).


emacs_pstep_msg(tr(T:N,D,_,A),
          ['by ',q(T),':',q(N),', ',q(D),', ',q(A)]).
emacs_pstep_msg(op(P,subprf(Id,_,_,_)),
          ['by subproof {{{', w(Id), '}}} on ',q(P)]).
emacs_pstep_msg(wr(wf(TacId,Msg,_),_,_,subprf(Id,_,_,_)),
          ['by subproof {{{', w(Id), '}}} using tactic ', q(TacId),': '|Msg]) :-
    Msg = [_|_].
emacs_pstep_msg(wr(wf(TacId,[],Tac),_,_,subprf(Id,_,_,_)),
          ['by subproof {{{', w(Id), '}}} using tactic ', q(TacId),': ',q(Tac)]).
emacs_pstep_msg(wr(wa(TacId,Msg,Arg),_,_,subprf(Id,_,_,_)),
          ['by subproof {{{', w(Id), '}}} using tactic ', q(TacId),'.',q(Arg),'  '|Msg]).

emacs_transcript(Field, Value) :-
    get_emacs_transcript_info(Info),
    member(Field, Info),
    !,
    emacs_transcript1(Field, Value).
emacs_transcript(_, _).

emacs_transcript1(window_level, L) :-
    writelist(['  level:   ', w(L),   nl]).
emacs_transcript1(window_id, I) :-
    writelist(['  window:  ', w(I),   nl]).
emacs_transcript1(focus, F) :-
    writelist(['  focus:   ', thy(F), nl]).
emacs_transcript1(newvars, Vs) :-
    writelist(['  newvars: ', w(Vs),  nl]).
emacs_transcript1(reln, R) :-
    writelist(['  reln:    ', w(R),   nl]).
emacs_transcript1(goal, G) :-
    writelist(['  goal:    ', thy(G), nl]).
emacs_transcript1(hyps, N) :-
    Retained is N-1,
    ( Retained > 0 ->
        writelist(['  ', 'hyps 1..', w(Retained), '  <unchanged>', nl])
    ;
        true
    ),
    ip_set(emacs_in_transcript, yes),
    show_hyps_ascending(N),
    ip_set(emacs_hyp_label, _).

?- dynamic(emacs_deferred_output/2).
emacs_defer_output(Kind, Data) :-
    retractall(emacs_deferred_output(Kind, _)),
    assert(emacs_deferred_output(Kind, Data)).

emacs_do_deferred_output :-
    retract(emacs_deferred_output(Kind, Data)),
    ( in_proof ->
        %%writelist(['doing deferred output: ', w(Kind), w(Data), nl]),
        emacs_do_deferred_output(Kind, Data)
    ;
        %%writelist(['discarding deferred output: ', w(Kind), w(Data), nl]),
        true
    ),
    fail.
emacs_do_deferred_output :-
    %%writelist(['done all deferred output', nl]),
    true.

emacs_do_deferred_output(focus, Panes) :-
    ctr_get(emacs_proofstep_id, DisplayedId),
    ctr_get(emacs_proofstep_num, DisplayedNum),
    is_subproof_id(PrfId),
    ( \+ member(subproof, Panes) ->
        true
    ; in_curr_subproof(DisplayedId) ->
        emacs_magic('SUBPROOF'),
        show_subproof(PrfId, DisplayedId, DisplayedNum, FinalNum),
        emacs_magic('NORMAL')
    ;
        emacs_magic('SUBPROOF'),
        write(clear),nl,
        show_subproof(PrfId, 0, 0, FinalNum),
        emacs_magic('NORMAL')
    ),
    is_focus_id(FinalId),
    ctr_set(emacs_proofstep_id, FinalId),
    ctr_set(emacs_proofstep_num, FinalNum),
    is_subproof_id(SID),
    ( ctr_get(emacs_subproof_id, SID) ->
        %% no need to update it.
        true
    ; member(subproof_stack, Panes) ->
        emacs_magic('CONTEXT'),
        show_subproof_stack,
        emacs_magic('NORMAL'),
        ctr_set(emacs_subproof_id, SID)
    ;
        ctr_set(emacs_subproof_id, SID)
    ),
    emacs_auto_hints,
    emacs_auto_rules.

emacs_do_deferred_output(allhyps, _) :-
    emacs_magic('HYPS'),
    write(clear), nl,
    show_hyps_ascending(0),
    emacs_magic('NORMAL').

emacs_do_deferred_output(changedhyps, Change) :-
    emacs_magic('HYPS'),
    writelist(['delete ', w(Change), nl]),
    show_hyps_ascending(Change),
    emacs_magic('NORMAL').

emacs_do_deferred_output(newvars, Vs) :-
    query_global(emacs_param(newvars, Label, artificial)),
    %% link up the names!
    ip_lookup(newvars, Vs),
    emacs_update_param(newvars, Label, artificial, w(Vs)).

emacs_do_deferred_output(reln, R) :-
    query_global(emacs_param(reln, Label, artificial)),
    emacs_update_param(reln, Label, artificial, w(R)).

emacs_do_deferred_output(goal, G) :-
    query_global(emacs_param(goal, Label, artificial)),
    %% link up the names!
    ip_lookup(goal, G),
    emacs_update_param(goal, Label, artificial, thy(G)).



general command
  emacs_button(Label, Command)
extends
  general_cmds
doc
  'Add or remove a button from the Emacs interface.',nl,
  ' Label is the label of the button to add or remove.',nl,
  ' Command is an atom giving the command the new button should generate',nl,
  '  or a variable if a button should be deleted.'
mode
  emacs_button(@atom, @atom),
  emacs_button(@atom, @var).

emacs_button(Label, Command) :-
    retractall_global(emacs_button(Label, _)),
    ( var(Command) ->
        Cmd = ''
    ;
        Cmd = Command,
        tactic_assert_global(emacs_button(Label, Cmd))
    ),
    emacs_add_button(Label, Cmd).

emacs_add_button(Label, Cmd) :-
    emacs_magic('BUTTON'),
    writelist([w(Label), :, w(Cmd), nl]),
    emacs_magic('NORMAL').

emacs_add_saved_buttons :-
    (
        query_global(emacs_button(Label, Cmd)),
        emacs_add_button(Label, Cmd),
        fail
    ;
        true
    ).

general command
  emacs_param(Name, Label, Kind)
extends
  general_cmds
doc
  'Add or remove a parameter from the Emacs interface.',nl,
  ' Name is the name of the parameter to add or remove.',nl,
  ' Label is the label that will appear in the parameter list.',nl,
  ' Kind is an atom: ''numeric'', ''normal'' or ''delete''.'
mode
  emacs_param(@atom, @atom, @atom).

emacs_param(Name, Label, Kind) :-
    retractall_global(emacs_param(Name, _, _)),
    ( Kind = delete ->
        true
    ;
        emacs_get_param(Name, Kind, Value),
        tactic_assert_global(emacs_param(Name, Label, Kind))
    ),
    emacs_update_param(Name, Label, Kind, w(Value)).

emacs_get_param(_, artificial, _) :-
    !.
emacs_get_param(_, delete, _) :-
    !.
emacs_get_param(verbosity, _, Val) :-
    !,
    %% ignore less_verbose!!
    msg_verbosity(Val).
emacs_get_param(Param, _, Val) :-
    get_parameter(Param, Val).

emacs_update_param(Name, Label, Kind, Value) :-
    emacs_magic('PARAM'),
    writelist([w(Label), ': ', w(Name), tab, w(Kind), tab, Value, nl]),
    emacs_magic('NORMAL').
emacs_update_param(Name, Label, Kind, Value).

emacs_add_saved_params :-
    (
        query_global(emacs_param(Name, Label, Kind)),
        emacs_get_param(Name, Kind, Value),
        emacs_update_param(Name, Label, Kind, w(Value)),
        fail
    ;
        true
    ).

%% emacs_log(WriteList).
%%
%%  Write WriteList into the proof script buffer, if there is one and logging
%%  is enabled.
%%
%%  mode emacs_log(@writelist).
%%
emacs_log(WL) :-
    emacs_magic('LOG'),
    writelist(WL),
    emacs_magic('NORMAL').

emacs_prover_hook(exit).
emacs_prover_hook(resume) :-
    emacs_prover_hook(startup).
emacs_prover_hook(startup) :-
    add_handler(reader_special_execute_hook(emacs(Call)),
                emacs_reader_special_execute_hook(Call)),
    load_global_tactic_file(emacs_index),
    load_global_tactic_file(emacs_browser),
    load_global_tactic_file(emacs_rules),
    load_global_tactic_file(emacs_elves),
    ( global_fact(emacs_button, 2) ->
        emacs_add_saved_buttons
    ;
        new_global_fact(emacs_button, 2, [assert(z),retract(true),save(true)]),
        emacs_button('Undo',  undo),
        emacs_button('Close', close),
        emacs_button('Goal',  goal)
    ),
    ( global_fact(emacs_auto_hints, 1) ->
        true
    ;
        new_global_fact(emacs_auto_hints, 1, [retract(true),save(true)])
    ),
    ( global_fact(emacs_rules_depth, 1) ->
        true
    ;
        new_global_fact(emacs_rules_depth, 1, [retract(true),save(true)])
    ),
    ( global_fact(emacs_auto_rules, 1) ->
        true
    ;
        new_global_fact(emacs_auto_rules, 1, [retract(true),save(true)])
    ),
    ( global_fact(emacs_param, 3) ->
        emacs_add_saved_params
    ;
        new_global_fact(emacs_param, 3, [assert(z),retract(true),save(true)]),
        emacs_param(goal,             'Goal',        artificial),
        emacs_param(newvars,          'New vars',    artificial),
        emacs_param(reln,             'Relation',    artificial),
        emacs_param(write_depth,      'Write depth', numeric),
        emacs_param(verbosity,        'Verbosity',   numeric)
    ).

general command
  emacs_refresh
extends
  proof_info_cmds
doc
  'Refresh all current proof information in the Emacs interface.',nl,
  'Currently, the Emacs interface cannot always detect when variables',nl,
  'have been instantiated or uninstantiated, so it is often necessary to',nl,
  'use this command, or the corresponding Emacs menu item, to update',nl,
  'the displayed portions of the proof after commands that instantiate',nl,
  'variables, or after undoing steps'
mode
  emacs_refresh.

emacs_refresh :-
    in_proof(Proof),
    !,
    emacs_proof_frame(Proof),
    show_window,
    emacs_do_deferred_output.
emacs_refresh :-
    emacs_proof_frame(none).

library command
  emacs_dialog(Prompt, Answers, Default, Answer)
extends
  proof_info_cmds
doc
  'Ask a question, with specified Prompt and possible Answers.',nl,
  'The answer Default will be highlighted.',nl,
  'and the selected answer will be returned as Answer.'
mode
  emacs_dialog(@writelist, @list(atom), @atom, -atom).

emacs_dialog(Prompt, Answers, Default, Answer) :-
    repeat,
      emacs_magic('DIALOG'),
      writelist(Prompt),
      writelist([nl]),
      emacs_dialog(Answers, Default),
      emacs_magic('NORMAL'),
      get_line(user_input, AnswerLine),
      string_to_atom(AnswerLine, Answer),
    member(Answer, Answers),
    !.

emacs_dialog([], _).
emacs_dialog([A|As], D) :-
    ( D == A ->
        writelist(['{{{default}}} ', w(A), nl])
    ;
        writelist(['{{{answer}}} ', w(A), nl])
    ),
    emacs_dialog(As, D).

library command
  emacs_send_string(WL)
extends
  proof_info_cmds
doc
  'Send WL as an echoed command to Ergo, via Emacs.'
mode
  emacs_send_string(@writelist).

emacs_send_string(WL) :-
    emacs_magic('SENDSTRING'),
    writelist(WL),
    emacs_magic('NORMAL').  

?- add_listener(prover(Event), emacs_prover_hook(Event)).
