%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##


%% Emacs proof browser.
%% $Id: emacs_browser.ql,v 1.18 2004/07/25 04:16:13 itp Exp $

#include "../prover/lib/ergo.h"

database command
  emacs_browse_proof(Name)
doc
  'Browse the (saved) proof of theorem Name.'
extends
  general_cmds
mode
  emacs_browse_proof(@atom).

emacs_browse_proof(Name) :-
    concat_atom([Name,'~%~'], Name1),
    with_local_exception_handler(
%    catch(
        (
            %% cf. rerun_proof/2
            add_theorem(Name1, [], Stmt, no,
                        (
                            demote_var(Stmt),
                            msg(browser, ['rerunning proof of ', w(Name)]),
                            rerun(true,Name),
                            msg(browser, ['browsing proof of ', w(Name)]),
                            emacs_browse_subproof(1),
                            get_command('', user, _),
                            emacs_quit_browser
                       ))),
%        signal(_,_),
         sigint_exception(_,_,_),
        (
            emacs_magic('NORMAL')
       )).

general command
  emacs_quit_browser
mode
  emacs_quit_browser.

emacs_quit_browser :-
    emacs_magic('BROWSEQUIT'),
    emacs_magic('NORMAL').

proof command
  emacs_browse_subproof(Id)
doc
  'Browse the subproof numbered Id.'
mode
  emacs_browse_subproof(@integer).

emacs_browse_subproof(N) :-
    find_subproof(N, Proofs, Steps, Hyps, Reln),
    !,
    emacs_magic('BROWSEANCS'),
    emacs_browse_subproof_ancestors(Proofs),
    emacs_magic('BROWSEHYPS'),
    reverse(Hyps, HypsR),
    emacs_browse_subproof_hyps(HypsR),
    emacs_magic('BROWSESUBP'),
    writelist(['{{{', w(N), '}}}', nl]),
    reverse(Steps, RSteps),
    show_pretty_reln(Reln, PReln),
    emacs_browse_subproof_steps(RSteps, PReln),
    emacs_magic('NORMAL').
emacs_browse_subproof(_) :-
    handle_error([nl,'Cannot find a subproof to browse!']).

emacs_browse_subproof_ancestors([subpf(_,_,_,_,_,Reason)]) :-
    !,
    ( Reason = new ->
        true
    ;
        writelist([tab, tab, tab, w(Reason), nl])
    ).
emacs_browse_subproof_ancestors([subpf(Id,_Hyps,Steps,_Reln,_Step,Reason),
                                 Anc | Ancs]) :-
    ( Reason = new ->
        true
    ;
        writelist([tab, tab, tab, w(Reason), nl])
    ),
    Anc = subpf(_,_,_,_,NextStep,_),
    %% If the next step is a backward step (i.e. open_pos or wrapped_tactic),
    %% we can show the correct focus.
    %% If the next step is a forward step (i.e. open_hyp or conjecture), we do
    %% not know what the parent focus was when the step took place (like as
    %% not, we do not care), so we show the initial focus.
    ( append(_, [pstep(NextStep,_,_), pstep(_,Focus,_) | _], Steps) ->
        true
    ;
        append(_, [pstep(_,Focus,_)], Steps)
    ),
    !,
    writelist(['{{{', w(Id), '}}}:   ', thy(Focus), nl]),
    emacs_browse_subproof_ancestors([Anc|Ancs]).

emacs_browse_subproof_hyps([]).
emacs_browse_subproof_hyps([hrec(Num,Term,DelFlag,Just,_CD,_UD)|Hyps]) :-
    writelist([w(Num), ':', tab, thy(Term), nl, tab]),
    ( DelFlag == deleted ->
        writelist(['(deleted)',tab])
    ;
        writelist([tab,tab])
    ),
    ( Just = i ->
        write(inherited)
    ; Just = cj(V), var(V) ->
        write('undischarged conjecture')
    ; Just = cj(subprf(JId,_,_,_)) ->
        writelist(['conjecture discharged by subproof {{{', w(JId), '}}}'])
    ; Just = cj(tr(JRule,JDir,_,JAnts)) ->
        writelist(['conjecture discharged by d_trans',
                   '(hyp:', w(Num), ', ', w(JRule), ', ',
                     w(JDir), ', ', w(JAnts), ')'])
    ; Just = new ->
        write('introduced by opening rule.')
    ; Just = sl(SN) ->
        writelist(['by splitting hyp ', w(SN)])
    ; Just = sr(SN) ->
        writelist(['by splitting hyp ', w(SN)])
    ; Just = oh(FN, subprf(FId, _, _, _)) ->
        writelist(['by subproof {{{', w(FId), '}}} from hyp ', w(FN)])
    ; Just = ht(HN, HR, HD, HA) ->
        writelist(['by ', q(h_trans(HN, HR, HD, HA))])
    ; Just = tr(PN) ->
        writelist(['by translation of parent hyp ', w(PN)])
    ),
    nl,
    emacs_browse_subproof_hyps(Hyps).

emacs_browse_subproof_steps([], _).
%% TO DO: Generate line numbers sequentially in emacs_brwose_subproof_steps.
emacs_browse_subproof_steps([pstep(Num,Foc,Just)|Steps], Reln) :-
    ( Just = new ->
        true
    ;
        emacs_pstep_msg(Just, Msg),
        writelist(['    ', w(Reln), tab, tab | Msg])
    ),
    writelist([nl, w(Num), ': ', thy(Foc), nl]),
    emacs_browse_subproof_steps(Steps, Reln).
      

general command
  summarize_proof(Name),
  summarize_proof(Name, File)
extends
  general_cmds
doc
  'Write a transcript of saved proof Name in File (default: standard output).'
mode
  summarize_proof(@atom),
  summarize_proof(@atom, @filename).

summarize_proof(Name) :-
    current_output(Strm),
    summarize_proof(Name, Strm).
summarize_proof(Name, File) :-
    atom(File),
    !,
    open(File, write, Strm),
    unwind_protect(summarize_proof(Name, Strm),
                   close(Strm)
                  ).
summarize_proof(Name, Strm) :-
    concat_atom([Name,'~%~'], Name1),
    with_local_exception_handler(   
    % catch(
        (
            %% cf. rerun_proof/2
            add_theorem(Name1, [], Stmt, no, (
                demote_var(Stmt),
                msg(summarize, ['rerunning proof of ', w(Name)]),
                rerun(true,Name),
                msg(summarize, ['summarizing proof of ', w(Name)]),
                writelist(Strm, ['Summary of proof of ',w(Name),nl,nl,
                            '----------------------------------------',nl,nl]),
                summarize_subproof(1, Strm)
                                            ))),
%        signal(_,_),
         sigint_exception(_,_,_)
       ).

proof command
  summarize_subproof,
  summarize_subproof(Id)
extends
  proof_info_cmds
doc
  'Write a transcript of subproof Id (default: current)',nl,
  'and all its descendents.'
mode
  summarize_subproof,
  summarize_subproof(@integer).

summarize_subproof :-
    current_output(Strm),
    ip_lookup(proofid, Id),
    ip_lookup(hyplist, Hyps),
    ip_lookup(proof, Steps),
    ip_lookup(relation, Reln),
    summarize_proof(Id, Hyps, Steps, Reln, [], Sub),
    sort(Sub, Sub1),
    write_proof_summary(Sub1, Strm).

summarize_subproof(Id) :-
    current_output(Strm),
    summarize_subproof(Id, Strm).

summarize_subproof(Id, Strm) :-
    find_subproof(Id, _, Steps, Hyps, Reln),
    summarize_proof(Id, Hyps, Steps, Reln, [], Sub),
    sort(Sub, Sub1),
    write_proof_summary(Sub1, Strm).

write_proof_summary([], _).
write_proof_summary([subproof(Id, HLines, SLines)|SP], Strm) :-
    writelist(Strm, ['Subproof ', w(Id), nl]),
    reverse(HLines, HLinesR),
    write_proof_lines(HLinesR, Strm), nl(Strm),
    reverse(SLines, SLinesR),
    write_proof_lines(SLinesR, Strm), nl(Strm),
    writelist(Strm, ['----------------------------------------',nl,nl]),
    write_proof_summary(SP, Strm).

write_proof_lines([], _).
write_proof_lines([Line|Lines], Strm) :-
    writelist(Strm, Line), nl(Strm),
    write_proof_lines(Lines, Strm).

summarize_proof(Id, Hyps, Steps, Reln, Sub0, Sub) :-
    Sub1 = [subproof(Id, HLines, SLines)|Sub0],
    summarize_proof_hyps(Id, Hyps, HLines, Sub1, Sub2),
    show_pretty_reln(Reln, PReln),
    summarize_proof_steps(Id, Steps, PReln, SLines, Sub2, Sub).

summarize_proof_steps(Id, [], _, [], Sub, Sub).
summarize_proof_steps(Id, [Step|Steps], Reln, [Line|Lines], Sub0, Sub) :-
    summarize_proof_step(Id, Step, Reln, Line, Sub0, Sub1),
    summarize_proof_steps(Id, Steps, Reln, Lines, Sub1, Sub).

summarize_proof_step(Id, pstep(Num, Term, Just), Reln, Line, Sub0, Sub) :-
    summarize_proof_just(Just, Reln, JLine, Sub0, Sub),
    append(JLine, [nl, w(Id), '.', w(Num), '  ', thy(Term)], Line).

summarize_proof_just(new, _, [], Sub, Sub) :-
    !.
summarize_proof_just(Trans, Reln, [q(Reln),tab|Msg], Sub, Sub) :-
    Trans = tr(_,_,_,_),
    !,
    pstep_msg(Trans, Msg).
summarize_proof_just(Open, Reln, [q(Reln),tab|Msg], Sub0, Sub) :-
    Open = op(_, subprf(Id, Hyps, Steps, SReln)),
    !,
    pstep_msg(Open, Msg),
    summarize_proof(Id, Hyps, Steps, SReln, Sub0, Sub).
summarize_proof_just(Wraptac, Reln, [q(Reln),tab|Msg], Sub0, Sub) :-
    Wraptac = wr(_,_,_,subprf(Id, Hyps, Steps, SReln)),
    !,
    pstep_msg(Wraptac, Msg),
    summarize_proof(Id, Hyps, Steps, SReln, Sub0, Sub).

summarize_proof_hyps(Id, [], [], Sub, Sub).
summarize_proof_hyps(Id, [hrec(Num, Term, Del, Just, _, _)|Hyps],
                    [Line|Lines], Sub0, Sub) :-
    Line = [w(Id), '.h', w(Num), ': ', thy(Term), nl, DelMsg, tab | JLine],
    ( var(Del) ->
        DelMsg = tab
    ;
        DelMsg = '(deleted)'
    ),
    summarize_proof_hyp_just(Just, JLine, Sub0, Sub1),
    summarize_proof_hyps(Id, Hyps, Lines, Sub1, Sub).

summarize_proof_hyp_just(i,
                         ['inherited'],
                         Sub, Sub).
summarize_proof_hyp_just(cj(J),
                         ['undischarged conjecture'],
                         Sub, Sub) :-
    var(J),
    !.
summarize_proof_hyp_just(cj(tr(JRule,JDir,_,JAnts)),
                         ['conjecture discharged by d_trans',
                          '(hyp:', w(Num), ', ', w(JRule), ', ',
                            w(JDir), ', ', w(JAnts), ')'],
                         Sub, Sub).
summarize_proof_hyp_just(cj(subprf(JId, JHyps, JSteps, JReln)),
                         ['conjecture discharged by subproof ', w(JId)],
                         Sub0, Sub) :-
    summarize_proof(JId, JHyps, JSteps, JReln, Sub0, Sub).
summarize_proof_hyp_just(new,
                         ['introduced by opening rule.'],
                         Sub, Sub).
summarize_proof_hyp_just(sl(N),
                         ['by splitting hyp ', w(N)],
                         Sub, Sub).
summarize_proof_hyp_just(sr(N),
                         ['by splitting hyp ', w(N)],
                         Sub, Sub).
summarize_proof_hyp_just(oh(N, subprf(Id, Hyps, Steps, Reln)),
                         ['by subproof ', w(Id), ' from hyp ', w(N)],
                         Sub0, Sub) :-
    summarize_proof(Id, Hyps, Steps, Reln, Sub0, Sub).
summarize_proof_hyp_just(ht(HN, HR, HD, HA),
                         ['by ', q(h_trans(HN, HR, HD, HA))],
                         Sub, Sub).
summarize_proof_hyp_just(tr(PN),
                         ['by translation of parent hyp ', w(PN)],
                         Sub, Sub).
