%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##


%% Support for listing rules and building queries.
%% $Id: emacs_rules.ql,v 1.22 2004/07/25 04:16:14 itp Exp $

#include "../prover/lib/ergo.h"

?- ( query_global(request(list_matching_rule(_,_,_,_,_,_,_,_,_,_), _)) ->
       true
   ;
       add_request(list_matching_rule(_,_,_,_,_,_,_,_,_,_),
                   ['list_matching_rule(Kind, Thy, Name, Dirn, Cond,',nl,
                    '               CTerms, ITerm, RuleReln, OTerm, Body).',nl,
                    'Send the details of a matched rule to emacs.'
                   ])
   ).

proof command
  emacs_focus_rules,
  emacs_focus_rules(Path)
doc
  'List all rules that tightly match the current focus',nl,
  'or the specified subterm theoreof.'
mode
  emacs_focus_rules,
  emacs_focus_rules(@subterm_path).

emacs_focus_rules :-
    emacs_focus_rules([]).

emacs_focus_rules(Path) :-
    is_focus(F),
    !,
    subterm_path(F, Path, Input),
    %% Don't specify relation unless Path=[]
    %%   (should really calculate correct relation using openrule???)
    ( Path = [] ->
        is_reln(Reln),
        cutoff_theory(Input, Reln, Cutoff, _)
    ;
        Cutoff = ergo
    ),
    %% Restrict to tight matching unless input is a variable
    ( var(Input) ->
        Matching = []
    ;
        Matching = [matching(tight)]
    ),
    emacs_rules([iterm(Input),
                 reln(Reln),
                 thys([hyp,upto(Cutoff)]) |
                 Matching]).
%% This can be called from outside a proof by the emacs interface.
emacs_focus_rules(_) :-
    emacs_rules([thys([])]).

general command
  emacs_rules(Query)
doc
  'List all rules matching Query.'
mode
  emacs_rules(@query).

emacs_rules(Query) :-
    emacs_magic('QUERY'),
    emacs_preprocess_query(Query, Query1),
    emacs_write_query(Query1),
    emacs_magic('RULES'),
    with_local_exception_handler(
        thawing_new_vars(emacs_rules1(Query1)),
        %     signal(_Sev,_Goal,_Msg,'SIGINT'),
        sigint_exception(_,_,_),
        (
            writelist([nl, '[[Interrupted]]', nl]),
            get0(0'c)
       )),
    emacs_magic('NORMAL').

emacs_rules1(Query) :-
    query_init(emacs_rules),
    query_process([input_name(null)]),
    query_process(Query),
    query_choose_rule(Thy, Name),
    itrans_conditional(Cond),
    itrans_direction(Dirn),
    get_emacs_rules_depth(Depth),
    itrans_finish(browse(Depth, _Res)),   %make this a ergo-parameter
%    itrans_finish(null), 
    itrans_kind(Kind),
    itrans_input(ITerm),
    itrans_rule_relation(RuleReln),
    show_pretty_reln(RuleReln, RuleRelnPretty),
    itrans_output(OTerm),
    itrans_condlist(CTerms),
    query_check_matching,
    itrans_input_shape(IShape),
    itrans_output_shape(OShape),
    itrans_body(Body),
    list_matching_rule(Kind, Thy, Name, Dirn, Cond, IShape, OShape,
                       CTerms, ITerm, RuleRelnPretty, OTerm, Body), 
    fail.
emacs_rules1(_).

%% private emacs_preprocess_query(Query0, Query).
%%
emacs_preprocess_query([], []).
emacs_preprocess_query([C|Q0], Q) :-
    ( C=Thy:Name ->
        Q = [thys([Thy]), name(Name) | Q1]
    ; member(C, [c,u]) ->
        Q = [cond(C) | Q1]
    ; member(C, [f,r,t,rt]) ->
        Q = [dirn(C) | Q1]
    ; atom(C) ->
        Q = [name(C) | Q1]
    ; is_chars(C) ->
        Q = [name(C) | Q1]
    ;
        Q = [C | Q1]
    ),
    emacs_preprocess_query(Q0, Q1).

%% private emacs_write_query(Query).
%% private emacs_write_query_constraint(Constraint).
%% private emacs_write_query_constraint(Constraint, Quote, Value, Values).
%% private emacs_write_query_constraint(Values, Quote).
%% private emacs_insert_constraint(Value, Values0, Values).
%% 
emacs_write_query([]).
emacs_write_query([C|Q]) :-
    emacs_write_query_constraint(C),
    emacs_write_query(Q).

emacs_write_query_constraint(thys(Thys)) :-
    emacs_write_query_constraint(thys, w, Thys).
emacs_write_query_constraint(name(Name)) :-
    emacs_write_query_constraint(name, q, Name).
emacs_write_query_constraint(dirn(Dirn)) :-
    emacs_write_query_constraint(dirn, w, Dirn).
emacs_write_query_constraint(cond(Cond)) :-
    emacs_write_query_constraint(cond, w, Cond).
emacs_write_query_constraint(reln(Reln)) :-
    emacs_write_query_constraint(reln, q, Reln).
emacs_write_query_constraint(frozen(Flag)) :-
    emacs_write_query_constraint(frozen, w, Flag).
emacs_write_query_constraint(matching(Matching)) :-
    emacs_write_query_constraint(matching, w, Matching).
emacs_write_query_constraint(imatch(Min,Max)) :-
    emacs_write_query_constraint(imatch, w, (Min,Max)).
emacs_write_query_constraint(omatch(Min,Max)) :-
    emacs_write_query_constraint(omatch, w, (Min,Max)).
emacs_write_query_constraint(i(Rule)) :-
    emacs_write_query_constraint(i, q, Rule).
emacs_write_query_constraint(c(Rule)) :-
    emacs_write_query_constraint(c, q, Rule).
emacs_write_query_constraint(iterm(IT)) :-
    emacs_write_query_constraint(iterm, thy, IT).
emacs_write_query_constraint(oterm(OT)) :-
    emacs_write_query_constraint(oterm, thy, OT).
emacs_write_query_constraint(cterm(CT)) :-
    emacs_write_query_constraint(cterm, thy, CT).
emacs_write_query_constraint(subterm(ST)) :-
    emacs_write_query_constraint(subterm, thy, ST).
emacs_write_query_constraint(conjs(C)) :-
    emacs_write_query_constraint(conjs, w, C).

emacs_write_query_constraint(C, Q, V) :-
    writelist([w(C), ': ']),
        ( thawed_var(V) ->
        writelist(['_'])
    ;
        writelist([Q(V)])
    ),
    nl.


%% list_matching_rule(Kind, Thy, Name, Dirn, Cond, IShape, OShape,
%%                    CTerms, ITerm, RuleReln, OTerm, Body).
%%
%% Write all rule details (listing and command) for Emacs.
%% 
list_matching_rule(K, T, N, D, C, _IS, _OS, CTs, IT, RR, OT, B) :-
    handle_event(list_matching_rule(K, T, N, D, C, CTs, IT, RR, OT, B)),
    !.
list_matching_rule(K, T, N, D, C, IS, OS, CTs, IT, RR, OT, B) :-
    list_matching_rule_data(K, T, N, D, C, IS, OS, CTs, IT, RR, OT, B),
    writelist(['end', nl]).

%% list_matching_rule_data(Kind, Thy, Name, Dirn, Cond, IShape, OShape,
%%                         CTerms, ITerm, RuleReln, OTerm, Body).
%%
%% Write data for rule listing for emacs.
%% 
list_matching_rule_data(K, T, N, D, C, IS, OS, CTs, IT, RR, OT, B) :-
    writelist(['kind: ',  w(K), nl,
               'rule: ',  w(T), ':', q(N), nl,
               'dirn: ',  w(D), nl,
               'cond: ',  w(C), nl]),
    list_matching_rule_conds(CTs),
    writelist(['iterm: ', thy(IT), nl,
               'ishape: ', w(IS), nl,
               'oshape: ', w(OS), nl,
               'rreln: ', thy(RR), nl,
               'oterm: ', thy(OT), nl]),
    ( B = true ->
        true
    ;
        writelist(['body: ',  code(B), nl])
    ).

%% private list_matching_rule_conds(CTerms).
%% 
list_matching_rule_conds([]).
list_matching_rule_conds([CT|CTs]) :-
    writelist(['cterm: ', thy(CT), nl]),
    list_matching_rule_conds(CTs).

%% private cutoff_theory(Focus, Reln, Cutoff, Kind).
%%
%%  Uses Focus and/or Reln to decide the Cutoff theory.
%%  The hope is that most of the interesting rules about Focus
%%  and Reln will be in Cutoff or its descendent theories.
%%  Also returns Kind=nonvar or Kind=var to indicate
%%  whether or not Focus is some kind of variable.
%%
%%  mode cutoff_theory(@thy(term), @ground, ?theory, ?atom).
%%
cutoff_theory(Focus, Reln, Cutoff, nonvar) :-
    outermost_name(Focus, FocusName, _FocusArity),
    !,
    internal_name(FocusName, FThy, _),
    reln_thy(Reln, RThy),
    ( ancestor_of(RThy, FThy) ->
        Cutoff = RThy
    ;
        Cutoff = FThy
    ).
cutoff_theory(_, Reln, RThy, var) :-
    %% Focus is a var/obvar.
    reln_thy(Reln, RThy).

outermost_name(T, T, 0) :-
    atom(T).
outermost_name(T, Fun, Ar) :-
    compound(T),
    functor(T, Fun, Ar),
    atom(Fun).
outermost_name(T, Q, quant) :-
    quant(T),
    %% strip off any subs, since quantify/4 doesn't like them.
    substitute(T, _, T2),
    quantify(T2, Q, _, _),
    atom(Q).


%% reln_thy(Reln, Thy).
%%
%%  Reln is a window relation, Thy is the theory it comes from.
%%
%%  mode reln_thy(@nonvar, -thyname).
%%
reln_thy(Rel, Thy) :-
    atom(Rel),
    !,
    internal_name(Rel, Thy, _).
reln_thy(rev(Rel), Thy) :-
    !,
    reln_thy(Rel, Thy).

emacs_auto_rules parameter
  [set_emacs_auto_rules, get_emacs_auto_rules]
doc
  'Global parameter that controls what theories are automatically',nl,
  'searched for rules by the emacs interface at each proof step.',nl,
  'Value is a ', xref(theory_search_list, global, theory_search_list), '.',nl,
  'The default value is ''[]'', so no searching is done.',nl
extends
  parameters.

set_emacs_auto_rules(TSL) :-
    retractall_global(emacs_auto_rules(_)),
    tactic_assert_global(emacs_auto_rules(TSL)).

get_emacs_auto_rules(TSL) :-
    ( query_global(emacs_auto_rules(Val)) ->
        TSL = Val
    ;
        TSL = []
    ).

emacs_auto_rules :-
    get_emacs_auto_rules(TSL),
    emacs_auto_rules(TSL).

emacs_auto_rules([]) :-
    !,
    emacs_magic('RULES'),
    writelist(['[[Out of date]]', nl]),
    emacs_magic('NORMAL').
emacs_auto_rules(TSL) :-
    is_focus(Input),
    is_reln(Reln),
    emacs_rules([iterm(Input), thys(TSL), reln(Reln), matching(tight)]).

emacs_rules_depth parameter
  [set_emacs_rules_depth, get_emacs_rules_depth]
extends
  parameters
doc
  'The emacs_rules_depth parameter determines how deep the',nl,
  'the unification problems should be solved via incomplete_retry_delays.',nl,
  'The default level is 5 for a fast search.'.


set_emacs_rules_depth(N) :-
    retractall_global(emacs_rules_depth(_)),
    tactic_assert_global(emacs_rules_depth(N)).

get_emacs_rules_depth(N) :-
    ( query_global(emacs_rules_depth(N)) ->
        true
    ;
        N = 5
    ).

%
%%%%%%%%%%%
%% Hints %%
%%%%%%%%%%%
proof command
  emacs_hints,
  emacs_hints(N)
extends
  elves_cmds
doc
  'Displays all suggestions from the first N elves in the current',nl,
    xref(elves,global,'parameter:elves'),
  ' list, without trying to execute those',nl,
  'suggestions.  If N is not given, then all elves in the current',nl,
  'elves list are tried.'
mode
  emacs_hints,
  emacs_hints(@integer).

emacs_auto_hints parameter 
  [set_emacs_auto_hints, get_emacs_auto_hints]
doc
  'Global parameter that controls the maximum number of elves hints that',nl,
  'are automatically displayed by the emacs interface at each proof step.',nl,
  'The default value is 0, so no hints are displayed by default.'
extends
  parameters.

set_emacs_auto_hints(N) :-
    retractall_global(emacs_auto_hints(_)),
    tactic_assert_global(emacs_auto_hints(N)).

get_emacs_auto_hints(N) :-
    ( query_global(emacs_auto_hints(N)) ->
        true
    ;
        N = 0
    ).

%% This cache is duplicated on the emacs side, and minimizes retransmission of
%% elf documentation.
?- dynamic(emacs_elf_doc_cache/2).

emacs_auto_hints :-
    get_emacs_auto_hints(N),
    get_active_elves(M),
    N > M,
    !,
    emacs_hints(M, N).
emacs_auto_hints.

emacs_hints :-
    get_elves_list(EL),
    length(EL, N),
    emacs_hints(0, N).
emacs_hints(N) :-
    emacs_hints(0, N).

emacs_hints(M, N) :-
    N > M,
    M1 is M + 1,
    emacs_magic('HINTS'),
    get_elves_list(EL),
    is_focus(F),
    is_goal(G),
    between(M1, N, Num),
    nth1(Num,EL,Elf),
    call_predicate(Elf(F,G,Msg,Action)),
    elf_documentation(Elf, Help),
    writelist(['elf: ', w(Elf), nl]),
    writelist(['action: ', w(Action), nl]),
    writelist(['msg: ' | Msg]),
    writelist([nl]),
    ( emacs_elf_doc_cache(Elf, Help) ->
        writelist(['{{{cached}}}'])
    ;
        assert(emacs_elf_doc_cache(Elf, Help)),
        writelist(Help)
    ),
    writelist([nl, 'end: ', nl]),
    fail.
emacs_hints(_, _) :-
    emacs_magic('NORMAL').

elf_documentation(Elf, Help) :-
    split_atom(Elf, Thy, :, Name),
    atom_concat2('elf:', Name, Node),
    query_theory(doc_node(Node,_,Help,_), Thy),
    !.
