%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% General purpose commands that should be available in all
%% user interfaces.
%%
%% CVS $Id: gencmds.ql,v 1.9 2007/03/27 23:40:57 itp Exp $

#include "../prover/lib/ergo.h"

general command
  shell(String)
extends
  general_cmds
mode
  shell(@chars)
doc
  'Executes the command in String by passing it to a shell subprocess.'
implementation
  is_chars(String),
  string_to_atom(String, Cmd),
  os(system(Cmd)).


general command
  statistics
extends
  general_cmds
mode
  statistics
doc
  'Print all Qu-Prolog statistics for the current Ergo session.'.


general command
  benchmark(Count, Code)
doc
  'Executes (Code,fail) Count times and reports the total',nl,
  'CPU time taken to do this (which includes the loop overhead).'
mode
  benchmark(@integer, +code).

benchmark(Count, Code) :-
    benchmark1(Count, Code, Time1),
    benchmark1(Count, benchmark_control_case(_), Time2),
    Time is Time1 - Time2,
    writelist([w(Time),'ms (',w(Time1),' - ',w(Time2),') for ',
               w(Count),' iterations of: ',code(Code),nl]).

benchmark_control_case(_).

benchmark1(Count, Code, _) :-
    statistics(runtime, _),
    between(1, Count, _),
    call_predicate(Code),
    fail.
benchmark1(_, _, Time) :-
    statistics(runtime, [_, Time]).


general command
  runtime
extends
  general_cmds
doc
  'Displays total user cputime for the current Ergo session and',nl,
  'the relative cputime since the last call to runtime/0 or',nl,
  'statistics(runtime,...).'
mode
  runtime.

runtime :-
    statistics(runtime, [Tot,Rel]),
    msg(runtime, ['Total cputime = ',w(Tot),'ms.  Delta = ',w(Rel),'ms.']).
