%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: htmldoc.ql,v 1.4 2002/06/27 05:10:26 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file provides commands for generating HTML manuals
%% from doc_node's.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

?- ( global_tactic_fact(html_doc_done, 1) ->
       true
   ;
       new_global_fact(html_doc_done, 1, [save(false)])
   ).

database command
  html_doc_clear_memory
extends
  misc_cmds
doc
  'Clears all the html_doc_done/1 facts, which record which doc_nodes',nl,
  'have been printed out by the htmldoc procedures.',nl,
  'This is done automatically whenever the htmldoc packages is loaded.'
mode
  html_doc_clear_memory.

html_doc_clear_memory :-
    tactic_retractall_global(html_doc_done(_)).

?- html_doc_clear_memory.


database command
  html_doc_rest
doc
  'Lists all doc_nodes that have not already been written out by html_doc_tree.'
mode
  html_doc_rest.

html_doc_rest :-
    setof(Node, html_unseen_node(Node), Nodes),
    html_doc_rest(Nodes).

html_doc_rest(Nodes) :-
    member(N, Nodes),
    msg(html_doc_rest, [q(N)]),
    fail.
html_doc_rest(_).

html_unseen_node(Node) :-
    (Thy = global; ancestor(Thy)),
    query_theory(doc_node(Node,_,_,_), Thy),
    \+ query_global(html_doc_done(Node)).

database command
  html_doc_tree(Dir, RootList)
extends
  misc_cmds
doc
  'Writes a tree of documentation nodes into Dir in HTML format.',nl,
  'Dir should be a quoted atom.  For example: ''../doc''.',nl,
  'RootList must be a list of names of a global doc_node.',nl,
  'The files produced are: contents.html, toc.html and index.html which',nl,
  'contain the documentation, a table of contents and an index.'
mode
  html_doc_tree(@atom, @list(docnode)).

html_doc_tree(Dir, Nodes) :-
    concat_atom([Dir, '/', 'contents.html'], Contents),
    concat_atom([Dir, '/', 'toc.html'], TOC),
    concat_atom([Dir, '/', 'INDEX.html'], Index),
    open(Contents, write, ConS),
    open(TOC, write, TOCS),
    open(Index, write, IS),
    add_handler(write(Strm, _Depth, Term),
                html_write(Term, Strm)),
    unwind_protect(html_doc_tree2(ConS, TOCS, IS, 1,Nodes),
                   (close(ConS), close(TOCS), close(IS),
                    delete_handlers(write(_,_,_),html_write(_,_)))
                  ).

html_write(xref([], Thy, Name), Strm) :-
    !,
    msg(html_write, ['Bad [] xref: ',q(Thy),':',q(Name)]),
    writelist(Strm, Name).
/*
html_write(xref(['{G}'], global, 'gumtree:{}/1'), Strm) :-
    !,
    writelist(Strm, ['<A HREF="contents.html#braces">{G}</A>']).
*/
html_write(xref(WL, Thy, Name), Strm) :-
    atom(WL),
    !,
    html_replace_slashes(Name, SName),
    writelist(Strm, ['<A HREF="contents.html#',SName,'">',WL,'</A>']).
html_write(xref(WL, Thy, Name), Strm) :-
    list(WL),
    member(W, WL),
    atom(W),
    \+ W = nl,
    !,
    html_replace_slashes(Name, SName),
    writelist(Strm, ['<A HREF="contents.html#',SName,'">',W,'</A>']).
html_write(xref(WL, Thy, Name), Strm) :-
    msg(html_write, ['bad xref: ',q(xref(WL,Thy,Name))]),
    writelist(Strm, [w(Name),' (????)']).
html_write(begin(list), Strm) :-
    writelist(Strm, ['<UL type="disc">',nl]).
html_write(end(list), Strm) :-
    writelist(Strm, [nl, '</UL>', nl]).
html_write(begin(verb), Strm) :-
    writelist(Strm, ['<pre>']).
html_write(end(verb), Strm) :-
    writelist(Strm, [nl, '</pre>', nl]).
html_write(item(Tag), Strm) :-
    list(Tag),
    !,
    writelist(Strm, [nl, '<LI>']),
    writelist(Strm, Tag).
html_write(item(Tag), Strm) :-
    atomic(Tag),
    writelist(Strm, [nl, '<LI>', w(Tag)]).
html_write(emph(WL), Strm) :-
    write(Strm, '<i>'),
    writelist(Strm, WL),
    write(Strm, '</i>').    
html_write(verb(WL), Strm) :-
    write(Strm, '<tt>'),
    writelist(Strm, WL),
    write(Strm, '</tt>').    
html_write(code(T), Strm) :-
    writelist(Strm, ['<pre>', T, '</pre>']).
html_write(para, Strm) :-
    writelist(Strm, ['<P>']).
html_write(br, Strm) :-
    writelist(Strm, ['<br/>']).

%% html_doc_tree2(Contents, TOC, Index, Level, DocNodes).
%%
%% Write out documentation about DocNodes in to the Contents, TOC and Index
%% streams. Level gives the level of the doc node in the tree.
%%
%%  mode html_doc_tree2(@stream, @stream, @stream, @integer, @list(docnode)).
%%
html_doc_tree2(Contents, TOC, Index, Level, Nodes) :-
    writelist(Contents, 
       ['<HTML>',nl,'<HEAD>',nl,
       '<!-- Automatically generated from Ergo. -->',nl,
       '<TITLE>',nl,
       'Ergo Reference Manual - Contents', nl,
       '</TITLE>',nl,
       '<BODY>',nl,
       '<H2>Contents</H2>',nl, '<P>',nl]),
    writelist(TOC, 
       ['<HTML>',nl,'<HEAD>',nl,
       '<!-- Automatically generated from Ergo. -->',nl,
       '<TITLE>',nl,
       'Ergo Reference Manual - Table of Contents', nl,
       '</TITLE>',nl,
       '<BODY>',nl,
       '<H2>Table of Contents</H2>',nl, '<P>',nl]),
    htmllevel2hdr(1, LHdr, EndLHdr),
    writelist(TOC, [nl,'<BLOCKQUOTE>',nl]),
    writelist(TOC, ['<BR>', LHdr, '<a href="intro.html">Introduction</a>',EndLHdr,nl]),
    writelist(TOC, [nl,'</BLOCKQUOTE>',nl]),
    init_index,
    writelist(Index, 
       ['<HTML>',nl,'<HEAD>',nl,
       '<!-- Automatically generated from Ergo. -->',nl,
       '<TITLE>',nl,
       'Ergo Reference Manual - Index', nl,
       '</TITLE>',nl,
       '<BODY>',nl,
       '<H2>Index</H2>',nl, '<P>',nl,'<UL>', nl]),
    html_doc_tree3(Nodes, Contents, TOC, Index, Level),
    writelist(Contents, ['</BODY>',nl,'</HTML>',nl]),
    writelist(TOC, ['</BODY>',nl,'</HTML>',nl]),
    dump_sorted_index(Index),
    writelist(Index, ['</UL>',nl,'</BODY>',nl,'</HTML>',nl]).

html_doc_tree3([], _Contents, _TOC, _Index, _Level).
html_doc_tree3([Node|Nodes], Contents, TOC, Index, Level) :-
    msg(html_doc_tree,[q(Node)]),
    ( get_doc_node(Node, Hdr, Doc, Cmds, Thy) ->
        true
    ;
        writelist(['WARNING: Could not find doc_node called: ',q(Node),nl]),
        Hdr = [Node],
        Doc = ['No such node.'],
        Thy = global
    ),
    get_doc_extns(Node, Extns),
    tactic_assert_global(html_doc_done(Node)),
    html_writelist_list_xrefs(Extns, XRefs, []),
    strip_leading_nls(Hdr, Hdr1),
    map(nl_to_spaces, [Hdr1, Hdr2]),
    htmllevel2hdr(Level, LHdr, EndLHdr),
    html_replace_slashes(Node, SNode),
    open_string(write, SS, []),
    writelist(SS, [max_depth(0, Hdr2)]),
    stream_to_atom(SS, AtomBHdr2),
    update_index(AtomBHdr2, SNode),
    writelist(Contents, ['<BR>',nl,'<HR>',nl,'<BR>',nl]),
    writelist(Contents,
      [LHdr, '<a name="',SNode,'"></a>', AtomBHdr2, EndLHdr,nl, '<P>',nl]),
    writelist(Contents, [nl,'<BLOCKQUOTE>',nl]),
    writelist(Contents, Doc),
    writelist(Contents, [nl,'<P>',nl]),
    html_writelist_modes_theory(Thy, Cmds, Contents),
    writelist(Contents, [nl,'<P>',nl]),
    (
        member(E, Extns),
        writelist(Contents, E),
        writelist(Contents, [nl,'<P>',nl]),
        fail
    ;
        true
    ),
    Level1 is Level+1,
    ( XRefs = [] ->
        true
    ;
        writelist(TOC, [nl,'<BLOCKQUOTE>',nl]),
        writelist(TOC, ['<BR>', LHdr, '<a href="contents.html#',SNode,'">',AtomBHdr2,'</a>',EndLHdr,nl])
    ),

    html_doc_tree3(XRefs, Contents, TOC, Index, Level1),
    writelist(Contents, [nl,'</BLOCKQUOTE>',nl]),
    ( XRefs = [] ->
        true
    ;
        writelist(TOC, [nl,'</BLOCKQUOTE>',nl])
    ),
    html_doc_tree3(Nodes, Contents, TOC, Index, Level).


%%
%% Managing indesx information for html documents. Indexes are stored in an
%% implicit parameter as they are found.
%% When all the index info has been collected it is sorted and written to the index file.
%%
init_index :-
    ip_set('html@index@list', []).

update_index(AtomBHdr2, SNode) :-
    ip_lookup('html@index@list', L),
    ip_set('html@index@list', [(AtomBHdr2, SNode)|L]).

dump_sorted_index(Stream) :-
    ip_lookup('html@index@list', L),
    sort(L, SL),
    dump_index(SL, Stream).

dump_index([],_).
dump_index([(Hdr, Ref)|Ind], Stream) :-
    writelist(Stream, ['<LI><a href="contents.html#',Ref,'">',Hdr,'</a>',nl]),
    dump_index(Ind, Stream).
    
%%
%% Choose the HTML heading size depending on the depth in the doc tree.
%%
htmllevel2hdr(1, '<H2>', '</H2>') :- !.
htmllevel2hdr(2, '<H3>', '</H3>') :- !.
htmllevel2hdr(_, '<H4>', '</H4>').

%%
%% Turn a atom with separators into a 'better' HTML string
%%
html_replace_slashes(Atom, SAtom) :-
    atom_codes(Atom, ACodes),
    map(html_replace_slash, [ACodes, SCodes]),
    atom_codes(SAtom, SCodes).

html_replace_slash(47, X) :- !, X = 95.
html_replace_slash(58, X) :- !, X = 95.
html_replace_slash(X, X).
    
%%
%% Convert newlines to HTML hard returns.
%%
nl_to_BR([], []).
nl_to_BR([nl|T], ['\n<BR>\n'|BT]) :-
    !,
    nl_to_BR(T,BT).
nl_to_BR([H|T], [H|BT]) :-
    nl_to_BR(T,BT).
    
%% html_writelist_list_xrefs(WriteLists, Xrefs, XrefsTail).
%%
%%  Xrefs is the list of all xref/3 terms that appear in
%%  WriteLists, which must be a list of writelists.
%%  XrefsTail is the tail of Xrefs.
%%  Records the xref'd nodes as having been done, as they have certainly
%%  been scheduled for writing.
%%
%%  mode html_writelist_list_xrefs(@list(writelist), ?list(compound),
%%                                              ?list(compound)).
%%
html_writelist_list_xrefs([], Tail, Tail).
html_writelist_list_xrefs([W|WL], Xrefs, Tail) :-
    html_writelist_xrefs(W, Xrefs, Tail1),
    html_writelist_list_xrefs(WL, Tail1, Tail).


%% html_writelist_xrefs(WriteList, Xrefs, XrefsTail).
%%
%%  Xrefs contains all the nodes referenced by xref/3 terms that appear in
%%  WriteList, and XrefsTail is the tail of Xrefs.
%%
%%  mode html_writelist_xrefs(@list(writelist), ?list(atom), ?list(atom)).
%%
html_writelist_xrefs([], Xtail, Xtail).
html_writelist_xrefs([I|WL], X, Xtail) :-
    html_writelist_item_xrefs(I, X, Xtail1),
    html_writelist_xrefs(WL, Xtail1, Xtail).

html_writelist_item_xrefs(xref(_,_,Node), X, Xtail) :-
    !,
    ( query_global(html_doc_done(Node)) ->
        X = Xtail
    ;
        tactic_assert_global(html_doc_done(Node)),
        X = [Node|Xtail]
    ).
html_writelist_item_xrefs(xinc(Thy, Frag), X, Xtail) :-
    query_theory(doc_fragment(Frag, Doc), Thy),
    !,
    html_writelist_xrefs(Doc, X, Xtail).
html_writelist_item_xrefs(_, X, X).

%% html_writelist_modes_theory(Thy, Cmds, Out).
%%
%% Writes all mode descriptors for each command in Cmds to Out.
%% When a command is specific to a theory then the theory name is
%% also written to Out.
%%
%% mode html_writelist_modes_theory(@theory, @list(cmd), @stream).
%%
html_writelist_modes_theory(_, [], _).
html_writelist_modes_theory(global, Cmds, Out) :- 
    html_writelist_modes(Out, Cmds),
    writelist(Out, [nl,nl]).
html_writelist_modes_theory(Thy, Cmds, Out) :-
    writelist(Out, [nl,'This command is specific to theory ',
                    q(Thy),' and descendants.',nl,nl]),
    html_writelist_modes(Out, Cmds),
    writelist(Out, [nl,nl]).

%% html_writelist_modes(Out, Cmds).
%%
%% Writes all mode descriptors for each command in Cmds to Out.
%%
%% mode html_writelist_modes(@stream, @list(cmd)).
%%
html_writelist_modes(_, []).
html_writelist_modes(Out, [cmd(Thy, Table, Name, Arity)|Cmds]) :-
    html_writelist_modes(Out, Cmds),
    functor(Cmd, Name, Arity),
    (
      query_global(gumtree_tactic_declaration(Cmd, Mode, IO)),
      Modes = [arity(Mode, IO)]
    ;
      find_command(Cmd, Table, Thy, Modes, _, _)
    ),
    html_writelist_modes_modes(Out, Modes).

html_writelist_modes_modes(Out, Modes) :-
    member(Mode, Modes),
    writelist(Out, ['<b>mode</b> &nbsp;&nbsp;']),
    html_writelist_modes_mode(Out, Mode),
    writelist(Out, ['.',nl,'<BR>',nl]),
    fail.
html_writelist_modes_modes(_, _).


html_writelist_modes_mode(Out, arity(Mode, Arity)) :-
    !,
    functor(Mode, Fun, Ar),
    writelist(Out, [w(Fun)]),
    ( Ar =:= 0 ->
        true
    ;
        writelist(Out, ['(']),
        html_writelist_modes_args(Out, Mode, 1, Ar),
        writelist(Out, [')'])
    ),
    writelist(Out, [' &nbsp;&nbsp;<b>arity</b>&nbsp;&nbsp; ', w(Arity)]).
html_writelist_modes_mode(Out, Mode) :-
    functor(Mode, Fun, Ar),
    writelist(Out, [w(Fun)]),
    ( Ar =:= 0 ->
        true
    ;
        writelist(Out, ['(']),
        html_writelist_modes_args(Out, Mode, 1, Ar),
        writelist(Out, [')'])
    ).

html_writelist_modes_args(Out, Mode, N, Ar) :-
    arg(N, Mode, Arg),
    Arg = Dir(Type),
    writelist(Out, [w(Dir), w(Type)]),
    ( N =:= Ar ->
        true
    ;
        writelist(Out, [', ']),
        N1 is N + 1,
        html_writelist_modes_args(Out, Mode, N1, Ar)
    ).


%%nl_to_spaces(nl, ' &nbsp;&nbsp;&nbsp;&nbsp;') :- !.
nl_to_spaces(nl, ' <BR>') :- !.
nl_to_spaces(X,X).
