%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% CVS $Id: newtypes5.ql,v 1.6 2000/11/15 06:29:50 itp Exp $


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file provides some additional type declarations which are synonyms 
%% of existing types. Use of these synonyms aids readability of mode 
%% declarations.
%% This file is specific for Ergo 5.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Define the help doc_node for the types which are specific to Ergo5.
modes_help_ergo5 doc
   'Types specific to Ergo5 are:'
header
    'Ergo5 types'
extends
  modes_help.

%% Declare some new types which are synonyms of existing types.
%% These types are used in mode declarations.
%% 
type_theory header
  'theory'
doc
  'theory is the name of a theory.  Eg. ''ergo''.'
extends
  modes_help_ergo5.

?- add_type(theory, [atom], type_theory, notypecheck).

type_thy_Type header
  'thy(Type)'
doc
  'thy(Type) specifies some term from the ''current'' theory, whose type is Type.'
extends
  modes_help_ergo5.

?- add_type(thy(Type), [term], type_thy_Type, satisfies_type(Type)).

type_bool header
  'bool'
doc
  'must be ''true'' or ''false''.'
extends
  modes_help_ergo5.

?- add_type(bool, [atom], type_bool, notypecheck).
?- add_type(truefalse, [atom], type_bool, notypecheck).

type_stream header
  'stream'
doc
  'Any stream.'
extends
  modes_help_ergo5.

?- add_type(stream, [ground], type_stream, notypecheck).

type_chars header
  'chars'
doc
  'Any list of chars.'
extends
  modes_help_ergo5.

?- add_type(chars, [list(integer)], type_chars, notypecheck).

type_docnode header
  'docnode'
doc
  'The name of a doc_node.'
extends
  modes_help_ergo5.

?- add_type(docnode, [atom], type_docnode, notypecheck).

type_cmd header
  'cmd'
doc
  'Something that satisfies is_legal_cmd/1.',nl,
  'That is, an atom or a compound term whose functor is an atom.'
extends
  modes_help_ergo5.

?- add_type(cmd, [nonvar], type_cmd, is_legal_cmd).

type_code header
  'code'
doc
  'Prolog code (not containing uninstantiated variables).'
extends
  modes_help_ergo5.

?- add_type(code, [nonvar], type_code, notypecheck).

type_tactic header
  'tactic'
doc
  'Tactic code, possibly containing uninstantiated variables',nl,
  'that will be instantiated by call_tac.'
extends
  modes_help_ergo5.

?- add_type(tactic, [term], type_tactic, notypecheck).

type_writelist header
  'writelist'
doc
  'List of atoms and output terms for printing by writelist/1.'
extends
  modes_help_ergo5.

?- add_type(writelist, [list(nonvar)], type_writelist, is_legal_writelist).

type_cmdtab header
  'cmdtab'
doc
  'cmdtab is the name of a command table.'
extends
  modes_help_ergo5.

?- add_type(cmdtab, [atom], type_cmdtab, notypecheck).

type_filename header
  'filename'
doc
  'The name of a file.'
extends
  modes_help_ergo5.

?- add_type(filename, [term], type_filename, notypecheck).

type_rootnode header
  'rootnode'
doc
  'Proof node identifier for root of a proof.'
extends
  modes_help_ergo5.

?- add_type(rootnode, [integer], type_rootnode, notypecheck).

type_proofnode header
  'proofnode'
doc
  'Proof node identifier/name.'
extends
  modes_help_ergo5.

?- add_type(proofnode, [integer], type_proofnode, notypecheck).

type_nodeid header
  'nodeid'
doc
  'Obsolete equivalent to ', xref(proofnode,global,type_proofnode),'.'
extends
  modes_help_ergo5.

?- add_type(nodeid, [integer], type_nodeid, notypecheck).

type_cname header
  'cname'
doc
  'The name of a context list.'
extends
  modes_help_ergo5.

?- add_type(cname, [atom], type_cname, notypecheck).

type_cdiff header
  'cdiff'
doc
  'Context difference between two proofnodes.'
extends
  modes_help_ergo5.

?- add_type(cdiff, [nonvar], type_cdiff, notypecheck).

type_cspec header
  'cspec'
doc
  'Specification of context list difference.'
extends
  modes_help_ergo5.

?- add_type(cspec, [nonvar], type_cspec, notypecheck).

type_cterm header
  'cterm'
doc
  'A context `term'' (ie. an (id,term) pair).'
extends
  modes_help_ergo5.

?- add_type(cterm, [compound], type_cterm, notypecheck).

type_ctermid header
  'ctermid'
doc
  'Identity of a term in a context list.'
extends
  modes_help_ergo5.

?- add_type(ctermid, [integer], type_ctermid, notypecheck).

type_rulename header
  'rulename'
doc
  'rulename specifies a database rule.',nl,
  'Eg. thmname, thy:thmname, thmname(args).'
extends
  modes_help_ergo5.

?- add_type(rulename, [nonvar], type_rulename, check_rulename).

check_rulename(N) :-
    atom(N).
check_rulename(N) :- % covers the thy:name and name(args) cases.
    compound(N),
    \+ list(N).
