%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##


%% $Id: pp.ql,v 1.22 2000/11/15 06:29:50 itp Exp $
%% Pretty-printer.

%%%% 
%% The procedures in this file manipulate a TokTree, which is a structure with
%% tokens at the leaves.
%% The text of the tokens is kept in a separate token string.
%% Each node in the tree can generate a single line or a list of lines; the
%% type-setter determines which by instantiating variables in the nodes.
%% 
%%  TokTree ::= toks(Path, StartPos, Length, Indent, NLTaken, [TokTree, ...])
%%            | tok(Path, StartPos, Length)
%%            | sp.
%%  Path is the path from the root to this node.
%%  StartPos is an index into the token string where this subtree starts.
%%  Length is the total size of the subtree (including inter-node spaces). 
%%  Indent is the extra amount of indentation needed if nls are taken here.
%%  NLTaken is 'yes' when nls are to be taken here; 'no' if nls are not
%%    allowed; a variable when nls may be requested by the type-setter.
%%%%

#include "../prover/lib/ergo.h"

general command
  pp(Term, Width)
doc
  'Pretty-print Term in Width character-spaces.'
extends
  general_cmds
mode
  pp(@thy(term), @integer)
implementation
  abbreviate(Term, Abbr),
  pp(Abbr, Width).

proof command
  pp_focus(Width)
doc
  'Pretty-print current focus in Width character-spaces.'
extends
  misc_cmds
mode
  pp_focus(@integer)
implementation
  is_focus(Focus),
  abbreviate(Focus, Abbr),
  pp(Abbr, Width).

proof command
  pp_focus
doc
  'Pretty-print current focus in 80 character-spaces.'
extends
  misc_cmds
mode
  pp_focus
implementation
  is_focus(Focus),
  abbreviate(Focus, Abbr),
  pp(Abbr, 80).

%% Ergo 4 only :-(
general command
  pp_rule(Rule, Width)
doc
  'Pretty-print statement of Rule in Width character-spaces.'
extends
  general_cmds
mode
  pp_rule(@rulename, @integer)
implementation
  find_rule(Rule, Stmt, _),
  abbreviate(Stmt, Abbr),
  pp(Abbr, Width).

general command
  pp_rule(Rule)
doc
  'Pretty-print statement of Rule in 80 character-spaces.'
extends
  general_cmds
mode
  pp_rule(@rulename)
implementation
  find_rule(Rule, Stmt, _),
  abbreviate(Stmt, Abbr),
  pp(Abbr, 80).

% 
% A trick to save memory usage when printing out huge terms with the
% pretty printer.
% When printing we want to name meta variables consistently
% previously writeRT did that - but right at the bottom of the
% computation (last thing before write) so all of the stack used
% to pretty print huge terms must be kept.
% Instead, we use name_vars ourselves write at the top of the computation
% and then after doing the pretty printing we can fail, reclaiming all the
% space used, but without forgetting our variable names.
% So,
% inside tokenizer use writeT instead of writeRT and at the top
% level we fail (once) after the pretty printing is complete. (Luke)
% NOTE any other callers to tokenize (not via pp/2) should also do this
% notably emacs_index:noindprint/1

pp(Term, Width) :-
    dbg(pp),
    !,
    name_vars(Term),
    pp_debug(Term, Width).

pp(Term, Width) :-
    name_vars(Term),
    nondebug_pp(Term, Width).

pp_debug(Term, Width) :-
    once(do_pp_debug(Term, Width)),
    fail.

pp_debug(_,_).

do_pp_debug(Term, Width):-
    check_current_theory(CD),
    tokenize(Term, CD, TokTree, TokChars),
    debug_write_tok_tree(TokTree, TokChars, 0, 0),
    typeset(TokTree, Width),
    nl,
    debug_write_tok_tree(TokTree, TokChars, 0, 0),
    nl,
    write_tokens(TokTree, TokChars, 0, _Index),
    nl,
    write(_Index),
    nl.


nondebug_pp(Term, Width):-
    once(do_nondebug_pp(Term, Width)),
    fail.

nondebug_pp(_,_).

do_nondebug_pp(Term, Width):-
    check_current_theory(CD),
    tokenize(Term, CD, TokTree, TokChars),
    typeset(TokTree, Width),
    write_tokens(TokTree, TokChars, 0, _Index),
    nl.

%%!database command
%%!  layout(Term, Tree)
%%!doc
%%!  'Set up customized layout for instances of Term.',nl,
%%!  'Tree must be a list of subtrees, each of which is either a tree',nl,
%%!  'a literal (double-quoted character string), or a variable from Term.',nl,
%%!  'The result of traversing the tree must produce the same sequence of',nl,
%%!  'tokens as the standard tokenizer for Term; a warning is issued if it',nl,
%%!  'does not.'
%%!mode
%%!  layout(@term, @term).
%%!
%%!layout(Term, Tree) :-
%%!    tokenize(Term, Tree0) with variables,
%%!    ( traverse(Tree) = traverse(Tree0) ->
%%!        true
%%!    ;
%%!        error
%%!    ),
%%!    do_custom_layout(Tree, TokTree),
%%!    assert_theory(layout(Term, TokTree)).

debug_write_tok_tree(sp, _, DebInd, _) :-
    writelist([tab(DebInd), 'Space', nl]).
debug_write_tok_tree(tok(Path,Pos,Len),
                     Chars, DebInd, Col) :-
    debug_write_tok(DebInd, 'Leaf:   ', Path, Pos, Len, Chars, Col).
debug_write_tok_tree(toks(Path,Pos,Size,Indent,NLTaken,SubTrees),
                     Chars, DebInd, Col) :-
    ( NLTaken == yes ->
        Tag = 'NLNode: '
    ; NLTaken == no ->
        Tag = 'UBNode: '
    ;
        Tag = 'Node:   '
    ),
    debug_write_tok(DebInd, Tag, Path, Pos, Size, Chars, Col),
    DebInd1 is DebInd + 1,
    Col1 is Col + Indent,
    debug_write_tok_tree(SubTrees, Chars, DebInd1, Col1).

debug_write_tok_tree([], _, _, _).
debug_write_tok_tree([SubTree|SubTrees], Chars, DI, C) :-
    debug_write_tok_tree(SubTree, Chars, DI, C),
    debug_write_tok_tree(SubTrees, Chars, DI, C).

debug_write_tok(DebInd, Tag, _Path, Pos, Len, Chars, TrueInd) :-
    writelist([tab(DebInd), w(Tag), w(TrueInd), ' ''']),
    write_token(Pos, Len, Chars),
    writelist(['''', nl]).

%
%% typeset(TokTree, Width).
%%
%%  Just decides which NLs to take.
%% 
%%  mode typeset(+compound, @integer).
%%
typeset(sp, _).
typeset(tok(_Path,_Pos,Length), Width) :-
    Length < Width,
    !.
typeset(toks(_Path,_Pos,Size,_Indent,_NL,_Subs), Width) :-
    PRE('typeset: Size =< Width', (integer(Size), integer(Width))),
    Size =< Width,
    !.
typeset(toks(_Path,_Pos,Size,Indent,NL,Subs), Width) :-
    PRE('typeset: Size > Width', (integer(Size), integer(Width), Size > Width)),
    NL=yes,
    !,
    Width1 is Width-Indent,
    typeset(Subs, Width1).
typeset([], _) :-
    !.
typeset([Tree|Trees], Width) :-
    !,
    typeset(Tree, Width),
    typeset(Trees, Width).
typeset(_TokTree, _) :-
    %% TO DO: Why is no warning issued here? MB 7/9/1998.
    %%writelist(['WARNING: failed to typeset ', w(_TokTree), nl]),
    true.

%% write_tokens(TokTree, TokChars, Column).
%% write_tokens(TokTree, TokChars, Column, Index).
%% write_token(StartPos, Len, TokChars).
%%
%%  mode write_tokens(@compound, @chars, @integer).
%%  mode write_tokens(@compound, @chars, @integer, -list(compound)).
%%  mode write_token(@integer, @integer, @chars).
%%

write_tokens(TokTree, Chars, Column) :-
    write_tokens(TokTree, Chars, Column, [], 0, _, _, _).

write_tokens(TokTree, Chars, Column, Index) :-
    write_tokens(TokTree, Chars, Column, [], 0, _, Index, []).

write_token(StartPos, Len, TokChars) :-
    write_token(StartPos, Len, TokChars, []).

write_tokens(tok(Path,Pos,Len), Chars, _Col, Sep, Off0, Off) -->
    {write_token(Pos, Len, Chars, Sep),
     ( integer(Sep) ->
         Off is Off0 + Sep + 1 + Len
     ;
         Off is Off0 + Len
     )
    },
    [i(Path,Off0,Off)].
write_tokens(toks(Path,Pos,Size,Indent,NLTaken,SubTrees),
             Chars, Col, Sep, Off0, Off) -->
    {NLTaken == yes},
    !,
    %%??[i(Path,Off0,Off)],
    {Col1 is Col + Indent},
    write_nl_tokens(SubTrees, Path, Chars, Col1, Sep, Off0, Off).
write_tokens(toks(Path,Pos,Size,_,_,SubTrees), Chars, Col, Sep, Off0, Off) -->
    % no nl taken,
    %%??[i(Path,Off0,Off)],
    write_tokens(SubTrees, Path, Chars, Col, Sep, Off0, Off).

write_tokens([], _, _, _, _, Off, Off) -->
    [].
write_tokens([sp], Path, Chars, Col, Sep, Off0, Off) -->
    !,
    write_sp(Sep, Path, Off0, Off).
write_tokens([Tree], _Path, Chars, Col, Sep, Off0, Off) -->
    !,
    write_tokens(Tree, Chars, Col, Sep, Off0, Off).
write_tokens([sp|Trees], Path, Chars, Col, Sep, Off0, Off) -->
    !,
    write_sp(Sep, Path, Off0, Off1),
    write_tokens(Trees, Path, Chars, Col, Sep, Off1, Off).
write_tokens([Tree|Trees], Path, Chars, Col, Sep, Off0, Off) -->
    write_tokens(Tree, Chars, Col, Sep, Off0, Off1),
    write_tokens(Trees, Path, Chars, Col, [], Off1, Off).

write_nl_tokens([], _, _, _, _, Off, Off) -->
    [].
write_nl_tokens([sp], _, _, _, _, Off, Off) -->
    !.
write_nl_tokens([Tree], _, Chars, Col, Sep, Off0, Off) -->
    !,
    write_tokens(Tree, Chars, Col, Sep, Off0, Off).
write_nl_tokens([sp|Trees], Path, Chars, Col, Sep, Off0, Off) -->
    !,
    write_sp(Sep, Path, Off0, Off1),
    write_nl_tokens(Trees, Path, Chars, Col, Col, Off1, Off).
write_nl_tokens([Tree|Trees], Path, Chars, Col, Sep, Off0, Off) -->
    write_tokens(Tree, Chars, Col, Sep, Off0, Off1),
    write_nl_tokens(Trees, Path, Chars, Col, Col, Off1, Off).

write_sp(Sep, Path, Off0, Off) -->
    ( {Sep = []} ->
        {
         write_atom(' '),
         Off is Off0 + 1
        },
        [i(Path,Off0,Off)]
    ;
        {Off = Off0}
    ).

write_token(0, 0, _, _) :-
    !.
write_token(0, Len, Chars, []) :-
    PRE('write_token: []', Len > 0),
    !,
    write_token(Len, Chars).
write_token(0, Len, Chars, Sep) :-
    PRE('write_token: Sep', Len > 0),
    PRE('write_token: Sep', integer(Sep)),
    !,
    nl,
    tab(Sep),
    write_token(Len, Chars).
write_token(SPos, Len, [_Char|Chars], Sep) :-
    PRE('write_token: [_Char|Chars]', SPos > 0),
    SPos1 is SPos - 1,
    write_token(SPos1, Len, Chars, Sep).

write_token(0, _) :-
    !.
write_token(Len, [Char|Chars]) :-
    PRE('write_token(2)', Len > 0),
    !,
    put(Char),
    Len1 is Len - 1,
    write_token(Len1, Chars).

%
%% tokenize(Term, Thy, TokTree, TokChars).
%% 
%%  TokChars is the flat output string form of Term in Thy,
%%  and TokTree is the token tree for it.
%%
%%  mode tokenize(@term, @atom, -compound, -chars).
%% 
tokenize(Term, Thy, TokTree, TokChars) :-
    ip_set(tokenize_theory, Thy),
    ip_set(tokenize_current_prec, 1000),
    open_string(write, Strm),
    tokenize(Term, [], 0, _Pos, Strm, TokTree),
    stream_to_chars(Strm, TokChars).
%    close(Strm).

%% tokenize(Term, Path, Pos0, Pos, Strm) --> TokTrees.
%% tokenize(Term, Path, Pos0, Pos, Strm, TokTree).
%% 
%%  mode tokenize(@term, @subterm_path,
%%                @integer, -integer, @stream,
%%                ?list(compound), ?list(compound)).
%%  mode tokenize(@term, @subterm_path,
%%                @integer, -integer, @stream, -compound).
%% 
tokenize(Term, Path, Pos0, Pos, Strm) -->
    {tokenize(Term, Path, Pos0, Pos, Strm, Tree)},
    [Tree].

%%!tokenize(Term, Path, Pos0, Pos, Strm, Tree) :-
%%!    query_ancestors(layout(Term1, Tree1)),
%%!    match(Term, Term1),
%%!    !,
%%!    ...
                    

tokenize(Term, Path, Pos0, Pos, Strm, Tree) :-
    sub(Term),
    !,
    substitute(Term, Subs, Body),
    ip_lookup(tokenize_current_prec, Prec),
    ip_set(tokenize_current_prec, 0),
    phrase(tokenize_subs(Subs, Path, Pos0, Pos1, Strm),
           Trees, [BodyTree]),
    tokenize(Body, Path, Pos1, Pos, Strm, BodyTree),
    make_tok_tree(Path, Pos0, Pos, 1, Trees, Tree),
    ip_set(tokenize_current_prec, Prec).
tokenize(Term, Path, Pos0, Pos, Strm, Tree) :-
    any_variable(Term),
    !,
    ip_lookup(tokenize_theory, Thy),
    writeT(Strm, Term, Thy),
    stream_property(Strm, position(Pos)),
    make_token(Path, Pos0, Pos, Tree).
tokenize(Term, Path, Pos0, Pos, Strm, Tree) :-
    atomic(Term),
    !,
    writeq(Strm, Term),
    stream_property(Strm, position(Pos)),
    make_token(Path, Pos0, Pos, Tree).
tokenize(Term, Path, Pos0, Pos, Strm, Tree) :-
    list(Term),
    !,
    tokenize_list(Term, ["["], [",", sp], ["]"], eolpush, list,
                  Path, Pos0, Pos, Strm, Tree).
tokenize(Term, Path, Pos0, Pos, Strm, Tree) :-
    compound(Term),
    !,
    functor(Term, Fun, Ar),
    tokenize_cpd(Term, Fun, Ar, Path, Pos0, Pos, Strm, Tree).
tokenize(Term, Path, Pos0, Pos, Strm, Tree) :-
    quant(Term),
    !,
    quantify(Term, Quant, BList, Body),
    tokenize_quant(Quant, BList, Body, Path, Pos0, Pos, Strm, Tree).
tokenize(_, _, _, _, _, _) :-
    write_term_list(['Bad tokenize term!',nl]),
    fail.

%% tokenize_cpd(Term, Fun, Ar, Path,
%%              Pos0, Pos, Strm, Tree).
%% 
%%  mode tokenize_cpd(@term, @term, @integer, @subterm_path,
%%                    @integer, -integer, @stream, -compound)).
%% 
tokenize_cpd(Term, Fun, Ar, Path, Pos0, Pos, Strm, Tree) :-
    ip_lookup(tokenize_theory, Thy),
    get_current_op(Thy, Fun, Ar, OP, LP, RP),
    !,
    ip_lookup(tokenize_op_style, Style),
    tokenize_op(Style, Term, OP, LP, RP, Fun, Ar, Path, Pos0, Pos, Strm, Tree).
tokenize_cpd(Term, Fun, _Ar, Path, Pos0, Pos, Strm, Tree) :-
    ip_lookup(tokenize_current_prec, Prec),
    ip_set(tokenize_current_prec, 1000),
    Term =.. [Fun|List],
    tokenize_list(List, tree(no, [toks(Fun),"("]), [",",sp], ")", eolpush, cpd,
                  Path, Pos0, Pos, Strm, Tree),
    ip_set(tokenize_current_prec, Prec).

%% tokenize_quant(Quant, BList, Body, Path, Pos0, Pos, Strm, Tree).
%% 
%%  mode tokenize_quant(@term, @term, @list(term), @subterm_path,
%%            @integer, -integer, @stream, -compound).
%%
tokenize_quant(Quant0, BList, Body, Path, Pos0, Pos, Strm, Tree) :-
    ip_lookup(tokenize_theory, Thy),
    get_current_quant(Quant0, Thy, OP, Quant),
    ip_lookup(tokenize_current_prec, Prec),
    phrase(tokenize_quant_left(Quant, BList, OP, Prec, Path, Pos0, Pos1, Strm),
           LTrees),
    phrase(tokenize_quant_right(Body, OP, Prec, Path, Pos1, Pos, Strm),
           RTrees),
    make_tok_tree(Path, Pos0, Pos1, 1, LTrees, LTree),
    make_tok_tree(Path, Pos1, Pos,  1, RTrees, RTree),
    make_tok_tree(Path, Pos0, Pos,  1, [LTree,RTree], Tree),
    ip_set(tokenize_current_prec, Prec).

tokenize_quant_left(Quant, BList, OP, Prec, Path, Pos0, Pos, Strm) -->
    tokenize_paren(l, OP, Prec, Path, Pos0, Pos1, Strm),
    {
     ip_set(tokenize_current_prec, 1000)
    },
    tokenize(Quant, [0|Path], Pos1, Pos2, Strm),
    make_sp(Pos2, Pos3, Strm),
    tokenize_list(BList, ["["], [",",sp], ["]"], eolpush, blist,
                  Path, Pos3, Pos, Strm).

tokenize_quant_right(Body, OP, Prec, Path, Pos0, Pos, Strm) -->
    make_sp(Pos0, Pos1, Strm),
    {
     ip_set(tokenize_current_prec, OP)
    },
    tokenize(Body, [2|Path], Pos1, Pos2, Strm),
    tokenize_paren(r, OP, Prec, Path, Pos2, Pos, Strm).

%% tokenize_list(List, Open, Sep, Close, Style, Kind,
%%               Path, Pos0, Pos, Strm) --> Trees.
%% tokenize_list(List, Open, Sep, Close, Style, Kind,
%%               Path, Pos0, Pos, Strm, Tree).
%%
%%  Tokenize an arbitrary non-empty List.
%%  Open, Sep and Close are lists of token descriptors; they can include
%%    sp         -- a discardable space
%%    "string"   -- a literal
%%    toks(Term) -- a term to be tokenized (many times, if necessary!)
%%    tree(Nl,[TokDec,...]) 
%%               -- a list of descriptors to be tokenized as a tree;
%%                  Nl is `yes', `no' or unbound if breaks are respectively
%%                  required, forbidden or allowed.
%% 
%%  Style determines placement of separators:
%%    bol -- at the beginning of each element but the first
%%    eol -- at the end of each element but the last
%%    bolpush -- pushed *inside* the beginning of each element but the first
%%    eolpush -- pushed *inside* the end of each element but the last.
%%   **Only eolpush is currently supported.**
%%
%%  Kind says what sort of list it is, and controls how paths are generated:
%%    list  -- paths are [1], [2,1], [2,2,1], etc.
%%    cpd   -- paths are [1], [2], [3], etc.
%%    blist -- paths are [1,1,2], [1,2,1,2], [1,2,2,1,2] etc.
%%             (for type positions only).
%% 
%%  The generated list has one node per list element,
%%   with Open coupled with the first element,
%%        Close coupled with the last element, 
%%        and Sep coupled with interior elements.  
%% 
%%  E.g. the list [a,b,c] has two primary breaks -- after the commas.
%%
%%  mode tokenize_list(@list(term),
%%                     @list(nonvar), @list(nonvar), @list(nonvar),
%%                     @atom, @atom,
%%                     @subterm_path, @integer, -integer, @stream,
%%                     +list(compound), -list(compound)).
%%  mode tokenize_list(@list(term),
%%                     @list(nonvar), @list(nonvar), @list(nonvar),
%%                     @atom, @atom,
%%                     @subterm_path, @integer, -integer, @stream,
%%                     -compound).
%% 

tokenize_list(List, Open, Sep, Close, Style, Kind,
              Path, Pos0, Pos, Strm, Tree) :-
    phrase(tokenize_list(List, Open, Sep, Close, Style, Kind,
                         Path, Pos0, Pos, Strm),
           Trees),
    make_tok_tree(Path, Pos0, Pos, 1, Trees, Tree).

%% [A,   B,C,  D]          [A,   B,    C|T]       [     A|B]
%% head  body  tail        head  body  tail       head  tail
tokenize_list([X|Y], Open, _Sep, Close, eolpush, Kind,
              Path, Pos0, Pos, Strm) -->
    { Y == [] },
    !,
    tokenize_list_head(Open, X, Close, Kind, Path, Pos0, Pos, Strm).
tokenize_list([X|Xs], Open, Sep, Close, eolpush, Kind,
              Path, Pos0, Pos, Strm) -->
    { list(Xs) },
    !,
    tokenize_list_head(Open, X, Sep, Kind, Path, Pos0, Pos1, Strm),
    tokenize_list_body(Xs, Y, Sep, 2, N, Kind, Path, Pos1, Pos2, Strm),
    tokenize_list_tail(Y, Close, N, Kind, Path, Pos2, Pos, Strm).
tokenize_list([X|Xs], Open, Sep, Close, eolpush, Kind,
              Path, Pos0, Pos, Strm) -->
    tokenize_list_toks(Open, Path, Pos0, Pos1, Strm),
    tokenize_list_tail([X|Xs], Close, 0, Kind, Path, Pos1, Pos, Strm).

tokenize_list_head(Open, X, Sep, Kind, Path, Pos0, Pos, Strm) -->
    {
     phrase(tokenize_list_toks(Open, Path, Pos0, Pos1, Strm), OpToks),
     make_tok_tree(Path, Pos0, Pos1, 1, OpToks, OpTree),
     list_position(Kind, Path, 1, NPath),
     tokenize(X, NPath, Pos1, Pos2, Strm, XTree),
     push_tok_tree_head(XTree, OpTree, Path, Pos0, OpXTree),
     phrase(tokenize_list_toks(Sep, Path, Pos2, Pos, Strm), SepToks),
     make_tok_tree(Path, Pos2, Pos, 1, SepToks, SepTree)
    },
    push_tok_tree_tail(OpXTree, SepTree, Path, Pos).

tokenize_list_body([X|Xs], Y, Sep, N0, N, Kind, Path, Pos0, Pos, Strm) -->
    {
     nonvar(Xs),
     Xs = [_|_],
     !,
     N1 is N0 + 1,
     list_position(Kind, Path, N0, NPath),
     tokenize(X, NPath, Pos0, Pos1, Strm, XTree),
     phrase(tokenize_list_toks(Sep, Path, Pos1, Pos2, Strm), SepToks),
     make_tok_tree(Path, Pos1, Pos2, 1, SepToks, SepTree)
    },
    push_tok_tree_tail(XTree, SepTree, Path, Pos2),
    tokenize_list_body(Xs, Y, Sep, N1, N, Kind, Path, Pos2, Pos, Strm).
tokenize_list_body(Tail, Tail, _, N, N, _, _, Pos, Pos, _) -->
    !.

tokenize_list_tail([X|Y], Close, N, Kind, Path, Pos0, Pos, Strm) -->
    {
     Y == [],
     !,
     list_position(Kind, Path, N, NPath),
     phrase((tokenize(X, NPath, Pos0, Pos1, Strm),
             tokenize_list_toks(Close, Path, Pos1, Pos, Strm)),
            Trees)
    },
    make_tok_tree(Path, Pos0, Pos, 1, Trees).
tokenize_list_tail([X|Y], Close, N, Kind, Path, Pos0, Pos, Strm) -->
    {
     list_position(Kind, Path, N, NPath),
     phrase((tokenize(X, NPath, Pos0, Pos1, Strm),
             tokenize_list_toks([sp,"|",sp], Path, Pos1, Pos2, Strm)),
            XTrees),
     NPath = [1|NPath1], % yuk! we put a spurious 1 on; take it off again.
     phrase((tokenize(Y, [2|NPath1], Pos2, Pos3, Strm),
             tokenize_list_toks(Close, Path, Pos3, Pos, Strm)),
            YTrees)
    },
    make_tok_tree(Path, Pos0, Pos2, 1, XTrees),
    make_tok_tree(Path, Pos2, Pos, 1, YTrees).

list_position(cpd,  Path, N, [N|Path]) :-
    !.
list_position(list, Path, 0, [1|Path]) :-
%% TO DO: Overcome need for this hack for improper list in list_position.
    !.
list_position(list, Path, 1, [1|Path]) :-
    !.
list_position(list, Path0, N, Path) :-
    N>1,
    N1 is N - 1,
    list_position(list, [2|Path0], N1, Path).
list_position(blist, Path0, N, Path) :-
    list_position(list, [1|Path0], N, Path).

tokenize_list_toks([], Path, Pos, Pos, Strm) -->
    !.
tokenize_list_toks(sp, _Path, Pos0, Pos, Strm) -->
    !,
    make_sp(Pos0, Pos, Strm).
tokenize_list_toks(Chars, Path, Pos0, Pos, Strm) -->
    {is_chars(Chars)},
    !,
    make_literal(Chars, Path, Pos0, Pos, Strm).
tokenize_list_toks(toks(Term), Path, Pos0, Pos, Strm) -->
    !,
    tokenize(Term, Path, Pos0, Pos, Strm).
tokenize_list_toks(tree(Nl,Toks), Path, Pos0, Pos, Strm) -->
    !,
    {
     phrase(tokenize_list_toks(Toks, Path, Pos0, Pos, Strm),
            Trees)
    },
    make_nl_tok_tree(Nl, Path, Pos0, Pos, 0, Trees).
tokenize_list_toks([Tok|Toks], Path, Pos0, Pos, Strm) -->
    tokenize_list_toks(Tok, Path, Pos0, Pos1, Strm),
    tokenize_list_toks(Toks, Path, Pos1, Pos, Strm).

%% tokenize_subs(SubsListList, Path, Pos0, Pos, Strm) --> Trees.
%% tokenize_sub(SubsList, Path, Pos0, Pos, Strm) --> Trees.
%%
%%  mode tokenize_subs(@list(list(compound)), @subterm_path,
%%                     @integer, -integer, @stream,
%%                     +list(compound), -list(compound)).
%%  mode tokenize_subs(@list(term), @subterm_path,
%%                     @integer, -integer, @stream,
%%                     +list(compound), -list(compound)).
%% 
tokenize_subs([], _, Pos, Pos, _) -->
    [].
tokenize_subs([Sub|Subs], Path, Pos0, Pos, Strm) -->
    make_literal("[", Path, Pos0, Pos1, Strm),
    {
     phrase(tokenize_sub(Sub, Path, Pos1, Pos2, Strm),
            SubTrees)
    },
    make_tok_tree(Path, Pos1, Pos2, 1, SubTrees),
    make_literal("]", Path, Pos2, Pos3, Strm),
    tokenize_subs(Subs, Path, Pos3, Pos, Strm).

tokenize_sub([], _, Pos, Pos, _) -->
    [].
tokenize_sub([(T/V)], Path, Pos0, Pos, Strm) -->
    tokenize_sub_pair(T, V, Path, Pos0, Pos, Strm).
tokenize_sub([(T/V)|Sub], Path, Pos0, Pos, Strm) -->
    tokenize_sub_pair(T, V, Path, Pos0, Pos1, Strm),
    make_literal(",", Path, Pos1, Pos2, Strm),
    tokenize_sub(Sub, Path, Pos2, Pos, Strm).

tokenize_sub_pair(T, V, Path, Pos0, Pos, Strm) -->
    tokenize(T, Path, Pos0, Pos1, Strm),
    make_literal("/", Path, Pos1, Pos2, Strm),
    tokenize(V, Path, Pos2, Pos, Strm).

%% tokenize_op(Style, Term,
%%             OP, LP, RP, Fun, Ar,
%%             Path, Pos0, Pos, Strm, Tree).
%%
%%  mode tokenize_op(?atom, @term,
%%                   @integer, @integer, @integer, @term, @integer,
%%                   @subterm_path, @integer, -integer, @stream, -compound).
%% 

%% TO DO: Flatten associative lists, e.g. A and B and C, in tokenize_op.

tokenize_op(_, -(N), OP, _LP, _RP, -, 1, Path, Pos0, Pos, Strm, Tree) :-
    integer(N),
    %% Special case: always requires inner parentheses
    !,
    ip_lookup(tokenize_current_prec, Prec),
    phrase(tokenize_minus_ltrees(Prec, OP, Path, Pos0, Pos1, Strm),
           LeftTrees),
    phrase(tokenize_minus_rtrees(Prec, OP, N, Path, Pos1, Pos, Strm),
           RightTrees),
    make_tok_tree(Path, Pos0, Pos1, 1, LeftTrees, LeftTree),
    make_tok_tree(Path, Pos1, Pos, 1, RightTrees, RightTree),
    make_tok_tree(Path, Pos0, Pos, 1, [LeftTree,RightTree], Tree),
    ip_set(tokenize_current_prec, Prec).
tokenize_op(bol, Term, OP, LP, RP, Fun, Ar, Path, Pos0, Pos, Strm, Tree) :-
    !,
    ip_lookup(tokenize_current_prec, P),
    ( LP =:= 0 ->
        phrase(tokenize_paren(l, OP, P, Path, Pos0, Pos1, Strm),
               LeftTrees)
    ;
        phrase(tokenize_bolop_ltrees(P, OP, LP, Term, Path, Pos0, Pos1, Strm),
               LeftTrees)
    ),
    make_tok_tree(Path, Pos0, Pos1, 1, LeftTrees,  LeftTree),
    ( 0<LP, LP<RP ->
        %% right associative binary
        arg(2, Term, RArg),
        phrase(tokenize_bolop_rightassoc_trees(P, OP, LP, RP, Fun, RArg,
                                               Path, Pos1, Pos, Strm),
               RightTrees),
        make_tok_tree(Path, Pos0, Pos, 1, [LeftTree|RightTrees], Tree)
    ;
        phrase(tokenize_bolop_rtrees(P, OP, RP, Term, Fun, Ar,
                                     Path, Pos1, Pos, Strm),
               RightTrees),
        make_tok_tree(Path, Pos1, Pos,  1, RightTrees, RightTree),
        make_tok_tree(Path, Pos0, Pos,  1, [LeftTree,RightTree], Tree)
    ),
    ip_set(tokenize_current_prec, P).
tokenize_op(eol, Term, OP, LP, RP, Fun, Ar, Path, Pos0, Pos, Strm, Tree) :-
    !,
    ip_lookup(tokenize_current_prec, P),
    phrase(tokenize_eolop_ltrees(P, OP, LP, Term, Fun, Path, Pos0, Pos1, Strm),
           LeftTrees),
    ( RP =:= 0 ->
        Pos2 = Pos1,
        phrase(tokenize_paren(r, OP, P, Path, Pos2, Pos, Strm),
               RightTrees)
    ;
        write_atom(Strm, ' '),
        Pos2 is Pos1 + 1,
        phrase(tokenize_eolop_rtrees(P, OP, RP, Term, Ar,
                                     Path, Pos2, Pos, Strm),
               RightTrees1),
        RightTrees = [sp|RightTrees1]
    ),
    make_tok_tree(Path, Pos0, Pos1, 1, LeftTrees,  LeftTree),
    make_tok_tree(Path, Pos2, Pos,  1, RightTrees, RightTree),
    make_tok_tree(Path, Pos0, Pos,  1, [LeftTree,RightTree], Tree),
    ip_set(tokenize_current_prec, P).

tokenize_minus_ltrees(Prec, OP, Path, Pos0, Pos, Strm) -->
    tokenize_paren(l, OP, Prec, Path, Pos0, Pos1, Strm),
    make_literal("-(", Path, Pos1, Pos, Strm).

tokenize_minus_rtrees(Prec, OP, N, Path, Pos0, Pos, Strm) -->
    {
     write(Strm, N),
     stream_property(Strm, position(Pos1))
    },
    make_token([1|Path], Pos0, Pos1),
    make_literal(")", Path, Pos1, Pos2, Strm),
    tokenize_paren(r, OP, Prec, Path, Pos2, Pos, Strm).

tokenize_eolop_ltrees(Prec, OP, LP, Term, Fun, Path, Pos0, Pos, Strm) -->
    tokenize_paren(l, OP, Prec, Path, Pos0, Pos1, Strm),
    ( {LP =:= 0} ->
        {Pos2 = Pos1}
    ;
        tokenize_op_left(Term, LP, Path, Pos1, Pos2, Strm)
    ),
    {
     ip_set(tokenize_current_prec, 1000)
    },
    tokenize(Fun, [0|Path], Pos2, Pos, Strm).
    
tokenize_eolop_rtrees(Prec, OP, RP, Term, Ar, Path, Pos0, Pos, Strm) -->
    tokenize_op_right(Term, RP, Ar, Path, Pos0, Pos1, Strm),
    tokenize_paren(r, OP, Prec, Path, Pos1, Pos, Strm).

tokenize_bolop_ltrees(Prec, OP, LP, Term, Path, Pos0, Pos, Strm) -->
    tokenize_paren(l, OP, Prec, Path, Pos0, Pos1, Strm),
    tokenize_op_left(Term, LP, Path, Pos1, Pos, Strm).

tokenize_bolop_rtrees(Prec, OP, RP, Term, Fun, Ar,
                      Path, Pos0, Pos, Strm) -->
    {
     ip_set(tokenize_current_prec, 1000)
    },
    tokenize(Fun, [0|Path], Pos0, Pos1, Strm),
    ( {RP =:= 0} ->
        {Pos2 = Pos1}
    ;
        make_sp(Pos1, Pos1A, Strm),         
        tokenize_op_right(Term, RP, Ar, Path, Pos1A, Pos2, Strm)
    ),
    tokenize_paren(r, OP, Prec, Path, Pos2, Pos, Strm).

tokenize_bolop_rightassoc_trees(Prec, OP, LP, RP, Fun, Term,
                                Path, Pos0, Pos, Strm) -->
    {
     functor(Term, AFun, 2),
     Fun == AFun,
     arg(1, Term, LArg),
     arg(2, Term, RArg)
    },
    !,
    {
     ip_set(tokenize_current_prec, 1000),
     tokenize(Fun, [0|Path], Pos0, Pos1, Strm, FunTree),
     phrase(make_sp(Pos1, Pos1A, Strm), [FunSP]),         
     ip_set(tokenize_current_prec, RP),
     tokenize(LArg, [1,2|Path], Pos1A, Pos2, Strm, ArgTree),
     phrase(make_sp(Pos2, Pos2A, Strm), [ArgSP])
    },
    make_tok_tree(Path, Pos0, Pos2A, 1, [FunTree,FunSP,ArgTree,ArgSP]),
    tokenize_bolop_rightassoc_trees(Prec, OP, LP, RP, Fun, RArg,
                                    [2|Path], Pos2A, Pos, Strm).
tokenize_bolop_rightassoc_trees(Prec, OP, LP, RP, Fun, Term,
                                Path, Pos0, Pos, Strm) -->
    {
     ip_set(tokenize_current_prec, 1000),
     tokenize(Fun, [0|Path], Pos0, Pos1, Strm, FunTree),
     phrase(make_sp(Pos1, Pos1A, Strm), [FunSP]),
     ip_set(tokenize_current_prec, RP),
     tokenize(Term, [2|Path], Pos1A, Pos2, Strm, ArgTree)
    },
    make_tok_tree(Path, Pos0, Pos2, 1, [FunTree,FunSP,ArgTree]),
    tokenize_paren(r, OP, Prec, Path, Pos2, Pos, Strm).

tokenize_op_left(Term, LP, Path, Pos0, Pos, Strm) -->
    {
     arg(1, Term, ArgL),
     ip_set(tokenize_current_prec, LP)
    },
    tokenize(ArgL, [1|Path], Pos0, Pos1, Strm),
    make_sp(Pos1, Pos, Strm).
 
tokenize_op_right(Term, RP, Ar, Path, Pos0, Pos, Strm) -->
    {
     arg(Ar, Term, ArgR),
     ip_set(tokenize_current_prec, RP)
    },
    tokenize(ArgR, [Ar|Path], Pos0, Pos, Strm).

tokenize_paren(Paren, OP, Prec, Path, Pos0, Pos, Strm) -->
    {OP>Prec},
    !,
    { Paren = l ->
        Tok = "("
    ;
        Tok = ")"
    },
    make_literal(Tok, Path, Pos0, Pos, Strm).
tokenize_paren(_Paren, _OP, _Prec, _Path, Pos, Pos, _Strm) -->
    [].

%
%% make_token(Path, Pos0, Pos, Token).
%% make_token(Path, Pos0, Pos) --> Tokens.
%% make_sp(Pos0, Pos, Strm) --> Tokens.
%% make_literal(Chars, Path, Pos0, Pos, Strm) --> Tokens.
%%
%%  mode make_token(@subterm_path, @integer, -integer, -compound).
%%  mode make_token(@subterm_path, @integer, -integer,
%%                  ?list(compound), ?list(compound)).
%%  mode make_sp(@integer, -integer, @stream, ?list(compound), ?list(compound)).
%%  mode make_literal(@chars, @subterm_path, @integer, -integer, @stream,
%%                  ?list(compound), ?list(compound)).
%% 
make_token(Path, Pos0, Pos,
           tok(Path, Pos0, Len)) :-
    Len is Pos - Pos0.

make_token(Path, Pos0, Pos) -->
    [tok(Path, Pos0, Len)],
    {Len is Pos - Pos0}.

make_sp(Pos0, Pos, Strm) -->
    {
     write_atom(Strm, ' '),
     Pos is Pos0 + 1
    },
    [sp].

make_literal(Chars, Path, Pos0, Pos, Strm) -->
    {
     make_literal(Chars, Strm, Pos0, Pos)
    },
    make_token(Path, Pos0, Pos).

make_literal([], _, P, P).
make_literal([C|Cs], Strm, P0, P) :-
    put(Strm, C),
    P1 is P0 + 1,
    make_literal(Cs, Strm, P1, P).

%% make_tok_tree(Path, Pos0, Pos, Indent, SubTrees, TokTree).
%% make_tok_tree(Path, Pos0, Pos, Indent, SubTrees) --> TokTrees.
%% make_nl_tok_tree(NlHere, Path, Pos0, Pos, Indent, SubTrees) --> TokTrees.
%%
%%  mode make_tok_tree(@subterm_path, @integer, -integer, @integer,
%%                     +list(compound), -compound).
%%  mode make_tok_tree(@subterm_path, @integer, -integer, @integer,
%%                     +list(compound), ?list(compound), ?list(compound)).
%%  mode make_nl_tok_tree(@atom, @subterm_path, @integer, -integer, @integer,
%%                     +list(compound), ?list(compound), ?list(compound)).
%% 
make_tok_tree(Path, Pos0, Pos, Indent, SubTrees,
              toks(Path,Pos0,Size,Indent,_NLTaken,SubTrees)) :-
    Size is Pos - Pos0.
make_tok_tree(Path, Pos0, Pos, Indent, SubTrees) -->
    [toks(Path,Pos0,Size,Indent,_NLTaken,SubTrees)],
    {Size is Pos - Pos0}.
make_nl_tok_tree(NlHere, Path, Pos0, Pos, Indent, SubTrees) -->
    [toks(Path,Pos0,Size,Indent,NlHere,SubTrees)],
    {Size is Pos - Pos0}.

%% push_tok_tree_head(Tree0, Push, Path, Pos, Tree).
%% push_tok_tree_head(Tree0, Push, Path, Pos) --> Trees.
%%
%%  The toktree Push occurs textually immediately before Tree0.
%%  Push it down inside Tree0 so that it is attached to the leftmost leaf of
%%  Tree0.  Path and Pos are the path and start position of Push (and hence of
%%  the resulting tree).
%%
%%  mode push_tok_tree_head(@compound, @compound, @subterm_path, @integer,
%%                          -compound).
%%  mode push_tok_tree_head(@compound, @compound, @subterm_path, @integer,
%%                          +list(compound), -list(compound)).
%% 
push_tok_tree_head(Tree0, Push, Path, Pos) -->
    {
     push_tok_tree_head(Tree0, Push, Path, Pos, Tree)
    },
    [Tree].

push_tok_tree_head(Tree0, toks(_,_,_,_,_,[]), _, _, Tree) :-
    !,
    Tree = Tree0.
push_tok_tree_head(Tok, Push, Path, Pos0, Tree) :-
    Tok = tok(_,TPos,TLen),
    !,
    Pos is TPos+TLen,
    make_tok_tree(Path, Pos0, Pos, 0, [Push,Tok], Tree).
push_tok_tree_head(Toks, Push, Path, Pos0, Tree) :-
    Toks = toks(_,TPos,TLen,TInd,TNL,[TTree|TTrees]),
    !,
    Pos is TPos+TLen,
    push_tok_tree_head(TTree, Push, Path, Pos0, PTree),
    make_nl_tok_tree(TNL, Path, Pos0, Pos, TInd, [PTree|TTrees], [Tree], []).
push_tok_tree_head(Toks, Push, _, _, Tree) :-
    Toks = toks(_,_,_,_,_,[]),
    !,
    Tree = Push.

%% push_tok_tree_tail(Tree0, Push, Path, Pos, Tree).
%% push_tok_tree_tail(Tree0, Push, Path, Pos) --> Trees.
%%
%%  The toktree Push occurs textually immediately after Tree0.
%%  Push it down inside Tree0 so that it is attached to the rightmost leaf of
%%  Tree0.  Path and Pos are the path and end position of Push (and hence of
%%  the resulting tree).
%%
%%  mode push_tok_tree_tail(@compound, @compound, @subterm_path, @integer,
%%                          -compound).
%%  mode push_tok_tree_tail(@compound, @compound, @subterm_path, @integer,
%%                          +list(compound), -list(compound)).
%% 
push_tok_tree_tail(Tree0, Push, Path, Pos) -->
    {
     push_tok_tree_tail(Tree0, Push, Path, Pos, Tree)
    },
    [Tree].

push_tok_tree_tail(Tree, toks(_,_,_,_,_,[]), _, _, Tree) :-
    !.
push_tok_tree_tail(Tok, Push, Path, Pos, Tree) :-
    Tok = tok(_,TPos,_TLen),
    !,
    make_tok_tree(Path, TPos, Pos, 0, [Tok,Push], Tree).
push_tok_tree_tail(Toks, Push, Path, Pos, Tree) :-
    Toks = toks(_,TPos,_TLen,TInd,TNL,TTrees),
    !,
    append(TFront, [TTail], TTrees),
    append(TFront, [PTail], PTrees),
    push_tok_tree_tail(TTail, Push, Path, Pos, PTail),
    make_nl_tok_tree(TNL, Path, TPos, Pos, TInd, PTrees, [Tree], []).
push_tok_tree_tail(Toks, Push, _, _, Tree) :-
    Toks = toks(_,_,_,_,_,[]),
    !,
    Tree = Push.


%
%% get_current_op(Thy, Fun, Ar, OP, LP, RP).
%% 
%%  Fun/Ar is an operator in Thy with precedence OP, left precedence LP and
%%  right precedence RP.
%%
%%  mode get_current_op(@thy, @atom, @integer, -integer, -integer, -integer).
%% 
get_current_op(Thy, Fun, Ar, OP, LP, RP) :-
    current_op(Thy, OP, Type, Fun),
    get_current_op_data(Type, Ar, OP, LP, RP).

get_current_op_data(fx, 1, OP, 0, RP) :-
    RP is OP - 1.
get_current_op_data(fy, 1, OP, 0, RP) :-
    RP = OP.
get_current_op_data(xf, 1, OP, LP, 0) :-
    LP is OP - 1.
get_current_op_data(yf, 1, OP, LP, 0) :-
    LP = OP.
get_current_op_data(xfx, 2, OP, LP, RP) :-
    LP is OP - 1,
    RP = LP.
get_current_op_data(xfy, 2, OP, LP, RP) :-
    LP is OP - 1,
    RP = OP.
get_current_op_data(yfx, 2, OP, LP, RP) :-
    LP = OP,
    RP is OP - 1.

%% get_current_quant(Quant0, Thy, OP, Quant).

get_current_quant(Quant0, Thy, Prec, Quant) :-
    current_op(Thy, Prec, quant, Quant0),
    !,
    Quant = Quant0.
get_current_quant(Quant0, _Thy, Prec, Quant) :-
    atom_concat('!!', Quant0, Quant),
    Prec = 600.

