%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% Procedures for displaying Ergo 4 proofs in HTML format.
%%
%% This file has yet to be integrated into Ergo 4.
%%
%% It was adapted from emacs_browser.ql in 1995 by Tapan Shah
%% <tapan@cs.uq.oz.au> as part of an IV honours project.
%%
%% CVS $Id: proof4_to_html.ql,v 1.3 2000/11/15 06:29:50 itp Exp $


% proof command
%   summarize_subproof,
%   summarize_subproof(Id)
% doc
%   'Write a transcript of subproof Id (default: current)',nl,
%   'and all its descendents.'
% mode
%   summarize_subproof,
%   summarize_subproof(@integer).
?- op(global, 50, xfx, :).
?- op(global, 400, xfx, /).
?- op(700, xfx, =).
?- op(810, xfy, and).
?- op(830, xfy, =>).
?- op(840, xfy, <=>). 
summarize_subproof :-
    open('any.html', write, Strm),
    writelist(Strm,['<title> Ergo Proof to Html Translator </title>']),
    find_subproof(Id,Steps, Hyps, Reln),
    summarize_proof(Id, Hyps, Steps, Reln, [], Sub),
    sort(Sub, Sub1),
    write_proof_summary(Sub1, Strm),
    close(Strm).

summarize_subproof(Id) :-
    current_output(Strm),
    summarize_subproof(Id, Strm).

summarize_subproof(Id, Strm) :-
    find_subproof(Id,Steps, Hyps, Reln),
    summarize_proof(Id, Hyps, Steps, Reln, [], Sub),
    sort(Sub, Sub1),

    write_proof_summary(Sub1, Strm).

write_proof_summary([], _).
write_proof_summary([subproof(Id, HLines, SLines)|SP], Strm) :-
    writelist(Strm, ['<h2> <a name = "sprf',w(Id),'">This is Subproof ', w(Id)]),
    writelist(Strm,['</h2>',nl]),
    reverse(HLines, HLinesR),
    write_proof_lines(HLinesR, Strm), nl(Strm),
    writelist(Strm,['<hr>',nl]),
    reverse(SLines, SLinesR),
    write_proof_lines(SLinesR, Strm), nl(Strm),
    writelist(Strm,[nl,nl]),
    writelist(Strm,['[<a href = "#sprf1"> First Proof </a>]',nl]),
    writelist(Strm,['<p><b> Click the BACK button on WEB MENU to go to the Parent Proof </b>']),
    writelist(Strm, ['<pre>',nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,nl,'</pre>']),
    write_proof_summary(SP, Strm).
  

write_proof_lines([], _).
write_proof_lines([Line|Lines], Strm) :-
    writelist(Strm, Line), nl(Strm),
    write_proof_lines(Lines, Strm).

summarize_proof(Id, Hyps, Steps, Reln, Sub0, Sub) :-
    Sub1 = [subproof(Id, HLines, SLines)|Sub0],
    summarize_proof_hyps(Id, Hyps, HLines, Sub1, Sub2),
    show_pretty_reln(Reln, PReln),
    summarize_proof_steps(Id, Steps, PReln, SLines, Sub2, Sub).
summarize_proof_steps(Id, [], _, [], Sub, Sub).
summarize_proof_steps(Id, [Step|Steps], Reln, [Line|Lines], Sub0, Sub) :-
    summarize_proof_step(Id, Step, Reln, Line, Sub0, Sub1),
    summarize_proof_steps(Id, Steps, Reln, Lines, Sub1, Sub).

summarize_proof_step(Id, pstep(Num, Term, Just), Reln, Line, Sub0, Sub) :-
    summarize_proof_just(Just, Reln, JLine, Sub0, Sub),
    append(JLine, [nl,'<b>','  ', thy(Term),'</b>','<p>'], Line).

summarize_proof_just(new, _, [], Sub, Sub) :-
    !.
summarize_proof_just(Trans, Reln, ['<pre>    ',q(Reln),tab|Msg], Sub, Sub) :-
    Trans = tr(_,_,_,_),
    !,
    pstep_msg(Trans, Msg).
summarize_proof_just(Open, Reln, ['<pre>    ',q(Reln),tab|Msg], Sub0, Sub) :-
    Open = op(_, subprf(Id, Hyps, Steps, SReln)),
    !,
    pstep_msg(Open, Msg),
    summarize_proof(Id, Hyps, Steps, SReln, Sub0, Sub).
summarize_proof_just(Wraptac, Reln, ['<pre>    ',q(Reln),tab|Msg], Sub0, Sub) :-
    Wraptac = wr(_,_,_,subprf(Id, Hyps, Steps, SReln)),
    !,
    pstep_msg(Wraptac, Msg),
    summarize_proof(Id, Hyps, Steps, SReln, Sub0, Sub).

summarize_proof_hyps(Id, [], [], Sub, Sub).
summarize_proof_hyps(Id, [hrec(Num, Term, Del, Just, _, _)|Hyps],
                    [Line|Lines], Sub0, Sub) :-
    Line = ['<b>','h', w(Num), ': ', thy(Term),'</b>','</p>', nl, DelMsg, tab | JLine],
    ( var(Del) ->
        DelMsg = tab
    ;
        DelMsg = '(deleted)'
    ),
    summarize_proof_hyp_just(Just, JLine, Sub0, Sub1),
    summarize_proof_hyps(Id, Hyps, Lines, Sub1, Sub).

summarize_proof_hyp_just(i,
                         ['<pre>     <i>inherited</i></pre>','<p>'],
                         Sub, Sub).
summarize_proof_hyp_just(cj(J),
                         ['<pre>     <i>undischarged conjecture</i></pre>','<p>'],
                         Sub, Sub) :-
    var(J),
    !.
summarize_proof_hyp_just(cj(tr(JRule,JDir,_,JAnts)),
                         ['<pre>     <i>conjecture discharged by d_trans',
                          '(hyp:', w(Num), ', ', w(JRule), ', ',
                            w(JDir), ', ', w(JAnts), ')</i></pre>','<p>'],
                         Sub, Sub).
summarize_proof_hyp_just(cj(subprf(JId, JHyps, JSteps, JReln)),
                         ['<pre>     <i>conjecture discharged by <a href = "#sprf',w(JId),'"','>','subproof ', w(JId),'</a></i></pre>','<p>'],
                         Sub0, Sub) :-
    summarize_proof(JId, JHyps, JSteps, JReln, Sub0, Sub).
summarize_proof_hyp_just(new,
                         ['<pre>     <i>introduced by opening rule.</i></pre>','<p>'],
                         Sub, Sub).
summarize_proof_hyp_just(sl(N),
                         ['<pre>     <i>by splitting hyp ', w(N),'</i></pre>','<p>'],
                         Sub, Sub).
summarize_proof_hyp_just(sr(N),
                         ['<pre>     <i>by splitting hyp ', w(N),'</i></pre>','<p>'],
                         Sub, Sub).
summarize_proof_hyp_just(oh(N, subprf(Id, Hyps, Steps, Reln)),
                         ['<pre>     <i>by <a href = "#sprf',w(Id),'"','>','subproof ', w(Id),' </a>', ' from hyp ', w(N),'</i></pre>','<p>'],
                         Sub0, Sub) :-
    summarize_proof(Id, Hyps, Steps, Reln, Sub0, Sub).
summarize_proof_hyp_just(ht(HN, HR, HD, HA),
                         ['<pre>     <i>by ', q(h_trans(HN, HR, HD, HA)),'</i></pre>','<P>'],
                         Sub, Sub).
summarize_proof_hyp_just(tr(PN),
                         ['<pre>     <i>by translation of parent hyp ', w(PN),'</i></pre>','<p>'],
                         Sub, Sub).


%%
%% THIS IS A MODIFIED VERSION OF pstep_msg FROM showwin.ql
%% =======================================================
%%
pstep_msg(tr(T:N,D,_,A),
          ['<i>by ',q(T),':',q(N),', ',q(D),', ',q(A),'</i></pre>','<p>']).
pstep_msg(op(P,subprf(Id,_,_,_)),
          ['<i>by <a href = "#sprf',w(Id),'"','>','subproof ', w(Id),'</a>', ' on ',q(P),'</i></pre>','<p>']).
pstep_msg(wr(wf(TacId,Msg,_),_,_,subprf(Id,_,_,_)),
          ['<i>by <a href = "#sprf',w(Id),'"','>','subproof ', w(Id),'</a>', ' using tactic ', q(TacId),': ','</i></pre>','<p>'|Msg]) :-
    Msg = [_|_].
pstep_msg(wr(wf(TacId,[],Tac),_,_,subprf(Id,_,_,_)),
          ['<i>by <a href = "#sprf',w(Id),'"','>','subproof ', w(Id),'</a>', ' using tactic ', q(TacId),': ',q(Tac),'</i></pre>','<p>']).
pstep_msg(wr(wa(TacId,Msg,Arg),_,_,subprf(Id,_,_,_)),
      ['<i>by <a href = "#sprf',w(Id),'"','>','subproof ', w(Id),'</a>', ' using tactic ', q(TacId),'.',q(Arg),'  ','</i></pre>','<p>'|Msg]).


