%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: prooffile.ql,v 1.41 2006/02/01 23:36:26 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file defines how 'gumtree' proofs are saved into disk files,
%% reloaded and rerun etc.
%%
%% The basic idea is that we turn a proof (which may have unproven
%% nodes) into a tactic, then save that tactic into a file.
%% To enable the proof to be rerun, we must also record any extra
%% information that is needed by proof_complete, such as the desired
%% order of the upper sequents and any extra constraints.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../prover/lib/ergo.h"

?- require_global_tactic_file(treetac).


%% This version number should be increased whenever the format of *.proof
%% files is changed, especially when the changes mean that some older
%% proofs may not rerun without changes, or may lack some information
%% that newer proofs have.
%%
%% When you change this, you must also change the `latest version'
%% comment in the declaration of the read_proof[_file] commands.
%%
proof_file_version('5.0').

display_prooftree(Node) :-
    handle_event(display_prooftree(Node)),
    !.
display_prooftree(Node) :-
    prooftreeIDs(Node, Tree),
    show_prooftreeIDs(Tree, 0),
    writelist([nl,nl]).



?- (\+ query_global(request(display_prooftree(Node),_))
    ->
        add_request(display_prooftree(Node),
                    ['display_prooftree(Node).',nl,
                     'Display the proof tree.'])
   ;
       true
   ).


proof command
  prooftreeIDs(Root, Tree)
extends
  proof_tree_cmds
doc
  'Tree is a representation of the tree of proofnodes below Root.'
mode
  prooftreeIDs(@nodeid, ?compound).

prooftreeIDs(PNode, Tree) :-
    ( proof_node_rule_and_children(PNode, _, _, _, Children) ->
        map(prooftreeIDs, [Children, ChildTrees]),
        Tree = (PNode;ChildTrees)
    ;
        Tree = PNode
    ).
    
proof command
  show_prooftree(Root)
extends
  prooftree_info_cmds
doc
  'Display a representation of the tree of proofnodes below Root.'
mode
  show_prooftree(@nodeid).

show_prooftree(Root) :-
    display_prooftree(Root).

%% private show_prooftreeIDs(Tree, Indent).
show_prooftreeIDs(;(T1,T2), Indent) :-
    !,
    writelist([nl, tab(Indent), q(T1)]),
    Indent2 is Indent+2,
    show_prooftreeIDs(T2, Indent2).
show_prooftreeIDs([], _Indent) :-
    !.
show_prooftreeIDs([T|Ts], Indent) :-
    !,
    show_prooftreeIDs(T, Indent),
    show_prooftreeIDs(Ts, Indent).
show_prooftreeIDs(T, Indent) :-
    writelist([nl, tab(Indent), q(T)]).

proof command
  show_prooftree_tactic(Root)
extends
  prooftree_info_cmds
doc
  'Display the proof tree below Root as a tactic. Note that the',nl,
  'tactic displayed does not provide any information about the order',nl,
  'in which the tree was generated. To do this use show_subproof.'
mode
  show_prooftree_tactic(@nodeid).

show_prooftree_tactic(Root) :-
    prooftree_tactic(Root, Tac),
    show_tactic(false, Tac).

prooftree_tactic(PNode, Tac) :-
    ( proof_node_rule_and_children(PNode, Thy, Rule, RuleType, Children) ->
        map(prooftree_tactic, [Children,Tac3]),
        simplify_term(Rule, SRule),
        Tac = (RuleType(Thy:SRule);Tac3)
     ;
        Tac = skip
    ).

    

proof command
  proof_tactic(Ann, NameMap, Root, Tac)
extends
  proof_tree_cmds
doc
  'Tac is a tactic that will regenerate the proof tree given by PNode.',nl,
  'PNode can be the name of any proof node -- it does not have to',nl,
  'be the root node of the current proof.',
  para,
  'Ann must be true or false.  If it is true, then Tac includes commands',nl,
  'to regenerate all the annotations of the proof tree as well.  If it',nl,
  'is false, no annotations will be recorded.',
  para,
  'NameMap must be a closed list of NodeId=Var pairs.  The generated',nl,
  'tactic will include a name(Var) command for each node NodeId.',nl,
  'This gives a way of naming the nodes that will be created',nl,
  'when the tactic runs.'
mode
  proof_tactic(@bool, @list(compound), @nodeid, ?gumtreetactic).

proof_tactic(Ann, NameMap, PNode, Tac) :-
    generate_proof_tactic([PNode], Ann, NameMap, Tac).

generate_proof_tactic([], _, _, skip) :- !.
generate_proof_tactic(Nodes, Ann, NameMap, Tac) :-
    setof(T+N, (member(N, Nodes), proof_node_timestamp(N,T)), S),!,
    S = [(_ + N1)|_],
    append(NL1, [N1|NL2], Nodes),
    length(NL1, NL1Len),
    proof_node_rule_and_children(N1, Thy, Rule, RuleType, Children),
    simplify_term(Rule, SRule),
    add_context_to_rule_tac(N1, Thy, SRule, RuleType, RuleTac),
    proof_tactic_ann(Ann, N1, Tac1, RuleTac),
    ( member(N1 = Var, NameMap) ->
        Tac2 = (name(Var);Tac1)
    ;
        Tac2 = Tac1
    ),
    ( NL1Len = 0 ->
	Tac = ([Tac2, skip];Rest)
      ;
        Tac = ([fix(NL1Len), Tac2, skip];Rest)
    ),
    append(Children, NL2, NL3),
    append(NL1, NL3, NewNodes),
    generate_proof_tactic(NewNodes, Ann, NameMap, Rest).
generate_proof_tactic([], _Ann, _NameMap, skip) :- !.
generate_proof_tactic(Nodes, Ann, NameMap, Tac) :-
    generate_proof_tactic_open(Nodes, Ann, NameMap, Tac).

generate_proof_tactic_open([], Ann, NameMap, []).
generate_proof_tactic_open([PNode|Nodes], Ann, NameMap, [HTac|Tac]) :-
    Tac2 = skip,
    proof_tactic_ann(Ann, PNode, Tac1, Tac2),
    ( member(PNode = Var, NameMap) ->
        HTac = (name(Var);Tac1)
    ;
        HTac = Tac1
    ),
    generate_proof_tactic_open(Nodes, Ann, NameMap, Tac).
    

/*
proof_tactic(Ann, NameMap, PNode, Tac) :-
    ( proof_node_rule_and_children(PNode, Thy, Rule, RuleType, Children) ->
        %% build the list of subtree tactics
        map(proof_tactic(Ann, NameMap), [Children,Tac3]),
	simplify_term(Rule, SRule),
        add_context_to_rule_tac(PNode, Thy, SRule, RuleType, RuleTac),
        Tac2 = (RuleTac;Tac3)
    ;
        Tac2 = skip
    ),
    %% prefix Tac2 with annotations, if desired.
    proof_tactic_ann(Ann, PNode, Tac1, Tac2),
    ( member(PNode = Var, NameMap) ->
        Tac = (name(Var);Tac1)
    ;
        Tac = Tac1
    ).

*/

proof_tactic_ann(true, PNode, Tac0, Tac) :-
    annotations(PNode, Anns),
    Anns = [_|_],
    !,
    Tac0 = (annotate(Anns);Tac).
proof_tactic_ann(_, _, Tac0, Tac0).


proof command
  tactic_annotations(Tac, Patt, Anns)
extends
  proof_tree_cmds
doc
  'Extracts annotations from the gumtree tactic Tac.',nl,
  'Anns is the list of all annotations within Tac that match Pattern.'
mode
  tactic_annotations(@gumtreetactic, @term, ?list(term)).

tactic_annotations(Tac, Patt, List) :-
    tac_anns(Tac, Patt, List, []).

tac_anns((T1;T2), P)      --> !, tac_anns(T1,P), tac_anns(T2,P).
tac_anns('|'(T1,T2), P)   --> !, tac_anns(T1,P), tac_anns(T2,P).
tac_anns([T1|T2], P)      --> !, tac_anns(T1,P), tac_anns(T2,P).
tac_anns('!'(T1), P)      --> !, tac_anns(T1,P).
%
% The DCG expander cannot handle this:
%
%   tac_anns(annotate(L), P)  --> {list(L), !, filter('?='(P), L, L2)}, L2.
%
% so we handle the annotate(List) case with the next two clauses:
%
tac_anns(annotate([A|AS]), P) --> {P ?= A}, !, [A], tac_anns(annotate(AS),P).
tac_anns(annotate([_|AS]), P) --> !, tac_anns(annotate(AS),P).

tac_anns(annotate(A), P)  --> {P ?= A}, !, [A].
tac_anns(_, _)            --> [].


general command
  show_proof(RuleName),
  show_proof_file(File)
extends
  general_cmds
doc
  'Converts the proof file for RuleName into a proof tactic and prints it.',nl,
  'Annotations are ignored; only the primitive rules are shown,',nl,
  'combined by sequential and parallel composition operators according to',nl,
  'the structure of the proof.'
mode
  show_proof(@rulename),
  show_proof_file(@filename).

show_proof(RuleName) :-
    read_proof(RuleName, _, Tac, _),
    show_tactic(false, Tac).

show_proof_file(File) :-
    read_proof_file(File, _, Tac, _),
    show_tactic(false, Tac).

proof command
  show_subproof(Node)
extends
  prooftree_info_cmds
doc
  'Display the subtree of the current proof rooted at Node.',nl,
  'Annotations are ignored; only the primitive rules are shown,',nl,
  'combined by sequential and parallel composition operators according to',nl,
  'the structure of the proof.'
mode
  show_subproof(@nodeid).

show_subproof(Node) :-
    proof_tactic(true, [], Node, Tac),
    show_tactic(false, Tac).

general command
  show_proof_script(RuleName),
  show_proof_script_file(File)
extends
  general_cmds
doc
  'Converts the proof file for RuleName into a proof script and prints it.',nl,
  'It does this by processing the ',
  xref('usercmd annotations',global,usercmd_annotation),nl,
  'that appear within the proof file.  The show_proof_script_file form',nl,
  'uses the given File directly, without searching.'
mode
  show_proof_script(@rulename),
  show_proof_script_file(@filename).

show_proof_script(RuleName) :-
    read_proof(RuleName, _, Tac, _),
    show_tactic_script(Tac).

show_proof_script_file(File) :-
    read_proof_file(File, _, Tac, _),
    show_tactic_script(Tac).


proof command
  show_subproof_script(Node)
extends
  prooftree_info_cmds
doc
  'Prints a proof script of the Node subtree of the current proof.',nl,
  'It does this by processing the ',
  xref('usercmd annotations',global,usercmd_annotation),nl,
  'that appear within the subtree.  If Node was proved by a',nl,
  'command outside the subtree, then the resulting proof script',nl,
  'will be incomplete.  To alert you to this possibility, a warning',nl,
  'message is printed if Node has no usercmd annotation.'
implementation
  call_passive_cmd(show_subproof_script(Node))
mode
  show_subproof_script(@nodeid).

show_subproof_script(Node) :-
    %% A more efficient implementation would be to scan the
    %% proof tree directly (extracting usercmd annotations), rather
    %% than turning it into a tactic first.
    proof_node_term(Node, Concl),
    proof_tactic(true, [], Node, Tac),
    writelist([
        '%% Proof script from subproof ',q(Node),'.',nl,
        '%%',nl,
        '%% Conclusion: ',thy(Concl),nl,
%% TO DO: Does show_subproof_script correctly handle multi-line terms here?
%% Ray's last comment: It needs to invoke the pretty-printer.
%% (perhaps by defining a new writelist-clean.)
        '%%',nl]),
    ( annotation(Node, usercmd(_,_,_,_,_)) ->
        true
    ;
        writelist([
            '%% WARNING: Node ',q(Node),' was not proved interactively, ',nl,
            '%%          so proof script may be incomplete.',nl,
            '%%',nl])
    ),
    show_tactic_script(Tac).


general command
  show_tactic(Anns, Tac)
extends
  general_cmds
doc
  'Pretty-prints the gumtree tactic Tac.',nl,
  'Anns=true means the tactic includes annotations; otherwise, they',nl,
  'are discarded.'
mode
  show_tactic(@bool, @gumtreetactic).

show_tactic(Anns, Tac0) :-
    simplify_tac(Tac0, Anns, Tac),
    show_tactic1(Tac, 0),
    nl.

%% private simplify_tac(Tac0, Anns, Tac).
simplify_tac(;(T1,T2), Anns, Tac) :-
    !,
    simplify_tac(T1, Anns, T1S),
    simplify_tac(T2, Anns, T2S),
    ( T1S = [] ->
        Tac = T2S
    ; T2S = [] ->
        Tac = T1S
    ;
        Tac = ;(T1S,T2S)
    ).
simplify_tac(List0, Anns, Tac) :-
    list(List0),
    !,
    simplify_tac_list(List0, Anns, List),
    ( List = [] ->
        Tac = []
    ; List = [Tac1] ->
        Tac = Tac1
    ;
        Tac = List
    ).
simplify_tac(annotate(_), false, Tac) :-
    !,
    Tac = [].
simplify_tac(Tac, _, Tac).

simplify_tac_list([], _, []).
simplify_tac_list([T1|Ts], Anns, List) :-
    simplify_tac(T1, Anns, T1S),
    simplify_tac_list(Ts, Anns, List1),
    ( T1S = [] ->
        List = List1
    ;
        List = [T1S|List1]
    ).

%% private show_tactic1(Tac, Indent).
/*
show_tactic1(;(T1,T2), _) :-
    show_tactic1(T1, _),
    writelist([';', nl]),
    show_tactic1(T2, _).
show_tactic1(Tac, _) :-
    writeRq(Tac).
*/
 
show_tactic1(;(T1,T2), Indent) :-
    !,
    show_tactic1(T1, Indent),
    writelist([';', nl, tab(Indent)]),
    show_tactic1(T2, Indent).
show_tactic1(Tac, _Indent) :-
    Tac = [fix(_N),T,skip],
    compound(T),
    functor(T,F,_),
    (F = rule ; F = srule),
    !,
    writeRq(Tac).
show_tactic1(Tac, _Indent) :-
    Tac = [T,skip],
    compound(T),
    functor(T,F,_),
    (F = rule ; F = srule),
    !,
    writeRq(Tac).
show_tactic1([T], Indent) :-
    !,
    show_tactic1(T, Indent).
show_tactic1([T|Ts], Indent) :-
    !,
    Indent1 is Indent + 2,
    writelist(['[',nl,tab(Indent1)]),
    show_tactic1(T, Indent1),
    show_tactic_list(Ts, Indent, Indent1),
    writelist([nl, tab(Indent), ']']).
show_tactic1(Tac, _) :-
    writeRq(Tac).



%% private show_tactic_list(List, Indent, Indent1).
show_tactic_list([], _, _).
show_tactic_list([T|Ts], Indent, Indent1) :-
    writelist([nl, tab(Indent), ',', nl, tab(Indent1)]),
    show_tactic1(T, Indent1),
    show_tactic_list(Ts, Indent, Indent1).

general command
  show_tactic_script(Tac)
extends
  general_cmds
doc
  'Converts the gumtree tactic Tac into a proof script and prints it.',nl,
  'It does this by processing the ',
  xref('usercmd annotations',global,usercmd_annotation),nl,
  'that appear within Tac.'
mode
  show_tactic_script(@gumtreetactic).

show_tactic_script(Tac) :-
    tactic_annotations(Tac, usercmd(_,_,_,_,_), UCmds),
    tactic_annotations(Tac, name(_,_), UNames),
    append(UNames, _, Names),
    sort(UCmds, UCmds2),
    show_proof_script_lines(UCmds2, []=[], Names).

%% private show_proof_script_lines(Anns, PrevCmdOutputs, NodeNames).
%%
show_proof_script_lines([], _, _).
show_proof_script_lines([U1,U2|US], PrevOut, Names) :-
    U1 = usercmd(Num,_,_,_,_),
    U2 = usercmd(Num,_,_,_,_),
    !,
    %% don't output multiple copies of same usercmd.
    show_proof_script_lines([U2|US], PrevOut, Names).
show_proof_script_lines([U|US], PrevOut, Names) :-
    show_proof_script_line(U, PrevOut, UOut, Names),
    show_proof_script_lines(US, UOut, Names).

show_proof_script_line(usercmd(_,_,In,Out,Cmd), 
                       PrevNodes=PrevVars, Out=OutVars, Names) :-
    map(num_to_name(Names), [In,InVars]),
    map(num_to_name(Names), [Out,OutVars]),
    %%
    %% See if we need to name any outputs of the previous command for
    %% future reference.  We don't bother naming them if this command
    %% takes a single input node and that node was the first output
    %% node of the immediately preceeding command.
    ( InVars == PrevVars, InVars = [_] ->
        true
    ;
        proof_script_name_outputs(PrevVars)
    ),
    ( In = [] ->
        %% no input nodes to select!
        writelist([tab])
    ; In = [_], In = PrevNodes ->
        %% No need to select the In node.
        writelist([tab])
    ; InVars = [V] ->
        writelist([w(V),':::',tab])
    ;
        writelist([w(InVars),':::',nl,tab])
    ),
    writelist([chars(Cmd)]).

%% num_to_name(Names, Num, Name).
%% num_to_var(Names, Num, Var).
%%
%%  Names is an open list of name(Node,Name) pairs.  The names may
%%  be named variables or atoms.
%%  num_to_name can return non-variable names; num_to_var will always generate
%%  named variables.
%%  Either version will extend NameMap with a name(node,var) pair if necessary.
%%  
num_to_name(Map, Num, Name) :-
    member(name(Num,Name), Map),
    ( nonvar(Name) ->
        true
    ; get_var_name(Name, _) ->
        true
    ;
        set_var_name(Name, 'N')
    ).

num_to_var(Map, Num, Var) :-
    member(name(Num,Name), Map),
    ( nonvar(Name) ->
        set_var_name(Var, Name)
    ; get_var_name(Name, _) ->
        Var = Name
    ;
        set_var_name(Var, 'N')
    ).

proof_script_name_outputs([]) :-
    !.
proof_script_name_outputs([V]) :-
    !,
    writelist([tab,'  ---> ',w(V),'.',nl]).
proof_script_name_outputs(L) :-
    writelist([tab,'  ---> ',w(L),'.',nl]).


proof_file_properties header
  'Proof File Properties'
doc
  'It is often useful to store additional information when a proof',nl,
  'is saved into a disk file.  For this reason, the read and write',nl,
  'procedures for proof files allow a ''property list'' to be saved',nl,
  'with the proof.  The property list is a closed list of',nl,
  'PropertyName=Value pairs.',
  para,
  'Callers of these procedures may define new ''PropertyName''s, but',nl,
  'the following PropertyName''s have predefined meanings:',nl,
  begin(list),
  item('annotations=Bool.'),' Annotations are saved iff Bool=true.',nl,
  '    If this property is not specified explicitly, the default is true.',nl,
  item('rulename=Term.'),' If this proof has been saved into a theory,',nl,
  '    then Term is the name of the resulting rule.  The outermost',nl,
  '    functor of Term is usually the same as the name of the file',nl,
  '    in which this proof is saved.',nl,
  item('premises=List.'),' List is a list of proof node names.  When',nl,
  '    the proof file is read back in again, List will be a list',nl,
  '    of variables, which will be instantiated to the corresponding',nl,
  '    node names when the proof tactic is rerun.  This is usually',nl,
  '    used to record the unproved nodes of the proof.',nl,
  item('concl=Term.'),' Term is the conclusion of PNode.',nl,
  '    This is not essential for rerun, but is useful documentation.',nl,
  item('constraints=Code.'),' Code is a comma-separated list of',nl,
  '    user-supplied constraints on the proof.  These are in addition',nl,
  '    to the constraints that are generated by the proof itself.',nl,
  end(list)
extends
  concepts.


proof command
  write_proof_file(File, PNode, Properties)
doc
  'Writes the proof subtree PNode into File',
  para,
  'File is typically an atom of the form ''Theory/Name.proof''.',nl,
  'Any directories that File refers to must already exist.',
  para,
  'PNode can be the name of any proof node -- it does not have to',nl,
  'be the root node of the current proof, though it usually is.',
  para,
  'Properties must be a closed list of PropertyName=Value pairs.',nl,
  'See ',xref('Proof File Properties',global,proof_file_properties),'.'
extends
  save_close_treecmds
mode
  write_proof_file(@atom, @nodeid, @list(compound)).

write_proof_file(File, PNode, Props0) :-
    check_atom(File),
    ( member(annotations=Ann, Props0) ->
        true
    ;
        Ann = true
    ),
    msg(write_proof_file, 2, ['writing proof to ',q(File),
                              ' with annotations=',q(Ann)]),
    ( delete(premises=Ids, Props0, Props1) ->
        %% replace Ids by a list of new variables.
        map_domain_range(NameMap, Ids, Vars),
        Props = [premises=Vars|Props1]
    ;
        Props = Props0
    ),
    proof_tactic(Ann, NameMap, PNode, Tac),
    gumtree_get_abbrev_premises(Prems),
    proof_file_version(Ver),
    name_vars((Prems, Tac)),
    ( open(File, write, Stream) ->
        unwind_protect(
            encoded_write_term(Stream,proof(Ver,Props,(premises(Prems);Tac)),[remember_name(true), quoted(true)]),
            close(Stream))
    ;
        handle_error(['cannot create proof file: ',q(File)])
    ).
    

library command
  read_proof(RuleName, Version, Tac, Properties),
  read_proof_file(File, Version, Tac, Properties)
doc
  'These both read a proof record (tactic + property list) from a file.',
  para,
  'The read_proof/4 command uses the ',
  xref(load_path,global,'parameter:load_path'),' parameter',nl,
  'to search for the proof file, whereas the read_proof_file/4 command',nl,
  'requires a precise path to a proof file.',
  para,
  'RuleName may be of the form Thy:Name, or just Name, in which case',nl,
  'Thy defaults to the current theory.  In both cases, the tactic is',nl,
  'read from a file called ''Thy/Name.proof'', and the ',
  xref(load_path,global,'parameter:load_path'),nl,
  'is used to determine the directories under which this file may appear.',
  para,
  'File must be the full path (relative or absolute) to a proof file.',nl,
  'Proof files typically have a .proof extension.',
  para,
  'Version is an atom that describes the version of the proof file',nl,
  'format that the proof was read from.  The latest version is ''5.0''.',
  para,
  'Tac is the gumtree tactic that is read from the proof file.',nl,
  'Properties is a closed list of PropertyName=Value pairs.',nl,
  'See ',xref('Proof File Properties',global,proof_file_properties),'.'
extends
  save_close_treecmds                                                 
mode
  read_proof(@rulename, ?atom, ?gumtreetactic, ?propertylist),
  read_proof_file(@filename, ?atom, ?gumtreetactic, ?propertylist).


read_proof(Name, Version, Tac, Props) :-
    atom(Name),
    !,
    check_current_theory(Thy),
    read_proof(Thy:Name, Version, Tac, Props).
read_proof(Thy:Name, Version, Tac, Props) :-
    check_atom(Thy),
    check_atom(Name),
    concat_atom([Thy,'/',Name,'.proof'], PathEnd),
    get_load_path(PathList),
    member(P, PathList),
    full_file_name(P, PathEnd, _, FullPath),
    msg(read_proof, 2, ['trying: ',q(FullPath)]),
    access(FullPath, 4, 0),
    !,
    read_proof_file(FullPath, Version, Tac, Props).
read_proof(Name, _, _, _) :-
    handle_error(['cannot find ',q(Name),' proof file in load_path.']).

                                   
read_proof_file(File, Version, Tac, Props) :-
    check_atom(File),
    msg(read_proof_file, 2, ['reading ',q(File)]),
    ( open(File, read, Stream) ->
        unwind_protect(
            read_proof_term(File, Stream, Version, Props, Tac),
            close(Stream))
    ;
        handle_error(['cannot open proof file: ',q(File)])
    ).

read_proof_term(File, Stream, Version, Props, Tac) :-
    encoded_read_term(Stream,Proof,[]),
    ( Proof = proof(V,P,T), list(P) ->
        Version = V,
        Props = P,
        Tac = T
    ;
        handle_error(['illegal proof file: ',q(File)])
    ).


library command
  map_domain_range(Map, Left, Right)
doc
  'Map is a list of equality terms [L1=R1,L2=R2,...],',nl,
  'Left is a list of the left-hand sides [L1,L2,...] (the domain) and',nl,
  'Right is a list of the right-hand sides [L1,L2,...] (the range).',nl,
  'At least one of the three lists must be closed when this is called.'
mode
  map_domain_range(+list(compound), ?list(term), ?list(term)),
  map_domain_range(?list(compound), +list(term), ?list(term)),
  map_domain_range(?list(compound), ?list(term), +list(term)).

map_domain_range([], [], []).
map_domain_range([L=R|MS], [L|LS], [R|RS]) :-
    map_domain_range(MS, LS, RS).
    


gumtree command
  save_proof(Name, Root)
implementation
  call_gumtree_tactic({save_proof(Name, Root, [], true)})
command
  save_proof(Name, Root, OpenNodes, Constraints)
implementation
  call_gumtree_tactic({save_proof(Name, Root, OpenNodes, Constraints)})
extends
  save_close_treecmds
doc
  'Saves a proof into the current theory (as a new rule) and into a file.',nl,
  'That is, this command invokes ',
  xref('proof_complete/4', global, 'proof:proof_complete/4'),', then ',nl,
  xref('write_proof_file/3',global,'proof:write_proof_file/3'),
      ' which saves the proof into the file',nl,
  '`./Theory/Name.proof)''.',
  para,
  'Name is the name of the new rule in the current theory.',nl,
  'The outermost functor of Name, say F, is also used to determine the',nl,
  'file in which the proof is saved: ''Thy/F.proof''.',
  para,
  'Root must be the name of the root node of a proof.',
  para,
  'OpenNodes determines the order of the upper sequents of the new',nl,
  'inference rule.  Constraints is any additional constraints that',nl,
  'the new inference rule should have.  These two arguments are described',nl,
  'more fully in the ',xref(proof_complete,global,'proof:proof_complete/4'),
  ' command documentation.',
  para,
  'Name, OpenNodes and Constraints are saved in the proof file so that',nl,
  'the same inference rule can be generated when the proof is rerun.',
  para,
  'The two argument form sets OpenNodes=[] and Constraints=true.',nl,
  'This is useful for saving simple theorems.',
  para,
  'Note the difference between this command and the ',
  xref(save_postulate_proof,global,'gumtree:save_postulate_proof/1'),nl,
  'command.  This command can be called at any time, during any proof,',nl,
  'and it always creates a ',emph(new),' rule in the current theory.',nl,
  'The save_postulate_rule command must be called within the proof of a',nl,
  'postulate (started via ',
    xref('prove_rule/1',global,'database:prove_rule/2'),
  ') and it does not create a',nl,
  'new rule, but just marks the postulate as having been proved.'
mode
  save_proof(@rulename, @nodeid),
  save_proof(@rulename, @nodeid, ?list(nodeid), +constraints).

save_proof(Name, Root) :-
    save_proof(Name, Root, [], true).

save_proof(Name, Root, OpenNodes, Constraints) :-
    check_current_theory(Thy),
    check_or_error(proof_root(Root),
                   [q(Root),' is not the root node of a proof.']),
    check_writable_directory(Thy),
    std_nonvar(Name),
    functor(Name, F, _),
    check_legal_filename(F),
    proof_complete(Name, Root, OpenNodes, Constraints),
    save_proof_to_file(Thy,Name,Root,OpenNodes,Constraints).


gumtree command
  save_postulate_proof(Root)
implementation
  call_gumtree_tactic({save_postulate_proof(Root)})
extends
  save_close_treecmds
doc
  'Completes the proof of a postulate and saves the proof in a file.',nl,
  'That is, this command invokes ',
  xref('proof_complete_postulate/1',
       global,
       'proof:proof_complete_postulate/2'),', then ',nl,
  xref('write_proof_file/3',global,'proof:write_proof_file/3'),
  'which saves the proof into the file `./Theory/Name.proof)''.',
  para,
  'Root must be the name of the root node of the proof of a postulate.',nl,
  'If the unproved nodes and constraints of the proof do not match',nl,
  'the postulate, then this command will fail with an error message.',
  para,
  'Note the difference between this command and the ',
  xref(save_proof,global,'gumtree:save_proof/4'),nl,
  'command.  This command can only be called within the proof of',nl,
  'a postulate (started via ',
  xref('prove_rule/1',global,'database:prove_rule/2'),
  ') and it does not create a',nl,
  'new rule, but just marks the postulate as having been proved.',nl,
  'The save_proof command can be called at any time',nl,
  'during any proof and always creates a ',emph(new),' rule in the',nl,
  'current theory.'
mode
  save_postulate_proof(@nodeid).

save_postulate_proof(Root) :-
    check_or_error(proof_root_postulate(Root,Thy,Name),
                   [q(Root),' is not the root node of a postulate proof.']),
    check_writable_directory(Thy),
    proof_open_nodes(Root,Open),
    proof_complete_postulate(Root, Open),
    save_proof_to_file(Thy,Name,Root,Open,true).


gumtree command
  save_proof_to_file(Thy,Name,Node,OpenNodes,Constraints)
extends
  save_close_treecmds
doc
  'Writes the proof subtree Node into File, but does not create a theorem.',nl,
  'This command is invoked automatically by the higher level proof saving',nl,
  'commands, such as ',
  xref('save_postulate_proof/1',global,'gumtree:save_postulate_proof/1'),
  ' and ',
  xref('save_proof/4',global,'gumtree:save_proof/4'),'.',nl,
  'However, you may occasionally find it useful to call it directly',nl,
  'instead.  For instance, if you want to save an incomplete proof',nl,
  'for further work later, or if you wish to save just one subtree of',nl,
  'a proof (the higher level commands can only save entire proofs).',
  para,
  'The proof is saved into the file ''Thy/Name.proof'' under the current',nl,
  'directory.  Thy and Name should both be atoms that do not contain',nl,
  'any characters that might be illegal within filenames',nl,
  'However, if Name is a compound term, then the functor of that term',nl,
  'is used as the filename instead of Name.',
  para,
  'Node can be the name of any proof node -- it does not have to',nl,
  'be the root node of the current proof, though it often is.',
  para,
  'OpenNodes and Constraints are saved as additional information in',nl,
  'the proof file.  They can be useful for rerunning the proof.',nl,
  'OpenNodes must be a closed list of proof node names, or a variable.',nl,
  'If it is a variable, then it is instantiated to all the open nodes',nl,
  'in the subtree specified by Node.'
implementation
  call_gumtree_tactic({save_proof_to_file(Thy,Name,Node,
                                          OpenNodes,Constraints)})
mode
  save_proof_to_file(@atom, @rulename, @nodeid, ?list(nodeid), @constraints).

save_proof_to_file(Thy,Name,Node,OpenNodes,Constraints) :-
    var(OpenNodes),
    !,
    proof_open_nodes(Node, OpenNodes),
    save_proof_to_file(Thy,Name,Node,OpenNodes,Constraints).
save_proof_to_file(Thy,Name,Node,OpenNodes,Constraints) :-
    %% Thy may not actually be a theory name, but it is 
    %% sufficent for it to be a legal filename component.
    check_legal_filename(Thy),
    check_writable_directory(Thy),
    check_std_nonvar(Name),
    functor(Name, File, _),
    check_legal_filename(Name),
    check_or_error(
        ( closed_list(OpenNodes),
          map(integer, [OpenNodes]) ),
        ['illegal list of open nodes: ',q(OpenNodes)]),
    check_or_error(
        proof_node_term(Node, Concl),
        ['illegal proof node: ',q(Node)]),
    concat_atom([Thy,'/',File,'.proof'], Path),
    write_proof_file(Path, Node,
                     [premises=OpenNodes,
                      annotations=true,
                      rulename=Name,
                      concl=Concl,
                      constraints=Constraints]),
    msg(save_proof, ['Proof saved in: ',q(File)]).


check_writable_directory(Dir) :-
    ( access(Dir, 7, 0) ->
        true
    ; access(Dir, 0, -1) ->
        writelist(['Creating proofs directory, ',q(Dir),'...',nl]),
        make_directory(Dir)
    ;
        handle_error(['directory ',q(Dir),' is not readable/writable.'])
    ).



database command
  rerun_proof(RuleName)
extends
  theory_extension_cmds
doc
  'Reruns the proof file determined by RuleName.',nl,
  nl,
  'RuleName may be of the form Thy:Name, or just Name, in which case',nl,
  'Thy defaults to the current theory.  In both cases, a proof is',nl,
  'read from the file called ''Thy/Name.proof'' and the ',
  xref(load_path,global,'parameter:load_path'),nl,
  'is used to determine the directories in which this file may appear.'
mode
  rerun_proof(@rulename).

rerun_proof(RuleName) :-
    check_current_theory(Thy),
    read_proof(RuleName,_,(premises(_);Tac),Props),
    member(premises=Premises,       Props),
    member(rulename=Name,           Props),
    member(constraints=Constraints, Props),
    !,
    msg(rerun, ['starting proof of: ',q(Thy:Name)]),
    proof_start(Concl, true, Open),
    thaw_term(Concl),
    Open = [Root|OT],
    check_or_error(
        call_gumtree_tactic(Tac, Open, OT, _Output, _OT),
        ['tactic from proof file ',q(RuleName),' failed.']),
    ( proof_complete(Name, Root, Premises, Constraints) ->
        msg(rerun, ['proved rule ',q(Thy:Name)])
    ;
        handle_error(['tactic succeeded, but proof_complete/4 failed.'])
    ).
    
database command
  rerun_all_postulate_proofs
extends
  theory_extension_cmds
doc
  'Reruns all proofs for postulates in the current theory.'
mode
  rerun_all_postulate_proofs.

rerun_all_postulate_proofs :-
    check_current_theory(Thy),
    rule_query(Thy, Name, _,_,_,_,_),
    get_rule_kind(Thy, Name, postulate),
    once(rerun_postulate_proof(Thy:Name)),
    fail.
rerun_all_postulate_proofs.

%% TO DO: The use of once() in rerun_all_postulates is a hack.
%% It fixes rerunning a proof in a number of ways and hides a bug
%% that lives further down in the rerun_postulate_proof mechanism
%% where a deterministic course of events should be followed. The
%% once() can probably be replaced by one ! somewhere.

database command
  rerun_postulate_proof(RuleName)
extends
  theory_extension_cmds
doc
  'Reruns the proof file determined by RuleName.',nl,
  nl,
  'RuleName must be an existing postulated inference rule.',nl, 
  'RuleName may be of the form Thy:Name, or just Name, in which case',nl,
  'Thy defaults to the current theory.  In both cases, a proof is',nl,
  'read from the file called ''Thy/Name.proof'' and the ',
  xref(load_path,global,'parameter:load_path'),nl,
  'is used to determine the directories in which this file may appear.'
mode
  rerun_postulate_proof(@rulename).

rerun_postulate_proof(RuleName) :-
    decode_rule(RuleName, Thy, Name),
    check_current_theory(Thy),
    read_proof(RuleName,_,Tac,Props),
    member(premises=Premises,       Props),
    member(rulename=Name,           Props),
    member(concl=Concl,             Props),
    msg(rerun, ['starting proof of: ',q(Thy:Name)]),
    proof_start_postulate(Name, [Root]),
    Open = [Root|OT],
    proof_node_term(Root, Concl1),
    structural_unify(Concl, Concl1),
    check_or_error(
        ( call_gumtree_tactic(Tac, Open, OT, Output, OT),
          incomplete_retry_delays,
          proof_complete_postulate(Root, Premises) 
        ),
        ['tactic from proof file ',q(RuleName),' failed.']),
    msg(rerun, ['proved rule ',q(Thy:Name)]).
    


gumtree command
  rerun_proof(RuleName)
extends
  proof_tree_cmds
doc
  'Applies the tactic from a proof file to the current proof node.',
  para,
  'RuleName may be of the form Thy:Name, or just Name, in which case',nl,
  'Thy defaults to the current theory.  In both cases, the tactic is',nl,
  'read from the file called ''Thy/Name.proof'' and the ',
  xref(load_path,global,'parameter:load_path'),nl,
  'is used to determine the directories in which this file may appear.'
implementation
  call_gumtree_tactic(({read_proof(RuleName,_,Tac,_),
                        ip_set(update_annotations, yes)};
                       Tac;
                       {ip_set(update_annotations, no)}))
mode
  rerun_proof(@rulename).


gumtree command
  rerun_proof_file(File)
extends
  proof_tree_cmds
doc
  'Applies the tactic from the given proof file to the current proof node.',
  para,
  'File must be the full path (relative or absolute) to the proof file',nl,
  '(these typically have a .proof extension).'
implementation
  call_gumtree_tactic(({read_proof_file(File,_,Tac,_),
                        ip_set(update_annotations, yes)};
                       Tac;
                       {ip_set(update_annotations, no)}))
mode
  rerun_proof_file(@filename).


%
% For rule applications that refer to a context ID, preceed the
% rule application by context_search to make the generated tactic
% independent of the numbering of context ID's.

add_context_to_rule_tac(_PNode, Thy, Rule, RuleType, RuleTac) :-
    atom(Rule), !,
    RuleTac = (RuleType(Thy:Rule)).
add_context_to_rule_tac(PNode, Thy, Rule, RuleType, RuleTac) :-
    functor(Rule, Name, N),
    rule_query(Thy, Name, FullName, _Concl, _ConclId, Ants, Constraints),
    check_or_error(\+ \+ mode_from_rule(FullName, Rule, _),
                   [q(Rule),' does not match ', q(FullName)]),
    extract_context_add_ids(Ants, AntIDs, []),
    simplify_rule_constraints(Constraints, [], Constraints1), 
    get_context_ids(Constraints1, CIDs),
    functor(ARule, Name, N),
    fix_ante_ids(N, Rule, ARule, FullName, AntIDs),
    functor(VR, Name, N),
    add_context_to_rule_tac_ids(PNode, CIDs, N, Thy, ARule, RuleType, 
                                FullName, VR, RuleTac).

fix_ante_ids(N, _Rule, _ARule, _FullName, _CIDs) :-
    N = 0, !.
fix_ante_ids(N, Rule, ARule, FullName, CIDs) :-
    arg(N, FullName, ':'(Arg,Type)),
    ( member_eq(Arg, CIDs) ->
        check_or_error((Type = ctermid), ['Incompatible type for context.'])
      ;
        arg(N, Rule, RA),
        arg(N, ARule, RA)
    ),
    N1 is N-1,
    fix_ante_ids(N1, Rule, ARule, FullName, CIDs).
    

add_context_to_rule_tac_ids(_, [], _, Thy, Rule, RuleType, 
                            _FullName, _VR, RuleTac) :-
    !,
    RuleTac = RuleType(Thy:Rule).
add_context_to_rule_tac_ids(Node, CIDs, N, Thy, Rule, RuleType, 
                            FN, VR, RT) :-
    add_context_to_rule_tac_ids_args(Node, CIDs, N, Thy, Rule, RuleType, 
                                     FN, VR, RT).

add_context_to_rule_tac_ids_args(_Node, _CIDs, N, Thy, _Rule, RuleType, 
                                 _FN, VR, RT) :-
    N = 0, !,
    RT = RuleType(Thy:VR).
add_context_to_rule_tac_ids_args(Node, CIDs, N, Thy, Rule, RuleType, 
                                 FN, VR, RT) :-
    arg(N, FN, ':'(A,_)),
    N1 is N-1,
    ( in_contexts(A, CIDs, KA) ->
        arg(N, Rule, RA),
        arg(N, VR, VRA),
        context_index(KA, Node, Index, RA),
        RT = (context_index(KA,Index,VRA);RT1),
        add_context_to_rule_tac_ids_args(Node, CIDs, N1, Thy, Rule, 
                                         RuleType, FN, VR, RT1)
    ;
        arg(N, Rule, RA),
        arg(N, VR, RA),
        add_context_to_rule_tac_ids_args(Node, CIDs, N1, Thy, Rule, 
                                         RuleType, FN, VR, RT)
    ).

extract_context_add_ids([], Link, Link).
extract_context_add_ids([seq(_, Contexts, _)|Ants], IDs, Link) :-
    extract_context_add_ids_1(Contexts, IDs, Link1),
    extract_context_add_ids(Ants, Link1, Link).

extract_context_add_ids_1([], Link, Link).
extract_context_add_ids_1([(_ = Delta)|Contexts], IDs, Link) :-
    extract_context_add_ids_2(Delta, IDs, Link1),
    extract_context_add_ids_1(Contexts, Link1, Link).

extract_context_add_ids_2(use(_,_), Link, Link).
extract_context_add_ids_2(filter(_,Delta), IDs, Link) :-
    extract_context_add_ids_2(Delta, IDs, Link).
extract_context_add_ids_2(add(ID, _, Delta), [ID|IDs], Link) :-
    extract_context_add_ids_2(Delta, IDs, Link).
extract_context_add_ids_2(delete(_,_, Delta), IDs, Link) :-
    extract_context_add_ids_2(Delta, IDs, Link).

get_context_ids([], []).
get_context_ids([context_search(K,_,ID,_)|R], [c(K,ID)|IDs]) :-
    !,
    get_context_ids(R, IDs).
get_context_ids([_|R], IDs) :-
    get_context_ids(R, IDs).
    
in_contexts(A, [c(K,ID)|_], KA) :-
    A == ID,!,
    KA = K.
in_contexts(A, [_|Rest], KA) :-
    in_contexts(A, Rest, KA).
