%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: prooftacs.ql,v 1.9 2000/11/15 06:29:50 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  This file defines various useful tactics for use within proofs.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../prover/lib/ergo.h"

proof command
  true
extends
  meta_proofcmds
doc
  'Does nothing.'
mode
  true.


proof command
  postulate,
  postulate(Hyps)
extends
  focus_proofcmds
doc
  'Generates a postulate that will discharge the current focus.',nl,
  'The goal must be true for this command to be valid.',nl,
  'The generated postulate will be:',nl,
  nl,
  '    Cond => Focus  provided  NFIs',nl,
  nl,
  'where Cond is the conjunction of all the hypotheses that are',nl,
  'listed in Hyps and NFIs is all the constraints in the current',nl,
  'proof that relate to the variables in Cond and Focus.',nl,
  nl,
  'Hyps must be a list of hypothesis names or a variable.  If it is',nl,
  'a variable, then all the hypotheses in the current window are used.'
mode
  postulate,
  postulate(?list(hypname)).

postulate :-
    postulate(_).

postulate(_) :-
    is_goal(Goal),
    Goal \== 'ergo.true',
    !,
    handle_error(['postulate command requires goal to be true']).
postulate(Hyps) :-
    var(Hyps),
    !,
    findall(Num, is_hyp(_,Num,_), Hyps),
    postulate(Hyps).
postulate(Hyps) :-
    closed_list(Hyps),
    in_proof(Thy:Name),
    retry_delays,
    get_postulate_hyps(Hyps, Terms),
    is_focus(Focus),
    ( Terms = [] ->
        Thm = Focus
    ;
        list_to_conjunction(Terms, Cond),
        Thm = 'ergo.=>'(Cond, Focus)
    ),
    collect_vars(Thm, Vars),
    collect_constraints(Vars, ObvarNFIs, NFIs, Funnies),
    check_or_error(
        Funnies = [],
        ['Cannot generate postulate because there are ',
         'unresolved constraints: ', q(Funnies)]),
    append(ObvarNFIs, NFIs, AllNFIs),
    list_to_comma(AllNFIs, Body),
    get_postulate_name(Thy, Name, Name2),
    add_postulate(Thy:Name2, [], Thm, Body),
    check_or_error(
        trans(Thy:Name2, t, Hyps),
        ['postulate trans failed: ', q(trans(Thy:Name2, t, Hyps))]).
postulate(Hyps) :-
    handle_error(['postulate command: illegal hyp list: ',q(Hyps)]).

get_postulate_name(Thy, Name0, Name) :-
    atom_codes(Name0, Codes1),
    append(Codes1, "_lemma_", Codes2),
    between(1,1000,Num),
    number_codes(Num, NumCodes),
    append(Codes2, NumCodes, Codes),
    atom_codes(Name,Codes),
    \+ query_theory(rule(Name,_,_,_,_), Thy),
    !.

get_postulate_hyps([], []).
get_postulate_hyps([H|HS], [T|TS]) :-
    is_hyp(H,T),
    !,
    get_postulate_hyps(HS, TS).

