%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: rerun.ql,v 1.33 2001/06/04 01:24:11 itp Exp $

#include "../prover/lib/ergo.h"

?- require_global_tactic_file(showwin).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file defines procedures for reading proof files and
%% attempting to rerun proofs.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% rerun_proof(TheoremName).
%% rerun_proof(TheoremName, ProofName).
%%
%%  Reruns the proof ProofName and places the resulting
%%  theorem (called TheoremName) into the current theory.
%%  
%%  If ProofName is Thy:Name, the proof record will be read from
%%  the theory Thy.  This can be useful for moving theorems up into
%%  an ancestor theory (or down into a descendant theory).
%%
%%  The single argument version does set_save_proof(false), whereas
%%  the two argument form leaves this decision to the caller.
%%
%%  mode rerun_proof(@rulename).
%%  mode rerun_proof(@rulename, @rulename).
%%
rerun_proof(Thm) :-
    set_save_proof(false),
    rerun_proof(Thm, Thm).

rerun_proof(Thm, Proof) :-
    %% Stmt will be instantiated during the rerun.
    add_theorem(Thm, [], Stmt, yes, (demote_var(Stmt),rerun(true,Proof))).


database command
  rerun_all,
  rerun_all(Delete)
extends
  theory_extension_cmds
doc
  'Try to rerun as many proofs as possible in the current theory and its',nl,
  'subtheories.  If Delete is true, then any remaining postulates are',nl,
  'deleted, otherwise they remain in the theory as postulates.',nl,
  'The zero argument version defaults Delete to false.'
mode
  rerun_all,
  rerun_all(@truefalse).

rerun_all :-
    rerun_all(false).

rerun_all(_) :-
    global_state_lookup(current_dir,Thy), Thy \= none,
    (Thy1 = Thy ; query_global(subthy_owner(Thy1, Thy))),
    % prove_postulate deletes and adds postulates so we need to collect a
    % list of them.
    setof(Thy1:Name,
          (Shape,Stmt,Body)^
           query_theory(rule(Name,postulate,Shape,Stmt,Body), Thy1),
          Postulates),
    member(Postulate, Postulates),
    msg(rerun,['Attempting to discharge: ',q(Postulate)]),
    prove_postulate(Postulate, rerun(true), yes,_),
    fail.
rerun_all(false) :-
    writelist(['The following postulates have yet to be proved:', nl]),
    global_state_lookup(current_dir,Thy), Thy \= none,
    (Thy1 = Thy ; query_global(subthy_owner(Thy1, Thy))),
    query_theory(rule(Name,postulate,_Shape,_Stmt,_Body), Thy1),
    writelist(['   ', Thy, ':', Name, nl]),
    fail.
rerun_all(Delete) :-
    Delete == true,
    writelist(['The following unproved postulates have been deleted:', nl]),
    global_state_lookup(current_dir,Thy), Thy \= none,
    (Thy1 = Thy ; query_global(subthy_owner(Thy1, Thy))),
    query_theory(rule(Name,postulate,Shape,Stmt,Body), Thy1),
    writelist(['   ', Thy, ':', Name, nl]),
    retract_theory(rule(Name,postulate,Shape,Stmt,Body), Thy1),
    fail.
rerun_all(_).


proof command
  rerun,
  rerun(SameVars),
  rerun(SameVars, Proof)
extends
  focus_proofcmds
doc
  'This rerun tactic must be invoked from within a proof',nl,
  '(normally from within the outermost level of the proof).',nl,
  'It reruns the stored proof record from the *.pf file.',nl,
  'If Proof is of the form Thy:Name, then the *.pf file is',nl,
  'read from the theory Thy, otherwise it is read from the current theory.',nl,
  'This is one way of (attempting to) rerun a proof in a different',nl,
  'theory from the one in which it was originally proved.',nl,
  nl,
  'The rerun process unifies the current focus with the statement of the',nl,
  'stored theorem (with newvars thawed), so if they are different, the',nl,
  'rerun will simply fail.',nl,
  nl,
  'SameVars must be true or false.  If it is true, then the names of',nl,
  'variables in the proof record are remembered.  This may cause them',nl,
  'to be unified with named variables in the current state.  If it is',nl,
  'false, then fresh names are used.  Consequently, SameVars should',nl,
  'always be false when there is any chance that variables in the focus',nl,
  'of the current proof have names (that is, they have been printed out)',nl,
  'and thus false is taken as the default value for SameVars.',nl,
  nl,
  'The zero and one argument versions do set_save_proof(false), whereas',nl,
  'the two argument version leaves this decision to the caller.'
mode
  rerun,
  rerun(@truefalse),
  rerun(@truefalse, @rulename).

rerun :-
    rerun(false).

rerun(SameVars) :-
    in_proof(Thm),
    set_save_proof(false),
    rerun(SameVars, Thm).

rerun(SameVars, ProofName) :-
    in_proof,
    decode_rule(ProofName, Thy, Name), 
    check_atom(Name),
    ( atom(Thy) ->
        true
    ;
        check_current_theory(Thy)
    ),
    read_proof_record(SameVars, Thy, Name, Ver, _, _Ass, Term, Body, Proof),
    ver_msg(Ver),
    call_predicate(Body),
    rerun_msg(['Unifying focus with proof statement ',thy(Term)]),
    thawing_new_vars(
        ( is_focus(Term),
          incomplete_retry_delays)),
    rerun_msg(['About to accept constraints...']),
    accept_constraints,
    rerun_msg(['Starting to rerun proof steps...']),
    %% locally disable auto hyp splitting (per elves.ql)
    ip_set(split_hyp_disable, disable),
    rerun_proof_steps(Ver, Proof),
    ip_set(split_hyp_disable, _).

%% private ver_msg(Version).
%%
%%  If the proof record to be rerun is stored in an obsolete format,
%%  this predicate will inform the user and ask if they want to update
%%  the stored record (or decide to do so automatically if we are
%%  running in batch mode).
%%
%%  mode ver_msg(@atom).
%%
ver_msg('2.0') :-
    !.
ver_msg(V) :-
    handle_error(['version ',w(V),
                  ' proof records are not supported by rerun']).



%% private check_update.
%%
%%  Asks the user whether they want to overwrite the stored proof record.
%%
%%  mode check_update.
%%
check_update :-
    (question('Update stored proof? ', yes) ->
        set_save_proof(true)
    ;
        set_save_proof(false)
    ).


%% private rerun_proof_steps(Version, Proof).
%%
%%  Select the correct rerun code for Version.
%%
%%  mode rerun_proof_steps(@atom, @proof).
%%
rerun_proof_steps('2.0', FHL) :-
    !,
    rerun20([], FHL, []).




%%%%%%%%%%%%%% Rerun code for proof records of version 2.0 %%%%%%%%%%%


%% private rerun20(+InheritedHyps, +ActionList, +OldNewNums).
%%
%%  Rerun ActionList (focus and hyp proof steps) using OldNewNums
%%  to translate between the stored and actual names of hypotheses.
%%  OldNewNums is a list of hyp(OldNum, NewNum, Term) terms.
%%  (In fact, the Term component of these translation tuples
%%  is not needed for the translation.  It is only used for printing
%%  out more understandable error messages.)
%%
%%  InheritedHyps is a list of the (stored) numbers of hypotheses
%%  that were inherited (without transformation) into this window.
%%
rerun20(IHyps, L, OldNewNums) :-
    %%  Note that OldNewNums is actually the hyp. translation table
    %%  for the parent window, which may contain more hypotheses
    %%  than are needed for this window.  However, it does no harm
    %%  to leave the extra ones in there, since they are always
    %%  accessed via number (rather than by matching terms) and
    %%  the (stored) numbers of parent hypotheses will all be
    %%  smaller than the (stored) numbers of hypotheses generated
    %%  in the current window.
    rerun_ihyp20(IHyps, OldNewNums),
    rerun_lines20(L, OldNewNums, _).


%% private rerun_lines20(Lines, OldNewNums0, OldNewNums).
%%
%%  This executes the commands in Lines, thus constructing
%%  the current proof.  Whenever it adds new hypotheses, it adds
%%  translation information for those hypotheses to OldNewNums0,
%%  to give OldNewNums.
%%
%%  mode rerun_lines20(+list(compound), +list(compound), -list(compound)).
%%
rerun_lines20([], HL, HL).
rerun_lines20([L|Lines], HL, HL2) :-
    rerun_line20(L, HL, HL1),
    !,
    rerun_lines20(Lines, HL1, HL2).
rerun_lines20([L|_], HL, _) :-
    less_verbose(-4, rerun_msg(['In the above proof state'],HL)),
    handle_error(['failed rerun command: ',max_depth(20,[q(L)]),nl,
                  'current hyptransfunc: ',max_depth(20,[q(HL)])]).


%% private rerun_msg(Msg).
%% private rerun_msg(Msg,HypTransList).
%%
%%  Prints out detailed information about the current state of the
%%  proof, then prints Msg and HypTransList (if non-empty).
%%
%%  mode rerun_msg(@writelist, @list(compound)).
%%
rerun_msg(Msg) :-
    rerun_msg(Msg,[]).

rerun_msg(Msg,HL) :-
    get_verbosity(V),
    V >= 2,
    ( V >= 3 ->
        nl,
        show_window,
        nl
    ;
        true
    ),
    ( V >= 4 ->
        show_proofvar_constraints
    ;
        true
    ),
    %% Build a prettier form of HL (just the hyp. numbers that have changed).
    findall((Old,New),
            (member(hyp(Old,New,_), HL), Old =\= New),
            HLNums),
    append(Msg, [nl,'    with hypothesis translation: ',w(HLNums)], Msg2),
    msg(rerun, 2, Msg2),
    fail.
rerun_msg(_,_).


%% private rerun_line20(+ActionList, OldNewNums0, OldNewNums).
%%
%%  Rerun a single line and extend the hypothesis number translation
%%  table OldNewNums0 to reflect any new hypotheses created by
%%  the line.
%%
%%  mode rerun_line20(+list(compound), +list(compound), -list(compound)).
%%
rerun_line20(Cmd, HL, _) :-
    rerun_msg(['Next cmd: ',max_depth(5,[q(Cmd)])],HL),
    fail.
rerun_line20(nil, HL, HL).
rerun_line20(tr(R,D,G,A), HL, HL) :-
    fix_TR_BUG(A),
    translate_rulenames([R|A], [Rule|Ant], HL),
    trans(Rule, D, Ant, G).
rerun_line20(op(P,I,J), HL, HL) :-
    open_pos(P, [], rerun20(I, J, HL)).
rerun_line20(wr(wf(Id,Msg,Tac),_,_,J), HL, HL2) :-
    wrap_func(rerun_lines20(J, HL, HL2), Id, Msg, Tac).
rerun_line20(wr(wa(Id,Msg,ArgNum),_,_,J), HL, HL2) :-
    wrap_arg(rerun_lines20(J, HL, HL2), Id, Msg, ArgNum).
rerun_line20(oh(New,Old,I,J), HL, [hyp(New,NewNow,G)|HL]) :-
    translate_rulenames([Old], [OldNow], HL),
    open_hyp(OldNow, rerun20(I, J, HL), NewNow, G).
rerun_line20(cj(New,G,tr(R,D,_,A)), HL, [hyp(New,NewNow,G)|HL]) :-
    fix_TR_BUG(A),
    conjecture(G,NewNow),
    translate_rulenames([R|A], [Rule|Ant], HL),
    d_trans(NewNow, Rule, D, Ant).
rerun_line20(cj(New,G,I,J), HL, [hyp(New,NewNow,G)|HL]) :-
    conjecture(G,NewNow),
    discharge_conjecture(NewNow, [rerun20(I, J, HL)]).
rerun_line20(sl(Old,New), HL, [hyp(New,NewNow,G)|HL]) :-
    translate_rulenames([Old], [OldNow], HL),
    split_left(OldNow, G, NewNow).
rerun_line20(sr(Old,New), HL, [hyp(New,NewNow,G)|HL]) :-
    translate_rulenames([Old], [OldNow], HL),
    split_right(OldNow, G, NewNow).
rerun_line20(ht(In,Rule,Dir,G,Ant,New), HL, [hyp(New,NewNow,G)|HL]) :-
    translate_rulenames([In,Rule|Ant], [In2,Rule2|Ant2], HL),
    h_trans(In2,Rule2,Dir,Ant2,G,NewNow).
rerun_line20(new(New,G), HL, [hyp(New,NewNow,G)|HL]) :-
    is_hyp(NewNow,G),
%% TO DO: A better way of doing the following might be to check that
%% the justification of NewNow is indeed 'new'. (However, there is no
%% kosher way for tactics to do such a check at the moment).
%% The same fix is needed in the next case of rerun_line20 too.
    %% Now check that G has really found us a hypothesis that has been
    %% newly introduced by an opening rule, *not* one that this rerun
    %% process has already generated itself.
    \+ member(hyp(_,NewNow,_), HL).
rerun_line20(tr(_Old,New,G), HL, [hyp(New,NewNow,G)|HL]) :-
    is_hyp(NewNow,G),
%% TO DO: The fix in the last case of rerun_line20 must be done here too.
    \+ member(hyp(_,NewNow,_), HL).

%% TO DO: Overcome the following temporary fix applied by fix_TR_BUG.
%% Some old proofs contain tr(_,_,_,A) where A is a variable, but should be [].
%% Ray's last comment: Probably no longer needed.
fix_TR_BUG(A) :-
    var(A),
    !,
    msg(rerun, ['WARNING: illegal condlist found in proof!',nl,
                'Please inform ergo-support@it.uq.edu.au about this.']),
    A = [].
fix_TR_BUG(_).


%% private rerun_ihyp20(InheritedHyps, OldNewNums).
%%
%%  This checks that all hyps whose numbers appear in
%%  InheritedHyps have indeed been inherited into the current window.
%%
%%  mode rerun_ihyp20(@list(integer), @list(compound)).
%%
rerun_ihyp20([], _).
rerun_ihyp20([I|IHL], HL) :-
    translate_rulename(I, hyp:INow, HL),
    is_hyp(INow,_),
    !,
    rerun_ihyp20(IHL, HL).
rerun_ihyp20([I|_], PHL) :-
    less_verbose(-4, rerun_msg(['In the above proof state'],PHL)),
    handle_error(['failed to inherit hypothesis: ',q(I),nl,
                  'current hyptransfunc: ',max_depth(20,q(PHL))]).



%% private translate_rulenames(Rules0, Rules, OldNewNums).
%%
%%  Translates rulenames (esp. hypothesis numbers) that appear in the
%%  proof into the corresponding name/number in the current proof.
%%
translate_rulenames([], [], _Lookup).
translate_rulenames([A|As], [B|Bs], Lookup) :-
    translate_rulename(A, B, Lookup),
    !,
    translate_rulenames(As, Bs, Lookup).


%% private translate_rulename(Rule0, Rule, Lookup).
%%
translate_rulename(hyp:HypNum, hyp:HypNumNow, Lookup) :-
    integer(HypNum),
    !,
    member(hyp(HypNum,HypNumNow,_), Lookup),
    !.
translate_rulename(HypNum, hyp:HypNumNow, Lookup) :-
    integer(HypNum),
    !,
    member(hyp(HypNum,HypNumNow,_), Lookup),
    !.
translate_rulename(N, N, _).
