%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: rulebrowse5.ql,v 1.8 2002/06/27 05:10:27 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  This file defines tactics for rule browsing (and application).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%
%% Transform the current open node using a rule satisfying Query.
%%
qiuse(Q, Id, Child) :-
    query_init(quse, [Id], []),
    query_process(Q),
    query_execute(Child).

qiuse(Q,Id,Output,Child) :-
    query_init(quse, [Id], []),
    query_process([prem(Output)]),
    query_process(Q),
    query_execute(Child).

quse(Q, Id, Child) :-
    query_init(quse, [Id], []),
    query_process([frozen(true)]),
    query_process(Q),
    query_execute(Child).

quse(Q,Id,Output,Child) :-
    query_init(quse, [Id], []),
    query_process([frozen(true)]),
    query_process([prem(Output)]),
    query_process(Q),
    query_execute(Child).

%% 
%% Print out all rules matching Query.
%%
rulebrowse(Q, X, XT) :-
    query_init(rulebrowse, X, XT),
    query_process([frozen(true)]),
    query_process(Q),
    less_verbose(rule2).
rulebrowse(_, _, _).

rule2 :-
    query_choose_rule(Thy,Name,FullName, Kind, Concl, ConclId, Ants, Constraints),
    (Kind = defn -> 
        show_defnbrowse(Name, Thy, Concl, Constraints)
    ;
        show_rulebrowse(Thy,Name,FullName, Kind, Concl, ConclId, Ants, Constraints) 
    ),
    fail.

show_defnbrowse(Name, Thy, Stmt, Body) :-
    writelist(['definition ', q(Thy),': ',q(Name), ' === ']),
    show_topic_rule(Stmt,[],['    ']),
    show_topic_body(Body, ['    ']).

show_rulebrowse(Thy,_,FullName, Kind, Concl, ConclId, Ants, Constraints) :-
    set_var_name(ConclId, 'C'),
    show_rule_name_sequents(Ants),
    writelist([w(Kind),' ',q(Thy),':',w(FullName), ' ===', nl]),
    simplify_rule_constraints(Constraints, [], Constraints1),
    show_rule_ants(Ants, ConclId),
    show_rule_sequent(ConclId, _, [], Concl),
    show_rule_constraints(Constraints1),
    writelist(['.',nl]).

rule3 :- query_check_matching, !.


%%
%%  Prints out all instantiated rules that match Query AND the current
%%  open proof node.
%%
qirule(Q, X, XT) :-
    query_init(qirule, X, XT),
    query_process([frozen(false)]),
    query_process(Q),
    less_verbose(rule2).
qirule(_,_, _).


%% private query_init(Caller).
%%
%%  Sets up a new query.  Caller is the name of the calling
%%  tactic, for use in output messages.
%%
%%  mode query_init(@nonvar).
%%
query_init(Caller, X, XT) :-
    ip_set(query_input_id, X - XT),
    ip_set(query_caller,Caller),
    ip_set(query_thys,[*]),
    ip_set(query_name,_),
    ip_set(query_subnames,[]),
    ip_set(query_subterms,[]),
    ip_set(query_frozen,false),
    ip_set(query_iu_depth,3),
    ip_set(query_premises, _),
    ip_set(query_premlist, []),
    ip_set(query_context, []),
    ip_set(query_conclusion, _).

%% private query_process(Query).
%%
%%  Preprocesses Query, ready for execution by query_execute.
%%  query_process may be called several times to build up
%%  a more complex query.  Some constraints (eg. thys/1) override
%%  similar but earlier constraints, while others (eg. subterm/1)
%%  are conjoined.  Thus, default constraints should be passed
%%  to query_process first.
%%
%%  Query may be a list of constraints, or a single constraint.
%%
%%  mode query_process(?query).
%%
query_process([]) :-
    !.
query_process(Chars) :-
    %% Special case for "..." constraints.
    Chars = [C|_],
    integer(C),
    !,
    do_constraint(Chars).
query_process([C|CS]) :-
    !,
    do_constraint(C),
    query_process(CS).
query_process(and(Q1,Q2)) :-
    !,
    query_process(Q1),
    query_process(Q2).
query_process(C) :-
    do_constraint(C).


%% query_execute(Children).
%%
%%  Completes the current itrans operation using the constraints
%%  that have been recorded by query_process.
%%  It assumes that the itrans_start operation has been done.
%%  It chooses the rule, conditionality and direction, then
%%  discharges all the conditions of the rule and calls itrans_finish.
%%
%%  mode query_execute(?integer).
%%
query_execute(Children) :-
    query_choose_rule(Thy,Name, _,_, Concl, _, Seqs, Body),
    ip_lookup(query_context, QC),
    ( QC = []
      ->
        Contexts = []
    ;
        get_contexts(Body, Contexts)
    ),
    query_rule_satisfy(Concl, Seqs, Contexts),
    ip_lookup(query_input_id, [Id] - []),
    proof_step(Thy, Name, Id, Children).

%%  
%% query_choose_rule(Thy,Name,FullName, Kind, Concl, Id, Seqs, Body).
%%
%%  Does the itrans_rule_name part of the current itrans operation,
%%  according to the constraints that have been built up in the
%%  query_* implicit parameters by query_process.
%%
%%  mode query_choose_rule(?theory,?term,?term,?term,?term,?term,?term).
%%
query_choose_rule(Thy,Name, FullName, Kind, Concl, Id, Seqs, Body) :-
    ip_lookup(query_thys, Thys),
    ip_lookup(query_name, Name),
    ip_lookup(query_subnames, Subnames),
    ip_lookup(query_subterms, Subterms),
    ip_lookup(query_frozen, Frozen),
    query_choose_theory(Thys, Thy),
    browse_find_rule(Kind, Thy, Name, FullName, Concl, Id, Seqs, Body),
    ip_lookup(query_context, QC),
    ( QC = []
      ->
        Contexts = []
    ;
        get_contexts(Body, Contexts)
    ),
    ( Frozen = true ->
       collect_vars([Concl, Seqs, Contexts], RVars)
    ;
       true
    ),
    query_rule_satisfy(Concl, Seqs, Contexts),
    query_check_subnames(Thy, Name, Subnames),
    query_check_subterms(Subterms, Concl, Seqs),
    ip_lookup(query_iu_depth, IUD),
    incomplete_retry_delays(IUD),
    ( Frozen = true ->
       collect_vars([Concl, Seqs, Contexts], NewRVars),
       map('==', [RVars, NewRVars])
    ;
       true
    ).

get_contexts(context_search(C, _, _, Term), R) :-
    !,
    R = [context(C, Term)].
get_contexts((Body1, Body2), Result) :-
    !,
    get_contexts(Body1, R1),
    get_contexts(Body2, R2),
    append(R1, R2, Result).
get_contexts(Body, []).

freeze_vars_in_Con(Concl) :-
    freeze_vars(Concl).
freeze_vars_in_Seqs([]).
freeze_vars_in_Seqs([S|Seqs]) :-
    S = seq(_,_,Term),
    freeze_vars(Term),
    freeze_vars_in_Seqs(Seqs).

browse_find_rule(Kind, Thy, Name, FullName, Concl, Id, Seqs, Body) :-
    rule_query(Thy, Name, FullName, Concl, Id, Seqs, Body),
    get_rule_kind(Thy, Name, Kind).
browse_find_rule(defn, Thy, Name, Name, Stmt, _, [], Body) :-
    query_theory(rule(Name,defn,_Shape,Stmt,Body), Thy).


query_rule_satisfy(Concl, Seqs, Body) :-
    ip_lookup(query_conclusion, In),
    Concl = In,
    ip_lookup(query_context, Context),
    query_check_context(Context, Body),
    ip_lookup(query_premises, N),
    length(Seqs, N),
    ip_lookup(query_premlist, AL),
    query_check_premlist(AL, Seqs).

query_check_premlist([], Seqs).
query_check_premlist([A|AA], Seqs) :-
    query_check_seqs(Seqs, A),
    query_check_premlist(AA, Seqs).

query_check_seqs([seq(_,_,A)|SS], A).
query_check_seqs([_|SS], A) :-
    query_check_seqs(SS, A).

query_check_context([], _).
query_check_context([C|CS], Body) :-
    member(C, Body),
    query_check_context(CS, Body).

query_check_constraints(context_search(C, _, _, Term), C, Term).
query_check_constraints((_, context_search(C, _, _, Term)), C, Term).
query_check_constraints((Body,_), C, Term) :-
    query_check_constraints(Body, C, Term).

query_check_subnames(_, Name, Subnames) :-
    query_check_subnames2(Subnames, Name).

query_check_subnames2([], _).
query_check_subnames2([S|SS], Name) :-
    atom_codes(Sub,S),
    atom_search(Name,1,Sub,_),
    !,
    query_check_subnames2(SS, Name).

query_check_subterms([], _, _).
query_check_subterms([S|SS], Stmt1, Stmt2) :-
    (subterm_path(Stmt1,_,S);subterm_path(Stmt2,_,S)),
    !,
    query_check_subterms(SS, Stmt1, Stmt2).

%% TO DO: Improve query_choose_theory so it doesn't return same theory twice.
%% See tactics/usefwd.ql for a similar task.

%% TO DO: Extend query_choose_theory to handle conj?
%% See tactics/usefwd.ql for a similar task.

%% query_choose_theory(ThySpecList, Thy).
%%
%%  mode query_choose_theory(@list(ground), ?atom).
%%
query_choose_theory([T|_], Thy) :-
    query_choose_theory2(T, Thy).
query_choose_theory([_|TS], Thy) :-
    query_choose_theory(TS, Thy).

query_choose_theory2('*', Thy) :-
    !,
    ancestor(Thy).
query_choose_theory2(upto(Max), Thy) :-
    !,
    check_or_error(
        ancestor(Max),
        ['unknown theory: ',q(Max)]),
    ancestor(Thy),
    ancestor_of(Thy,Max).
query_choose_theory2(curr, Thy) :-
    !,
    global_state_lookup(current_dir,Thy),
    Thy \= none.
query_choose_theory2(Thy, Thy).


query_language header
  'Query Language for Rules'
extends
  concepts
doc
  'Several commands take a `query'' argument, which specifies a rule of',nl,
  'a set of rules.  For example, the `use'' command attempts to find a',nl,
  'rule that satisfies the query and then applies it.  The',nl,
  '`rule'' command displays all rules that satisfy the query.',
  para,
  'A query may be a single constraint or a set of constraints (in a list',nl,
  'or combined with and/2) that specifies a set of rules..',nl,
  'For example, if the query is `imp_reflex'', then all ancestor',nl,
  'theories will be searched for a rule called ''imp_reflex''.',nl,
  'If the query is `[jprop,iprop]:_'', then all rules in the jprop and iprop',nl,
  'theories will be tried.',
  para,
  'The following constraints can be used within queries:',nl,
  begin(list),
  item('Thys:Name -- '), 'short for thys(Thys) and name(Name).',
  item('Name -- '), 'short for name(Name).',
  item('thys(Thys) -- '), 'Thys specifies the ',
                    xref('theory search list',global,theory_search_list),nl,
  'for the rule.',
  item('name(Atom) -- '), 'the name of the rule must be Atom.',
  item('name("Str") -- '), 'the name of the rule must contain Str.',
  item('frozen(Flag) -- '), 'Flag=true means freeze all variables in the',nl,
  'rule, thus preventing instantiations.',
  item('iu_depth(Number) -- '), 'Set the incomplete unification depth (default 3).',
  item('premises(Number) -- '), 'Constrains how many premises the rule produces.',
  item('concl(Term) -- '), 'the conclusion of the rule must unify with Term.',
  item('goal -- '), 'the conclusion of the rule must unify with the goal.',
  item('context(C, T) -- '), 'T must unify with a C context entry in the conclusion',nl,
  'of the rule.',
  item('prem(Prem) -- '), 'Prem must unify with some premise of the rule.',
  item('subterm(Term) -- '), 'some subterm of the rule must unify with Term.',
  end(list).


theory_search_list header
  'Theory Search List'
extends
  concepts
doc
  'One component of a ',xref(query,global,query_language),' for a rule can',nl,
  'be a `theory search list''.  A theory search list specifies a list of',nl,
  'theories which should be searched for a rule.',
  para,
  'A theory search list may be a single theory name, or a list of',nl,
  'the following terms, which specify alternative theories to search.',nl,
  begin(list),
  item('curr -- '),                 'search the current theory.',
  item('Thy -- '),                  'search the theory called Thy.',
  item('* -- '),                    'search all ancestor theories.',
  item('upto(Thy) -- '),            'search between the current theory and Thy.',
  item('i(N) -- '),                 'search the theories associated with the symbols',nl,
  'of the input term (N levels deep).',nl,
  'For example, if the input term is',br,
  '''TA.a''(''TB.b''(''TC.c'',X),''TD.d''),',br,
  'then',br,
  'i(1) gives [''TA''],',br,
  'i(2) gives [''TA'',''TB'',''TD''],',br,
  'i(3) gives [''TA'',''TB'',''TC'',''TD''].',
  item('o(N) -- '),                 'similar to i(N), but using the output term.',nl,
  'currently being discharged (if any).',nl,
  end(list),nl,
  'NB.  the i(N) o(N) specs are not implemented yet.'.


%% private do_constraint(Constraint).
%%
%%  This is a table of all the constraints that can appear
%%  in a query list.  Each Constraint is processed by 
%%  setting up some query_* implicit parameters for later processing.
%%
%%  mode do_constraint(?nonvar).
%%

%%
%% Constraints on the whole statement of the rule.
%%
do_constraint(iu_depth(T)) :-
    integer(T),
    !,
    ip_set(query_iu_depth, T).
do_constraint(subterm(T)) :-
    !,
    ip_lookup(query_subterms, L),
    ip_set(query_subterms, [T|L]).
do_constraint(frozen(TF)) :-
    member(TF, [true,false]),
    !,
    ip_set(query_frozen, TF).
%%
%% Constraints on the conclusion of the rule.
%%
do_constraint(concl(Term)) :-
    !,
    ip_set(query_conclusion, Term).
do_constraint(context(C, Term)) :-
    !,
    ip_lookup(query_context, AL),
    ip_set(query_context, [context(C, Term)|AL]).
%%
%% Constraints on the premises of the rule.
%%
do_constraint(premises(N)) :-
    !,
    ip_set(query_premises, N).
do_constraint(prem(T)) :-
    !,
    ip_lookup(query_premlist, AL),
    ip_set(query_premlist, [T|AL]).
%%
%% Constraints on the name and theory of the rule.
%%
do_constraint(G) :-
    G == goal,
    !,
    ip_lookup(query_input_id, X - _),
    nonvar(X),
    X = [Id|_],
    proof_node_term(Id,Term),
    ip_set(query_conclusion, Term).
do_constraint(Thys:Name) :-
    atom(Thys),
    Thys = hyp,
    !,
    ip_set(query_thys, [hyp]),
    ip_lookup(query_name, Name).
do_constraint(Thys:Name) :-
    %% Thys \== hyp.
    once(do_constraint(thys(Thys))),
    do_constraint(name(Name)),
    !.
do_constraint(thys(L)) :-
    var(L),
    !,
    ip_set(query_thys, [hyp,*]).
do_constraint(thys(L)) :-
    atom(L),
    !,
    ip_set(query_thys, [L]).
do_constraint(thys(L)) :-
    list(L),
    !,
    ip_set(query_thys, L).
do_constraint(name(S)) :-
    list(S),
    !,
    ip_lookup(query_subnames, L),
    ip_set(query_subnames, [S|L]).
do_constraint(name(S)) :-
    var(S),
    %% ignore, as it does not help to narrow the search.
    !.
do_constraint(name(S)) :-
    atom(S),  %% a complete rulename
    !,
    ip_lookup(query_name, S).
do_constraint(name(S)) :-
    integer(S),  %% a complete rulename
    !,
    ip_lookup(query_name, S).
do_constraint(S) :-
    list(S),
    is_chars(S),
    !,
    ip_lookup(query_subnames, L),
    ip_set(query_subnames, [S|L]).
do_constraint(S) :-
    %% This clause must come *after* the clauses for f,r,t,rt,c,u etc.
    %% We assume that any remaining atoms are rule names.
    atom(S),
    !,
    ip_lookup(query_name, S).
do_constraint(C) :-
    ip_lookup(query_caller, Caller),
    handle_error(['bad constraint ',q(C),' in ',q(Caller),' query.']).


