%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: texdoc4.ql,v 1.1 2002/02/11 04:34:12 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file provides commands for generating LaTeX manuals
%% from doc_node's.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

?- ( global_tactic_fact(tex_doc_done, 1) ->
       true
   ;
       new_global_fact(tex_doc_done, 1, [save(false)])
   ).

database command
  tex_doc_clear_memory
extends
  misc_cmds
doc
  'Clears all the tex_doc_done/1 facts, which record which doc_nodes',nl,
  'have been printed out by the texdoc procedures.',nl,
  'This is done automatically whenever the texdoc packages is loaded.'
mode
  tex_doc_clear_memory.

tex_doc_clear_memory :-
    tactic_retractall_global(tex_doc_done(_)).

?- tex_doc_clear_memory.


database command
  tex_doc_rest(File)
doc
  'Writes out all doc_nodes that have not already been written out',nl,
  'into File in LaTeX format.',nl,
  'More precisely, it writes out all doc_nodes whose names do not',nl,
  'appear as a global tex_doc_done/1 fact.',
  'File should be a quoted atom.  For example: ''../doc/refman.tex''.',nl
mode
  tex_doc_rest(@filename).

tex_doc_rest(File) :-
    open(File, write, Out),
    add_handler(write(Strm, _Depth, Term),
                latex_write(Term, Strm)),
    unwind_protect(tex_doc_rest2(Out),
                   ( close(Out),
                     delete_handlers(write(_,_,_),latex_write(_,_)))
                  ).

tex_doc_rest2(Out) :-
    setof(node(Thy, Name, Hdr, Doc, Cmds, Extns),
          Cmds ^
            ( (Thy = global ; ancestor(Thy)),
              query_theory(doc_node(Name,Hdr,Doc,Cmds), Thy),
              \+ query_global(tex_doc_done(Name)),
              get_doc_extns(Name, Extns)),
          Nodes),
    tex_doc_rest3(Nodes, Out).

tex_doc_rest3([], _).
tex_doc_rest3([node(Thy,Name,Hdr,Doc,Cmds,Extns)|Nodes], Out) :-
    msg(tex_doc_rest, ['\\subsection ',q(Name)]),
    tex_doc_label(Name, _, Label),
    tex_doc_minihdr(Hdr,Doc,MiniHdr),
    strip_leading_newlines(Hdr, Hdr2),
    map(nl_to_latex_nl, [Hdr2,Hdr3]),
    writelist(Out,
              [nl,
               nl,
               '%% doc_node name is ',q(Name),nl,
               '\\subsection[{',w(MiniHdr),'}]{',max_depth(0,Hdr3),'}',nl,
               '\\label{',chars(Label),'}',nl,
               '\\index{',w(MiniHdr),'}',nl,
               nl]),
    writelist(Out, Doc),
    writelist(Out, [nl,nl]),
    writelist_modes_theory(Thy, Cmds, Out),
    (
        member(E, Extns),
        writelist(Out, E),
        writelist(Out, ['\\\\ \\relax',nl]),
        fail
    ;
        true
    ),
    !,
    tex_doc_rest3(Nodes, Out).


database command
  tex_doc_tree(File, Root)
extends
  misc_cmds
doc
  'Writes a tree of documentation nodes into File in LaTeX format.',nl,
  'File should be a quoted atom.  For example: ''../doc/refman.tex''.',nl,
  'Root must be the name of a global doc_node, which will be the',nl,
  'root of the generated documentation tree.',nl,
  nl,
  'The subtrees of each node of the tree are found by following',nl,
  'cross references that appear in the doc_extension facts.',nl,
  ' requires code.sty. '
mode
  tex_doc_tree(@atom, @docnode).

tex_doc_tree(File, Root) :-
    open(File, write, Out),
    add_handler(write(Strm, _Depth, Term),
                latex_write(Term,Strm)),
    unwind_protect(tex_doc_tree2(['\\section',
                                        '\\subsection',
                                        '\\subsubsection',
                                        '\\paragraph'
                                 ],
                                       1,Out,Root),
                   ( close(Out),
                     delete_handlers(write(_,_,_),latex_write(_,_)))
                  ).

database command
  tex_doc_trees(File, Roots)
extends
  misc_cmds
doc
  'Writes a forest of documentation nodes into File in LaTeX format.',nl,
  'File should be a quoted atom.  For example: ''../doc/refman.tex''.',nl,
  'Roots must be a list of the names of global doc_nodes, which will be the',nl,
  'roots of the generated documentation trees.',nl,
  nl,
  'The subtrees of each forest are found by following',nl,
  'cross references that appear in the doc_extension facts.'
mode
  tex_doc_trees(@atom, @list(docnode)).

tex_doc_trees(File, Roots) :-
    open(File, write, Out),
    add_handler(write(Strm, _Depth, Term),
                latex_write(Term,Strm)),
    unwind_protect(tex_doc_tree2(['\\section',
                                        '\\subsection',
                                        '\\subsubsection',
                                        '\\paragraph'
                                 ],
                                       1,Out,Roots),
                   ( close(Out),
                     delete_handlers(write(_,_,_),latex_write(_,_)))
                  ).

latex_write(xref([], Thy, Name), Strm) :-
    !,
    msg(latex_write, ['Bad [] xref: ',q(Thy),':',q(Name)]),
    writelist(Strm, Name).
latex_write(xref(['{G}'], global, 'gumtree:{}/1'), Strm) :-
    !,
    writelist(Strm, '\\{G\\}'),
    writelist(Strm, [' (p\\pageref{gumtree:lbracerbrace/1})']).
latex_write(xref(WL, Thy, Name), Strm) :-
    atom(WL),
    !,
    writelist(Strm, WL),
    tex_doc_label(Name, Thy, Label),
    writelist(Strm, [' (p\\pageref{',chars(Label),'})']).
latex_write(xref(WL, Thy, Name), Strm) :-
    list(WL),
    member(W, WL),
    atom(W),
    \+ W = nl,
    !,
    writelist(Strm, W),
    tex_doc_label(Name, Thy, Label),
    writelist(Strm, [' (p\\pageref{',chars(Label),'})']).
latex_write(xref(WL, Thy, Name), Strm) :-
    msg(latex_write, ['bad xref: ',q(xref(WL,Thy,Name))]),
    writelist(Strm, [w(Name),' (????)']).
latex_write(begin(list), Strm) :-
    writelist(Strm, ['\\begin{description}']).
latex_write(end(list), Strm) :-
    writelist(Strm, [nl, '\\end{description}', nl]).
latex_write(begin(verb), Strm) :-
    writelist(Strm, ['\\begin{bigcode}{c}',nl]).
latex_write(end(verb), Strm) :-
    writelist(Strm, [nl, '\\end{bigcode}', nl]).
latex_write(item(Tag), Strm) :-
    list(Tag),
    !,
    writelist(Strm, [nl, '\\item[{']),
    writelist(Strm, Tag),
    writelist(Strm, ['}] ']).
latex_write(item(Tag), Strm) :-
    atomic(Tag),
    writelist(Strm, [nl, '\\item[{', w(Tag), '}] ']).
latex_write(emph(WL), Strm) :-
    write(Strm, '{\\em '),
    writelist(Strm, WL),
    write(Strm, '\\/}').    

%% tex_doc_tree2(SectionNames, Depth, OutStream, DocNode).
%%
%%  Writes out DocNode as a LaTeX section, using the head of
%%  SectionNames as the command to generate that section, then does
%%  all xref's in DocNode extensions recursively, using the
%%  tail of SectionNames.
%%  DocNode may also be a list of docnodes; each is written in turn.
%%
%%  mode tex_doc_tree2(@list(atom), @integer, @stream, @docnode).
%%  mode tex_doc_tree2(@list(atom), @integer, @stream, @list(docnode)).
%%
tex_doc_tree2(_, _, _, []) :-
    !.
tex_doc_tree2([Sect|SubSects], Depth, Out, [Node|Nodes]) :-
    !,
    tex_doc_tree2([Sect|SubSects], Depth, Out, Node),
    tex_doc_tree2([Sect|SubSects], Depth, Out, Nodes).
tex_doc_tree2([Sect|SubSects], Depth, Out, Node) :-
    msg(tex_doc_tree,[w(Depth),': ',w(Sect),' ',q(Node)]),
    ( get_doc_node(Node, Hdr, Doc, Cmds, Thy) ->
        true
    ;
        writelist(['WARNING: Could not find doc_node called: ',q(Node),nl]),
        Hdr = [Node],
        Doc = ['No such node.'],
        Thy = global
    ),
    get_doc_extns(Node, Extns),
    tactic_assert_global(tex_doc_done(Node)),
    writelist_list_xrefs(Extns, XRefs, []),
    tex_doc_label(Node, _, Label),
    tex_doc_minihdr(Hdr,Doc,MiniHdr),
    strip_leading_newlines(Hdr, Hdr2),
    map(nl_to_latex_nl, [Hdr2,Hdr3]),
    writelist(Out,
              [nl,
               nl,
               '%% ',q(Depth),':  doc_node name is ',q(Node),nl,
               w(Sect),'[{',w(MiniHdr),'}]{',max_depth(0,Hdr3),'}',nl,
               '\\label{',chars(Label),'}',nl,
               '\\index{',w(MiniHdr),'}',nl,
               nl]),
    writelist(Out, Doc),
    writelist(Out, [nl,nl]),
    writelist_modes_theory(Thy, Cmds, Out),
%%    ( Cmds = [] ->
%%        true
%%    ;
%%        writelist_modes(Out, Cmds)
%%    ),
%%    ( Thy = global ->
%%        true
%%    ;
%%       writelist(Out, [nl,'This command is specific to theory ',q(Thy),nl,nl])
%%    ),
    (
        member(E, Extns),
        writelist(Out, E),
        writelist(Out, ['\\\\ \\relax',nl]),
        fail
    ;
        true
    ),
    Depth1 is Depth + 1,
    tex_doc_subtrees(XRefs, SubSects, Depth1, Out).
tex_doc_tree2([], Depth, _, Node) :-
    writelist(['WARNING: tree is too deep.  Depth=',w(Depth),
               '  doc_node=',q(Node),nl,'Ignoring that subtree.',nl]).


tex_doc_subtrees([], _, _, _).
tex_doc_subtrees([Node|Nodes], Sects, Depth, Out) :-
    tex_doc_tree2(Sects, Depth, Out, Node),
    tex_doc_subtrees(Nodes, Sects, Depth, Out).

nl_to_latex_nl(nl, ' \\\\ \\relax ') :- !.
nl_to_latex_nl(X, X).


database command
  tex_doc_cmdtables(File, Tables)
extends
  misc_cmds
doc
  'Writes all the commands in Tables out into File, in LaTeX format.',nl,
  'Command show(command_table) shows possible values for Tables.'

mode
  tex_doc_cmdtables(@atom, @atom),
  tex_doc_cmdtables(@atom, @list(atom)).

tex_doc_cmdtables(File,Tables) :-
    open(File, write, Out),
    add_handler(write(Strm, _Depth, Term),
                latex_write(Term, Strm)),
    unwind_protect(tex_doc_cmdtables2(Tables, Out),
                   ( close(Out),
                     delete_handlers(write(_,_,_),latex_write(_,_)))
                  ).

tex_doc_cmdtables2(Tables, Out) :-
    setof(C,
          T ^ (elem(T,Tables),get_cmd_and_doc(T,C)),
          Cmds),
    member(cmd(Hdr,Doc,Mode), Cmds),
    tex_doc_minihdr(Hdr,Doc,MiniHdr),
    strip_leading_newlines(Hdr, Hdr2),
    map(nl_to_latex_nl, [Hdr2,Hdr3]),
    writelist(Out, [nl,nl,'\\subsection[',w(MiniHdr),']{',
                    max_depth(0,Hdr3),'}',nl,nl]),
    writelist(Out, Doc),
    writelist(Out, [nl,nl,'\\MODE{',q(Mode),'}',nl,nl]),
    fail.
tex_doc_cmdtables2(_, _).

get_cmd_and_doc(T, cmd(Hdr,Doc,Mode)) :-
    Cmd =.. [T,_Call,Mode,_Impl,DocNode],
    (
        theory_fact(T,4),
        query_ancestors(Cmd)
    ;
        global_fact(T,4),
        query_global(Cmd)
    ),
    get_doc_node(DocNode, Hdr, Doc, _, _),
    tactic_assert_global(tex_doc_done(DocNode)).


%% private tex_doc_minihdr(Hdr, Doc, MiniHdr).
%%
%%  MiniHdr is a one-line short form of Hdr (or Doc), suitable for
%%  putting into a table of contents.
%%
tex_doc_minihdr(Hdr,_,MiniHdr) :-
    tex_doc_minihdr(Hdr, MiniHdr),
    !.
tex_doc_minihdr(_,Doc,MiniHdr) :-
    tex_doc_minihdr(Doc, MiniHdr),
    !.
tex_doc_minihdr(Hdr,Doc,unknown_topic) :-
    msg(tex_doc_tree, ['WARNING: could not make a TOC entry.',nl,
                           'Hdr = ',q(Hdr),nl,
                           'Doc = ',q(Doc)]).

tex_doc_minihdr(['{G}'], '\\{G\\}') :- !.
tex_doc_minihdr('{G}', '\\{G\\}') :- !.
tex_doc_minihdr(Hdr, MiniHdr) :-
    atom(Hdr),
    \+ Hdr = nl,
    !,
    %% tex_minihdr_convert(Hdr, MiniHdr).
    MiniHdr = Hdr.
tex_doc_minihdr(Hdr,MiniHdr) :-
    list(Hdr),
    member(E,Hdr),
    atom(E),
    \+ E = nl,
    !,
    %% tex_minihdr_convert(E, MiniHdr).
    MiniHdr = E.

%% tex_minihdr_convert(H, H).

%% strip_leading_newlines(WL0, WL).
%%
%%  WL is the same writelist as WL0, except with sequences of 
%%  `nl' atoms removed from the front. Also occurences of `{G}' 
%%  are replaced by `\{G\}'.
%% 
%%  mode strip_leading_newlines(@writelist, ?writelist).
%%
strip_leading_newlines(['{G}'], ['\\{G\\}']).
strip_leading_newlines(WL0, WL) :-
    strip_leading_nls(WL0, WL).

%% strip_leading_trailing_newlines(WL0, WL).
%%
%%  WL is the same writelist as WL0, except with sequences of
%%  `nl' atoms removed from the front and the rear.
%%
%%  mode strip_leading_trailing_newlines(@writelist, ?writelist).
%%
strip_leading_trailing_newlines(WL0, WL) :-
    atom(WL0),
    !,
    WL = WL0.
strip_leading_trailing_newlines(WL0, WL) :-
    list(WL0),
    strip_leading_nls(WL0, WL1),
    reverse(WL1, WL2),
    strip_leading_nls(WL2, WL3),
    reverse(WL3, WL).


%% private tex_doc_label(Name, Thy, Label).
%%
%%  Strips the '^' and '_' characters from a label for TeX purposes.
%%  Translates cmd(_,_,_,_) references to their real name.
%%
%%  mode tex_doc_label(@atom, ?theory, ?chars).
%%  mode tex_doc_label(@gcomp, @theory, ?chars).
%%
tex_doc_label(cmd(Cmd/Arity), Thy, Label) :-
    !,
    query_theory(doc_node(RealName, _, _, Cmds), Thy),
    member(cmd(_, _, Cmd, Arity), Cmds),
    tex_doc_label(RealName, Thy, Label).
tex_doc_label(Node, _, Label) :-
    atom_codes(Node, L),
    filter(good_tex_label_char, L, Label1),
    tex_label_convert(Label1, Label).

good_tex_label_char(C) :- C > 32, \+ member(C, [94,95]).

tex_label_convert([], []).	
tex_label_convert([H|T], NewHT) :-
    tex_label_char_convert(H, NewH),
    tex_label_convert(T, NewT),
    append(NewH, NewT, NewHT).	

%% private tex_label_char_convert(InChar, OutChars).
%%
%%  convert some special latex characters to a 
%%  sequence of nonspecial characters.
%%
%%  mode tex_label_char_convert(@chars, ?list(chars)).
%%
tex_label_char_convert(124, [118, 98, 97, 114]).		%% |
tex_label_char_convert(123, [108, 98, 114, 97, 99, 101]).	%% {
tex_label_char_convert(125, [114, 98, 114, 97, 99, 101]).	%% }
tex_label_char_convert(45, [109, 105, 110, 117, 115]).		%% -
tex_label_char_convert(62, [103, 114, 101, 97, 116, 101, 114, 108, 101, 115, 115]).	%% >
tex_label_char_convert(C, [C]).

%% writelist_list_xrefs(WriteLists, Xrefs, XrefsTail).
%%
%%  Xrefs is the list of all xref/3 terms that appear in
%%  WriteLists, which must be a list of writelists.
%%  XrefsTail is the tail of Xrefs.
%%  Records the xref'd nodes as having been done, as they have certainly
%%  been scheduled for writing.
%%
%%  mode writelist_list_xrefs(@list(writelist), ?list(compound),
%%                                              ?list(compound)).
%%
writelist_list_xrefs([], Tail, Tail).
writelist_list_xrefs([W|WL], Xrefs, Tail) :-
    writelist_xrefs(W, Xrefs, Tail1),
    writelist_list_xrefs(WL, Tail1, Tail).


%% writelist_xrefs(WriteList, Xrefs, XrefsTail).
%%
%%  Xrefs contains all the nodes referenced by xref/3 terms that appear in
%%  WriteList, and XrefsTail is the tail of Xrefs.
%%
%%  mode writelist_xrefs(@list(writelist), ?list(atom), ?list(atom)).
%%
writelist_xrefs([], Xtail, Xtail).
writelist_xrefs([I|WL], X, Xtail) :-
    writelist_item_xrefs(I, X, Xtail1),
    writelist_xrefs(WL, Xtail1, Xtail).

writelist_item_xrefs(xref(_,_,Node), X, Xtail) :-
    !,
    ( query_global(tex_doc_done(Node)) ->
        X = Xtail
    ;
        tactic_assert_global(tex_doc_done(Node)),
        X = [Node|Xtail]
    ).
writelist_item_xrefs(xinc(Thy, Frag), X, Xtail) :-
    query_theory(doc_fragment(Frag, Doc), Thy),
    !,
    writelist_xrefs(Doc, X, Xtail).
writelist_item_xrefs(_, X, X).

%% writelist_modes_theory(Thy, Cmds, Out).
%%
%% Writes all mode descriptors for each command in Cmds to Out.
%% When a command is specific to a theory then the theory name is
%% also written to Out.
%%
%% mode writelist_modes_theory(@theory, @list(cmd), @stream).
%%
writelist_modes_theory(_, [], _).
writelist_modes_theory(global, Cmds, Out) :- 
    writelist_modes(Out, Cmds),
    writelist(Out, [nl,nl]).
writelist_modes_theory(Thy, Cmds, Out) :-
    writelist(Out, [nl,'This command is specific to theory ',
                    q(Thy),' and descendants.',nl,nl]),
    writelist_modes(Out, Cmds),
    writelist(Out, [nl,nl]).

%% writelist_modes(Out, Cmds).
%%
%% Writes all mode descriptors for each command in Cmds to Out.
%%
%% mode writelist_modes(@stream, @list(cmd)).
%%
writelist_modes(_, []).
writelist_modes(Out, [cmd(Thy, Table, Name, Arity)|Cmds]) :-
    writelist_modes(Out, Cmds),
    functor(Cmd, Name, Arity),
    find_command(Cmd, Table, Thy, Modes, _, _),
    writelist_modes_modes(Out, Modes).

writelist_modes_modes(Out, Modes) :-
    member(Mode, Modes),
    writelist(Out, ['{\\bf mode}\\ \\ ']),
    writelist_modes_mode(Out, Mode),
    writelist(Out, ['.\\\\',nl]),
    fail.
writelist_modes_modes(_, _).

writelist_modes_mode(Out, arity(Mode, Arity)) :-
    functor(Mode, {}, 1),
    !,
    writelist(Out, ['\\{\\}']),
    writelist(Out, ['(']),
    writelist_modes_args(Out, Mode, 1, 1),
    writelist(Out, [')']),
    writelist(Out, ['\\ \\ {\\bf arity}\\ \\ ', w(Arity)]).
writelist_modes_mode(Out, arity(Mode, Arity)) :-
    !,
    functor(Mode, Fun, Ar),
    writelist(Out, [w(Fun)]),
    ( Ar =:= 0 ->
        true
    ;
        writelist(Out, ['(']),
        writelist_modes_args(Out, Mode, 1, Ar),
        writelist(Out, [')'])
    ),
    writelist(Out, ['\\ \\ {\\bf arity}\\ \\ ', w(Arity)]).
writelist_modes_mode(Out, Mode) :-
    functor(Mode, Fun, Ar),
    writelist(Out, [w(Fun)]),
    ( Ar =:= 0 ->
        true
    ;
        writelist(Out, ['(']),
        writelist_modes_args(Out, Mode, 1, Ar),
        writelist(Out, [')'])
    ).

writelist_modes_args(Out, Mode, N, Ar) :-
    arg(N, Mode, Arg),
    Arg = Dir(Type),
    writelist(Out, [w(Dir), w(Type)]),
    ( N =:= Ar ->
        true
    ;
        writelist(Out, [', ']),
        N1 is N + 1,
        writelist_modes_args(Out, Mode, N1, Ar)
    ).

