%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##


%% Compatability functions so Ergo 5 can parse Ergo 4 theory files.

#include "../prover/lib/ergo.h"

%% Ergo 4 axioms are converted to a fundamental inference rule and a
%% postulated theorem.
%% Ergo 4 window opening rules are converted to a series of postulated
%% inference rules.
%%
%% The following changes need to be done by hand:
%%   (maintain this list as new incompatibilities are discovered!!!)
%% 1. Each Ergo 5 theory should include_theory(stdops, share). 
%% 2. The include list should be changed so that:
%%   a. there is one include_theory(thy, all) command per included theory
%%      (i.e., avoid the list form, because the Ergo 5 Makefile cannot cope);
%%   b. only directly relevant theories are included -- the Ergo 5 theories
%%      tended to include everything instead of relying on the transitivity of
%%      the theory graph.  An example is the `iarith' theory: it builds on
%%      typed intuitionistic predicate logic, so it should include ipred_t but
%%      not jprop, jpred, iprop etc.
%% 3. Ergo 5 doesn't use the C preprocessor: just remove the # lines.  The
%%    following functionality is thereby lost: (maintain this list!)
%%     -- CHECK_THY option
%%     -- RERUN_PROOFS option
%%     -- INTERPRETABLE option
%% 4. The following features do not YET have compatibility functions, so they
%%    should be commented out of the thy files (maintain this list!):
%%   rulegroups --- to be done using tables
%%   elves      --- to be done using tables?
%%   param_axioms

database command
  openrule(Names, Term, Path, NewHyps, HypFn, RelFn, GoalFn)
implementation
  deabbreviate_relation_map(RelFn, RelFn1),
  openrule(Names, Term, Path, NewHyps, RelFn1)
command
  openrule(Term, Path, NewHyps, HypFn, RelFn, GoalFn)
implementation
  deabbreviate_relation_map(RelFn, RelFn1),
  openrule(Term, Path, NewHyps, RelFn1)
doc
  'Postulate a series of Ergo 5 inference rules whose effect is the same',nl,
  'as the Ergo 4 openrule.',nl,
  'Does not handle quantifiers, requires HypFn=id, and ignores GoalFn.',nl,
  'For each pair R1--->R2 in RelFn, it postulates an inference rule',nl,
  '  hyp NewHyp1, ..., hyp NewHypN |- ST1 R2 ST2',nl,
  '  -------------------------------------------',nl,
  '                      T1 R1 T2',nl,
  'where T1=Term, ST1 is the subterm at Path, ST2 is a fresh metavariable,',nl,
  ' and T2 is T1 with ST2 in place of ST1.',nl,
  'Names is a list of names for the generated inference rules; it must be',nl,
  'the same length as RelFn, since a rule is generated for each RelFn pair',nl,
  'generates an inference rule.',nl,
  'If names is omitted, names will be generated automatically, by:',nl,
  '  `open_<op>_<path>_<rel>'',',nl,
  'where <op>   is the name of the outermost operator of Term,',nl,
  '      <path> is the position numbers in Path, as `_''-separated digits,',nl,
  '      <rel>  is the name of the relation R.'
mode
  openrule(@list(atom), @thy(compound), @subterm_path, @list(thy(term)),
           @atom, @list(compound), @list(compound)),
  openrule(@thy(compound), @subterm_path, @list(thy(term)),
           @atom, @list(compound), @list(compound)).

openrule(Term, Path, NewHyps, RelFn) :-
    openrule(_, Term, Path, NewHyps, RelFn).
openrule(Names, Term, Path, NewHyps, RelFn) :-
    subterm_change_path(Term, Path, SubTerm, SubTerm1, Term1),
    functor(Term, Op, _),
    openrule_generate_postulates(RelFn, Names, Op, Path, NewHyps,
                                 Term, SubTerm, Term1, SubTerm1).


openrule_generate_postulates([], [], _, _, _, _, _, _, _).
openrule_generate_postulates([(R1--->R2)|RF], [Name|Names], Op, P, NH,
                             T1, ST1, T2, ST2) :-
    openrule_name(Op, P, R1, Name),
    hyplist_to_cterm(NH, ConclId, HypAdds),
    rule_add(Name, R1(T1,T2), ConclId,
             [seq(_,[hyp=HypAdds],R2(ST1,ST2))],
             true),
    openrule_generate_postulates(RF, Names, Op, P, NH, T1, ST1, T2, ST2).

openrule_name(Op, Path, Rel, Name) :-
    var(Name),
    !,
    internal_name(Op, _, OpN),
    internal_name(Rel, _, RelN),
    append([[open, OpN], Path, [RelN]], NameList),
    concat_atom(NameList, '_', Name).
openrule_name(_, _, _, Name).

%% TO DO: Handle constraints (`provided' clause) in the command axiom.

database command
  axiom(===(Name, Stmt))
doc
  'Generate a fundamental inference rule and a theorem.',nl,
  'The fundamental inference rule is called Name_rule, and the theorem is',nl,
  'called Name.',nl,
  'If Stmt does not have the form Prem=>Concl, the fundamental rule',nl,
  'and theorem are identical, with conclusion Stmt and no premisses.',nl,
  'If Stmt has the form Prem=>Concl, the fundamental rule has conclusion',nl,
  'Concl; the premisses are the conjuncts of Prem, except that conjuncts',nl,
  'with form Ants=>Succ form sequents with ''hyp'' antecedents the',nl,
  'conjuncts of Ants, and succedent Succ.'
mode
  axiom(@compound)
implementation
  deabbreviate(Stmt,Stmt2),
  axiom(Name, Stmt2).

axiom(Name, Stmt) :-
    axiom_infrule_name(Name, RuleName),
    axiom_infrule(Stmt, Concl, ConclId, Prems),
    rule_add(RuleName, Concl, ConclId, Prems, true),
    rule_make_fundamental(RuleName),
    rule_add(Name, Stmt, _, [], true).

axiom_infrule_name(Name, RuleName) :-
    atom_concat2(Name, '_rule', RuleName).

axiom_infrule(Stmt, Concl, ConclId, Seqs) :-
    compound(Stmt),
    Stmt = 'jprop.=>'(Prem,Concl),
    !,
    right_assoc_to_list(Prem, 'jprop.and', PremTerms),
    map(axiom_premterm_to_sequent(ConclId), [PremTerms,Seqs]).
axiom_infrule(Stmt, Stmt, _, []).

axiom_premterm_to_sequent(ConclId, PremTerm, Sequent) :-
    compound(PremTerm),
    PremTerm = 'jprop.=>'(Ante,Succ),
    !,
    right_assoc_to_list(Ante, 'jprop.and', Antes),
    hyplist_to_cterm(Antes, ConclId, CSpec),
    Sequent = seq(_,[hyp=CSpec],Succ).
axiom_premterm_to_sequent(_, PremTerm, seq(_,[],PremTerm)).

hyplist_to_cterm([], ConclId, use(hyp,ConclId)).
hyplist_to_cterm([Hyp|Hyps], ConclId, add(_,Hyp,CSpec)) :-
    hyplist_to_cterm(Hyps, ConclId, CSpec).
