%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: treetac.ql,v 1.56 2006/02/13 04:17:02 itp Exp $

#include "../prover/lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The Ergo 5 tactic language.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gumtree_tactic
header
  'Gumtree Tactics'
doc
  'A tactic is a function that maps a sequence of `m'' open proof nodes',nl,
  'to a new sequence of `n'' open proof nodes.  We say such a tactic',nl,
  'has `arity m--->n''.',
  para,
  'Suppose: ',br,
  '`T'' is a tactic with arity `m--->n'';',br,
  '`U'' is a tactic with arity `p--->q'';',br,
  '`R'' is an inference rule with `n'' premises',
  ' and one conclusion; and',br,
  '`L'' is a list of length k.',para,
  'Primitive gumtree_tactics are:',nl,
  begin(list),
  item('Tactic:'),xref(['rule R'], global, cmd(rule/1)),
  begin(list),
  item('  Arity:'),' 1--->n',
  item('  Meaning:'),' map conclusion to premises.',
  end(list),
  item('Tactic:'),xref(['srule R'], global, cmd(srule/1)),
  begin(list),
  item('  Arity:'),' 1--->n',
  item('  Meaning:'),' map conclusion to premises.',
  end(list),
  item('Tactic:'),xref(['zero'], global, cmd(zero/0)),
  begin(list),
  item('  Arity:'),' 0--->0',
  item('  Meaning:'),' do nothing -- zero === [].',
  end(list),
  item('Tactic:'),xref(['one'], global, cmd(one/0)),
  begin(list),
  item('  Arity:'),' 1--->1',
  item('  Meaning:'),' do nothing -- one  === term(_).',
  end(list),
  item('Tactic:'),xref(['two'], global, cmd(two/0)),
  begin(list),
  item('  Arity:'),' 2--->2',
  item('  Meaning:'),' do nothing -- two  === [one,one].',
  end(list),
  item('Tactic:'),xref(['fix(N)'], global, cmd(fix/1)),
  begin(list),
  item('  Arity:'),' N--->N',
  item('  Meaning:'),' do nothing -- fix(N)  === [one,...,one] N times.',
  end(list),
  item('Tactic:'),xref(['skip'], global, cmd(skip/0)),
  begin(list),
  item('  Arity:'),' N--->N',
  item('  Meaning:'),' do nothing -- skip === !([one,skip]|zero).',
  end(list),
  item('Tactic:'),xref(['fail'], global, cmd(fail/0)),
  begin(list),
  item('  Arity:'),' N--->N',
  item('  Meaning:'),' fail',
  end(list),
  item('Tactic:'),xref(['T;U'], global, cmd(;/2)),
  begin(list),
  item('  Arity:'),' m--->q (n=p)',
  item('  Meaning:'),' do T then U',
  end(list),
  item('Tactic:'),xref(['T|U'], global, cmd('|'/2)),
  begin(list),
  item('  Arity:'),' m--->n or p--->q',
  item('  Meaning:'),' do T or U',
  end(list),
  item('Tactic:'),xref(['[T,U]'], global, cmd('.'/2)),
  begin(list),
  item('  Arity:'),' m+p--->n+q',
  item('  Meaning:'),' do T and U to parallel sets of subgoals',nl,
  '           (generalized to an arbitrary list of parallel tactics)',
  end(list),
  item('Tactic:'),xref(['!(T)'], global, cmd('!'/1)),
  begin(list),
  item('  Arity:'),' m--->n',
  item('  Meaning:'),' do T but discard alternatives.',
  end(list),
  item('Tactic:'),xref(['fails(T)'], global, cmd(fails/1)),
  begin(list),
  item('  Arity:'),' n--->n',
  item('  Meaning:'),' The tactic T would fail.',
  end(list),
  item('Tactic:'),xref(['succs(T)'], global, cmd(succs/1)),
  begin(list),
  item('  Arity:'),' n--->n',
  item('  Meaning:'),' The tactic T would succeed.',
  end(list),
  item('Tactic:'),xref(['if_tac(Cond, Then, Else)'], global, cmd(if_tac/3)),
  begin(list),
  item('  Arity:'),' n--->m',
  item('  Meaning:'),' If tactic Cond succeeds then remove alternatives',nl,
  'and do tactic Then, otherwise do tactic Else.',nl,
  'This has the same semantics as Prolog''s if-then-else.',nl,
  'The required arities are Cond : n--->p, Then : p --->m and Else : n--->m.',
  end(list),
  item('Tactic:'),xref(['name(N)'], global, cmd(name/1)),
  begin(list),
  item('  Arity:'),' 1--->1',
  item('  Meaning:'),' The current subgoal is named N (N is a variable).',
  end(list),
  item('Tactic:'),xref(['term(T)'], global, cmd(term/1)),
  begin(list),
  item('  Arity:'),' 1--->1',
  item('  Meaning:'),' The term of the current subgoal is T.',
  end(list),
  item('Tactic:'),xref(['premises(T)'], global, cmd(premises/1)),
  begin(list),
  item('  Arity:'),' N--->N',
  item('  Meaning:'),' The premises of the rule being proved.',
  end(list),
  item('Tactic:'),xref(['constraints(V,Dist,NFI,Other)'], global, cmd(constraints/4)),
  begin(list),
  item('  Arity:'),' N--->N',
  item('  Meaning:'),' The constraints of the rule being proved - V is the list
of variables in the rule, Dist is the list of disjointness constraints,
NFI is the list of not-free-in constraints and Other is the list of other constraints.',
  end(list),
  item('Tactic:'),xref(['context(K,N,T)'], global, cmd(context/3)),
  begin(list),
  item('  Arity:'),' 1--->1',
  item('  Meaning:'),' The current subgoal has a context of kind K, number N, term T.',
  end(list),
  item('Tactic:'),xref(['context_terms(T)'], global, cmd(context_terms/1)),
  begin(list),
  item('  Arity:'),' 1--->1',
  item('  Meaning:'),' The current subgoal has context terms T.',
  end(list),
  item('Tactic:'),xref(['context_index(K,I,CId)'], global, cmd(context_index/3)),
  begin(list),
  item('  Arity:'),' 1--->1',
  item('  Meaning:'),' The current subgoal has a context of kind K, index I and ID CId.',
  end(list),
  item('Tactic:'),xref(['context_terms(T)'], global, cmd(context_terms/1)),
  begin(list),
  item('  Arity:'),' 1--->1',
  item('  Meaning:'),' The current subgoal has context terms T.',
  end(list),
  item('Tactic:'),xref(['names(L)'], global, cmd(names/1)),
  begin(list),
  item('  Arity:'),' k--->k',
  item('  Meaning:'),' The first k subgoals are named L.',
  end(list),
  item('Tactic:'),xref(['terms(L)'], global, cmd(terms/1)),
  begin(list),
  item('  Arity:'),' k--->k',
  item('  Meaning:'),' The terms of the first k subgoals are L.',
  end(list),
  item('Tactic:'),xref(['annotate(L)'], global, cmd(annotate/1)),
  begin(list),
  item('  Arity:'),' 1--->1',
  item('  Meaning:'),' add the annotation(s) L to the first open subproof',
  end(list),
  item('Tactic:'),xref(['{G}'], global, cmd('{}'/1)),
  begin(list),
  item('  Arity:'),' N--->N',
  item('  Meaning:'),' execute Prolog goal G (must not affect proof!)',
  end(list),
  end(list),
  nl
extends
  concepts.

?- require_global_tactic_file(operator),
   ( query_global(global_operator(tactic, _, _)) ->
       true
   ;
       add_global_operator(1190, prefix, tactic),
       add_global_operator(1180, prefix, delete),
       add_global_operator(900, rightassoc, ';'),
       add_global_operator(910, rightassoc, '|')
       %% add_global_operator(710, prefix, 'rule')
   ).
       
call_gumtree_tactic(Tac, Input, IT, Output, OT) :-
    expand_tac_body_and_call(Tac, Input, IT, Output, OT).


gumtree command
  rule(R)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'invoke an inference rule.'
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic(rule(R))
mode
  rule(?rulename) arity 1 ---> _.

gumtree command
  srule(R)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'invoke an inference rule. Structural unify is first attempted', nl,
  'on the rule and proof terms and if that fails normal unify is',nl,
  'used. Structural unify avoids changes of bound variables and',nl,
  'and so reduces the number of substitutions and delayed not-free-in',nl,
  'constraints.'
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic(srule(R))
mode
  srule(?rulename) arity 1 ---> _.

gumtree command
  curNodes(R)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'returns the current proof nodes.'
implementation
  call_gumtree_tactic(curNodes(R))
mode
  curNodes(?list(integer)).

gumtree command
  zero
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Ensure there is no current subgoal',nl,
  '(this is a left- and right-identity for parallel composition).'
extends
  proof_tree_cmds
mode
  zero arity 0 ---> 0
implementation
  call_gumtree_tactic(zero).

gumtree command
  one
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Ensure there is a single current subgoal.'
extends
  proof_tree_cmds
mode
  one arity 1 ---> 1
implementation
  call_gumtree_tactic(one).

gumtree command
  two
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Ensure there are exactly two current subgoals.'
extends
  proof_tree_cmds
mode
  two arity 2 ---> 2
implementation
  call_gumtree_tactic(two).

gumtree command
  fix(N)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Ensure there are exactly N current subgoals.'
extends
  proof_tree_cmds
mode
  fix(?integer) arity  N ---> N
implementation
  call_gumtree_tactic(fix(N)).

gumtree command
  skip
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Accepts all current subgoals',nl,
  '(this is a left- and right-identity for sequential composition).'
extends
  proof_tree_cmds
mode
  skip arity _ ---> _
implementation
  call_gumtree_tactic(skip).

gumtree command
  use(Query)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Transform the current open node using a rule satisfying Query.'
extends
  proof_tree_cmds
mode
  use(?query) arity 1 ---> _
implementation
  call_gumtree_tactic(use(Query)).

gumtree command
  use(Query, Output)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Transform the current open node using a rule satisfying Query.'
extends
  proof_tree_cmds
mode
  use(?query, ?thy(term)) arity 1 ---> _
implementation
  call_gumtree_tactic(use(Query,Output)).

gumtree command
  iuse(Query)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Transform the current open node using an instantiated rule satisfying Query.'
extends
  proof_tree_cmds
mode
  iuse(?query) arity 1 ---> _
implementation
  call_gumtree_tactic(iuse(Query)).

gumtree command
  iuse(Query, Output)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Transform the current open node using an instantiated rule satisfying Query.'
extends
  proof_tree_cmds
mode
  iuse(?query, ?thy(term)) arity 1 ---> _
implementation
  call_gumtree_tactic(iuse(Query,Output)).

gumtree command
  rules(Query)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Print out all rules matching Query.'
extends
  query_treecmds
mode
  rules(?query)
implementation
  call_gumtree_tactic(rules(Query)).

gumtree command
  irules(Query)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Print out all instantiated rules matching Query and current open proof node.'
extends
  query_treecmds
mode
  irules(?query)
implementation
  call_gumtree_tactic(irules(Query)).

%%gumtree command
%%  emacs_focus_rules
%%doc
%%  'List all rules that match the current focus.'
%%extends
%%  proof_tree_cmds
%%mode
%%  emacs_focus_rules
%%implementation
%%  call_gumtree_tactic(emacs_focus_rules).

gumtree command
  trace
mode
  trace
implementation
  call_gumtree_tactic(trace(trace, 1, ['']))
command
  trace(Caller)
mode
  trace(@nonvar)
implementation
  call_gumtree_tactic(trace(Caller, 1, ['']))
command
  trace(Caller, Level, Msg)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Accepts all current subgoals.',nl,
  'If the current ', xref(verbosity,global,'parameter:verbosity'),
  ' is greater than or equal to Level,',nl,
  'It displays a message of the form:',nl,
  '  Caller[subgoals]: Msg.',nl,
  '''trace'' is equivalent to ''trace(trace, 1, [''''])''.'
extends
  proof_tree_cmds
mode
  trace(@nonvar, @integer, @writelist)
implementation
  call_gumtree_tactic(trace(Caller, Level, Msg)).

gumtree command
  fail
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), '.'
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic(fail)
mode
  fail arity _ ---> _.

gumtree command
  ;(T,U)
doc
  'A ', xref(gumtree_tactic,global,gumtree_tactic), ' combinator:',nl,
  'do T then U.'
extends
  tactic_combine_cmds
implementation
  call_gumtree_tactic(;(T,U))
mode
  ;(+tactic, +tactic) arity _ ---> _.

gumtree command
  '|'(T,U)
doc
  'A ', xref(gumtree_tactic,global,gumtree_tactic), ' combinator:',nl,
  'do T or U.'
extends
  tactic_combine_cmds
implementation
  call_gumtree_tactic('|'(T,U))
mode
  '|'(+tactic, +tactic) arity _ ---> _.

gumtree command
  exec_gumtree_tactic(T)
doc
  'Execute tactic.',nl
extends
  tactic_combine_cmds
implementation
  call_gumtree_tactic(exec_gumtree_tactic(T))
mode
  exec_gumtree_tactic(+tactic).

gumtree command
  deabbrev_and_call_gumtree_tactic(T)
doc
  'Deabbreviate and call the gumtree tactic.',nl
extends
  tactic_combine_cmds
implementation
  call_gumtree_tactic(deabbrev_and_call_gumtree_tactic(T))
mode
  deabbrev_and_call_gumtree_tactic(+tactic).

gumtree command
  []
%% not documented: equivalent to `zero'
implementation
  call_gumtree_tactic([])
mode
  [] arity 0 ---> 0.

gumtree command
  [T|Ts]
header
  '[T1,T2,...]'
doc
  'A ', xref(gumtree_tactic,global,gumtree_tactic), ' combinator:',nl,
  'do each Ti in parallel on a subsequence of the current open nodes.'
extends
  tactic_combine_cmds
implementation
  call_gumtree_tactic([T|Ts])
mode
  [+tactic | +list(tactic)] arity _ ---> _.

gumtree command
  !(T)
doc
  'A ', xref(gumtree_tactic,global,gumtree_tactic), ' combinator:',nl,
  'do T, but discard alternatives.'
extends
  tactic_combine_cmds
implementation
  call_gumtree_tactic(!(T))
mode
  !(+tactic) arity _.

gumtree command
  with_context_delaying(Patt,T)
doc
  'A ', xref(gumtree_tactic,global,gumtree_tactic), ' combinator:',nl,
  'it executes tactic T in a mode which allows implicit context to',nl,
  'be added to the current proof.  That is, within T procedures that',nl,
  'search the context of a proof node will add a required context term',nl,
  'if it is not already there and it is unifiable with Patt.',nl,
  'All context terms that are added in this way become constraints',nl,
  'on any inference rules that result from this proof.'
extends
  tactic_combine_cmds
implementation
  call_gumtree_tactic(with_context_delaying(Patt,T))
mode
  with_context_delaying(@thy(term), +tactic) arity _ ---> _.

gumtree command
  fails(T)
doc
  'A ', xref(gumtree_tactic,global,gumtree_tactic), ' combinator:',nl,
  'The tactic fails(T) succeeds if the tactic T would fail.'
extends
  tactic_combine_cmds
implementation
  call_gumtree_tactic(fails(T))
mode
  fails(+tactic) arity N ---> N.

gumtree command
  succs(T)
doc
  'A ', xref(gumtree_tactic,global,gumtree_tactic), ' combinator:',nl,
  'The tactic succs(T) succeeds if the tactic T would succeed.'
extends
  tactic_combine_cmds
implementation
  call_gumtree_tactic(succs(T))
mode
  succs(+tactic) arity N ---> N.

gumtree command
  if_tac(Cond, Then, Else)
doc
  'A ', xref(gumtree_tactic,global,gumtree_tactic), ' combinator:',nl,
  'Do Cond (removing choices) followed by Then, otherwise do Else.',nl,
  'It has the same semantics as Prolog''s if-then-else.'
extends
  tactic_combine_cmds
implementation
  call_gumtree_tactic(if_tac(Cond, Then, Else))
mode
  if_tac(+tactic,+tactic,+tactic) arity _ ---> _.

gumtree command
  name(N)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'The variable N is bound to the name of the current subgoal.'
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic(name(N))
mode
  name(?nodeid) arity 1 ---> 1.

gumtree command
  names(L)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'L are the names of some prefix of the current subgoal list.',nl,
  '  names([a,b,c]) == [name(a), name(b)), name(c)].'  
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic(names(L))
mode
  names(?list(nodeid)) arity N ---> N.

gumtree command
  term(T)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'T is the term of the current subgoal.',nl,
  'The subgoal is not affected; like ', xref(['one'], global, cmd(one/0)), ', this tactic',nl,
  'passes a single subgoal unchanged to the next tactic',nl,
  'in a sequential composition.'
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic(term(T))
mode
  term(?thy(term)) arity 1 ---> 1.

gumtree command
  premises(Prems)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Prems is the list of premises of the rule.'
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic(premises(Prems))
mode
  premises(?list(term)) arity N ---> N.

gumtree command
  constraints(Vars, Dist, NFI, Other)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'The constraints of the rule being proved - ',br,
  ' Vars is the list of variables in the rule',br,
  ' Dist is the list of disjointness constraints',br,
  ' NFI is the list of not-free-in constraints',br,
  ' Other is the list of other constraints'
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic(constraints(Vars, Dist, NFI, Other))
mode
  constraints(?list(thy(term)), ?list(thy(term)), ?list(thy(term)),
              ?list(thy(term))) arity N ---> N.

gumtree command
  context(K,N,T)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'The current subgoal has a context of kind K, number N, term T.',nl,
  'The subgoal is not affected; like `one'', this tactic',nl,
  'passes a single subgoal unchanged to the next tactic',nl,
  'in a sequential composition.',nl,
  'It is nondeterministic: alternative contexts are returned on backtracking.'
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic(context(K,N,T))
mode
  context(@cname, ?ctermid, ?thy(term)) arity 1 ---> 1.

gumtree command
  context_index(K,I,CId)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'The current subgoal has a context of kind K, index I and ID CId.',nl,
  'The subgoal is not affected; like `one'', this tactic',nl,
  'passes a single subgoal unchanged to the next tactic',nl,
  'in a sequential composition.',nl,
  'This tactic is not normally expected to be used by users, it is used',nl,
  'for constructing and rerunning proofs to avoid searching for context terms.'
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic(context_index(K,I,CId))
mode
  context_index(@cname, ?number, ?ctermid) arity 1 ---> 1.

gumtree command
  context_terms(T)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'The current subgoal has context terms T.',nl,
  'The context terms is a list of terms of the form ContextType = ContextList',
  nl,'where ContextList is a list of the form c(ID,Term).',nl,
  'ID is the id of the context term Term.',nl,
  'The subgoal is not affected; like `one'', this tactic',nl,
  'passes a single subgoal unchanged to the next tactic',nl,
  'in a sequential composition.'
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic(context_terms(T))
mode
  context_terms(?list(cterm)) arity 1 ---> 1.

gumtree command
  terms(L)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'L is a list of terms of some prefix of the current subgoal list.',nl,
  '  terms([a,b,c]) == [term(a), term(b)), term(c)].'  
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic(terms(L))
mode
  terms(?list(thy(term))) arity N ---> N.

gumtree command
  annotate(L)
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Add the annotation(s) L to the current subgoal.',nl,
  'The subgoal is not affected; like `one'', this tactic',nl,
  'passes a single subgoal unchanged to the next tactic',nl,
  'in a sequential composition.',nl
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic(annotate(L))
mode
  annotate(@list(annotation)) arity 1 ---> 1.
  
gumtree command
  {G}
header
  '{G}'
doc
  'A primitive ', xref(gumtree_tactic,global,gumtree_tactic), ':',nl,
  'Execute Prolog goal G (must not affect proof!),',nl,
  'As with `skip'', any number of subgoals are passed unchanged',nl,
  'to the next tactic in a sequential composition.'
extends
  proof_tree_cmds
implementation
  call_gumtree_tactic({G})
mode
  {}(?term) arity N ---> N.


expand_tac_body_and_call(Body, X, XT, Y, YT) :-
    (std_nonvar(Body) ; list(Body)),
    !,
    msg(call_gumtree_tactic, 2, ['Calling tactic: ', q(Body)]),
    expand_tac_body_and_exec(Body, X, XT, Y, YT).
expand_tac_body_and_call(Body, _, _, _, _) :-
    handle_error(['uninstantiated tactic: ',q(Body)]).


%% private expand_tac_body_and_exec(Body, X, XT, Y, YT).
%%
%%  Expands the tactic body of the call Body and executes it.
%%  X-XT is a difference list representing the input nodes, 
%%  Y-YT is a difference list representing the output nodes.
%%
expand_tac_body_and_exec(Body, _, _, _, _) :-
    var(Body),
    !,
    handle_error(['uninstantiated tactic: ',q(Body)]).
expand_tac_body_and_exec(rule(R), X, XT, Y, YT) :-
    !,
    nonvar(X),
    X = [X0|XT],
    decode_rule(R, Thy, Name),
    ip_set('rule@unify@flag', unify),
    proof_step(Thy, Name, X0, Y1),
    append(Y1, YT, Y).
expand_tac_body_and_exec(srule(R), X, XT, Y, YT) :-
    !,
    nonvar(X),
    X = [X0|XT],
    decode_rule(R, Thy, Name),
    ip_set('rule@unify@flag', sunify),
    proof_step(Thy, Name, X0, Y1),
    append(Y1, YT, Y).
expand_tac_body_and_exec(curNodes(R), X, XT, Y, YT) :-
    !,
    Y = X,
    YT = XT,
    findall(Z, (Z = X, XT = []), ZL),
    [R] = ZL.
expand_tac_body_and_exec(zero, X, XT, Y, YT) :-
    !,
    X = XT,
    Y = X,
    YT = XT.
expand_tac_body_and_exec(one, X, XT, Y, YT) :-
    !,
    nonvar(X),
    X = [_|XT],
    Y = X,
    YT = XT.
expand_tac_body_and_exec(two, X, XT, Y, YT) :-
    !,
    nonvar(X),
    X = [_,_|XT],
    Y = X,
    YT = XT.
expand_tac_body_and_exec(fix(L), X, XT, Y, YT) :-
    !,
    \+(\+((XT = [], length(X, L)))),
    Y = X,
    YT = XT.
expand_tac_body_and_exec(skip, X, XT, Y, YT) :-
    !,
    Y = X,
    YT = XT.
expand_tac_body_and_exec(use(Q), X, XT, Y, YT) :-
    !,
    nonvar(X),
    X = [X0|XT],
    quse(Q, X0, Y1),
    append(Y1, YT, Y).
expand_tac_body_and_exec(use(Q, O), X, XT, Y, YT) :-
    !,
    nonvar(X),
    X = [X0|XT],
    quse(Q, X0, O, Y1),
    append(Y1, YT, Y).
expand_tac_body_and_exec(iuse(Q), X, XT, Y, YT) :-
    !,
    nonvar(X),
    X = [X0|XT],
    qiuse(Q, X0, Y1),
    append(Y1, YT, Y).
expand_tac_body_and_exec(iuse(Q, O), X, XT, Y, YT) :-
    !,
    nonvar(X),
    X = [X0|XT],
    qiuse(Q, X0, O, Y1),
    append(Y1, YT, Y).
expand_tac_body_and_exec(rules(Q), X, XT, Y, YT) :-
    !,
    Y = X,
    YT = XT,
    rulebrowse(Q, X, XT).
expand_tac_body_and_exec(irules(Q), X, XT, Y, YT) :-
    !,
    Y = X,
    YT = XT,
    qirule(Q, X, XT).
expand_tac_body_and_exec(trace(Caller,Level,Msg), X, XT, Y, YT) :-
    !,
    Y = X,
    XT = YT,
    msg(Caller(X), Level, Msg).
expand_tac_body_and_exec(fail, _, _, _, _) :-
    !,
    fail.
expand_tac_body_and_exec(;(T1,T2), X, XT, Y, YT) :-
    !,
    expand_tac_body_and_exec(T1, X, XT, I, IT),
    expand_tac_body_and_exec(T2, I, IT, Y, YT).
expand_tac_body_and_exec('|'(T1,T2), X, XT, Y, YT) :-
    !,
    ( expand_tac_body_and_exec(T1, X, XT, Y, YT)
    ; expand_tac_body_and_exec(T2, X, XT, Y, YT)
    ).
expand_tac_body_and_exec([], X, XT, Y, YT) :-
    !,
    X = XT,
    Y = X,
    YT = XT.
expand_tac_body_and_exec([T1], X, XT, Y, YT) :-
    !,
    expand_tac_body_and_exec(T1, X, XT, Y, YT).
expand_tac_body_and_exec([T1|T2], X, XT, Y, YT) :-
    !,
    split_open_lists(X, XT, X1, X1T, X2, X2T),
    expand_tac_body_and_exec(T1, X1, X1T, Y, Y1T),
    expand_tac_body_and_exec(T2, X2, X2T, Y1T, YT).
expand_tac_body_and_exec(!(T), X, XT, Y, YT) :-
    !,
    once(expand_tac_body_and_exec(T, X, XT, Y, YT)).
expand_tac_body_and_exec(with_context_delaying(Patt,T), X, XT, Y, YT) :-
    !,
    with_context_delaying(Patt, expand_tac_body_and_exec(T, X, XT, Y, YT)).
expand_tac_body_and_exec(fails(T), X, XT, Y, YT) :-
    !,
    Y = X,
    YT = XT,
    ( \+ expand_tac_body_and_exec(T, X, XT, _, _) ).
expand_tac_body_and_exec(succs(T), X, XT, Y, YT) :-
    !,
    Y = X,
    YT = XT,
    ( \+ \+ expand_tac_body_and_exec(T, X, XT, _, _) ).
expand_tac_body_and_exec(if_tac(Cond, Then, Else), X, XT, Y, YT) :-
    !,
    (  expand_tac_body_and_exec(Cond, X, XT, X1, X1T) ->
         expand_tac_body_and_exec(Then, X1, X1T, Y, YT)
       ;
         expand_tac_body_and_exec(Else, X, XT, Y, YT)
    ).
expand_tac_body_and_exec(name(Id), X, XT, Y, YT) :-
    !,
    nonvar(X),
    X = [Id1|XT],
    Y = X,
    YT = XT,
    ( record_proof_node_names([Id1], [Id]),
      Id1=Id
    ).
expand_tac_body_and_exec(term(T), X, XT, Y, YT) :-
    !,
    nonvar(X),
    X = [Id|XT],
    Y = X,
    YT = XT,
    proof_node_term(Id,T1),
    ip_set('rule@unify@flag', sunify),
    rule_unify(T1, T).
expand_tac_body_and_exec(premises(T), X, XT, Y, YT) :-
    !,
    Y = X,
    YT = XT,
    gumtree_get_abbrev_premises(T1),
    structural_unify(T, T1).
expand_tac_body_and_exec(constraints(Vars, Dist, NFI, Other), X, XT, Y, YT) :-
    !,
    Y = X,
    YT = XT,
    theorem_constraints(Vars, Dist, NFI, Other).
expand_tac_body_and_exec(context(K,I,T), X, XT, Y, YT) :-
    !,
    X = [Id|XT],
    Y = X,
    YT = XT,
    context_search(K, Id, I, T1),
    ip_set('rule@unify@flag', sunify),
    rule_unify(T1, T).
expand_tac_body_and_exec(context_index(K,Index,CID), X, XT, Y, YT) :-
    !,
    X = [Id|XT],
    Y = X,
    YT = XT,
    context_index(K, Id, Index, CID).
expand_tac_body_and_exec(context_terms(T), X, XT, Y, YT) :-
    !,
    X = [Id|XT],
    Y = X,
    YT = XT,
    collect_context_terms(Id, T).
expand_tac_body_and_exec(names(L), X, XT, Y, YT) :-
    !,
    Y = X,
    YT = XT,
    diff_list_to_list(X, XT, Nodes),
    ( record_proof_node_names(Nodes, L),
      Nodes=L
    ).
expand_tac_body_and_exec(terms(L), X, XT, Y, YT) :-
    !,
    Y = X,
    YT = XT,
    diff_list_to_list(X, XT, Nodes),
    proof_node_terms(Nodes,L).
expand_tac_body_and_exec(annotate(L), X, XT, Y, YT) :-
    !,
    X = [Id|XT],
    Y = X,
    YT = XT,
    annotate(Id,L).
expand_tac_body_and_exec({Goal}, X, XT, Y, YT) :-
    !,
    Y = X,
    YT = XT,
    call_predicate(Goal).
expand_tac_body_and_exec(call_gumtree_tactic(Tac), X, XT, Y, YT) :-
    !,
    expand_tac_body_and_exec(Tac, X, XT, Y, YT).
expand_tac_body_and_exec(exec_gumtree_tactic(Tac), X, XT, Y, YT) :-
    !,
    uncurry(Tac(X,XT,Y,YT), UTac),
    call_predicate(UTac).
expand_tac_body_and_exec(deabbrev_and_call_gumtree_tactic(Cmd), 
                         X, XT, Y, YT) :-
    !,
    deabbrev_and_call_gumtree_tactic(Cmd, X, XT, Y, YT).
expand_tac_body_and_exec(Head, X, XT, Y, YT) :-
    query_global(gumtree_tactic_declaration(Head, _, '--->'(In, _Out))),
    !,
    ( integer(In) ->
        length(InList, In),
        append(InList, XT, X)
    ;
        true
    ),
    gumtree_2_predicate(Head, X, XT, Y, YT, Call),
    call_predicate(Call).
expand_tac_body_and_exec(Head, X, XT, Y, YT) :-
    functor(Head, F, N),
    functor(HeadCopy, F, N),
    find_command(HeadCopy, gumtree, _, call_gumtree_tactic(Tac), _),
    !,
    Head = HeadCopy,
    expand_tac_body_and_exec(Tac, X, XT, Y, YT).
expand_tac_body_and_exec(Head, _, _, _, _) :-
    handle_error(['unknown gumtree tactic: ',q(Head)]).

deabbrev_and_call_gumtree_tactic(deabbrev_and_call_gumtree_tactic(Cmd), 
                                 _X, _XT, _Y, _YT) :-
    !,
    handle_error(['unknown gumtree tactic: ',q(Cmd)]).
deabbrev_and_call_gumtree_tactic(Cmd, X, XT, Y, YT) :-
    gumtree_deabbrev_cmd(gumtree, Cmd, [], Call),
    once(( 
        Call = call_gumtree_tactic(_, Tac) 
     ; 
        Call = call_gumtree_tactic_basic(_, Tac)
    )),
    expand_tac_body_and_exec(Tac, X, XT, Y, YT).

%% expand_append(X, Y, Z, Code).
%%
%%  Code ensures that append(X,Y,Z) is true.
%%  Deterministic calls are done at expansion time (in which case Code=true).
%% 
expand_append(X, Y, Z, Code) :-
    ( closed_list(X) ->
        append(X, Y, Z),
        Code = true
    ; Y == [] ->
        X = Z,
        Code = true
    ; closed_list(Y), closed_list(Z) ->
        append(X, Y, Z),
        Code = true
    ;
        Code = append(X, Y, Z)
    ).

%% record_proof_node_names(Nodes, Vars).
%%
%%  Nodes have just been named with Vars.
%%  If the Vars themselves have names, annotate the Nodes with those Names.
%% 
record_proof_node_names(N, V) :- var(N), !, V = [].
record_proof_node_names([], []) :- !.
record_proof_node_names([Node|Nodes], [Var|Vars]) :-
    ( get_var_name(Var, Name) ->
        annotate(Node, name(Node,Name))
    ;
        true
    ),
record_proof_node_names(Nodes, Vars).

%% write_proof_node(Strm, Node).
%% write_proof_nodes(Strm, Nodes).
%%
%%  Write Node(s) from the current proof tree, using names for those nodes with
%%  names, and numbers for those without.
%%
write_proof_node(Strm, Node) :-
    proof_node_names([Node], [Name]),
    writelist2([w(Name)], Strm, 0, 0).

write_proof_nodes(Strm, Nodes) :-
    proof_node_names(Nodes, Names),
    writelist2([w(Names)], Strm, 0, 0).

proof_node_names([], []).
proof_node_names([Node|Nodes], [Name|Names]) :-
    ( nonvar(Node), annotation(Node, name(Node,Name)) ->
        true
    ;
        Name = Node
    ),
    proof_node_names(Nodes, Names).

%% Collect all the context information as a list of terms of the form
%% ContextName = TermList where TermList is a list of the form c(ID,Term)
%% where ID is the id of context term Term.

collect_context_terms(Id, T) :-
   proof_node_context(Id, Contexts),
   collect_context_terms2(Contexts, Id, T).

collect_context_terms2([], _, []).
collect_context_terms2([Name=IDList|Rest], Id, [Name=CList|Result]) :-
   collect_context_terms3(IDList, Name, Id, CList),
   collect_context_terms2(Rest, Id, Result).

collect_context_terms3([], _, _, []).
collect_context_terms3([CID|Rest], Name, Id, [c(CID,Term)|Result]) :-
   context_retrieve(Name, Id, CID, Term),
   collect_context_terms3(Rest, Name, Id, Result).
   



split_open_lists(X, XT, X1, X1, X, XT).
split_open_lists([H|X], XT, [H|X1], X1T, X2, X2T) :-
    var(XT),
    split_open_lists(X, XT, X1, X1T, X2, X2T).

diff_list_to_list(X, XD, Y) :- append(Y, XD, X), !.

?- ( query_global(handler(write(_,_,_), write_proof_nodes(_,_))) ->
       true
   ;
       add_handler(write(S,_,proof_node(N)), write_proof_node(S,N)),
       add_handler(write(S,_,proof_nodes(N)), write_proof_nodes(S,N))
   ).
