%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: usefwd.ql,v 1.53 2001/06/04 01:24:11 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  This file defines the 'use', 'fwd' and 'rule' tactics.
%%  They are higher level interfaces to the 'incremental trans'
%%  commands (itrans_*), which allow a single transformation to
%%  be built up incrementally by specifying theory name, rule name,
%%  input, output, relation and conditions etc. in a flexible order.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../prover/lib/ergo.h"

%% TO DO: Get defaults from the database (ancestors) in get_use_defaults,
%% get_fwd_defaults([]), get_rule_defaults([]) and get_irule_defaults([]).
%% See tactics/rulebrowse.ql for a similar comment.

get_use_defaults([]).
get_fwd_defaults([]).
get_rule_defaults([]).
get_irule_defaults([]).


proof command
  use(Query)
wrapper
  []  % create a wrapper, but don't display messages.
command
  use(Query,Output)
wrapper
  []  % create a wrapper, but don't display messages.
extends
  focus_proofcmds
doc
  'Transforms the focus using a rule that satisfies Query ',
  '(see ', xref(query_language,global,query_language), ').',nl,
  nl,
  'The one argument form does not constraint the output of the rule,',nl,
  'whereas the two argument form requires it to match Output.'
mode
  use(?query),
  use(?query,?thy(term)).

use(Q) :-
    query_init(use),
    get_use_defaults(Defs),
    query_process([i(focus)|Defs]),
    query_process(Q),
    thawing_new_vars(query_execute(focus)).

use(Q,Output) :-
    query_init(use),
    get_use_defaults(Defs),
    query_process([i(focus),oterm(Output)|Defs]),
    query_process(Q),
    thawing_new_vars(query_execute(focus)).


proof command
  fwd(Query)
wrapper
  []
command
  fwd(Query,Out)
wrapper
  []
extends
  hyp_proofcmds
doc
  'Deduces a new hypothesis using a rule that satisfies Query ',
  '(see ', xref(query_language,global,query_language), ').',nl,
  nl,
  'The two argument form requires the new hypothesis to unify',nl,
  'with the given Output term.'
mode
  fwd(?query),
  fwd(?query,?thy(term)).


fwd(Q) :-
    query_init(fwd),
    get_fwd_defaults(Defs),
    query_process(Defs),
    query_process(Q),
    ( itrans_input_name(true) ->
        %% default input name is true.
        true
    ;
        %% Q has already specified i(...).
        true
    ),
    thawing_new_vars(query_execute(hyp:_)).

fwd(Q,Output) :-
    query_init(fwd),
    get_fwd_defaults(Defs),
    query_process([oterm(Output)|Defs]),
    query_process(Q),
    ( itrans_input_name(true) ->
        %% default input name is true.
        true
    ;
        %% Q has already specified i(...).
        true
    ),
    thawing_new_vars(query_execute(hyp:_)).


general command
  rule(Query)
extends
  general_cmds
doc
  'Prints out all rules that match Query ',
  '(see ', xref(query_language,global,query_language), ').',nl,
  'By default, all rules are frozen, so that the printed rules are',nl,
  'identical to the rule in the database, not instantiated in some way.'
mode
  rule(?query).

rule(Q) :-
    query_init(rule),
    get_rule_defaults(Defs),
    query_process([input_name(null),frozen(true)|Defs]),
    query_process(Q),
    less_verbose(thawing_new_vars(rule2)).
rule(_).

rule2 :-
    query_choose_rule(Thy,Name),
    rule3,
    itrans_kind(Kind),
    itrans_stmt(Stmt),
    itrans_body(Body),
    writelist([q(Kind),' ',q(Thy),':',q(Name),' ===',nl,
               tab,thy(Stmt)]),
    rule_show_body(Body),
    fail.

rule_show_body(true) :-
    !,
    writelist(['.',nl]).
rule_show_body(Body) :-
    writelist([nl,tab,'provided  ',code(Body),'.',nl]).

rule3 :-
    itrans_conditional(_),
    itrans_direction(_),
    query_check_matching,
    itrans_finish(null),
    !.


proof command
  irule(Query)
doc
  'Prints out all ',emph(instantiated),' rules that match Query AND the current',nl,
  'window''s focus.'
extends
  proof_info_cmds
mode
  irule(?query).

irule(Q) :-
    query_init(irule),
    get_irule_defaults(Defs),
    query_process([i(focus),frozen(false)|Defs]),
    query_process(Q),
    less_verbose(thawing_new_vars(rule2)).
irule(_).


%% private query_init(Caller).
%%
%%  Sets up a new query.  Caller is the name of the calling
%%  tactic, for use in output messages.
%%
%%  mode query_init(@nonvar).
%%
query_init(Caller) :-
    itrans_start,
    ip_set(query_caller,Caller),
    ip_set(query_thys,[hyp,*]),
    ip_set(query_name,_),
    ip_set(query_subnames,[]),
    ip_set(query_subterms,[]),
    ip_set(query_frozen,false),
    ip_set(query_imatch,any),
    ip_set(query_omatch,any),
    ip_set(query_condlist,_),
    ip_set(query_conjterms, []),
    ip_set(query_conds,[]),
    ip_set(query_cthys,[hyp,*]).


%% private query_process(Query).
%%
%%  Preprocesses Query, ready for execution by query_execute.
%%  query_process may be called several times to build up
%%  a more complex query.  Some constraints (eg. thys/1) override
%%  similar but earlier constraints, while others (eg. subterm/1)
%%  are conjoined.  Thus, default constraints should be passed
%%  to query_process first.
%%
%%  Query may be a list of constraints, or a single constraint.
%%
%%  mode query_process(?query).
%%
query_process([]) :-
    !.
query_process(Chars) :-
    %% Special case for "..." constraints.
    Chars = [C|_],
    integer(C),
    !,
    do_constraint(Chars).
query_process([C|CS]) :-
    !,
    do_constraint(C),
    query_process(CS).
query_process(and(Q1,Q2)) :-
    !,
    query_process(Q1),
    query_process(Q2).
query_process(C) :-
    do_constraint(C).


%% query_execute(OutName).
%%
%%  Completes the current itrans operation using the constraints
%%  that have been recorded by query_process.
%%  It assumes that the itrans_start operation has been done.
%%  It chooses the rule, conditionality and direction, then
%%  discharges all the conditions of the rule and calls itrans_finish.
%%
%%  mode query_execute(@nonvar).
%%
query_execute(OutName) :-
    ip_lookup(query_caller, Caller),
    query_choose_rule(Thy,Name),
    %% Name is sometimes a theory term, but for simplicity we always
    %% write it as thy(_) -- atomic rulenames will rarely be abbreviated.
    msg(Caller, 2, ['Trying ',q(Thy),':',thy(Name)]),
    itrans_conditional(Cond),
    itrans_direction(Dir),
    msg(Caller, 3, ['    cond=',q(Cond),' dir=',q(Dir)]),
    query_check_matching,
    itrans_eval_body,
    query_discharge(Cond),
    itrans_dischargelist(DList),
    msg(Caller, 4, ['    satisfied conditions with: ',q(DList)]),
    itrans_finish(OutName),
    itrans_rule_name(FinalThy,FinalName),
    msg(Caller, 1, ['[',q(FinalThy),':',q(FinalName),',',Cond,',',Dir,']']).

    

%%  
%% query_choose_rule(Thy,Name).
%%
%%  Does the itrans_rule_name part of the current itrans operation,
%%  according to the constraints that have been built up in the
%%  query_* implicit parameters by query_process.
%%
%%  mode query_choose_rule(?theory,?term).
%%
query_choose_rule(Thy,Name) :-
    ip_lookup(query_thys, Thys),
    ip_lookup(query_name, Name),
    ip_lookup(query_subnames, Subnames),
    ip_lookup(query_subterms, Subterms),
    ip_lookup(query_frozen, Frozen),
    query_set_frozen(Frozen),
    query_choose_theory(Thys, Thy),
    itrans_rule_name(Thy, Name),
    query_check_subnames(Thy, Name, Subnames),
    query_check_subterms(Subterms).

query_set_frozen(true) :- itrans_frozen(true).
query_set_frozen(false).

query_check_subnames(hyp, _, Subnames) :-
    !,
    Subnames = [].
query_check_subnames(_, Name, Subnames) :-
    query_check_subnames2(Subnames, Name).

query_check_subnames2([], _).
query_check_subnames2([S|SS], Name) :-
    atom_codes(Sub,S),
    atom_search(Name,1,Sub,_),
    !,
    query_check_subnames2(SS, Name).

query_check_subterms([]).
query_check_subterms([S|SS]) :-
    itrans_stmt(Stmt),
    subterm_path(Stmt,_,S),
    !,
    query_check_subterms(SS).

%% query_check_matching.
%% query_check_matching(InOrOut, Match).
%%
%%  Check that rule satisfies any input/output matching constraints.
%%
%%  mode query_check_matching.
%%  mode query_check_matching(@atom, @nonvar).
%%
query_check_matching :-
    ip_lookup(query_imatch, IMatch),
    ip_lookup(query_omatch, OMatch),
    query_check_matching(input, IMatch),
    query_check_matching(output, OMatch).

query_check_matching(_, any) :-
    !.
query_check_matching(input, Min-Max) :-
    itrans_input_shape(Shape),
    integer(Shape),
    !,
    CHECK('query_check_matching(input)', integer(Min)),
    Min =< Shape,
    ( integer(Max) ->
        Shape =< Max
    ;
        true
    ).
query_check_matching(output, Min-Max) :-
    itrans_output_shape(Shape),
    integer(Shape),
    !,
    CHECK('query_check_matching(output)', integer(Min)),
    Min =< Shape,
    ( integer(Max) ->
        Shape =< Max
    ;
        true
    ).
query_check_matching(_, _) :-
    %% TO DO: Work out what this comment is talking about. MB 7/9/1998.
    %% this was an old-style rule with a bad shape value.
    %% issue a warning???
    true.

%% query_discharge(CondFlag).
%%
%%  Discharges all the conditions of the rule in the current
%%  itrans operation, using the constraints that have been
%%  recorded by query_process.  Note that itrans_rule_name,
%%  itrans_conditional and itrans_direction must have been
%%  done before this is called.
%%
%%  mode query_discharge(@atom).
%%
query_discharge(u) :-
    itrans_satisfy([], []).
query_discharge(c) :-
    ip_lookup(query_condlist, CondList),
    ip_lookup(query_conjterms, Conjs),
%% TO DO: In query_discharge, use query_conds if it is non-empty.
%% TO DO: In query_discharge, extend itrans_satisfy to use cthys.
    itrans_satisfy(CondList, Conjs).
%% TO DO: Allow conjectures in query_discharge.


%% TO DO: Improve query_choose_theory so it doesn't return same theory twice.
%% See tactics/rulebrowse.ql for a similar task.

%% TO DO: Extend query_choose_theory to handle conj?
%% See tactics/rulebrowse.ql for a similar task.

%% query_choose_theory(ThySpecList, Thy).
%%
%%  mode query_choose_theory(@list(ground), ?atom).
%%
query_choose_theory([T|_], Thy) :-
    query_choose_theory2(T, Thy).
query_choose_theory([_|TS], Thy) :-
    query_choose_theory(TS, Thy).

query_choose_theory2('*', Thy) :-
    !,
    ancestor(Thy).
query_choose_theory2(upto(Max), Thy) :-
    !,
    check_or_error(
        ancestor(Max),
        ['unknown theory: ',q(Max)]),
    ancestor(Thy),
    ancestor_of(Thy,Max).
query_choose_theory2(curr, Thy) :-
    !,
    global_state_lookup(current_dir,Thy), Thy \= none.
query_choose_theory2(Thy, Thy).


query_language header
  'Query Language for Rules'
extends
  concepts
doc
  'Several commands take a `query'' argument, which specifies a rule of',nl,
  'a set of rules.  For example, the `use'' command attempts to find a',nl,
  'rule that satisfies the query and transform the focus with it.  The',nl,
  '`rule'' command displays all rules that satisfy the query.',nl,
  nl,
  'A query may be a single constraint or a set of constraints (in a list',nl,
  'or combined with and/2) that specifies a set of rules..',nl,
  'For example, if the query is `eqv_reflex'', then all ancestor',nl,
  'theories will be searched for a rule called ''eqv_reflex''.',nl,
  'If the query is `[ergo,iprop]:_'', then all rules in the ergo and iprop',nl,
  'theories will be tried.  If the query is `[ergo:_,t,u]'', then all rules',nl,
  'in the ergo theory will be tried, with each rule being used as',nl,
  'an unconditional (u) rule in the true direction (t).',nl,
  nl,
  'The following constraints can be used within queries:',nl,
  nl,
  begin(verb),nl,
  'Constraint           Description',nl,
  '----------           -----------',nl,
  'Thys:Name            short for thys(Thys) and name(Name).',nl,
  'Name                 short for name(Name).',nl,
  'thys(Thys)           Thys specifies the ',
                    xref('theory search list',global,theory_search_list),nl,
  '                     for the primary rule of the transformation.',nl,
  'name(Atom)           the name of the primary rule must be Atom.',nl,
  'name("Str")          the name of the primary rule must contain Str.',nl,
  'f                    apply the rule with direction=f (forward).',nl,
  'r                    apply the rule with direction=r (reverse).',nl,
  't                    apply the rule with direction=t (true).',nl,
  'rt                   apply the rule with direction=rt (rev-true).',nl,
  'c                    apply the rule as being conditional.',nl,
  'u                    apply the rule as being unconditional.',nl,
  'reln(Reln)           rule relation must be as strong as Reln.',nl,
  'frozen(Flag)         Flag=true means freeze all variables in the',nl,
  '                     primary rule, thus preventing instantiations.',nl,
  'matching(Kind)       Constrains how input of rule is matched.',nl,
  '                     Kind=tight means rule input must be nonvar.',nl,
  '                     Kind=loose means rule input must be variable.',nl,
  '                     Kind=frozen freezes all vars in rule statement.',nl,
  '                     Kind=any leaves rule unconstrained and unfrozen.',nl,
  'imatch(Min,Max)      Constrains number of nonvar subterms of input of',nl,
  '                     rule to lie in range Min..Max (inclusive).',nl,
  'omatch(Min,Max)      Constrains number of nonvar subterms of output of',nl,
  '                     rule to lie in range Min..Max (inclusive).',nl,
  'i(Rule)              Rule must discharge the input of the rule.',nl,
  'c(Rule)              Rule must discharge one of the conditions.',nl,
  '                        NB. c(Rule) is not implemented yet.',nl,
  'c(RuleList)          use RuleList to discharge all of the conditions.',nl,
  'iterm(Term)          the input of the rule must unify with Term.',nl,
  'oterm(Term)          the output of the rule must unify with Term.',nl,
  'cterm(Term)          one of the conditions must unify with Term.',nl,
  'subterm(Term)        some subterm of the rule must unify with Term.',
  end(verb).


theory_search_list header
  'Theory Search List'
extends
  concepts
doc
  'One component of a ',xref(query,global,query_language),' for a rule can',nl,
  'be a `theory search list''.  A theory search list specifies a list of',nl,
  'theories which should be searched for a rule.',nl,
  nl,
  'A theory search list may be a single theory name, or a list of',nl,
  'the following terms, which specify alternative theories to search.',nl,
  nl,
  begin(verb),nl,
  'Theory Spec.         Meaning',nl,
  '------------         -------',nl,
  'curr                 search the current theory.',nl,
  'Thy                  search the theory called Thy.',nl,
  '*                    search all ancestor theories.',nl,
  'upto(Thy)            search between the current theory and Thy.',nl,
  'i(N)                 search the theories associated with the symbols',nl,
  '                     of the input term (N levels deep).',nl,
  '                     For example, if the input term is',nl,
  '                        ''TA.a''(''TB.b''(''TC.c'',X),''TD.d''),',nl,
  '                     then',nl,
  '                        i(1) gives [''TA''],',nl,
  '                        i(2) gives [''TA'',''TB'',''TD''],',nl,
  '                        i(3) gives [''TA'',''TB'',''TC'',''TD''].',nl,
  'o(N)                 similar to i(N), but using the output term.',nl,
  'c(N)                 similar to i(N), but using the condition that is',nl,
  '                     currently being discharged (if any).',nl,
  '    NB.  the i(N), o(N) and c(N) specs are not implemented yet.',
  end(verb).


%% private do_constraint(Constraint).
%%
%%  This is a table of all the constraints that can appear
%%  in a query list.  Each Constraint is processed by calling
%%  an itrans_* procedure immediately, or setting up some
%%  query_* implicit parameters for later processing.
%%
%%  mode do_constraint(?nonvar).
%%

%% These are for backwards compatibility only.
do_constraint(input_name(N)) :-
    !,
    do_constraint(i(N)).
do_constraint(input(N)) :-
    !,
    do_constraint(iterm(N)).
do_constraint(output(N)) :-
    !,
    do_constraint(oterm(N)).
do_constraint(matching(M)) :-
    ( M=tight ->
        ip_set(query_imatch, 1-inf)
    ; M=loose ->
        ip_set(query_imatch, 0-0)
    ; M=frozen ->
        do_constraint(frozen(true))
    ; M=any ->
        do_constraint(frozen(false)),
        ip_set(query_imatch, any)
    ;
        fail
    ),
    !.
%%
%% Constraints on the whole statement of the rule.
%%
do_constraint(subterm(T)) :-
    !,
    ip_lookup(query_subterms, L),
    ip_set(query_subterms, [T|L]).
do_constraint(frozen(TF)) :-
    member(TF, [true,false]),
    !,
    ip_set(query_frozen, TF).
%%
%% Constraints on the conditionality and the conditions.
%%
do_constraint(c) :-
    !,
    itrans_conditional(c).
do_constraint(u) :-
    !,
    itrans_conditional(u).
do_constraint(cond(C)) :-
    !,
    itrans_conditional(C).
do_constraint(c(Rules)) :-
    list(Rules),
    !,
    ip_lookup(query_condlist, Rules).
do_constraint(conjs(Terms)) :-
    (
        var(Terms)
    ;
        list(Terms)
    ),
    !,
    ip_set(query_conjterms, Terms).
do_constraint(c(Name0)) :-
    nonvar(Name0),
%% TO DO: Check that Name0 is unique enough in do_constraint.
    !,
    itrans_conditional(c),
    is_hyp_or_theorem(Name0, Name, Stmt),
    !,
    ip_lookup(query_conds, CS),
    ip_set(query_conds, [c(Name,Stmt)|CS]).
do_constraint(cterm(T)) :-
    !,
    itrans_conditional(c),
    ip_lookup(query_conds, CS),
    ip_set(query_conds, [c(_,T)|CS]).
do_constraint(cthys(TS)) :-
    list(TS),
    !,
    ip_set(query_cthys, TS).
%%
%% Constraints on the direction of the rule.
%%
do_constraint(f) :-
    !,
    itrans_direction(f).
do_constraint(r) :-
    !,
    itrans_direction(r).
do_constraint(t) :-
    !,
    itrans_direction(t).
do_constraint(rt) :-
    !,
    itrans_direction(rt).
do_constraint(dirn(Dir)) :-
    !,
    itrans_direction(Dir).
%%
%% Constraints on the relation of the rule.
%%
do_constraint(reln(R)) :-
    !,
    itrans_relation(R).
do_constraint(rulereln(R)) :-
    !,
    itrans_rule_relation(R).
%%
%% Constraints on the input of the rule.
%%
do_constraint(i(Name)) :-
    !,
    itrans_input_name(Name).
do_constraint(iterm(T)) :-
    !,
    itrans_input(T).
do_constraint(imatch(Min,Max)) :-
    !,
    ip_set(query_imatch, Min-Max).
%%
%% Constraints on the output of the rule.
%%
do_constraint(o(_Name)) :-
%% TO DO: Do something with this clause which is only a ! in do_constraint.
%% See tactics/rulebrowse.ql for a similar task.
    !.
do_constraint(oterm(T)) :-
    !,
    itrans_output(T).
do_constraint(omatch(Min,Max)) :-
    !,
    ip_set(query_omatch, Min-Max).
%%
%% Constraints on the name and theory of the rule.
%%
do_constraint(Thys:Name) :-
    atom(Thys),
    Thys = hyp,
    !,
    ip_set(query_thys, [hyp]),
    ip_lookup(query_name, Name).
do_constraint(Thys:Name) :-
    %% Thys \== hyp.
    once(do_constraint(thys(Thys))),
    do_constraint(name(Name)),
    !.
do_constraint(thys(L)) :-
    var(L),
    !,
    ip_set(query_thys, [hyp,*]).
do_constraint(thys(L)) :-
    atom(L),
    !,
    ip_set(query_thys, [L]).
do_constraint(thys(L)) :-
    list(L),
    !,
    ip_set(query_thys, L).
do_constraint(name(S)) :-
    list(S),
    !,
    ip_lookup(query_subnames, L),
    ip_set(query_subnames, [S|L]).
do_constraint(name(S)) :-
    var(S),
    %% ignore, as it does not help to narrow the search.
    !.
do_constraint(name(S)) :-
    atom(S),  %% a complete rulename
    !,
    ip_lookup(query_name, S).
do_constraint(name(S)) :-
    integer(S),  %% a complete rulename
    !,
    ip_lookup(query_name, S).
do_constraint(S) :-
    list(S),
    is_chars(S),
    !,
    ip_lookup(query_subnames, L),
    ip_set(query_subnames, [S|L]).
do_constraint(S) :-
    %% This clause must come *after* the clauses for f,r,t,rt,c,u etc.
    %% We assume that any remaining atoms are rule names.
    atom(S),
    !,
    ip_lookup(query_name, S).
do_constraint(C) :-
    ip_lookup(query_caller, Caller),
    handle_error(['bad constraint ',q(C),' in ',q(Caller),' query.']).
