%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##
 
%%
%% CVS $Id: wi_support.ql,v 1.8 2002/06/28 02:04:57 itp Exp $
%%
%% This file defines support for window inference.
    
%% Initialise the symbol name database.
%%

  
%%
%% Construct rule names from the shape of the goal
%%

%% Transitivity rules.
%%
get_trans_rule_name(Term, RuleName) :-
    table(symbols, Term, WRName),
    concat_atom([WRName, WRName, WRName], '__', RuleName).

%% Reflexivity rules.
%%
get_reflex_rule_name(Term, RuleName) :-
    table(symbols, Term, WRName),
    concat_atom([WRName, reflex], '_', RuleName).

%% Opening rules.
%%
get_open_rule_name(Term, Pos, RuleName) :-
    integer(Pos),!,
    Term = _WR(LHS, _),
    compound(LHS),
    table(symbols, Term, WRName),
    table(symbols, LHS, RName),
    concat_atom([open,RName,Pos,WRName], '_', RuleName).
get_open_rule_name(Term, Pos, RuleName) :-
    (  Pos = b([ObVar]) ; Pos = b(ObVar) ; (obvar(Pos), ObVar = Pos)  ), !,
    Term = _WR(LHS, _),
    LHS = !!_ _ _,
    table(symbols, Term, WRName),
    table(symbols, LHS, RName),
    concat_atom([open,RName,1,WRName], '_', QName),
    RuleName = QName(ObVar).
get_open_rule_name(Term, Pos, RuleName) :-
    Pos = t(N), 
    Term = _WR(LHS, _),
    LHS = !!_ _ _,
    table(symbols, Term, WRName),
    table(symbols, LHS, RName),
    concat_atom([open,RName,t,N,WRName], '_', RuleName).
    
get_trans_true_rule_name(Term, RuleName) :-
    table(symbols, Term, WRName),
    concat_atom([l_, WRName, true], '_', RuleName).

get_trans_rt_rule_name(Term, RuleName) :-
    table(symbols, Term, WRName),
    concat_atom([WRName, rt], '_', RuleName).

get_trans_rt_ante_rule_name(Term, RuleName) :-
    table(symbols, Term, WRName),
    concat_atom([WRName, rt, ante], '_', RuleName).

%%
%% Calculate the arguments for the application of expand_elim
%% to unfold a definition in a window
%%
unfold_defn_args(Window, BaseTerm, !x, SubTerm) :-
    Window = WR(OldTerm, X),
    BaseTerm = WR(!x, X),
    SubTerm = OldTerm,
    !x not_free_in X.
    
     
%%
%% Window top of stack style display
%%

show_top_window(Nodes) :-
    ( Nodes = [Node|_], 
      proof_node_term(Node, Goal),
      nonvar(Goal),
      Goal = WR(_Focus,RHS) ->
        (  WR \== 'jprop.=>', 
           WR \== 'jprop.<=', 
           WR \== 'jprop.<=>', 
           WR \== 'weak_eq.='
        ;
           nonvar(RHS),
           RHS \= 'jprop.true'
        )
     ; 
        true
    ),
    !,
    show_nodes_and_context(Nodes).

show_top_window([Node|_]) :-
    ( setof(c(CName,CId,Node), Node^CTerm^
            ( is_context_name(CName),
              context_search(CName, Node, CId, CTerm)
            ),
            Contexts) ->
        show_contexts(Contexts),
        show_focus(Node)
    ;
        show_focus(Node)
    ).
	
show_focus(Id) :-
    proof_node_term(Id, Term),
    Term = WR(Focus, _),
    get_write_depth(D),
    writelist([nl,'Window relation : ', thy(WR), nl, 'Focus : ', 
               max_depth(D, [thy(Focus)]), nl,nl]).



gumtree command
  start_wi
extends
  proof_tree_cmds
doc
  'Start a simple window inference interface that simply displays the',nl,
  'window relation and focus of the top-level window.'
mode
  start_wi.

start_wi :-
    set_parameter(gumtree_display, open(show_top_window)).

gumtree command
  stop_wi
extends
  proof_tree_cmds
doc
  'Stop the window inference interface.'
mode
  stop_wi.

stop_wi :-
    set_parameter(gumtree_display, open(show_nodes_and_context)).


