%% Support for Xergo interface.
 
?- add_handler(write(Strm, _, Term),
               xergo_write(Strm, Term)),
   new_annotation(comment(writelist), comment_annotation),
%   add_handler(write(Strm, _, emph(WL)),
%               xergo_write_emph(Strm, WL)),
   add_handler(showing_help(Action,Node),
               xergo_showing_help(Action,Node)),
%   add_handler(error(Class, Msg), xergo_error(Class, Msg)),
   add_handler(display_prooftree(Node), xergo_display_prooftree(Node)),
   add_handler(pre_prompt_cmd(R, N), xergo_pre_prompt_cmd(R, N)). 

xergo_pre_prompt_cmd(_,_) :-
    get_xergo_showtree(off), !.
xergo_pre_prompt_cmd(R, N) :-
    xergo_set_current_nodes(N),
    show_prooftree(R).

xergo_set_current_nodes(N) :-
    write('XERGOCURRSTART'),nl,
    length(N, L),
    write(L), write(' '),
    xergo_set_current_nodes_aux(N),
    write(' XERGOCURREND'),nl,
    flush_output.

xergo_set_current_nodes_aux([]).
xergo_set_current_nodes_aux([H|T]) :-
    write(H), write(' '),
    xergo_set_current_nodes_aux(T).
    
/*    
xergo_error(_,Msg) :-
    writelist(user_error, Msg),
    write(user_error, 'XERGO_END_ERROR').
*/

xergo_magic(Command) :-
    write('XERGO'),write(Command),nl.

xergo_showing_help(start,Node) :-
    xergo_magic('HELP'),
    writelist(['%%%%node: ' | Node]),
    nl.
xergo_showing_help(end,_) :- 
    xergo_magic('NORMAL').

xergo_write(Strm, Term) :-
    stream_property(Strm, alias(user_output)),
    xergo_write(Term).
 
xergo_write(xref(WL, Thy, Name)) :-
    writelist(['<a href="', 'silent_cmd(help(', q(Name), ', ', q(Thy), ')).', '">']),
    writelist(WL), 
    writelist(['</a>']).
xergo_write(para) :-
    writelist(['<P>']).
xergo_write(br) :-
    writelist(['<br/>']).
xergo_write(item(T)) :-
    writelist(['<li>']),
    writelist([T]).
xergo_write(begin(list)) :-
    writelist(['<ul>',nl]).
xergo_write(end(list)) :-
    writelist(['</ul>',nl]).
xergo_write(begin(verb)) :-
    writelist(['<pre>', nl]).
xergo_write(end(verb)) :-
    writelist(['</pre>', nl]).
xergo_write(emph(T)) :-
    writelist(['<em>']),
    xergo_write_emph(T), 
    writelist(['</em>']).
xergo_write(code(T)) :-
    writelist(['<code>', T, '</code>']).
   
xergo_write_emph([H|T]) :- 
    !,
    writelist([H|T]).
xergo_write_emph(A) :-
    writelist([A]).

xergo_display_prooftree(Node) :-
    prooftreeIDs(Node, Tree),
    write('XERGOPROOFTREESTART'),nl,
    xergo_dump_prooftree(Tree),
    write(' XERGOPROOFTREEEND'),nl,
    flush_output.

xergo_dump_prooftree(;(T1, T2)) :-
    !,
    write(T1), write(' '),
    length(T2, N),
    write(N), write(' '),
    xergo_dump_prooftree(T2).
xergo_dump_prooftree([]) :- !.
xergo_dump_prooftree([H|T]) :-
    !,
    xergo_dump_prooftree(H),
    xergo_dump_prooftree(T).
xergo_dump_prooftree(T) :-
    write(T),write(' '),write(-1),write(' ').


general command
  xergo_display_node(Node)
extends
  tactic_tables
extends
  general_cmds
doc
  'Display node information to xergo GUI.'
mode
  xergo_display_node(@nodeid).


xergo_display_node(N) :-
    write('XERGONODESTART'),nl,
    show_nodes_and_context([N]),
    nl,
    show_comment_context(N),
    write('XERGONODEEND'),nl.

show_comment_context(N) :-
   annotation(N, comment(WL)),
   !,
   writelist(WL).
show_comment_context(_).

?- assert(xergo_showtree(off)).

xergo_showtree parameter
  [set_xergo_showtree, get_xergo_showtree]
doc
  'This determines if the xergo interface displays the prooftree.'
extends
  parameters.

set_xergo_showtree(X) :-
    ( xergo_showtree(X) ->
        true
    ;
       retract(xergo_showtree(_)),
       assert(xergo_showtree(X))
    ).

get_xergo_showtree(Flag) :-
    xergo_showtree(Flag).

