%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: bools.thy,v 1.7 2002/05/22 05:35:20 itp Exp $

                                %%%%%%%%%%%%%%
				%% Booleans %%
                                %%%%%%%%%%%%%%

%% A theory of the boolean values (true and false).

add_theory(bools).
include_theory(stdops, share).
include_theory(cpred_eq_t, all).


                                %%%%%%%%%%%%%%%
				%% Constants %%
                                %%%%%%%%%%%%%%%

declare bools.   %% The type of true and false.

				 %%%%%%%%%%%%
                                 %% Axioms %%
                                 %%%%%%%%%%%%

axiom bools_def ===
 X:bools <=>
 X = true or X = false.

axiom and_bool ===
  (X and Y):bools.

axiom or_bool ===
   (X or Y):bools.

axiom imp_bool ===
   (X => Y):bools.

axiom all_bool ===
   (all x A):bools.

axiom ex_bool ===
   (ex x A):bools.

%% bools type theorems

theorem not_bool ===
  (not X):bools.

theorem iff_bool ===
   (A <=> B):bools.

theorem fol_bool ===
    (A <= B):bools.

theorem all_t_bool ===
  (all x:T B):bools
    provided x not_free_in T.

theorem ex_t_bool ===
  (ex x:T B):bools
    provided x not_free_in T.

theorem ex_one_bool ===
   (ex_one x A):bools.

theorem ex_one_t_bool ===
  (ex_one x:T B):bools
    provided x not_free_in T.


%% Truth table type theorems.

%% not
theorem not_tt_1 ===
  (A <=> true)  <=> (not A) = false.

theorem not_tt_2 ===
  (A <=> false) <=> (not A) = true.

%% =>
theorem imp_tt_1 ===
  (A <=> true) and (B <=> true)   => (A => B) = true.

theorem imp_tt_2 ===
  (A <=> true) and (B <=> false)  <=> (A => B) = false.

theorem imp_tt_3 ===
  (A <=> false)  and (B <=> true) => (A => B) = true.

theorem imp_tt_4 ===
  (A <=> false)  and (B <=> false) => (A => B) = true.

%% or
theorem or_tt_1 ===
  (A <=> true) and (B <=> true)   => (A or B) = true.

theorem or_tt_2 ===
  (A <=> true) and (B <=> false)  => (A or B) = true.

theorem or_tt_3 ===
  (A <=> false) and (B <=> true)  => (A or B) = true.

theorem or_tt_4 ===
  (A <=> false) and (B <=> false) <=> (A or B) = false.

%% and
theorem and_tt_1 ===
  (A <=> true) and (B <=> true)   <=> (A and B) = true.

theorem and_tt_2 ===
  (A <=> true) and (B <=> false)  => (A and B) = false.

theorem and_tt_3 ===
  (A <=> false) and (B <=> true)  => (A and B) = false.

theorem and_tt_4 ===
  (A <=> false) and (B <=> false) => (A and B) = false.

%% <=>
theorem iff_tt_1 ===
  (A <=> true) and (B <=> true)   => (A <=> B) = true.

theorem iff_tt_2 ===
  (A <=> true) and (B <=> false)  => (A <=> B) = false.

theorem iff_tt_3 ===
  (A <=> false) and (B <=> true)  => (A <=> B) = false.

theorem iff_tt_4 ===
  (A <=> false) and (B <=> false) => (A <=> B) = true.

                                %%%%%%%%%%%%%%
				%% Theorems %%
                                %%%%%%%%%%%%%%

theorem true_bool ===
  true:bools.

theorem false_bool ===
  false:bools.

theorem bools_type ===
  is_type(bools).

theorem eq_false_imp_not ===
   A = false => not A.

theorem eqv_imp_eql ===
  (A:bools and B:bools and (A <=> B)) => A = B.

theorem iff_equiv_class ===
  (A <=> true) or (A <=> false).

theorem not_true_eq_false ===
  (not true) = false.

theorem not_false_eq_true ===
  (not false) = true.

theorem true_and_true_eq_true ===
  (true and true) = true.

theorem and_false_eq_false ===
  (A and false) = false.

theorem false_and_eq_false ===
  (false and A) = false.

theorem true_or_eq_true ===
  (true or A) = true.

theorem or_true_eq_true ===
  (A or true) = true.

theorem false_or_false_eq_false ===
  (false or false) = false.

theorem imp_true_eq_true ===
  (A => true)   = true.

theorem true_imp_false_eq_false ===
  (true => false)  = false.

theorem false_imp_eq_true ===
  (false => A)  = true.

theorem iff_eq_true ===
  (A <=> A)   = true.

theorem true_iff_false_eq_false ===
  (true <=> false)  = false.

theorem false_iff_true_eq_false ===
  (false <=> true)  = false.

theorem open_and_1_eq ===
   hyp +++[C] ---> A <=> B
   ------------
   (A and C) = (B and C).

theorem open_and_2_eq ===
   hyp +++[C] ---> A <=> B
   ------------
   (C and A) = (C and B).

theorem open_or_1_eq ===
   hyp +++[not C] ---> A <=> B
   ------------
   (A or C) = (B or C).

theorem open_or_2_eq ===
   hyp +++[not C] ---> A <=> B
   ------------
   (C or A) = (C or B).

theorem open_imp_1_eq ===
   hyp +++[not C] ---> A <=> B
   ------------
   (A => C) = (B => C).

theorem open_imp_2_eq ===
   hyp +++[C] ---> A <=> B
   ------------
   (C => A) = (C => B).

theorem open_iff_1_eq ===
   A <=> B
   ------------
   (A <=> C) = (B <=> C).

theorem open_iff_2_eq ===
   A <=> B
   ------------
   (C <=> A) = (C <=> B).

theorem open_all_1_eq(x0:obvar) ===
  hyp ??? nfi(x0) ---> B <=> A
  -------------------
  (all x0 B) = (all x0 A).

theorem open_ex_1_eq(x0:obvar) ===
  hyp ??? nfi(x0) ---> B <=> A
  -------------------
  (ex x0 B) = (ex x0 A).

theorem open_allt_1_eq(x0:obvar) ===
  hyp ??? nfi(x0) +++ [x0:T] ---> B <=> A
  -------------------
  (all x0:T B) = (all x0:T A)
    provided x0 not_free_in T.

theorem open_ext_1_eq(x0:obvar) ===
  hyp ??? nfi(x0) +++ [x0:T] ---> B <=> A
  -------------------
  (ex x0:T B) = (ex x0:T A)
    provided x0 not_free_in T.

theorem open_all_t_1_eq(x:obvar) ===
  hyp ??? nfi(x) ---> A = B
  -------------------
  (all x:A X) = (all x:B X).
 
theorem open_ex_t_1_eq(x:obvar) ===
  hyp ??? nfi(x) ---> A = B
  ----------------
  (ex x:A X) = (ex x:B X).  


%%
%% Support for Window Inference - loads
%% bools specific information.
%%

require_tactic_file(wi_support,bools).

