%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: fin.thy,v 1.5 2002/09/09 00:43:56 itp Exp $

%%%%%%%%%%%%%%%%%
%% Finite Sets %%
%%%%%%%%%%%%%%%%%

%% NOTE: This is an old version of this theory and needs to be redone
%% using the new zfc theory.

%% This theory is a theory based on Z's theory of finite sets.

add_theory(fin).
include_theory(stdops, share). 
include_theory(rel, all).
include_theory(func, all).
include_theory(undef, all).
include_theory(localness, all).
include_theory(zfc, all).

%% Add some extra object variables.
declare_object_vars(w).
declare_object_vars(y).
declare_object_vars(z).

%%%%%%%%%%%%%%%
%% Constants %%
%%%%%%%%%%%%%%%

declare finite.

declare finite_partial_functions.

%%%%%%%%%%%%%%%
%% Functions %%
%%%%%%%%%%%%%%%

define up_to(L,U) ===
  (set_of x:ints (L =< x and x =< U)) provided_that
  L:ints and U:ints.

add_operator(525, nonassoc, up_to).

define card(X) ===
  (sm x:nats ex y:sets (y:((1 up_to x) '>-->>' X))) provided_that
  X:finite.

define finite_power(T) ===
  (set_of x:power(T) (x:finite)) provided_that
  T:sets.

define is_finite_partial_function(F,D,R) === %% F in (D -||-> R)
  F:finite and F:(D '-|->' R).

define '-||->'(D,R) ===
  (set_of x is_finite_partial_function(x,D,R)) provided_that
  D:sets and R:sets.

add_operator(525, nonassoc, '-||->').

define is_finite_partial_injection(F,D,R) === %% F in (D >-||-> R) 
  F:finite and F:(D '>-|->' R). %% - one to one

define '>-||->'(D,R) ===
  (set_of x is_finite_partial_injection(x,D,R)) provided_that
  D:sets and R:sets.

add_operator(525, nonassoc, '>-||->').

%% TO DO: In the theory of finite sets, add finite partial surjection and
%% finite total functions.

%%%%%%%%%%%%
%% Axioms %%
%%%%%%%%%%%%

axiom finite_def ===
  X:finite <=>
  X:sets and
  ex x:nats ex y:sets 
    y:(nats '-|->>' X).

axiom f_p_funcs_def ===
  F:finite_partial_functions <=>
  F:finite and
  F:partial_functions.

%%%%%%%%%%%%%%
%% Theorems %%
%%%%%%%%%%%%%%

%% Simplification rules
theorem up_to_undef_A ===
  ? up_to _ = (?).

theorem up_to_undef_B ===
  _ up_to ? = (?).

theorem not_is_finite_undef_A ===
  not (?):finite.

theorem not_is_finite_undef_B ===
  (?):finite <=> false.

theorem card_undef ===
  card(?) = (?).

theorem finite_power_undef ===
  finite_power(?) = (?).

theorem not_is_f_p_func_undef_A ===
  not (?):finite_partial_functions.

theorem not_is_f_p_func_undef_B ===
  (?):finite_partial_functions <=> false.

theorem not_is_f_p_func_undef_C ===
  not (?):(_ '-||->' _).

theorem not_is_f_p_func_undef_D ===
  (?):(_ '-||->' _) <=> false.

theorem f_p_func_undef_A ===
  ? '-||->' _ = (?).

theorem f_p_func_undef_B ===
  _ '-||->' ? = (?).

theorem not_is_finite_partial_injection_undef_A ===
  not (?):(_ '>-||->' _).

theorem not_is_finite_partial_injection_undef_B ===
  (?):(_ '>-||->' _) <=> false.

theorem finite_partial_injection_undef_A ===
  ? '>-||->' _ = (?).

theorem finite_partial_injection_undef_B ===
  _ '>-||->' ? = (?).

theorem card_base ===
  card(empty_set) = 0.

%% Localness theorems

theorem up_to_local_1 ===
  local x x up_to A
    provided x not_free_in A.

theorem up_to_local_2 ===
  local x A up_to x
    provided x not_free_in A.

theorem card_local ===
  local x card(x).

theorem finite_power_local ===
  local x finite_power(x).

theorem is_finite_partial_function_local_1 ===
  local x is_finite_partial_function(x, A, B)
    provided x not_free_in A, x not_free_in B.

theorem is_finite_partial_function_local_2 ===
  local x is_finite_partial_function(A, x, B)
    provided x not_free_in A, x not_free_in B.

theorem is_finite_partial_function_local_3 ===
  local x is_finite_partial_function(A, B, x)
    provided x not_free_in A, x not_free_in B.

theorem fp_fun_local_1 ===
  local x x '-||->' A
    provided x not_free_in A.

theorem fp_fun_local_2 ===
  local x A '-||->' x
    provided x not_free_in A.

theorem is_finite_partial_injection_local_1 ===
  local x is_finite_partial_injection(x, A, B)
    provided x not_free_in A, x not_free_in B.

theorem is_finite_partial_injection_local_2 ===
  local x is_finite_partial_injection(A, x, B)
    provided x not_free_in A, x not_free_in B.

theorem is_finite_partial_injection_local_3 ===
  local x is_finite_partial_injection(A, B, x)
    provided x not_free_in A, x not_free_in B.

theorem fp_inj_local_1 ===
  local x x '>-||->' A
    provided x not_free_in A.

theorem fp_inj_local_2 ===
  local x A '>-||->' x
    provided x not_free_in A.

%% Typing rules.
theorem finite_is_type ===
  is_type(finite).

theorem finite_subtype_of_sets ===
  finite subtype_of sets.

theorem f_p_funcs_is_type ===
  is_type(finite_partial_functions).

theorem f_p_funcs_subtype_of_sets ===
  finite_partial_functions subtype_of sets.

theorem f_p_funcs_subtype_of_finite ===
  finite_partial_functions subtype_of finite.

theorem f_p_funcs_subtype_of_p_funcs ===
  finite_partial_functions subtype_of partial_functions.

theorem is_finite_empty_set ===
  empty_set:finite.

theorem single_is_finite ===
  X:sets <=> single(X):finite.

theorem pair_is_finite ===
  X:sets and Y:sets <=> pair(X,Y):finite.

theorem ordered_pair_is_finite ===
  X:sets and Y:sets <=> ordered_pair(X,Y):finite.

theorem is_finite_is_set ===
  X:finite => X:sets.

theorem up_to_is_finite ===
  L:nats and U:nats <=> (L up_to U):finite.

theorem up_to_is_defined ===
  L:nats and U:nats <=> is_defined((L up_to U)).

theorem up_to_is_undefined_A ===
  is_undefined(L) => is_undefined((L up_to _)).

theorem up_to_is_undefined_B ===
  is_undefined(U) => is_undefined((_ up_to U)).

theorem card_is_nat ===
  X:finite <=> card(X):nats.

theorem card_is_defined ===
  X:finite <=> is_defined(card(X)).

theorem card_is_undefined ===
  is_undefined(X) => is_undefined(card(X)).

theorem finite_power_is_set ===
  T:sets <=> finite_power(T):sets.

theorem finite_power_is_defined ===
  T:sets <=> is_defined(finite_power(T)).

theorem finite_power_is_undefined ===
  is_undefined(T) => is_undefined(finite_power(T)).

theorem in_finite_power_is_finite ===
  X:finite_power(T) => X:finite.

theorem is_f_p_func_empty_set_A ===
  empty_set:finite_partial_functions.

theorem is_f_p_funcs_is_finite ===
  F:finite_partial_functions => F:finite.

theorem is_f_p_funcs_is_p_funcs ===
  F:finite_partial_functions => F:partial_functions.

theorem is_f_p_func_empty_set_B ===
  D:sets and R:sets <=> empty_set:(D '-||->' R).

theorem f_p_func_is_f_p_func ===
  D:sets and R:sets <=> (D '-||->' R):finite_partial_functions.

theorem f_p_func_is_defined ===
  D:sets and R:sets <=> is_defined((D '-||->' R)).

theorem f_p_func_is_undefined_A ===
  is_undefined(D) => is_undefined((D '-||->' _)).

theorem f_p_func_is_undefined_B ===
  is_undefined(R) => is_undefined((_ '-||->' R)).

theorem is_f_p_func_is_f_p_func ===
  F:(_ '-||->' _) => F:finite_partial_functions. %% Reverse direction.

theorem is_f_p_func_is_p_func ===
  F:(D '-||->' R) => F:(D '-|->' R).

theorem is_f_p_func_is_set_A ===
  (_):(D '-||->' _) => D:sets.

theorem is_f_p_func_is_set_B ===
  (_):(_ '-||->' R) => R:sets.

theorem is_finite_partial_injection_empty_set ===
  D:sets and R:sets <=> empty_set:(D '>-||->' R).

theorem finite_partial_injection_is_f_p_func ===
  D:sets and R:sets <=> (D '>-||->' R):finite_partial_functions.

theorem finite_partial_injection_is_defined ===
  D:sets and R:sets <=> is_defined((D '>-||->' R)).

theorem finite_partial_injection_is_undefined_A ===
  is_undefined(D) => is_undefined((D '>-||->' _)).

theorem finite_partial_injection_is_undefined_B ===
  is_undefined(R) => is_undefined((_ '>-||->' R)).

theorem card_finite_partial_injection ===
  F:finite_partial_functions and card(F) = card(range(F)) <=>
  F:(_ '>-||->' _).

theorem is_finite_partial_injection_is_set_A ===
  (_):(D '>-||->' _) => D:sets.

theorem is_finite_partial_injection_is_set_B ===
  (_):(_ '>-||->' R) => R:sets.

theorem is_finite_partial_injection_is_partial_injection ===
  F:(D '>-||->' R) => F:(D '>-|->' R).

theorem in_finite_partial_injection ===
  F:finite and F:(D '>-|->' R) <=>
  F:(D '>-||->' R).

%% Other theorems
theorem f_p_func_empty_set_A ===
  R:sets <=> (empty_set '-||->' R) = empty_set.

theorem f_p_func_empty_set_B ===
  D:sets <=> (D '-||->' empty_set) = empty_set.

theorem finite_partial_injection_empty_set_A ===
  R:sets <=> (empty_set '>-||->' R) = empty_set.

theorem finite_partial_injection_empty_set_B ===
  D:sets <=> (D '>-||->' empty_set) = empty_set.

theorem card_induction ===
  Y not_in X =>
  card(X union single(Y)) = succ(card(X)).

theorem nat_subset_homomorphism_finite ===
  ex x (x:nats and
        ex y y:((1 up_to x) '-|->>' X)) => X:finite.

theorem card_ints_equal_card_nats ===
  card(ints) = card(nats).

theorem card_pos_ints_equal_card_nats ===
  card(pos_ints) = card(nats).

theorem card_pos_ints_equal_card_ints ===
  card(pos_ints) = card(ints).

theorem card_single ===
  X:sets <=> card(single(X)) = 1.

theorem card_pair ===
  X:sets and (Y):sets and X /= Y <=> card(pair(X,Y)) = 2.

theorem card_ordered_pair ===
  X:sets and (Y):sets <=> card(ordered_pair(X,Y)) = 2.

theorem card_total_injection ===
  F:finite =>
  (F:(D '>-->' R) <=>
   F:(D '-->' R) and card(F) = card(range(F))).

theorem card_total_function ===
  F:finite =>
  (F:(D '-->' R) <=>
   F:partial_functions and card(F) = card(D)).

theorem card_bijection ===
  F:finite and F:(D '>-->>' R) =>
  card(F) = card(D) and card(F) = card(R).


%%%%%%%%%%%%%
%%  Types  %%
%%%%%%%%%%%%%

theorem up_to_type ===
  L:ints and U:ints => (L up_to U):power(ints).

theorem card_type ===
  X:finite_power(T) => card(X):nats.


%%%%%%%%%%%%%%%%%%%
%% Opening Rules %%
%%%%%%%%%%%%%%%%%%%

theorem open_up_to_1_eq ===
    A = B
    ----------
    (A up_to C) = (B up_to C).

theorem open_up_to_2_eq ===
    A = B
    ----------
    (C up_to A) = (C up_to B).

theorem open_card_1_eq ===
    A = B
    ----------
    card(A) = card(B).

theorem open_finite_power_1_eq ===
    A = B
    ----------
    finite_power(A) = finite_power(B).

theorem open_is_finite_partial_function_1_eq ===
    A = B
    ----------
    is_finite_partial_function(A, C, D) = is_finite_partial_function(B, C, D).

theorem open_is_finite_partial_function_2_eq ===
    A = B
    ----------
    is_finite_partial_function(C, A, E) = is_finite_partial_function(C, B, E).

theorem open_is_finite_partial_function_3_eq ===
    A = B
    ----------
    is_finite_partial_function(C, D, A) = is_finite_partial_function(C, D, B).

theorem open_fp_fun_1_eq ===
    A = B
    ----------
    (A '-||->' C) = (B '-||->' C).

theorem open_fp_fun_2_eq ===
    A = B
    ----------
    (C '-||->' A) = (C '-||->' B).

theorem open_is_finite_partial_injection_1_eq ===
    A = B
    ----------
    is_finite_partial_injection(A, C, D) = is_finite_partial_injection(B, C, D).

theorem open_is_finite_partial_injection_2_eq ===
    A = B
    ----------
    is_finite_partial_injection(C, A, E) = is_finite_partial_injection(C, B, E).

theorem open_is_finite_partial_injection_3_eq ===
    A = B
    ----------
    is_finite_partial_injection(C, D, A) = is_finite_partial_injection(C, D, B).

theorem open_fp_inj_1_eq ===
    A = B
    ----------
    (A '>-||->' C) = (B '>-||->' C).

theorem open_fp_inj_2_eq ===
    A = B
    ----------
    (C '>-||->' A) = (C '>-||->' B).


%%
%% Support for Window Inference - loads
%% fin specific information.
%%

require_tactic_file(wi_support,fin).



