%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##
%% 
%% CVS $Id: int_oracle.ql,v 1.4 2006/02/01 23:36:26 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file defines the `arith_oracle' oracle.
%% It uses a dc to decide various arithmetic formula.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% arith_oracle_qp(Target, Result).
%%
%%  Use dc to evaluate Target to give Result.  Legal values of Target
%%  are given by the following grammar:
%%
%%    Target <- IntExp < IntExp.
%%    Target <- IntExp = IntExp.
%%    Target <- IntExp =< IntExp.
%%    Target <- IntExp.
%%    IntExp <- succ(IntExp).
%%    IntExp <- - IntExp.
%%    IntExp <- IntExp + IntExp.
%%    IntExp <- IntExp * IntExp.
%%    IntExp <- IntExp div IntExp.
%%    IntExp <- IntExp mod IntExp.
%%    IntExp <- IntExp ** IntExp.
%%    IntExp <- Digit Digit*.
%%    Digit  <- 0|1|2|3|4|5|6|7|8|9
%%
%% mode arith_oracle_qp(@term, ?thy(term)).
%%

%%?- define_macro(dcprog, "/usr/bin/dc"),
%%   define_macro(rmprog, "/bin/rm"),
%%   define_macro(echoprog, "/bin/echo").

arith_oracle_qp(@thy("Exp1 < Exp2"), Result) :-
    arith_oracle_qp(@thy("Exp1 - Exp2"), Diff),
    (Diff = @thy("- param_const('nats.intconst', _)")
      ->
          Result = @thy("true")
    ;
          Result = @thy("false")
    ).
                 
arith_oracle_qp(@thy("Exp1 = Exp2"), Result) :-
    arith_oracle_qp(@thy("Exp1 - Exp2"), Diff),
    (Diff = @thy("param_const('nats.intconst', 0)")
      ->
          Result = @thy("true")
    ;
          Result = @thy("false")
    ).
                 
arith_oracle_qp(@thy("Exp1 =< Exp2"), Result) :-
    arith_oracle_qp(@thy("Exp1 - Exp2"), Diff),
    ((
         Diff = @thy("- param_const('nats.intconst', _)")
     ;
         Diff = param_const('nats.intconst', 0)
     )
      ->
          Result = @thy("true")
    ;
          Result = @thy("false")
    ).
                 
arith_oracle_qp(Formula, Result) :-
    mk_dc_arith_command(Formula, Cmd),
    os(mktemp('/tmp/iarith_oracleXXXXXX', TempFile)),
    atom_codes(TempFile, TempFileCodes),
    append(["/bin/echo '", Cmd, "' | /usr/bin/dc > '", TempFileCodes, "'"],
           ShellCmd),
    atom_codes(ShellCmd1, ShellCmd),
    !,
    os(system(ShellCmd1, 0)),
    !,
    open(TempFile, read, Stream),
    unwind_protect(
        get_line(Stream, Line),
        close(Stream)
                  ),
    
    append(["/bin/rm ", TempFileCodes], RemoveTempFile1),
    atom_codes(RemoveTempFile, RemoveTempFile1),
    os(system(RemoveTempFile)),
    
    (Line = [0'-|Line1]
      ->
        convert_to_intconst(Num, Line1),
        Result = @thy("- param_const('nats.intconst', Num)")
    ;
        convert_to_intconst(Num, Line),
        Result = @thy("param_const('nats.intconst', Num)")
    ).

convert_to_intconst(Num, Line) :-
    length(Line, Len),
    ( Len < 10 ->
         number_codes(Num, Line)
      ;
         atom_codes(Num, Line)
    ).

%% mk_dc_arith_command(Target, Cmd).
%%
%%  Construct a dc command to evaluate Target and printthe result.  Legal
%%  values of Target are given by the following grammar:
%%
%%    Target <- pred(Target).
%%    Target <- succ(Target).
%%    Target <- - Target.
%%    Target <- Target + Target.
%%    Target <- Target * Target.
%%    Target <- Target div Target.
%%    Target <- Target mod Target.
%%    Target <- Target ** Target.
%%    Target <- Digit Digit*.
%%    Digit  <- 0|1|2|3|4|5|6|7|8|9
%%
%% mode mk_dc_arith_command(@term, -list(char)).
%%
mk_dc_arith_command(Formula, Cmd) :-
    mk_dc_arith_command1(Formula, Cmd1),
    append(Cmd1, " p", Cmd).

%% private mk_dc_arith_command1(Target, Cmd).
%%
%%  Construct a dc command to evaluate Target.  Legal values of Target
%%  are given by the following grammar:
%%
%%    Target <- pred(Target).
%%    Target <- succ(Target).
%%    Target <- - Target.
%%    Target <- Target + Target.
%%    Target <- Target * Target.
%%    Target <- Target div Target.
%%    Target <- Target mod Target.
%%    Target <- Target ** Target.
%%    Target <- Digit Digit*.
%%    Digit  <- 0|1|2|3|4|5|6|7|8|9
%%
%% mode mk_dc_arith_command1(@term, -list(char)).
%%
mk_dc_arith_command1(Target, _Cmd) :-
    var(Target),
    !,
    fail.
mk_dc_arith_command1(@thy("succ(_X)"), Cmd) :-
    !,
    mk_dc_arith_command2(_X, '+', param_const('nats.intconst', 1), Cmd).

mk_dc_arith_command1(@thy("_X + _Y"), Cmd) :-
    !,
    mk_dc_arith_command2(_X, '+', _Y, Cmd).

mk_dc_arith_command1(@thy("- _X"), Cmd) :-
    !, % Turn -X into 0 X - (_X will not work because __X = _X in dc).
    mk_dc_arith_command2(param_const('nats.intconst', 0) , '-', _X, Cmd).
mk_dc_arith_command1(param_const('nats.intconst', X), Cmd) :-
    integer(X), X < 0,
    !,
    PX is -X,
    mk_dc_arith_command2(param_const('nats.intconst', 0) , '-', 
                         param_const('nats.intconst', PX), Cmd).
mk_dc_arith_command1(@thy("_X * _Y"), Cmd) :-
    !,
    mk_dc_arith_command2(_X, '*', _Y, Cmd).

mk_dc_arith_command1(@thy("_X div _Y"), Cmd) :-
    !,
    mk_dc_arith_command2(_X, '/', _Y, Cmd).

mk_dc_arith_command1(@thy("_X mod _Y"), Cmd) :-
    !,
    mk_dc_arith_command2(_X, '%', _Y, Cmd).

%mk_dc_arith_command1(@thy("_X ** _Y"), Cmd) :-
%    !,
%    mk_dc_arith_command2(_X, '^', _Y, Cmd).

mk_dc_arith_command1(param_const('nats.intconst', X), Cmd) :-
    atom(X),
    is_bigint(X),
    !,
    atom_codes(X, Cmd).
mk_dc_arith_command1(param_const('nats.intconst', X), Cmd) :-
    integer(X),
    !,
    number_codes(X, Cmd).

%% private mk_dc_arith_command2(X, Op, Y, Cmd).
%%
%%  Construct a dc command to evaluate X Op Y.  Legal values of Op
%%  are:  +, *, /, %, ^.
%%
%% mode mk_dc_arith_command1(@term, *atom, @term, -list(char)).
%%
mk_dc_arith_command2(X, Op, Y, Cmd) :-
    mk_dc_arith_command1(X, Cmd1),
    mk_dc_arith_command1(Y, Cmd2),
    atom_codes(Op, OpCodes),
    append([Cmd1, " ", Cmd2, " ", OpCodes], Cmd).



