%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: ints.thy,v 1.7 2003/12/23 22:13:30 itp Exp $
%%

% A theory about ints based on nats.
% The ints type is used to allow this theory to be used in other theories
% that contain terms of other types.

add_theory(ints).

include_theory(stdops,  share).
include_theory(nats,  all).

declare_object_vars(w).
declare_object_vars(y).
declare_object_vars(z).


declare ints.

declare '-'(_).
add_operator(200, prefix, '-').

% integers axioms

% Ints are nats or the negation of nats
axiom int_type ===
    X:ints <=> X:nats or ex x:nats X = -x
       provided x not_free_in X.

axiom neg0_eq_0 === 0 = -(0).

axiom undef_neg ===
    not X:ints <=> -X = ? .

axiom neg_idempotent_axiom ===
     X:nats
  ------------------
   - (- X) = X.

axiom neg_not_nat ===
     all x:nats all y:nats (y = -x) => x = 0
       provided x not_free_in y.

axiom eq_negs ===
    X:ints, Y:ints
    ----------------------
    X = Y <=> -X = -Y.

close_theory.

% definitions

define '=<'(A,B) === 
   if(A:nats and B:nats, le_nats(A,B),
    if(A:ints and B:nats, true,
     if(A:nats and B:ints, false,
      if(A:ints and B:ints, le_nats(-B, -A), false)))).

add_operator(700, nonassoc, '=<').

define '<'(A,B) === not (A = B) and A =< B.

add_operator(700, nonassoc, '<').

abbreviation '>'(A,B) === B < A.

add_operator(700, nonassoc, '>').

abbreviation '>='(X,Y) ===
  Y =< X.
add_operator(700, nonassoc, '>=').

define diff_nats(A,B) ===
     if(le_nats(B,A), subtract(A,B), if(le_nats(A,B), -subtract(B,A), ?)).

define '+'(A,B) ===
     if(0 =< A and 0 =< B, plus(A,B),
       if(A < 0 and 0 =< B, diff_nats(B,-A),
         if(0 =< A and B < 0, diff_nats(A,-B),
           if(A < 0 and B < 0, -(plus(-A, -B)), ?)))).
        
add_operator(500, leftassoc, '+').

abbreviation '-'(X, Y) === (X  + -(Y)).

add_operator(500, leftassoc, '-').

define '*'(A,B) ===
     if(0 =< A and 0 =< B, mult(A,B),
       if(A < 0 and 0 =< B, - mult(-A, B),
         if(0 =< A and B < 0, - mult(A, -B),
           if(A < 0 and B < 0, mult(-A, -B), ?)))).

add_operator(400, leftassoc, '*').

define div(A, B) ===
       if(0 =< A and 0 < B, div_nats(A,B),
        if(A < 0 and 0 < B, -div_nats(-A,B),
         if(0 =< A and B < 0, -div_nats(A,-B),
          if(A < 0 and B < 0, div_nats(-A,-B), ?)))).

add_operator(400, leftassoc, 'div').

define mod(A,B) === A - B*(A div B).

add_operator(400, leftassoc, 'mod').

define abs(A) === if(A >= 0, A, if(A < 0, -A, ?)).

% theorems

theorem int_induction ===
    (w_local x A) and
    [0/x]A and
    (all x:nats (A => [succ(x)/x]A)) and
    (all y:nats ([-y/x]A => [-succ(y)/x]A))
    =>
    all x:ints A
       provided x not_free_in y, y not_free_in A.

theorem induction_int ===
     w_local x A, 
     [0/x]A,
     hyp ??? nfi(x) +++ [x:nats, A] ---> [succ(x)/x]A,
     hyp ??? nfi(y) +++ [y:nats, [-y/x]A] ---> [-succ(y)/x]A
     --------------------------------------
      all x:ints A
           provided x not_free_in y, y not_free_in A.


theorem neg_undef_undef ===
    - (?) = ? .

theorem nat_is_int ===
     X:nats
   ------------
    X:ints.

theorem neg_ints ===
     X:ints
  ------------------
  not X:nats => ex x:nats not (x = 0) and X = -x
     provided x not_free_in X.

theorem le_imp_int1 ===
      A =< B
    -----------------
      A:ints.

theorem le_imp_int2 ===
      A =< B
   -----------------
      B:ints.

theorem lt_imp_int1 ===
      A < B
    -----------------
      A:ints.

theorem lt_imp_int2 ===
      A < B
   -----------------
      B:ints.

theorem diff_nats_type ===
     A:nats, B:nats
   --------------------
     diff_nats(A, B):ints.

theorem undef_diff_nats ===
     not (A:nats and B:nats) <=> diff_nats(A,B) = ? .

theorem add_undef ===
    not (X:ints and Y:ints) <=> X + Y = ? .

theorem undef_add_undef ===
    ? + X = ? .

theorem add_undef_undef ===
    X + ? = ? .

theorem neg_int_is_int ===
     X:ints
 -----------------
    -X:ints.

theorem neg_nat_not_nat ===
    X:nats, not (X = 0)
   --------------------
    not ((-X):nats).


theorem positive0_is_nat ===
     X >= 0
  -------------
    X:nats.

theorem positive_is_nat ===
     X > 0
  -------------
    X:nats.

theorem neg_is_int ===
    X < 0
 --------------
    X:ints.

theorem neg_negative_is_nat ===
     A < 0
  ----------------
   -A : nats.

theorem add_ints_is_int ===
    X:ints, Y:ints
  ------------------
   X+Y : ints.

theorem int_l__add_int ===
     X+Y:ints
  --------------
    X:ints.

theorem int_r__add_int ===
     X+Y:ints
  --------------
    Y:ints.

theorem diff_nats_size1 ===
     le_nats(A,B)
  -------------------
    diff_nats(A, B) =< 0.

theorem diff_nats_size2 ===
     le_nats(A,B)
  -------------------
    diff_nats(B, A) >= 0.

theorem open_neg_1_eq ===
     X = Y
  --------------
    -X = -Y.

theorem neg_idempotent ===
     X:ints
  ------------------
   - (- X) = X.

theorem neg_not_nat2 ===
    X:nats, Y:nats
  --------------------
    X = -Y => Y = 0.

theorem nat_and_neg_nat_iff_eq0 ===
    0 =< X and 0 =< -X <=> X = 0.

theorem le_nats ===
    X:nats, Y:nats
  --------------------
   X =< Y <=> ex x:nats plus(X, x) = Y
       provided x not_free_in X, x not_free_in Y.

theorem succ_is_add1 ===
    X:nats 
 ---------------------
   succ(X) = X + 1.

theorem lt0_iff_not_nat ===
   X < 0 <=> X:ints and not X:nats.

theorem lt_neg_nat ===
   X:ints, not X:nats, Y:nats
   -------------------------
      X < Y.

theorem not_le_nat_neg ===
      X:ints, not X:nats, Y:nats
   -------------------------
     not (Y =< X).


theorem le_neg_neg ===
    X:ints, not X:nats, Y:ints, not Y:nats
  --------------------
   X =< Y <=> ex x:nats plus(-Y, x) = -X
       provided x not_free_in X, x not_free_in Y.

theorem lt_imp_le ===
   X < Y => X =< Y.

theorem neg_nats_le_ge ===
     X:nats, Y:nats
    ------------------------
    X =< Y <=> -Y =< -X.

theorem neg_negs_le_ge ===
     X:ints, not X:nats, Y:nats
    ------------------------
    X =< Y <=> -Y =< -X.

theorem neg_le_ge ===
     X:ints, Y:ints
   ------------------------
    X =< Y <=> -Y =< -X.

theorem neg_lt_lt ===
     X:ints, Y:ints
   ------------------------
    X < Y <=> -Y < -X.

theorem le_iff_lt_or_eq ===
       X:ints, Y:ints
   -------------------------
    X =< Y <=> X < Y or X = Y.

theorem nats_are_positive ===
    X:nats
 ---------------
   X >= 0.

theorem negs_are_negative ===
    X:ints, not X:nats
  ---------------------
    X < 0.

theorem neg_negs_are_positive ===
    X:ints, not X:nats
  ---------------------
    -X > 0.

theorem ints_are_neg_or_pos ===
     X:ints
  -----------------
   0 =< X or X < 0.

theorem ints_iff_neg_or_pos ===
    X:ints <=> (0 =< X or X < 0).

theorem le_iff_not_gt ===
    X:ints, Y:ints
 --------------------
   X =< Y <=> not (X > Y).

theorem le_reflex ===
    X:ints
  ------------
   X =< X.

theorem add0 ===
    X:ints
 --------------
  X + 0 = X.

theorem eq0_iff_neg_eq0 ===
    X:ints
  ---------------------
   X = 0 <=> -X = 0.

theorem open_diff_nats_1_eq ===
    X = Y
  -----------------
  diff_nats(X,Z) = diff_nats(Y,Z).

theorem open_diff_nats_2_eq ===
    X = Y
  -----------------
  diff_nats(Z,X) = diff_nats(Z,Y).


theorem open_le_1_iff ===
    X = Y
  ---------------
    X =< Z <=> Y =< Z.

theorem open_le_2_iff ===
    X = Y
  ---------------
    Z =< X <=> Z =< Y.

theorem open_lt_1_iff ===
    X = Y
  ---------------
    X < Z <=> Y < Z.

theorem open_lt_2_iff ===
    X = Y
  ---------------
    Z < X <=> Z < Y.

theorem open_add_1_eq ===
       X = Y
  ---------------------
   X + Z = Y + Z.

theorem open_add_2_eq ===
       X = Y
  ---------------------
   Z + X = Z + Y.

theorem minus_eq_0 ===
       X:ints
  ----------------
     X - X = 0.

theorem nat_le_nat__le_nats ===
     X:nats, Y:nats, X =< Y
  -----------------------------
      le_nats(X,Y).


theorem diff_nats0 ===
     A:nats
  -------------------
   diff_nats(A,A) = 0.

theorem diff_nats_plus_eq ===
    A:nats, B:nats
  ------------------------
   diff_nats(plus(A,B), B) = A.

theorem positive_iff_neg_negitive ===
     A:ints
  ---------------------
  0 < A <=> -A < 0.

theorem negative_iff_neg_positive ===
     A:ints
  ---------------------
  A < 0 <=> -A > 0.

theorem id_add ===
    X:ints
 --------------------
   X+Y = X <=> Y = 0.

theorem add_nats_plus ===
    X >= 0, Y >= 0
   ----------------
    X+Y = plus(X,Y).

theorem add_nats_is_nat ===
    X:nats, Y:nats
  --------------------
    X+Y : nats.

theorem add_positives_is_nat ===
    X >= 0, Y >= 0
   ----------------
    X+Y : nats.

theorem nat_add_neg1 ===
    0 =< X, Y < 0, -Y =< X
   ---------------
    X+Y = subtract(X, -Y).

theorem nat_add_neg2 ===
    0 =< X, Y < 0, -Y > X
   ---------------
    X+Y = -subtract(-Y,X).

theorem neg_add_nat1 ===
    0 =< X, Y < 0, -Y =< X
   ---------------
    Y+X = subtract(X, -Y).

theorem neg_add_nat2 ===
    0 =< X, Y < 0, -Y > X
   ---------------
    Y+X = -subtract(-Y,X).

theorem negs_add ===
     X < 0, Y < 0
  -------------------
    X+Y = -plus(-X,-Y).
 
theorem negs_add_is_neg ===
    X < 0, Y < 0
  -----------------
    X + Y < 0.

theorem non_negs_add_is_non_neg ===
    X >= 0, Y >= 0
  -----------------
    X + Y >= 0.

theorem le_antisym ===
    A =< B and B =< A => A = B.

theorem diff_nats_eq_iff_eq ===
    le_nats(B,A), le_nats(C,A)
  -------------------------------
    diff_nats(A,B) = diff_nats(A,C) <=> B = C.

theorem diff_nats_eq_plus_iff_eq ===
    le_nats(B,A), C:nats
  ---------------------------
    diff_nats(A,B) = plus(A,C) <=> B = 0 and C = 0.

theorem add_eq1 ===
    A >= 0, B >= 0, D >= 0
   ------------------------
     A+B = A+D <=> B = D.

theorem add_eq2 ===
    A >= 0, B < 0, D:ints
  ------------------------
     A+B = A+D <=> B = D.

theorem add_eq ===
    A:ints, B:ints, D:ints
  -------------------------
    A+B = A+D <=> B = D.

theorem add_eq_neg ===
    A:ints, B:ints, D:ints
  -------------------------
    B+A = D <=> B = D-A.

theorem add_eq_neg1 ===
    A:ints, B:ints
  -------------------------
    B+A = D <=> B = D-A.

theorem add_eq_neg2 ===
    A:ints, D:ints
  -------------------------
    B+A = D <=> B = D-A.

theorem add_eq_neg3 ===
    B:ints, D:ints
  -------------------------
    B+A = D <=> B = D-A.

theorem le_iff_ex_nat_add ===
    A:ints, B:ints
  ------------------------
   A =< B <=> ex x:nats A+x = B
      provided x not_free_in A, x not_free_in B.

theorem lt_iff_ex_pos_add ===
    A:ints, B:ints
  ------------------------
   A < B <=> ex x:nats not (x = 0) and A+x = B
      provided x not_free_in A, x not_free_in B.

theorem lt_add1_iff_le ===
    A:ints, B:ints
  -------------------
   A =< B <=> A < B+1.

theorem le_le_transitive ===
    A =< B, B =< C
  -------------------
   A =< C.

theorem le_lt_transitive ===
    A =< B, B < C
  -------------------
   A < C.

theorem lt_le_transitive ===
    A < B, B =< C
  -------------------
   A < C.

theorem lt_lt_transitive ===
    A < B, B < C
  -------------------
   A < C.

theorem le_add ===
    A:ints, B:ints, C:ints
 ------------------------------
  A =< B <=> A + C =< B + C.

theorem le_add_neg ===
    A:ints, B:ints, C:ints
 ------------------------------
  A =< B + C <=> A - C =< B.

theorem le_neg_add ===
    A:ints, B:ints, C:ints
 ------------------------------
  A =< B - C <=> A + C =< B.

theorem lt_add ===
    A:ints, B:ints, C:ints
 ------------------------------
  A < B <=> A + C < B + C.

theorem lt_add_neg ===
    A:ints, B:ints, C:ints
 ------------------------------
  A < B + C <=> A - C < B.

theorem lt_neg_add ===
    A:ints, B:ints, C:ints
 ------------------------------
  A < B - C <=> A + C < B.


theorem le_add_le__le ===
    A =< B, C =< D
  ---------------------
   A + C =< B + D.

theorem le_add_lt__lt ===
    A =< B, C < D
  ---------------------
   A + C < B + D.

theorem lt_add_lt__lt ===
    A < B, C < D
  ---------------------
   A + C < B + D.

theorem gt_nats_imp_add_nat_gt ===
      Y:nats, Z:nats, X:nats
   ----------------------------
    Y > Z => Y + X > Z.

theorem add_nats_lt_imp_lt ===
     Y:nats, Z:nats, X:nats
   ----------------------------  
    X + Y < Z => X < Z.

theorem gt_and_non_neg_imp_add_gt ===
    Y > Z and X >= 0 => X + Y > Z.

theorem add_non_neg_lt_imp_lt ===
    X + Y < Z and Y >= 0 => X < Z.


theorem add_nats_lt_iff_nat_lt_subtract ===
    A >= 0, B >= 0, C > 0
   -----------------------
    A + B < C => A < subtract(C,B).

theorem add_nats_gt_iff_nat_gt_subtract ===
    A >= 0, B >= 0, C > 0, B =< C
   -----------------------
    A + B > C => A > subtract(C,B).
 

theorem add_nats0_iff_both0 ===
    X:nats, Y:nats
  ----------------------
   X+Y = 0 <=> X = 0 and Y = 0.

theorem add_commutes ===
    X + Y = Y + X.

theorem neg_add1 ===
    X >= 0, Y >= 0
  --------------------
   -(X+Y) = -X - Y.

theorem neg_add2 ===
    X >= 0, Y < 0, X >= -Y
  ------------------------
   -(X+Y) = -X - Y.

theorem neg_add3 ===
    X >= 0, Y < 0
  -------------------
   -(X+Y) = -X - Y.

theorem neg_add ===
    -(X + Y) = -X - Y.

theorem add_associates1 ===
    X >= 0, Y >= 0, Z >= 0
  ----------------------
   X + (Y + Z) = (X + Y) + Z.

theorem add_associates2 ===
    X >= 0, Y >= 0, Z < 0, Y > -Z
  ----------------------
   X + (Y + Z) = (X + Y) + Z.

theorem add_associates3 ===
    X >= 0, Y >= 0, Z < 0, X+Y < -Z
  ----------------------
   X + (Y + Z) = (X + Y) + Z.

theorem add_associates4 ===
    X < 0, Y < 0, Z < 0
  ----------------------
   X + (Y + Z) = (X + Y) + Z.

theorem add_associates5 ===
    X >= 0, Y >= 0, Z < 0, Y =< -Z, X+Y >= -Z
  ----------------------
   X + (Y + Z) = (X + Y) + Z.

theorem add_associates6 ===
    X >= 0, Y >= 0, Z < 0
  -------------------------
   X + (Y + Z) = (X + Y) + Z.

theorem add_associates7 ===
    X >= 0, Y >= 0, Z:ints
 -------------------------
   X + (Y + Z) = (X + Y) + Z.

theorem add_associates8 ===
   X >= 0, Y:ints, Z:ints
 -------------------------
   X + (Y + Z) = (X + Y) + Z.

theorem add_associates ===
   X + (Y + Z) = (X + Y) + Z.


%% Rules about *

theorem open_prod_1_eq ===
     A = B
  -----------------
   A*C = B*C.

theorem open_prod_2_eq ===
    A = B
  -----------------
   C*A = C*B.

theorem prod_undef ===
   not (A:ints and B:ints) <=> A*B = ? .

theorem undef_prod_undef ===
    ? * X = ? .

theorem prod_undef_undef ===
    X * ? = ? .

theorem prod_type ===
    X:ints, Y:ints
  ------------------
   X*Y : ints.

theorem prod0 ===
    A:ints
  ----------------
   A*0 = 0.

theorem prod1 ===
    A:ints
 ---------------
  A*1 = A.

theorem prod_nats_is_mult ===
    A:nats, B:nats
  ----------------------
   A*B = mult(A,B).

theorem prod_eq0__or_eq0 ===
   A:ints, B:ints
 -------------------------
   A*B = 0 <=> A = 0 or B = 0.

theorem prod_neg_eq_neg_prod ===
    A * -B = -(A * B).

theorem prod_commutes ===
    A * B = B * A.

theorem prod_associates1 ===
    A >= 0, B >= 0, C >= 0
  ------------------------
   A*(B*C) = (A*B)*C.

theorem prod_associates2 ===
    A >= 0, B >= 0, C:ints
  ------------------------
   A*(B*C) = (A*B)*C.

theorem prod_associates3 ===
    A >= 0, B:ints, C:ints
  ------------------------
   A*(B*C) = (A*B)*C.

theorem prod_associates4 ===
    A:ints, B:ints, C:ints
  ------------------------
   A*(B*C) = (A*B)*C.

theorem prod_associates ===
    A*(B*C) = (A*B)*C.

theorem prod_dist_add1 ===
   A >= 0, B >= 0, C >= 0
 ------------------------
    A*(B+C) = A*B + A*C.

theorem prod_dist_add2 ===
   A >= 0, B >= 0, C < 0, B >= -C
 ------------------------
    A*(B+C) = A*B + A*C.

theorem prod_dist_add3 ===
   A >= 0, B >= 0, C < 0
 ------------------------
    A*(B+C) = A*B + A*C.

theorem prod_dist_add4 ===
   A >= 0, B >= 0, C:ints
 ------------------------
    A*(B+C) = A*B + A*C.

theorem prod_dist_add5 ===
   A >= 0, B:ints, C:ints
 ------------------------
    A*(B+C) = A*B + A*C.

theorem prod_dist_add ===
    A*(B+C) = A*B + A*C.

% Theorems about abs

theorem abs_type ===
     A:ints
  ----------------
   abs(A):ints.

theorem abs_undef ===
   not(A:ints) <=> abs(A) = ? .

theorem abs_undef_undef ===
    abs(?) = ? .

theorem open_abs_1_eq ===
      A = B
   -------------------
     abs(A) = abs(B).

theorem abs_pos ===
     A >= 0
  ---------------
    abs(A) = A.

theorem abs_neg ===
     A < 0
  ----------------
    abs(A) = -A.

% Theorems about div and mod

theorem div_type ===
    A:ints, B:ints, not B=0
 ---------------------------
    (A div B):ints.

theorem div_undef ===
    not(A:ints and B:ints and not(B=0)) <=> A div B = ? .

theorem undef_div_undef ===
     ? div X = ? .

theorem div_undef_undef ===
     X div ? = ? .

theorem div0_undef ===
     X div 0 = ? .

theorem open_div_1_eq ===
     X = Y
  -------------------
  X div Z = Y div Z.

theorem open_div_2_eq ===
     X = Y
  -------------------
  Z div X = Z div Y.

theorem zero_div ===
    B:ints, not B = 0
 -----------------------
    0 div B = 0.

theorem div_neg_neg1 ===
    A >= 0, B > 0
 -------------------
  A div -B = -(A div B).

theorem div_neg_neg2 ===
    A >= 0, B:ints, not B = 0
 -------------------
  A div -B = -(A div B).

theorem div_neg_neg3 ===
    A < 0, B > 0
 -------------------
  A div -B = -(A div B).

theorem div_neg_neg4 ===
    A < 0, B:ints, not B = 0
 -------------------
  A div -B = -(A div B).



theorem neg_div_neg1 ===
    A > 0, B > 0
  -----------------------
   -A div B = -(A div B).

theorem neg_div_neg2 ===
    A:ints, B > 0
  -----------------------
   -A div B = -(A div B).

theorem neg_div_neg ===
    -A div B = -(A div B).

theorem div_neg_neg ===
     A div -B = -(A div B).

theorem mod_type ===
    A:ints, B:ints, not B=0
 ---------------------------
    (A mod B):ints.

theorem mod_undef ===
    not(A:ints and B:ints and not(B=0)) <=> A mod B = ? .

theorem undef_mod_undef ===
     ? mod X = ? .

theorem mod_undef_undef ===
     X mod ? = ? .

theorem mod0_undef ===
     X mod 0 = ? .

theorem open_mod_1_eq ===
     X = Y
  -------------------
  X mod Z = Y mod Z.

theorem open_mod_2_eq ===
     X = Y
  -------------------
  Z mod X = Z mod Y.


theorem neg_mod_neg ===
    -A mod B = -(A mod B).

theorem mod_neg_neg ===
     A mod -B = A mod B.

theorem divs_and_mods_eq1 ===
    A > 0, B =< D, C >= 0, E >= 0, C < A, E < A 
   ------------------------------
    A*B+C = A*D+E <=> B = D and C = E. 

theorem divs_and_mods_eq2 ===
    A > 0, B:ints, D:ints, C >= 0, E >= 0, C < A, E < A 
   ------------------------------
    A*B+C = A*D+E <=> B = D and C = E. 


theorem divs_and_mods_eq ===
    A:ints, B:ints, D:ints, C >= 0, E >= 0, C < abs(A), E < abs(A), not A = 0 
   ------------------------------
    A*B+C = A*D+E <=> B = D and C = E. 


theorem div_nats_eq0_1 ===
     A >= 0, B > 0
   ------------------------------
  A div B = 0 <=> A < B.

theorem div_nats_eq0 ===
     A:ints, B:ints, not B = 0
   ------------------------------
  A div B = 0 <=> abs(A) < abs(B).

theorem div_nats_add_prod ===
    A >= 0, C >= 0, C < B
  ---------------------------------
  (A*B+C) div B = A.

theorem add_div_nats_eq_div1 ===
    B > 0, C >= 0
  -------------------------
   (B+C) div B = 1 + (C div B).

theorem add_div_negs ===
    B < 0, C =< 0
  -------------------------
   (B+C) div B = 1 + (C div B).

theorem minnat_property ===
   w_local x P, ex x:nats P
   --------------------------
   all x:nats P => (minnat x P)  =< x.


%%
%% Support for Window Inference - loads
%% nats specific information.
%%

require_tactic_file(wi_support,ints).


