%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% Construct rule names from the shape of the term
%%


%%
%% Commutative rules.
%%

get_comm_rule_name(Term, RuleName) :-
    Term = WR(R,_),
    table(symbols, R, RName),
    table(symbols, Term, WRName),
    (concat_atom([RName, WRName, RName], '_', RuleName);
     concat_atom([RName, comm], '_', RuleName)),
    rule_query(_,RuleName,_,_,_,_,_).

%%
%% Associative rules.
%%

get_assoc_rule_name(Term, RuleName) :-
    Term = _WR(R,_),
    table(symbols, R, RName),
    concat_atom([RName, assoc], '_', RuleName),
    rule_query(_,RuleName,_,_,_,_,_).


%%
%% Flatten Tactic Generator.
%% Assumes AssocL, AssocR, Comm and Enter are _fully_ ground.
%% ac_flatten_r flattens to a right associative expression, and
%% ac_flatten_l flattens to a left associative expression.
%%

ac_flatten_r(Op, Op(Op(A,B),C), Result, Code, NewCode, AssocL, AssocR, Comm, Enter) :-
    !,
    add_code(Code, AssocR, Enter, NewCode0),
    ac_flatten_r(Op, Op(A,Op(B,C)), Result, NewCode0, NewCode, AssocL, AssocR, Comm, Enter).

ac_flatten_r(Op, Op(A, B), Op(A, Result), Code, NewCode, AssocL, AssocR, Comm, Enter) :-
    !,
    ac_flatten_r(Op, B, Result, skip, NewCode0, AssocL, AssocR, Comm, Enter),
    add_code(Code, Enter(2,NewCode0), Enter, NewCode).

ac_flatten_r(_, A, A, Code, Code, _, _, _, _).

ac_flatten_l(Op, Op(A,Op(B,C)), Result, Code, NewCode, AssocL, AssocR, Comm, Enter) :-
    !,
    add_code(Code, AssocL, Enter, NewCode0),
    ac_flatten_l(Op, Op(Op(A,B),C), Result, NewCode0, NewCode, AssocL, AssocR, Comm, Enter).

ac_flatten_l(Op, Op(A, B), Op(Result, B), Code, NewCode, AssocL, AssocR, Comm, Enter) :-
    !,
    ac_flatten_l(Op, A, Result, skip, NewCode0, AssocL, AssocR, Comm, Enter),
    add_code(Code, Enter(1,NewCode0), Enter, NewCode).

ac_flatten_l(_, A, A, Code, Code, _, _, _, _).


%%
%% Sort Tactic Generator.
%% Assumes AssocL, AssocR, Comm and Enter are _fully_ ground.
%%

ac_sort(Op, Op(Op(A,B),C), Result, Code, NewCode, AssocL, AssocR, Comm, Enter, Goal) :-
    !,
    add_code(Code, AssocR, Enter, NewCode0),
    ac_sort(Op, Op(A,Op(B,C)), Result, NewCode0, NewCode, AssocL, AssocR, Comm, Enter, Goal).

ac_sort(Op, Op(A, B), Result, Code, NewCode, AssocL, AssocR, Comm, Enter, Goal) :-
    !,
    ac_sort(Op, B, B0, skip, NewCode0, AssocL, AssocR, Comm, Enter, Goal),
    add_code(Code, Enter(2,NewCode0), Enter, NewCode1),
    ac_filter(Op, Op(A,B0), Result, NewCode1, NewCode, AssocL, AssocR, Comm, Enter, Goal).

ac_sort(_, A, A, Code, Code, _, _, _, _, _).

%%
%% Ordering Tactic Generator.
%% Assumes AssocL, AssocR, Comm and Enter are _fully_ ground.
%%

ac_filter(Op, Op(A,Op(B,C)), Result, Code, NewCode, AssocL, AssocR, Comm, Enter, Goal) :-
    !,
    ac_count_member_conj_r_list(Op, A, C, CountA),
    ac_count_member_conj_r_list(Op, B, C, CountB),
    (ac_compare(A, CountA, B, CountB, Goal)->
        Result = Op(A,Op(B,C)),
        NewCode = Code
     ;
        ac_filter(Op, Op(A,C), Result0, skip, InnerCode, AssocL, AssocR, Comm, Enter, Goal),
        add_code(Code, AssocL, Enter, NewCode0),
        add_code(NewCode0, Enter(1,Comm), Enter, NewCode1),
        add_code(NewCode1, AssocR, Enter, NewCode2),
        add_code(NewCode2, Enter(2,InnerCode), Enter, NewCode),
        Result = Op(B, Result0)
    ).

ac_filter(Op, Op(A, B), Result, Code, NewCode, _AssocL, _AssocR, Comm, Enter, Goal) :-
    !,
    (ac_compare(A,0,B,0,Goal)->
        Result = Op(A,B),
        NewCode = Code
     ;
        Result = Op(B,A),
        add_code(Code, Comm, Enter, NewCode)
    ).

ac_count_member_conj_r_list(Op, A, Op(A,C), Count) :-
    !,
    ac_count_member_conj_r_list(Op, A, C, Count0),    
    Count is 1 + Count0.

ac_count_member_conj_r_list(Op, A, Op(_B,C), Count) :-
    !,
    ac_count_member_conj_r_list(Op, A, C, Count).

ac_count_member_conj_r_list(_Op, A, A, 1) :-
    !.

ac_count_member_conj_r_list(_Op, _A, _B, 0).

% Shouldn't be called
% ac_filter(_, A, A, Code, Code, _, _, _, _, _).


%%
%% Term transformation (to Goal) tactic generator.
%%

ac_set_focus(Op, Term, Goal, Code, NewCode, AssocL, AssocR, Comm, Enter) :-
    (
        ac_unify(Op, Term, Goal);
        ac_unify(Op, Goal, Term)
    ),
    !,
    freeze_term(p(Term,Goal),Frozen),
    ac_flatten_r(Op, Term, Term0, Code, NewCode0, AssocL, AssocR, Comm, Enter),
    ac_unify_collect(Op, Goal, [], NV, [], _),
    ac_sort(Op, Term0, _Term1, NewCode0, NewCode1, AssocL, AssocR, Comm, Enter, NV),
    ac_regroup(Op, Goal, NewCode1, NewCode, AssocL, AssocR, Comm, Enter),
    thaw_term(Frozen).


%%
%% Free variable unifier.
%%

ac_unify(Op, Term1, Term2) :-
    ac_unify_collect(Op, Term1, [], NV1, [], V1),
    ac_unify_collect(Op, Term2, [], NV2, [], V2),
    ac_unify_extract_pairs(V1, V2, V12, V22, [], P),
    ac_unify_set(V12, V13),
    ac_unify_set(V22, V23),
    ac_unify_set(P, P2),
    append(V13, P2, V1F),
    append(V23, P2, V2F),
    ac_unify_do(Op, NV1, NV2, V1F, V2F).

ac_unify_collect(Op, Term, NV0, NV, V0, V) :-
    nonvar(Term),
    Op(A,B) = Term,
    !,
    ac_unify_collect(Op,A,NV0,NV1,V0,V1),
    ac_unify_collect(Op,B,NV1,NV,V1,V).

ac_unify_collect(_Op, Term, NV, NV, V0, V) :-
    thawed_var(Term),
    !,
    V = [Term|V0].

ac_unify_collect(_Op, Term, NV0, NV, V, V) :-
    NV = [Term|NV0].

ac_unify_extract_pairs([], [], [], [], VP, VP) :-
    !.

ac_unify_extract_pairs([H|T], V2, V12, V22, VP0, VP) :-
    (member_eq(H,V2);member_eq(H,VP0)),!,
    ac_unify_extract_pairs(T, V2, V12, V22, [H|VP0], VP).

ac_unify_extract_pairs([H|T], V2, [H|V12], V22, VP0, VP) :-
    !,
    ac_unify_extract_pairs(T, V2, V12, V22, VP0, VP).

ac_unify_extract_pairs([], V2, V12, V22, VP0, VP) :-
    !,
    ac_unify_extract_pairs(V2, [], V22, V12, VP0, VP).

ac_unify_extract_pairs([], [], [], [], VP, VP).

ac_unify_set([],[]) :-
    !.

ac_unify_set([H|T],S) :-
    member_eq(H,T),
    !,
    ac_unify_set(T,S).

ac_unify_set([H|T],[H|S]) :-
    \+ member_eq(H,T),
    !,
    ac_unify_set(T,S).

ac_unify_do(_Op, [], [], [], []) :-
    !.

ac_unify_do(_Op, [], [], [], _V) :-
    !,
    fail.

ac_unify_do(_Op, [], [], [H], [H2]) :-
    !,
    H=H2.

ac_unify_do(Op, [], [], [S], V3) :-
    !,
    ac_unify_do(Op, [], [], V3, [S]).

ac_unify_do(Op, [], [], [H|T], [S]) :-
    !,
    S = Op(H,S2),
    ac_unify_do(Op, [], [], T, [S2]).

ac_unify_do(Op, [], [], [H|T], [H|V2T]) :-
    !,
    ac_unify_do(Op, [], [], T, V2T).

ac_unify_do(Op, [H|T], NV2, V1, V2) :-
    delete(H,NV2,NV22),
    !,
    ac_unify_do(Op, T, NV22, V1, V2).

ac_unify_do(Op, [H], NV2, V1, [H]) :-
    !,
    ac_unify_do(Op, [], NV2, V1, []).

ac_unify_do(Op, [H|T], NV2, V1, [S]) :-
    !,
    S = Op(H,S2),
    ac_unify_do(Op, T, NV2, V1, [S2]).

ac_unify_do(Op, [H], NV2, V1, [H]) :-
    !,
    ac_unify_do(Op, NV2, [], [], V1).

ac_unify_do(Op, [H|T], NV2, V1, [H|V2T]) :-
    !,
    ac_unify_do(Op, T, NV2, V1, V2T).

ac_unify_do(Op, [], NV2, V1, V2) :-
    !,
    ac_unify_do(Op, NV2, [], V2, V1).

%%
%% Reconstruction Tactic Generator.
%% Given a Goal, and a right associative term, it will rearrange the term into
%% the same structure.
%%

ac_regroup(Op, Op(A,B), Code, NewCode, AssocL, AssocR, Comm, Enter) :-
    !,
    ac_regroup1(Op, A, Code, NewCode0, AssocL, AssocR, Comm, Enter),
    ac_regroup(Op, B, skip, InnerCode, AssocL, AssocR, Comm, Enter),
    add_code(NewCode0, Enter(2, InnerCode), Enter, NewCode).

ac_regroup(_Op, _B, Code, Code, _AssocL, _AssocR, _Comm, _Enter).

ac_regroup1(Op, Op(A,B), Code, NewCode, AssocL, AssocR, Comm, Enter) :-
    !,
    ac_regroup1(Op, A, Code, NewCode0, AssocL, AssocR, Comm, Enter),
    ac_regroup1(Op, B, skip, InnerCode, AssocL, AssocR, Comm, Enter),
    add_code(NewCode0, Enter(2,InnerCode), Enter, NewCode1),
    add_code(NewCode1, AssocL, Enter, NewCode).

ac_regroup1(_Op, _, Code, Code, _AssocL, _AssocR, _Comm, _Enter).


%%
%% Term comparator.
%% ac_compare(A,B,G) succeeds if A occurs before B in G.
%%

ac_compare(A,0,_B,_CountB,[H|_T]) :-
    \+ \+ H = A,
    !,
    fail.

ac_compare(_A,_CountA,B,0,[H|_T]) :-
    \+ \+ H = B,
    !.

ac_compare(A,CountA,B,CountB,[H|T]) :-
    \+ \+ H = A,
    !,
    CountA1 is CountA - 1,
    ac_compare(A,CountA1,B,CountB,T).

ac_compare(A,CountA,B,CountB,[H|T]) :-
    \+ \+ H = B,
    !,
    CountB1 is CountB - 1,
    ac_compare(A,CountA,B,CountB1,T).

ac_compare(A,CountA,B,CountB,[_H|T]) :-
    ac_compare(A,CountA,B,CountB,T).

%%
%% [Slightly Optimising] Code Combinator.
%%

add_code(Code, AddCode, Enter, NewCode) :-
    nonvar(Code),
    AddCode = Enter(_,skip),
    NewCode = Code,
    !.

add_code(Code, AddCode, _Enter, NewCode) :-
    nonvar(Code),
    Code = skip,
    NewCode = AddCode,
    !.

add_code(Code, AddCode, _Enter, NewCode) :-
    nonvar(Code),
    NewCode = ';'(Code,AddCode),
    !.

add_code(Code, AddCode, _Enter, NewCode) :-
    var(Code),
    NewCode = AddCode,
    !.
