%% ##include tactic.copyright##

%% Use any intro laws and assumptions.
tactic prop ===
  !(prop1 '|' one).

%% As prop, but fails unless it does something.
tactic prop1 ===
  term(T);
  trace(prop, 1, [thy(T)]);
   (     rule(jprop:assump)
     '|' trace(prop,2,['trying and_intro']); 
         rule(jprop:and_intro);
         trace(prop,1,['Using and_intro']); 
         [prop,prop]
     '|' trace(prop,2,['trying imp_intro']); 
         rule(jprop:imp_intro); 
         trace(prop,1,['Using imp_intro']); 
         [prop]
     '|' trace(prop,2,['trying l__or']); 
         rule(jprop:l__or);
         trace(prop,1,['Using l__or']);
         [prop]
     '|' trace(prop,2,['trying r__or']); 
         rule(jprop:r__or);
         trace(prop,1,['Using r__or']); 
         [prop]
   ).

%% More conservative than prop1: won't explore alternatives.
tactic prop2 ===
  term(T);
  trace(prop2, 1, [thy(T)]);
   (     rule(jprop:assump)
     '|' trace(prop,2,['trying and_intro']); 
         rule(jprop:and_intro);
         trace(prop,1,['Using and_intro']); 
         [prop2a,prop2a]
     '|' trace(prop,2,['trying imp_intro']); 
         rule(jprop:imp_intro); 
         trace(prop,1,['Using imp_intro']); 
         [prop2a]
   ).

%% Auxilliary for prop2
tactic prop2a ===
  !(prop2 '|' one).

%% Use a false assumption.
tactic false ===
  rule(jprop:false); rule(jprop:assump).

%% Use elim laws to build a term to match context hypothesis H,
%% then use the assump law.
tactic elim ===
  elim(_).

tactic elim(H) ===
  context(hyp, H, T);
  trace(elim, 1, ['trying hyp ', w(H), ', term ', thy(T)]);
  elim_term(T);
  [rule(jprop:assump), skip].

%% Use elim laws to build a term to match T.
tactic elim_term(T:thy(term)) ===
  !(     term(T)
     '|' {'jprop:elim_rule'(T, T1, R)};
         elim_term(T1); 
         trace(elim_term, 1, ['Using ', w(R)]);
         [rule(R),skip]
   ).

%% Use imp_elim to introduce and defer an antecedent Term called Name
tactic ante(Term:thy(term), Name) ===
  rule(jprop:imp_elim);
  [rule(jprop:imp_intro),
   name(Name); term(Term); defer
  ].

%% Introduce and defer (variable) antecedents called Names.
tactic antes(Names) ===
  {same_length(Names, Terms),
   list_to_right_assoc(Terms, 'jprop.and', Ante)
  };
  rule(jprop:imp_elim);
  [one,
   term(Ante); conjs(Names); defer
  ].

%% Introduce and defer antecedents Terms, called Names.
tactic antes(Terms:list(thy(term)), Names) ===
  {same_length(Names, Terms),
   list_to_right_assoc(Terms, 'jprop.and', Ante)
  };
  rule(jprop:imp_elim);
  [one,
   term(Ante); conjs(Names); terms(Terms); defer
  ].

%% Use and_intro to split a conjunction into nodes named Names.
tactic conjs(Names) ===
  !(
     {=(Names, [Name])};
     name(Name)
  '|'
     {=(Names, [Name|Names1])};
     rule(jprop:and_intro);
     [name(Name), conjs(Names1)]
  ).

require_tactic_file(prop).
