%% ##include tactic.copyright##


%% The one argument version of skip.

tactic pass( T ) === skip.

%% Some common combinations used in reporting information.

tactic rule_report( S, R ) ===
	rule( R ); trace(S,1,['Rule ',q(R)]).

tactic tactic_report( S, R ) ===
	R; trace(S,1,['Tactic ',q(R)]).

tactic begin_report( S ) ===
	trace(S,2,['Begin ',q(S)]).

tactic fail_report( S ) ===
	trace(S,3,['Failure ',q(S)]); fail.

tactic end_report( S ) ===
	trace(S,3,['End ',q(S)]).

%% Actual tactics.

tactic hyp_unfold( HypNo, Path ) ===
	trace(hyp_unfold,1,['Attempting hyp_replace']);
	rule( hyp_replace(HypNo, Old, New) );
	trace(hyp_unfold,1,['Attempting context search']);
	{context_search(hyp, _, HypNo, Old)};
	trace(hyp_unfold,1,['About to split into two']);
	[ one; skip, term(New); unfold(_); term( Old ); rule(assump(HypNo)) ].

tactic jprop_prim_1( T ) ===
	(
	    ( rule( imp_intro ); [one; T(T), T(T)] )
	'|' ( rule( imp_intro ); T(T) )
	'|' ( rule( expansion_elim( !x, !x, _ ) ); T(T) )
	'|' rule( assump )
	) '|' skip.

tactic jprop_prim_2 ===
	( jprop_prim_1 '|' skip ).



tactic jprop_1( T ) ===
	( trace(jprop_1,1,['Begin jprop_1']);
	  (
	    (
	      (
%% If we are finished, do nothing.
		  zero
%% TO DO: Add case where more than one node is in focus?

%% Get rid of true.
	      '|' ( rule_report( jprop_1, true ) )
%% Get rid of ( A and B ) as a hypothesis.
	      '|' ( rule_report( jprop_1, and_elim_c ); T(T) )
%% Swap ( A => B ) for ( B ) as a hypothesis if ( A ) is a hypothesis.
	      '|' ( rule_report( jprop_1, hyp_imp_intro ); T(T) )
%% Some simple theorems.
	      '|' ( rule_report( jprop_1, imp_imp ) )
	      '|' ( rule_report( jprop_1, imp_reflex ) )
	      '|' ( rule_report( jprop_1, iff_reflex ) )
	      '|' ( rule_report( jprop_1, imp_true ) )
	      '|' ( rule_report( jprop_1, l__and_true ); T(T) )
	      '|' ( rule_report( jprop_1, r__true_and ); T(T) )
	      '|' ( rule_report( jprop_1, r__true_imp ); T(T) )
	      '|' ( rule_report( jprop_1, l__iff_true ); T(T) )
	      '|' ( rule_report( jprop_1, r__true_iff ); T(T) )
%% Unfold definitions
%%	      '|' ( rule_report( jprop_1, 'expansion_elim( _, _, _ )' ); T(T) )
%% Begin rearrangement
	      '|' ( rule_report( jprop_1, imp__fol ); T(T) )
	      '|' ( rule_report( jprop_1, imp__imp__iff ); [one; T(T), T(T)] )
	      '|' ( rule_report( jprop_1, imp_intro ); T(T) )
%% Get rid of ( A or B ) as a hypothesis.
	      '|' ( rule_report( jprop_1, or_elim_c ); [one; T(T), T(T)] )
%% Use assump.
	      '|' ( rule_report( jprop_1, assump ) )
%% Attempt to prove conjuncts.
	      '|' ( rule_report( jprop_1, imp_intro ); [one; T(T), T(T)] )
%% Attempt to prove one of the disjuncts.
	      '|' ( rule_report( jprop_1, l__or ); T(T) )
	      '|' ( rule_report( jprop_1, r__or ); T(T) )
	      ); trace(jprop_1,1,['End jprop_1'])
	    ) %%%  '|' trace(jprop_1,1,['Failed jprop_1'])
	  )
	).

tactic jprop_2 ===
	( trace(jprop_2,1,['Begin jprop_2']);
	  ( jprop_1 '|' skip );
	  trace(jprop_2,1,['End jprop_2'])
	).

tactic jprop_window ===
	(
	    ( trace(jprop_window,1,['Attempting lift_unary']); lift_unary( jprop_2 ); trace(jprop_window,1,['Finished lift_unary']) )
	'|' ( trace(jprop_window,1,['Attempting lift_binary']); lift_binary( jprop_2 ); trace(jprop_window,1,['Finished lift_binary']) )
	).


