%% ##include tactic.copyright##

%% Window inference

%% TO DO: The rules imp_trans, fol_trans and so on must be updated to their
% new names.

require_global_tactic_file(tables).

declare_table(trans, wi_trans_table).
erase_table(trans).

wi_trans_table doc
  'Table of tactics for unconditional `trans'' steps.',nl,
  'Each tactic applies to a term of the form WRel(In,Out),',nl,
  'and produces two proof nodes RRel(LHS,RHS) and WRel(Mid,Out).',nl,
  'For the forward direction, RRel >= WRel (i.e., can strengthen the',nl,
  '  relation), In=LHS and Mid=RHS.',nl,
  'For the reverse direction, RRel >= rev(WRel), In=RHS and Mid=LHS.',nl,
  'The extra arg of the table is args(Dirn,Mid,RRel).'.

%% TO DO: Check if the following rules of transitivity are the ones that
%% are meant from the original version.
%% Change imp_trans --> imp__imp__imp and similarly for fol and iff.

% forwards
in_table(trans, rule(jprop:imp__imp__imp),     
         In=>Out,  args(f, Mid, 'jprop.=>')).
in_table(trans, rule(jprop:fol__fol__fol), 
         In<=Out,  args(f, Mid, 'jprop.<=')).
in_table(trans, rule(jprop:iff__iff__iff),     
         In<=>Out, args(f, Mid, 'jprop.<=>')).

% forwards with strengthening
%% TO DO: Prove derived rules for forwards window inference.
in_table(trans, (rule(jprop:imp__imp__imp);[rule(jprop:iff__imp),one]),
         In=>Out,  args(f, Mid, 'jprop.<=>')).
in_table(trans, (rule(jprop:fol__fol__fol);[rule(jprop:iff__fol),one]),
         In<=Out,  args(f, Mid, 'jprop.<=>')).

% backwards
%% TO DO: Prove derived rules for backwards window inference.
in_table(trans, (rule(jprop:imp__imp__imp);
                 [term(In=>Mid);rule(jprop:expansion_elim(x,x,_));term(Mid<=In),
                  one]),
         In=>Out,  args(r, Mid, 'jprop.<=')).
in_table(trans, (rule(jprop:fol__fol__fol);
                 [term(In<=Mid);rule(jprop:expansion_intro(x,x,_));term(Mid=>In),
                  one]),
         In<=Out,  args(r, Mid, 'jprop.=>')).

declare_table(open, wi_open_table).
erase_table(open).

wi_open_table doc
  'Table of window opening tactics.',nl,
  'Each tactic applies to a term of the form',nl,
  '   Rel(Fun(X1,...,Xi,...,Xn), Goal)',nl,
  'and leaves subproofs of the form',nl,
  '   Rel''(Xi,Yi) and Rel(Fun(X1,...,Yi,...,Xn),Goal).',nl,
  'The extra arg of the table is the argument position i.'.

in_table(open, rule(jprop:open_and_1_imp),     (A and B) =>  (A1 and B), 1).
in_table(open, rule(jprop:open_and_1_fol),     (A and B) <=  (A1 and B), 1).
in_table(open, rule(jprop:open_and_1_iff),         (A and B) <=> (A1 and B), 1).
in_table(open, rule(jprop:open_and_2_imp),     (A and B) =>  (A and B1), 2).
in_table(open, rule(jprop:open_and_2_fol),     (A and B) <=  (A and B1), 2).
in_table(open, rule(jprop:open_and_2_iff),         (A and B) <=> (A and B1), 2).

in_table(open, rule(jprop:open_or_1_imp),      (A or B)  =>  (A1 or B),  1).
in_table(open, rule(jprop:open_or_1_fol),      (A or B)  <=  (A1 or B),  1).
in_table(open, rule(jprop:open_or_1_iff),          (A or B)  <=> (A1 or B),  1).
in_table(open, rule(jprop:open_or_2_imp),      (A or B)  =>  (A or B1),  2).
in_table(open, rule(jprop:open_or_2_fol),      (A or B)  <=  (A or B1),  2).
in_table(open, rule(jprop:open_or_2_iff),          (A or B)  <=> (A or B1),  2).

in_table(open, rule(jprop:open_imp_1_imp), (A => B)  =>  (A1 => B),  1).
in_table(open, rule(jprop:open_imp_1_fol), (A => B)  <=  (A1 => B),  1).
in_table(open, rule(jprop:open_imp_1_iff),     (A => B)  <=> (A1 => B),  1).
in_table(open, rule(jprop:open_imp_2_imp), (A => B)  =>  (A =>  B1), 2).
in_table(open, rule(jprop:open_imp_2_fol), (A => B)  <=  (A =>  B1), 2).
in_table(open, rule(jprop:open_imp_2_iff),     (A => B)  <=> (A =>  B1), 2).

in_table(open, rule(jprop:open_not_1_imp),     not(A)    =>  not(A1),    1).
in_table(open, rule(jprop:open_not_1_fol),     not(A)    <=  not(A1),    1).
in_table(open, rule(jprop:open_not_1_iff),         not(A)    <=> not(A1),    1).

in_table(open, rule(jprop:open_iff_1_iff),         (A <=> B) <=> (A1 <=> B), 1).
in_table(open, rule(jprop:open_iff_2_iff),         (A <=> B) <=> (A <=> B1), 1).



%% Turn A into A<=true.
tactic wi_start ===
  rule(jprop:imp_elim);
  [
    rule(jprop:expansion_intro(x,x,'jprop.<='(_,'jprop.true')))
  ,
    rule(jprop:true)
  ].

%% Turn (A<=true) into A.
tactic wi_end ===
  term('jprop.<='(A,'jprop.true')); 
  rule(unfold(x,x,_)); 
  term('jprop.=>'('jprop.true',A)); 
  rule(imp_intro).

%% Rule is C and ... and C => RuleRel(LHS,RHS).
tactic trans(Rule) === 
  trans(Rule, _, _, _).
tactic trans(Rule, Dirn) === 
  trans(Rule, Dirn, _, _).
tactic trans(Rule, Dirn, Goal:thy(term)) === 
  trans(Rule, Dirn, Goal, _).
tactic trans(Rule, Dirn, Goal:thy(term), RuleRel) === 
  table(trans, args(Dirn,Goal,RuleRel));
  [
     rule(Rule)
  ,
     one
  ].

%% Rule is C and ... and C => RuleRel(LHS,RHS).
tactic c_trans(Rule, Dirn, Goal:thy(term), RuleRel) ===
  table(trans, args(Dirn,Goal,RuleRel));
  [
     rule(jprop:imp_elim);
     [
        rule(Rule)
     ,
        conjs(_)
     ] 
  ,
     one
  ].

%% Use a rule of the form A=>B to deduce B from A
tactic crule(Rule) ===
  rule(jprop:imp_elim);
  [rule(Rule), one].

tactic open(Pos) ===
  table(trans);
  [ table(open, Pos), one ].

tactic open(Pos, Goal:thy(term)) ===
  table(trans);
  [ table(open, Pos); term(_(_,Goal)), one ].
