%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: localness.thy,v 1.13 2002/08/05 01:46:12 itp Exp $
%%
%% Minimal Predicate Calculus with Weak Equality and Localness
%% ===========================================================
%%
%% This theory adds a conditional form of the replacement axiom of 
%% equality, giving a standard theory of equality for those operators
%% that satisfy the 'localness' conditions.
%%
%% We do not define multiple-position versions of these quantifiers
%% (eg. local [x,y] x+y), because it is better and more flexible to
%% specify the properties of each position separately.
%% That is, use (local x x+_) and (local y _+y).


add_theory(localness).
include_theory(stdops,  share).
include_theory(jpred_eq,   all).
close_theory.

%% `Weak' localness.
%%
define !!w_local [x] Body ===
    all x all x2 (
      x = x2 => 
      (Body <=> [x2/x]Body)).


%% The most common kind of localness.
%%
define !!local [x] Body ===
  all x all x2  (x2 = x => (Body = [x2/x]Body)).


%% `Strong' localness.
%%
define !!s_local [x] Body ===
  all x all x2 (
    (x2 <=> x) =>
      (Body = [x2/x]Body)).



theorem local_nfi ===
     local x A
       provided x not_free_in A.

%% Opening Rules
%% =============

theorem open_w_local_1_iff(x:obvar) ===
    hyp ??? nfi(x) ---> B <=> B2
    ------------------------------
    (w_local x B) <=> (w_local x B2).

theorem open_local_1_iff(x:obvar) ===
    hyp ??? nfi(x) ---> B = B2
    ------------------------------
    (local x B) <=> (local x B2).

theorem open_s_local_1_iff(x:obvar) ===
    hyp ??? nfi(x) ---> B = B2
    ------------------------------
    (s_local x B) <=> (s_local x B2).


%% Some one-point theorems hold for local places.
%%

theorem one_pt_all(x:obvar) ===
    (local x A)
    => ((all x (x = B => A))
        <=> [B/x]A)
    provided
      x not_free_in B.

theorem one_pt_all_imp(x:obvar) ===
    (local x C) and (local x D) 
    => ((all x (x = A) and C => D) <=> ([A/x]C => [A/x]D))
       provided x not_free_in A.

theorem one_pt_all_and_imp(x:obvar) ===
    (local x C) and (local x D) 
    => ((all x ((x = A) or B) and C => D)
        <=>
        (([A/x]C => [A/x]D) and all x B and C => D))
       provided x not_free_in A.

theorem one_pt_ex(x:obvar) ===
    (local x C) 
    => ((ex x (x = A) and C) <=> [A/x]C)
         provided x not_free_in A.
                                
theorem one_pt_ex_or(x:obvar) ===
    (local x C) =>
     ((ex x ((x = A) or B) and C) <=> ([A/x]C or ex x B and C))
       provided x not_free_in A.

theorem one_pt_ex_one(x:obvar) ===
    (local x C)
    => ((ex_one x ((x = A) and C)) <=> [A/x]C)
       provided x not_free_in A.
 
theorem one_pt_all_rev(x:obvar) ===
    (local x A)
    => ((all x (B = x => A))
        <=> [B/x]A)
    provided
      x not_free_in B.
 
theorem one_pt_ex_rev(x:obvar) ===
    (local x C)
    => ((ex x (A = x) and C) <=> [A/x]C)
         provided x not_free_in A.
 
theorem one_pt_ex_or_rev(x:obvar) ===
    (local x C) =>
     ((ex x ((A = x) or B) and C) <=> ([A/x]C or ex x B and C))
       provided x not_free_in A.    

theorem one_pt_ex_one_rev(x:obvar) ===
    (local x C)
    => ((ex_one x ((A = x) and C)) <=> [A/x]C)
       provided x not_free_in A.
                                                                                
theorem local_substitutivity(A:thy(term),x:obvar) ===
    (local x A),
    B = C
    --------------------
    [B/x]A = [C/x]A.

