%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: nats.thy,v 1.3 2003/05/23 01:42:32 itp Exp $
%%

%% A theory of natural numbers based on the Peano axioms.
%% The type nats and axioms about the nats type are supplied. 
%% The Peano axioms are all relative to the nats type so that
%% this theory can be used in other theories that contain other
%% types.

add_theory(nats).

include_theory(stdops,  share).
include_theory(ipred_eq_t,  all).
include_theory(undef,   all).
include_theory(sm,  all).
include_theory(localness,  all).

declare_object_vars(w).
declare_object_vars(y).
declare_object_vars(z).

require_tactic_file(big_int).

declare nats.

declare succ(_).

declare plus(_,_).

declare mult(_,_).

declare_param_const(intconst,I,is_bigint(I)).

abbreviation I ===
  param_const(intconst,I) provided
  is_bigint(I).

% Peano axioms (+ types)

axiom zero_is_nat === 0:nats.

axiom one_is_succ0 ===
     1 = succ(0).

axiom succ_nat_is_nat ===
       X:nats
   -----------------
    succ(X):nats.

axiom undef_succ ===
      not X:nats <=> succ(X) = ? .

axiom plus_nats_is_nat ===
     X:nats, Y:nats
  -----------------------
    plus(X,Y):nats.

axiom undef_plus ===
     not (X:nats and Y:nats) <=> plus(X,Y) = ? .

axiom mult_nats_is_nat ===
     X:nats, Y:nats
  -----------------------
    mult(X,Y):nats.

axiom undef_mult ===
     not (X:nats and Y:nats) <=> mult(X,Y) = ? .

axiom succ_local ===
      local x succ(x).

axiom plus_local_1 ===
      local x plus(x, A).

axiom plus_local_2 ===
      local x plus(A, x).

axiom mult_local_1 ===
      local x mult(x, A).

axiom mult_local_2 ===
      local x mult(A, x).

axiom succ_eq1 ===
      X:nats, Y:nats
    ------------------
     succ(X) = succ(Y) => X = Y.

axiom succ_nat ===
     X:nats
   -----------------
   not (succ(X) = 0).

axiom plus0 ===
     X:nats
  --------------
    plus(X, 0) = X.

axiom plus_succ ===
     X:nats, Y:nats
   ----------------------
    plus(X, succ(Y)) = succ(plus(X,Y)).

axiom mult0 ===
     X:nats
  --------------
    mult(X, 0) = 0.

axiom mult_succ ===
     X:nats, Y:nats
   ----------------------
     mult(X, succ(Y)) = plus(mult(X, Y), X).

axiom nat_induction ===
    (w_local x A) and
    [0/x]A and
    (all x:nats (A => [succ(x)/x]A))
    =>
    all x:nats A.
    
close_theory.

% definitions

%% less than or equal to for naturals
define le_nats(A,B) ===
     A:nats and B:nats and ex x:nats plus(A, x) = B.

%% subtracting naturals
define subtract(A,B) ===
     if(le_nats(B,A), sm x:nats plus(B, x) = A, ?).


%% div for naturals
define div_nats(A,B) ===
     if(B = 0, ?, 
       if(A:nats and B:nats, 
              sm x:nats ex y:nats 
                  not (y = B) and le_nats(y, B) and A = plus(mult(B,x),y), ?)).

%% minimum natural satisfying property

add_operator(890, quant, minnat).

define minnat x P ===
     sm x:nats (P and all y:nats ([y/x]P => le_nats(x, y))).

% theorems

theorem succ_eq ===
      X:nats, Y:nats
    ------------------
     X = Y <=> succ(X) = succ(Y).

theorem one_is_nat === 1:nats.

theorem nats_property ===
   X:nats <=> X = 0 or ex x:nats X = succ(x)
       provided x not_free_in X.

theorem open_succ_1_eq ===
      X = Y
   -----------------
    succ(X) = succ(Y).

theorem induction_nat ===
     w_local x A,
     [0/x]A,
     hyp ??? nfi(x) +++ [x:nats, A] ---> [succ(x)/x]A
     --------------------------------------
      all x:nats A.

theorem zero_plus ===
    X:nats
  ----------------
   plus(0, X) = X.

theorem plus_eq1 ===
    X:nats, Y:nats, Z:nats
   ------------------------
   X = Y <=> plus(X,Z) = plus(Y,Z).

theorem succ_is_plus1 ===
    X:nats
  ---------------
  succ(X) = plus(X,1).

theorem succ_not_id ===
    X:nats
 -----------------
  not (X = succ(X)).


theorem zero_is_plus0s ===
    X:nats, Y:nats
 -----------------------
  X = 0 and Y = 0 <=> plus(X,Y) = 0.

theorem zero_mult0 ===
    X:nats, Y:nats
 -------------------
  (X = 0 or Y = 0) <=> mult(X,Y) = 0.

theorem plus_eq2 ===
    X:nats, Y:nats, Z:nats
   ------------------------
   X = Y <=> plus(Z,X) = plus(Z,Y).

theorem mult_eq1 ===
    X:nats, Y:nats, Z:nats, not (Z = 0)
   ------------------------
   X = Y <=> mult(X,Z) = mult(Y,Z).

theorem mult_eq2 ===
    X:nats, Y:nats, Z:nats, not (Z = 0)
   ------------------------
   X = Y <=> mult(Z,X) = mult(Z,Y).


theorem open_natplus_1_eq ===
      X = Y
  ----------------------
    plus(X,Z) = plus(Y,Z).

theorem open_natplus_2_eq ===
      X = Y
   --------------------
    plus(Z,X) = plus(Z,Y).

theorem succ_plus ===
     X:nats, Y:nats
   --------------------
    plus(succ(X),Y) = succ(plus(X,Y)).

theorem plus_commutes ===
     X:nats, Y:nats
   --------------------
    plus(X,Y) = plus(Y,X).

theorem plus_associates ===
    X:nats, Y:nats, Z:nats
  --------------------------
   plus(X,plus(Y,Z)) = plus(plus(X,Y), Z).

theorem zero_mult ===
    X:nats
  ----------------
   mult(0, X) = 0.

theorem id_plus ===
   X:nats
 ---------------
  plus(X,Y) = X <=> Y = 0.

theorem open_natmult_1_eq ===
      X = Y
  ----------------------
    mult(X,Z) = mult(Y,Z).

theorem open_natmult_2_eq ===
      X = Y
   --------------------
    mult(Z,X) = mult(Z,Y).

theorem succ_mult ===
     X:nats, Y:nats
   --------------------
    mult(succ(X),Y) = plus(Y, mult(X,Y)).

theorem one_mult ===
    X:nats
  ----------------
   mult(1, X) = X.

theorem mult_commutes ===
     X:nats, Y:nats
   --------------------
    mult(X,Y) = mult(Y,X).

theorem mult_associates ===
    X:nats, Y:nats, Z:nats
  --------------------------
   mult(X,mult(Y,Z)) = mult(mult(X,Y), Z).

theorem mult_dist_plus ===
    X:nats, Y:nats, Z:nats
  --------------------------
  mult(X, plus(Y,Z)) = plus(mult(X,Y), mult(X,Z)).

theorem plus1_or_plus2 ===
    A:nats, B:nats, not (A = B)
  -------------------------------
   (ex x:nats plus(A,x) = B)
    <=>
   not (ex x:nats plus(B,x) = A)
     provided x not_free_in A, x not_free_in B.

theorem le_nats_or ===
     A:nats, B:nats
  --------------------------
   le_nats(A,B) or le_nats(B,A).

theorem le_nats_imp_nat1 ===
     le_nats(A,B)
   --------------------
       B:nats.

theorem le_nats_imp_nat2 ===
     le_nats(A,B)
   --------------------
       A:nats.

theorem subtract_type ===
     le_nats(B,A)
  --------------------
     subtract(A,B):nats.

theorem undef_subtract ===
     not le_nats(B,A) <=> subtract(A,B) = ? .


theorem zero_le_nats ===
     X:nats
  -----------------
   le_nats(0, X).

theorem le_nats_succ ===
     X:nats
 -------------------
   le_nats(X, succ(X)).

theorem zero_ge_nats ===
     X:nats
  ----------------
   le_nats(X, 0) <=> X = 0.

theorem le_nats_antisym ===
    X:nats, Y:nats, not (X = Y)
   -----------------------------
    le_nats(X,Y) <=> not le_nats(Y,X).

theorem open_le_nats_1_iff ===
    X = Y
  --------------
   le_nats(X,Z) <=> le_nats(Y,Z).

theorem open_le_nats_2_iff ===
    X = Y
  --------------
   le_nats(Z,X) <=> le_nats(Z,Y).

theorem open_subtract_1_eq ===
    X = Y
  -----------------
  subtract(X,Z) = subtract(Y,Z).

theorem open_subtract_2_eq ===
    X = Y
  -----------------
  subtract(Z,X) = subtract(Z,Y).



theorem le_nats_reflex ===
    X:nats
----------------
  le_nats(X,X).

theorem subtract_eq0 ===
    X:nats
  -------------
  subtract(X,X) = 0.

theorem subtract0 ===
    X:nats
 ---------------
  subtract(X,0) = X.

theorem plus_le_nats_imp_le_nats1 ===
    A:nats, B:nats, C:nats
  -------------------------
   le_nats(plus(A,B),C) => le_nats(A,C).

theorem plus_le_nats_imp_le_nats2 ===
    A:nats, B:nats, C:nats
  -------------------------
   le_nats(plus(A,B),C) => le_nats(B,C).


theorem le_nats_antisym2 ===
   le_nats(A,B) and le_nats(B,A) => A = B.

theorem le_nats_transitive ===
    le_nats(A,B), le_nats(B,C)
    --------------------------
        le_nats(A,C).

theorem le_nats_plus ===
    A:nats, B:nats, C:nats
   -------------------------
    le_nats(A,B) <=> le_nats(plus(A,C), plus(B,C)).

theorem le_nats_plus_le_nats ===
      le_nats(A,B), le_nats(C,D)
   --------------------------------
      le_nats(plus(A,C), plus(B,D)).

theorem le_nats_plus_nat_le_nats ===
       C:nats
  ---------------------------
   le_nats(A,B) => le_nats(A, plus(C,B)).

theorem le_nats_iff_le_nats_mult ===
    A:nats, B:nats, C:nats, not C = 0
 -------------------------------
   le_nats(A,B) <=> le_nats(mult(C,A), mult(C,B)).

theorem plus_subtract1 ===
     A:nats, B:nats, C:nats, le_nats(C, B)
    -------------------------------------
     subtract(plus(A,B),C) = plus(A,subtract(B,C)).


theorem plus_subtract2 ===
     A:nats, B:nats, C:nats, le_nats(B, C), le_nats(C, plus(A,B))
    -------------------------------------
     subtract(plus(A,B),C) = subtract(A,subtract(C,B)).

theorem subtract_plus ===
     A:nats, B:nats, C:nats, le_nats(plus(B,C), A)
    -----------------------------------------------
        subtract(A,plus(B,C)) = subtract(subtract(A,B),C).

theorem plus_le_imp_le_subtract ===
    A:nats, B:nats, C:nats
 -----------------------------
  le_nats(plus(A,B),C) => le_nats(A, subtract(C,B)).

theorem subtract_eq_iff_eq ===
     le_nats(B,A), le_nats(C,A)
  --------------------------------
   subtract(A,B) = subtract(A,C) <=> B = C.

theorem subtract_eq_iff_eq1 ===
     le_nats(A,B), le_nats(A,C)
  --------------------------------
   subtract(B,A) = subtract(C,A) <=> B = C.

theorem subtract_eq_plus_iff_eq ===
     le_nats(B,A), C:nats 
    -----------------------
      subtract(A,B) = plus(A,C) <=> B = 0 and C = 0.

theorem le_nats_subtract ===
    le_nats(C,A), le_nats(B,C)
 -------------------------------
    le_nats(subtract(C,B), A).

theorem mult_le_nats ===
    A:nats, le_nats(B, C)
   ------------------------
    le_nats(mult(A,B), mult(A,C)).

theorem mult_subtract ===
    A:nats, le_nats(C,B)
  --------------------------
   mult(A, subtract(B,C)) = subtract(mult(A,B), mult(A,C)).


theorem le_succ_iff_eq_or_le ===
    A:nats, B:nats
  ---------------------------
   le_nats(A, succ(B)) <=> A = succ(B) or le_nats(A,B).

theorem le_iff_succ_le ===
    A:nats, B:nats
 ----------------------
  le_nats(A, B) <=> le_nats(succ(A), succ(B)).

%% Theorems about div_nats

theorem le_succ_le ===
 le_nats(A, B) and le_nats(B, succ(A)) => A = B or B = succ(A).

theorem divs_and_mods_eq_aux ===
    B:nats, D:nats, not A = 0, le_nats(C,A), 
    le_nats(E,A), not C=A, not E=A, le_nats(B,D)
   ------------------------------------
    plus(mult(A,B),C) = plus(mult(A,D), E) <=> B = D and C = E.

theorem divs_and_mods_eq ===
    B:nats, D:nats, not A = 0, le_nats(C,A), le_nats(E,A), not C=A, not E=A
   ------------------------------------
    plus(mult(A,B),C) = plus(mult(A,D), E) <=> B = D and C = E.

theorem mult_bounds ===
    A:nats, not A = 0
 ------------------------
 all x:nats ex y:nats 
        le_nats(mult(A,y), x) 
        and not x = mult(A,succ(y))
        and le_nats(x, mult(A,succ(y)))
   provided
      x not_free_in A, y not_free_in A, x not_free_in y.

theorem ex_div_nats ===
    A:nats, B:nats, not B = 0
  -------------------------
  ex x:nats ex y:nats not y = B and le_nats(y, B) and A = plus(mult(B,x),y)
    provided
      x not_free_in A, y not_free_in A, x not_free_in y,
      x not_free_in B, y not_free_in B.


theorem div_nats_type ===
    A:nats, B:nats, not B=0
 ---------------------------
    div_nats(A,B):nats.

theorem div_nats_undef ===
    not(A:nats and B:nats and not(B=0)) <=> div_nats(A,B) = ? .

theorem undef_div_nats_undef ===
     div_nats(?, X) = ? .

theorem div_nats_undef_undef ===
     div_nats(X, ?) = ? .

theorem div_nats0_undef ===
     div_nats(X, 0) = ? .

theorem open_div_nats_1_eq ===
     X = Y
  -------------------
  div_nats(X,Z) = div_nats(Y,Z).

theorem open_div_nats_2_eq ===
     X = Y
  -------------------
  div_nats(Z,X) = div_nats(Z,Y).

theorem mult_le_iff_one_or_zero ===
    A:nats, B:nats, not B = 0
  ---------------------------
    le_nats(mult(B,A),B) <=> A = 0 or A = 1.

theorem div_nats_eq0 ===
    A:nats, B:nats, not B=0
 ---------------------------
    div_nats(A,B) = 0 <=> le_nats(A,B) and not (A=B).

theorem div_nats_add_prod ===
    A:nats, not B=0, le_nats(C,B), not C=B
 --------------------------
    div_nats(plus(mult(A,B),C), B) = A.
 

theorem plus_div_eq_div1 ===
    A:nats, B:nats, not B=0
 ---------------------------
   div_nats(plus(B,A), B) = succ(div_nats(A,B)).


%% Theorems for and about minnat

theorem minnat_lemma1 ===
    w_local x P
  ------------------------------------
    (all y:nats (ex x:nats le_nats(x, y) and P) =>
         ex x:nats le_nats(x, y) and P and all z:nats [z/x]P => le_nats(x, z))
     provided x not_free_in y, x not_free_in z,
              y not_free_in z, y not_free_in P, z not_free_in P.

theorem minnat_lemma2 ===
    w_local x P, ex x:nats P
   --------------------------
    (ex x:nats P and  all z:nats [z/x]P => le_nats(x, z))
      provided x not_free_in z, z not_free_in P.

theorem minnat_type ===
   w_local x P, ex x:nats P
   --------------------------
     (minnat x P) : nats.

theorem minnat_property1 ===
   w_local x P, ex x:nats P
   --------------------------
    [minnat x P/x]P.

theorem minnat_property2 ===
   w_local x P, ex x:nats P
   --------------------------
   all x:nats P => le_nats(minnat x P, x).


%%
%% Support for Window Inference - loads
%% nats specific information.
%%

require_tactic_file(wi_support,nats).


