%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: seq.thy,v 1.3 2002/09/09 00:43:56 itp Exp $

%%%%%%%%%%%%%%%%%%%
%% ZFC Sequences %%
%%%%%%%%%%%%%%%%%%%

%% NOTE: This is an old version of this theory and needs to be redone
%% using the new zfc theory.

%% This theory is a theory based on Z's theory of sequences.


add_theory(seq).
include_theory(stdops, share).
include_theory(rel, all).
include_theory(fin, all).
include_theory(undef, all).
include_theory(localness, all).
include_theory(zfc, all).

%% Add some extra object variables.
declare_object_vars(w).
declare_object_vars(y).
declare_object_vars(z).

%%%%%%%%%%%%%%%
%% Constants %%
%%%%%%%%%%%%%%%

define empty_seq ===
  empty_set.

%%%%%%%%%%%%%%%
%% Functions %%
%%%%%%%%%%%%%%%

define sequence(T) ===
  (set_of x:(nats '-||->' T) (domain(x) = (1 up_to card(x)))) provided_that
  T:sets.

define sequence_1(T) ===
  (set_of x:sequence(T) (card(x) > 0)) provided_that
  T:sets.

define isequence(T) ===
  (sequence(T) inter (nats '>-|->' T)) provided_that
  T:sets.

%%%%%%%%%%%%%%%
%% Constants %%
%%%%%%%%%%%%%%%

define sequences  ===
  class_of x ex y:sets (x = sequence(y)).

define sequences_1 ===
  class_of x ex y:sets (x = sequence_1(y)).

define isequences  ===
  class_of x ex y:sets (x = isequence(y)).

%%%%%%%%%%%%%%%
%% Functions %%
%%%%%%%%%%%%%%%

abbreviation is_empty_sequence(S) ===
  S = empty_seq.

declare seq(L).

define concat(S_1, S_2) ===
  (S_1 union set_of x ex y:domain(S_2) (
    x = ((y+card(S_1)) maps_to (S_2 at y)))) provided_that
  S_1:sequences and S_2:sequences.

add_operator(525, nonassoc, concat).

define head(S) ===
  (S at 1) provided_that
  S:sequences_1.

define tail(S) ===
  sm x:sequences (S = (seq([head(S)]) concat x)) provided_that
  S:sequences_1.

define front(S) ===
  ((1 up_to (card(S)-1)) domain_restrict S) provided_that
   S:sequences_1.

define last(S) ===
  (S at card(S)) provided_that
   S:sequences_1.

declare reverse(S).

define disjoint(S) ===
  (all x:domain(S) all y:domain(S) (
    x /= y =>
      ((S at x) inter (S at y)) = empty_set)) provided_that
  S:partial_functions.

define partitions(S,P) ===
  disjoint(S) and gen_union(range(S)) = P provided_that
  S:partial_functions and P:sets.

add_operator(525, nonassoc, partitions).

%%%%%%%%%%%%
%% Axioms %%
%%%%%%%%%%%%

axiom def_seq1 ===
  seq([]) = empty_seq.

axiom def_seq2 ===
  seq([H|T]) = set([1 maps_to H]) concat seq(T).

axiom def_reverse ===
  reverse(S) =
  (if(is_empty_sequence(S),
     empty_seq,
     seq([last(S)]) concat reverse(front(S))) provided_that
   S:sequences).

%%%%%%%%%%%%%%%%%%%
%% Opening Rules %%
%%%%%%%%%%%%%%%%%%%



theorem open_seq_1_eq ===
    A = B
    ----------
    seq(A) = seq(B).

theorem open_sequence_1_eq ===
    A = B
    ----------
    sequence(A) = sequence(B).

theorem open_sequence_1_1_eq ===
    A = B
    ----------
    sequence_1(A) = sequence_1(B).

theorem open_isequence_1_eq ===
    A = B
    ----------
    isequence(A) = isequence(B).

theorem open_concat_1_eq ===
    A = B
    ----------
    A concat C = B concat C.

theorem open_concat_2_eq ===
    A = B
    ----------
    C concat A = C concat B.

theorem open_head_1_eq ===
    A = B
    ----------
    head(A) = head(B).

theorem open_tail_1_eq ===
    A = B
    ----------
    tail(A) = tail(B).

theorem open_last_1_eq ===
    A = B
    ----------
    last(A) = last(B).

theorem open_front_1_eq ===
    A = B
    ----------
    front(A) = front(B).

theorem open_reverse_1_eq ===
    A = B
    ----------
    reverse(A) = reverse(B).

theorem open_disjoint_1_eq ===
    A = B
    ----------
    disjoint(A) = disjoint(B).

theorem open_partitions_1_eq ===
    A = B
    ----------
    A partitions C = B partitions C.

theorem open_partitions_2_eq ===
    A = B
    ----------
    C partitions A = C partitions B.


%%%%%%%%%%%%%%
%% Theorems %%
%%%%%%%%%%%%%%

%% Simplification rules
theorem not_is_sequence_undef_A ===
  not (?):sequences.

theorem not_is_sequence_undef_B ===
  (?):sequences <=> false.

theorem sequence_undef ===
  sequence(?) = (?).

theorem not_is_sequence_undef_C ===
  not (?):sequence(T).

theorem not_is_sequence_undef_D ===
  (?):sequence(T) <=> false.

theorem not_is_sequence_1_undef_A ===
  not (?):sequences_1.

theorem not_is_sequence_1_undef_B ===
  (?):sequences_1 <=> false.

theorem not_is_empty_sequence_undef_A ===
  not is_empty_sequence(?).

theorem not_is_empty_sequence_undef_B ===
  is_empty_sequence(?) <=> false.

theorem not_is_sequence_1_empty_sequence_A ===
  not empty_seq:sequences_1.

theorem not_is_sequence_1_empty_sequence_B ===
  empty_seq:sequences_1 <=> false.

theorem head_undef ===
  head(?) = (?).

theorem tail_undef ===
  tail(?) = (?).

theorem front_undef ===
  front(?) = (?).

theorem last_undef ===
  last(?) = (?).

theorem concat_undef_A ===
  ? concat S = (?).

theorem concat_undef_B ===
  S concat ? = (?).

theorem reverse_undef ===
  reverse(?) = (?).

theorem reverse_empty_sequence ===
  reverse(empty_seq) = empty_seq.

theorem disjoint_undef ===
  disjoint(?) = (?).

theorem partitions_undef_A ===
  ? partitions _ = (?).

theorem partitions_undef_B ===
  _ partitions ? = (?).

%% Localness theorems
theorem sequence_local ===
  local x sequence(x).

theorem sequence_1_local ===
  local x sequence_1(x).

theorem isequence_local ===
  local x isequence(x).

theorem seq_local ===
  local x seq(x).

theorem concat_local_1 ===
  local x x concat A
    provided x not_free_in A.

theorem concat_local_2 ===
  local x A concat x
    provided x not_free_in A.

theorem head_local ===
  local x head(x).

theorem tail_local ===
  local x tail(x).

theorem front_local ===
  local x front(x).

theorem last_local ===
  local x last(x).

theorem reverse_local ===
  local x reverse(x).

theorem disjoint_local ===
  local x disjoint(x).

theorem partitions_local_1 ===
  local x x partitions A
    provided x not_free_in A.

theorem partitions_local_2 ===
  local x A partitions x
    provided x not_free_in A.

%% Other theorems
theorem sequences_def ===
  S:sequences <=> 
  ex x:nats ex y:sets S:((1 up_to x) '-->' y).

theorem sequences_def_A ===
  S:sequence(T) <=> 
  T:sets and
  ex x:nats S:((1 up_to x) '-->' T).

theorem concat_def1 ===
  S_1 concat S_2 =
  (if(is_empty_sequence(S_1),
     S_2,
     seq([head(S_1)]) concat (tail(S_1) concat S_2)) provided_that
   S_1:sequences and S_2:sequences).

theorem sequence_induction_A ===
  [empty_seq/x]P and
  all y_1:sets all y_2:sequences ([y_2/x]P => [(seq([y_1]) concat y_2)/x]P)
  =>
    (X:sequences => [X/x]P).

theorem sequence_induction_B ===
  [empty_seq/x]P and
  all y_1:sets all y_2:sequences ([y_2/x]P => [(seq([y_1]) concat y_2)/x]P)
  =>
    (all x:sequences P).

theorem sequence_induction_C ===
  T subtype_of sets and
  [empty_seq/x]P and
  all y_1:T all y_2:sequence(T) ([y_2/x]P => [(seq([y_1]) concat y_2)/x]P)
  =>
    (X:sequence(T) => [X/x]P).

theorem sequence_induction_D ===
  T subtype_of sets and
  [empty_seq/x]P and
  all y_1:T all y_2:sequence(T) ([y_2/x]P => [(seq([y_1]) concat y_2)/x]P)
  =>
    all x:sequence(T) P.

theorem sequence_1_induction_A ===
  (all y:sets [seq([y])/x]P) and
  all y_1:sets all y_2:sequences_1 ([y_2/x]P => [(seq([y_1]) concat y_2)/x]P)
  =>
    (X:sequences_1 => [X/x]P).

theorem sequence_1_induction_B ===
  (all y (y:sets => [seq([y])/x]P)) and
  (all y_1 (y_1:sets => (all y_2 (
    y_2:sequences_1 and [y_2/x]P => [(seq([y_1]) concat y_2)/x]P)))) =>
  (all x x:sequences_1 => P).

theorem sequence_1_induction_C ===
  T subtype_of sets and
  (all y:T [seq([y])/x]P) and
  (all y_1:T all y_2:sequence_1(T)
      ([y_2/x]P => [(seq([y_1]) concat y_2)/x]P))
  =>
    (X:sequence_1(T) => [X/x]P).

theorem sequence_1_induction_D ===
  T subtype_of sets and
  (all y:T [seq([y])/x]P) and
  (all y_1:T all y_2:sequence_1(T)
      ([y_2/x]P => [(seq([y_1]) concat y_2)/x]P))
  =>
    (all x:sequence_1(T) P).

theorem head_sequence ===
  H:sets and seq(T):sequences =>
    head(seq([H|T])) = H.

theorem tail_sequence ===
  H:sets and seq(T):sequences =>
    tail(seq([H|T])) = seq(T).

theorem concat_associates ===
  S_1 concat (S_2 concat S_3) = (S_1 concat S_2) concat S_3.
  
theorem head_concat_tail ===
  S:sequences_1 =>
    S = seq([head(S)]) concat tail(S).
  
theorem front_concat_last ===
  S:sequences_1 =>
    S = (front(S) concat seq([last(S)])).

theorem concat_is_sequence_1_C ===
  S_1:sequences_1 or S_2:sequences_1 <=>
  (S_1 concat S_2):sequences_1.

theorem empty_sequence_concat ===
  S:sequences =>
    S = (empty_seq concat S).

theorem concat_empty_sequence ===
  S:sequences =>
    S = (S concat empty_seq).

theorem reverse_reverse ===
  S:sequences =>
     S = reverse(reverse(S)).

theorem reverse_sequence ===
  H:sets => 
    reverse(seq([H])) = seq([H]).

theorem reverse_front ===
  reverse(front(S)) = tail(reverse(S)).

theorem reverse_tail ===
  reverse(tail(S)) = front(reverse(S)).

theorem head_reverse ===
  head(reverse(S)) = last(S).

theorem last_reverse ===
  last(reverse(S)) = head(S).

theorem reverse_concat ===
  reverse(S_1 concat S_2) = reverse(S_2) concat reverse(S_1).

theorem def_front_1 ===
  front(S) =
  (if(is_empty_sequence(tail(S)),
     empty_seq,
     seq([head(S)]) concat front(tail(S))) provided_that
   S:sequences_1).

theorem def_last_1 ===
  last(S) =
  (if(is_empty_sequence(tail(S)),
     head(S),
     last(tail(S))) provided_that
   S:sequences_1).

theorem sequences_def_B ===
  S:sequences <=>
  (is_empty_sequence(S) or
   ex x (x:sets and ex y (y:sequences and S = seq([x]) concat y))).

theorem sequences_def_C ===
  S:sequence(T) <=>
  T subtype_of sets and
  (is_empty_sequence(S) or
   ex x (x:T and ex y (y:sequence(T) and S = seq([x]) concat y))).

theorem sequence_equal ===
  H_1:sets and seq(T_1):sequences and
  H_2:sets and seq(T_2):sequences =>
  (seq([H_1|T_1]) = seq([H_2|T_2]) <=> H_1 = H_2 and T_1 = T_2).

theorem sequences_def_D ===
  (S:sequences) =
  ex_set x S:sequence(x).

theorem sequences_1_def_A ===
  sequence_1(T) =
  (set_of x:(pos_ints '-||->' T) (
      card(x) > 0 and
      domain(x) = (1 up_to card(x))) provided_that
   T:sets).

theorem sequence_not_equal_empty_sequence ===
  seq([H|T]) /= empty_seq.


%% Typing rules

theorem empty_seq_is_sequences === 
  empty_seq : sequences.

theorem is_type_sequences ===
  is_type(sequences).

theorem is_type_sequences_1 ===
  is_type(sequences_1).

theorem is_type_sequences_A ===
  T:sets <=>
  is_type(sequence(T)).

theorem is_type_sequences_1_A ===
  T:sets <=>
  is_type(sequence_1(T)).

theorem empty_sequence_is_sequence ===
  T:sets <=>
  empty_seq:sequence(T).

theorem sequences_1_def_B ===
  S:sequences_1 <=>
  S:sequences and not is_empty_sequence(S).

theorem sequences_1_def_C ===
  S:sequence_1(T) <=>
  S:sequence(T) and not is_empty_sequence(S).

theorem sequences_1_def_D ===
  S:sequences_1 <=>
  S:sequences and S /= empty_seq.

theorem sequences_1_def_E ===
  S:sequence_1(T) <=> 
  S:sequence(T) and S /= empty_seq.

theorem sequence_is_sequence_B ===
  H:S and seq(T):sequence(S) <=>
  seq([H|T]):sequence(S).

theorem sequence_is_sequence_1_A ===
  H:S and seq(T):sequence(S) <=>
  seq([H|T]):sequence_1(S).

theorem sequence_is_undef_A ===
  not (H:sets and seq(T):sequences) <=>
  is_undefined(seq([H|T])).

theorem sequences_is_defined ===
  is_defined(sequences).

theorem sequences_subtype_of_sequences ===
  T:sets <=>
  sequence(T) subtype_of sequences.

theorem sequences_is_defined_A ===
  T:sets <=>
  is_defined(sequence(T)).

theorem subtypes_sequences_subtype_of_sequences ===
  T_1 subtype_of T_2 and T_2:sets <=>
  sequence(T_1) subtype_of sequence(T_2).

theorem is_sequence_is_defined ===
  S:sequences => is_defined(S).

theorem is_sequence_is_sequence ===
  S:sequence(T) => S:sequences.

theorem sequences_are_sets ===
  T:sets <=> sequence(T):sets.

theorem sequences_1_are_sets ===
  T:sets <=> sequence_1(T):sets.

theorem sequences_1_subtype_of_sequences_A ===
  sequences_1 subtype_of sequences.

theorem sequences_1_subtype_of_sequences_B ===
  T:sets <=> sequence_1(T) subtype_of sequence(T).

theorem is_sequence_1_is_sequence ===
  S:sequence_1(T) => S:sequence(T).

theorem sequence_is_defined ===
  H:sets and T:sequences <=> is_defined(seq([H]) concat T).

theorem sequence_is_sequence_A ===
  H:sets and T:sequences <=> seq(seq([H]) concat T):sequences.

theorem sequence_is_undef_B ===
  is_undefined(H) => is_undefined(seq([H|_])).

theorem sequence_is_undef_C ===
  is_undefined(T) => is_undefined(seq([_|T])).

theorem sequence_is_sequence_1_B ===
  H:sets and T:sequences <=>
  (seq([H]) concat T):sequences_1.

theorem is_sequence_is_finite_func ===
  S:sequence(T) => S:pos_ints '-||->' T.

theorem sequences_subtype_of_fp_funcs ===
  sequences subtype_of finite_partial_functions.

theorem head_is_defined ===
  S:sequences_1 <=> is_defined(head(S)).

theorem head_is_undef ===
  is_undefined(S) => is_undefined(head(S)).

theorem head_is_sets ===
  S:sequences_1 <=> head(S):sets.

theorem head_empty_sequence_is_undef ===
  is_undefined(head(empty_seq)).

theorem head_is_sequence_type ===
  S:sequence_1(T) => head(S):T.

theorem tail_is_defined ===
  S:sequences_1 <=> is_defined(tail(S)).

theorem tail_is_undef ===
  is_undefined(S) => is_undefined(tail(S)).

theorem tail_empty_sequence_is_undef ===
  is_undefined(tail(empty_seq)).

theorem tail_is_sequence_A ===
  S:sequences_1 <=> tail(S):sequences.

theorem tail_is_sequence_B ===
  S:sequence_1(T) => tail(S):sequence(T).

theorem front_is_defined ===
  S:sequences_1 <=> is_defined(front(S)).

theorem front_is_undef ===
  is_undefined(S) => is_undefined(front(S)).

theorem front_empty_sequence_is_undef ===
  is_undefined(front(empty_seq)).

theorem front_is_sequence_A ===
  S:sequences_1 => front(S):sequences.

theorem front_is_sequence_B ===
  S:sequence_1(T) => front(S):sequence(T).

theorem last_is_defined ===
  S:sequences_1 <=> is_defined(last(S)).

theorem last_is_undef ===
  is_undefined(S) => is_undefined(last(S)).

theorem last_is_sets ===
  S:sequences_1 <=> last(S):sets.

theorem last_empty_sequence_is_undef ===
  is_undefined(last(empty_seq)).

theorem last_is_sequence_type ===
  S:sequence_1(T) => last(S):T.

theorem concat_is_defined ===
  S_1:sequences and S_2:sequences <=> is_defined(S_1 concat S_2).

theorem concat_is_undef_A ===
  is_undefined(S) => is_undefined(S concat _).

theorem concat_is_undef_B ===
  is_undefined(S) => (_ concat S).

theorem concat_is_sequence_A ===
  S_1:sequences and S_2:sequences <=> (S_1 concat S_2):sequences.

theorem concat_is_sequence_1_A ===
  S_1:sequences_1 and S_2:sequences => (S_1 concat S_2):sequences_1.

theorem concat_is_sequence_1_B ===
  S_1:sequences and S_2:sequences_1 => (S_1 concat S_2):sequences_1.

theorem concat_is_sequence_B ===
  S_1:sequence(T) and S_2:sequence(T) <=> (S_1 concat S_2):sequence(T).

theorem concat_is_sequence_1_D ===
  S_1:sequence_1(T) and S_2:sequence(T) => (S_1 concat S_2):sequence_1(T).

theorem concat_is_sequence_1_E ===
  S_1:sequence(T) and S_2:sequence_1(T) => (S_1 concat S_2):sequence_1(T).

theorem reverse_is_defined ===
  S:sequences => is_defined(reverse(S)).

theorem reverse_is_undef ===
  is_undefined(S) => is_undefined(reverse(S)).

theorem reverse_is_sequence_A ===
  S:sequences => reverse(S):sequences.

theorem reverse_is_sequence_1_A ===
  S:sequences_1 => reverse(S):sequences_1.

theorem reverse_is_sequence_B ===
  S:sequence(T) => reverse(S):sequence(T).

theorem reverse_is_sequence_1_B ===
  S:sequence_1(T) => reverse(S):sequence_1(T).

theorem disjoint_is_bool ===
  S:partial_functions <=> disjoint(S):bools.

theorem disjoint_is_defined ===
  S:partial_functions <=> is_defined(disjoint(S)).

theorem partitions_is_bool ===
  S:partial_functions and P:sets <=> (S partitions P):bools.

theorem partitions_is_defined ===
  S:partial_functions and P:sets <=> is_defined(S partitions P).

