%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: undef.thy,v 1.6 2002/05/22 05:35:23 itp Exp $

                                %%%%%%%%%%%%%%%
				%% Undefined %%
                                %%%%%%%%%%%%%%%

%% Ergo does not provide any facilities for type checking hence we need to
%% guard against expressions like `12 and true'.  The solution taken here is
%% two-fold:
%%
%%  * Add a bottom element (`?').  The intention is that if a function has
%%    illtyped arguments it will return `?'.  Also, the typing predicate `:'
%%    will be false if an argument is ?.
%%
%%  * Add a function `provided' that can be used to express typing information
%%    together with tactics that automatically infer type information.

add_theory(undef).
include_theory(stdops, share).                                                  
include_theory(cpred_eq_t, all). 
include_theory(if, all).                                                  
include_theory(apart, all).                                                  


%% Add some extra object variables.
declare_object_vars(w).
declare_object_vars(y).
declare_object_vars(z).

                                %%%%%%%%%%%%%%%
				%% Constants %%
                                %%%%%%%%%%%%%%%

declare '?'.

                                %%%%%%%%%%%%%%%
				%% Functions %%
                                %%%%%%%%%%%%%%%

define is_defined(X) ===
  X apart '?'.

define is_undefined(X) ===
  X = '?'.

define provided_that(Body, Cond) ===
  if(Cond, Body, ?).

add_operator(850, nonassoc, provided_that).

                                 %%%%%%%%%%%%
				 %% Axioms %%
                                 %%%%%%%%%%%%

axiom undef_not1 ===
  not (?):T.

axiom not_undef1 ===
  not (?).

axiom not_undef_type1 ===
  not is_type(?).

axiom is_type_is_defined ===
  is_type(X) => is_defined(X).

axiom defined_or_undefined ===
  is_defined(X) or is_undefined(X).

                                %%%%%%%%%%%%%%
				%% Theorems %%
                                %%%%%%%%%%%%%%

%% Simplification rules
theorem not_undef2 ===
  ? <=> false.

theorem undef_not2 ===
  (?):T <=> false.

theorem not_undef_type2 ===
  is_type(?) <=> false.

theorem nothing_type_undef1 ===
  not X:(?).

theorem nothing_type_undef2 ===
  X:(?) <=> false.

theorem not_undef_subtype1 ===
  not ? subtype_of T.

theorem not_undef_subtype2 ===
  ? subtype_of T <=> false.

theorem not_subtype_undef1 ===
  not S subtype_of (?).

theorem not_subtype_undef2 ===
  S subtype_of (?) <=> false.

theorem provided_that_true ===
  (X provided_that true) = X.

theorem provided_that_false1 ===
  not C => (_ provided_that C) = (?).

theorem provided_that_false2 ===
  (_ provided_that false) = (?).

theorem not_provided_that_false ===
  not (_ provided_that false).

theorem not_undef_provided_that ===
  not (? provided_that Cond).

theorem undef_provided_that ===
  (? provided_that Cond) = (?).

theorem is_defined_undef ===
  is_defined(?) <=> false.

theorem imp_is_defined ===
  A => is_defined(A).

theorem type_is_defined1 ===
  X:T => is_defined(X).

theorem type_is_defined2 ===
  X:T => is_defined(T).

theorem not_is_defined_iff_is_undefined ===
  not is_defined(A) <=> is_undefined(A).

%theorem not_is_undefined_iff_is_defined ===
%  not is_undefined(A) <=> is_defined(A).

%% Typing theorems
theorem undef_is_undefined ===
  is_undefined(?).

theorem is_undefined_undef ===
  is_undefined(A) <=> A = (?).

theorem is_undefined_undef1 ===
  is_undefined(A) <=> ? = A.

                              %%%%%%%%%%%%%%%%%%%
			      %% Opening Rules %%
                              %%%%%%%%%%%%%%%%%%%

%theorem open_is_def_1_eq ===
%    A = B
%   ---------------
%   is_defined(A) = is_defined(B).

theorem open_is_def_1_imp ===
    A = B
   ---------------
   is_defined(A) => is_defined(B).

theorem open_is_def_1_iff ===
    A = B
   ---------------
   is_defined(A) <=> is_defined(B).


%theorem open_is_undef_1_eq ===
%    A = B
%   ---------------
%   is_undefined(A) = is_undefined(B).

theorem open_is_undef_1_imp ===
    A = B
   ---------------
   is_undefined(A) => is_undefined(B).

theorem open_is_undef_1_iff ===
    A = B
   ---------------
   is_undefined(A) <=> is_undefined(B).


theorem open_provided_1_eq ===
     hyp+++[C] ---> A = B
     --------------------
     (A provided_that C) = (B provided_that C).

theorem open_provided_1_imp ===
     hyp+++[C] ---> A => B
     --------------------
     (A provided_that C) => (B provided_that C). 

theorem open_provided_1_fol ===
     hyp+++[C] ---> A <= B
     --------------------
     (A provided_that C) <= (B provided_that C). 

theorem open_provided_1_iff ===
     hyp+++[C] ---> A <=> B
     --------------------
     (A provided_that C) <=> (B provided_that C). 

theorem open_provided_2_iff ===
     hyp+++[C] ---> A <=> B
     --------------------
     (C provided_that A) <=> (C provided_that B). 

theorem open_provided_2_imp ===
     hyp+++[C] ---> A => B
     --------------------
     (C provided_that A) => (C provided_that B). 

theorem open_provided_2_fol ===
     hyp+++[C] ---> A <= B
     --------------------
     (C provided_that A) <= (C provided_that B). 

theorem open_provided_2_eq ===
     hyp+++[not(C = ?)] ---> A <=> B
     --------------------
     (C provided_that A) = (C provided_that B). 

theorem provided_elim_c ===
     hyp---[CId:::A provided_that B]+++[A,B] ---> C
     ----------------------------------------
     Node ::: C
  provided
     context_search(hyp, Node, CId, A provided_that B).

%%
%% Support for Window Inference - loads
%% undef specific information.
%%

require_tactic_file(wi_support,undef).

