%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: zfc.thy,v 1.14 2003/07/07 00:01:13 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Zermelo Fraenkel Set Theory %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% A theory of ZFC set theory.
%% This theory is directly based on
%% "Axioms of set theory" by J.R. Shoenfield appearing in
%% "Handbook of Mathematical Logic" by Jon Barwise.
%%
%% In the development above, it is assumed all quantified variables
%% are sets. We make this explicit by declaring a type 'sets' and
%% use this to assert that some expression is a set.

add_theory(zfc).

include_theory(stdops, share).
include_theory(cpred_eq_t,all).
include_theory(localness, all).


%% Add some extra object variables.
declare_object_vars(w).
declare_object_vars(y).
declare_object_vars(z).

declare sets.           %% The type of sets

%% Primitive set operations.
declare in(_,_).

add_operator(700, nonassoc, in).

add_operator(890, quant, ex_set).

add_operator(890, quant, set_of).
declare set_of x _.

define ex_set x P === ex y:sets all x:sets x in y <=> P.

%% Because we are inheriting the weak equality theory rather
%% than the strong equality theory we add axioms that allow us
%% to replace equal terms within the arguments of 'in'

axiom open_in_1_iff ===
      A = B
   --------------
    A in C <=> B in C.

axiom open_in_2_iff ===
      A = B
   ---------------
    C in A <=> C in B.

axiom  open_set_of_1_eq(x:obvar) ===
    hyp ??? nfi(x) +++ [x:sets] ---> A <=> B
    ---------------------------
      (set_of x A) = (set_of x B).

axiom extensionality ===
   all x:sets all y:sets
    (all z:sets z in x <=> z in y) <=> x = y
        provided z not_free_in x, z not_free_in y, x not_free_in y.

axiom separation ===
    all y:sets (all x:sets P => x in y) => ex_set x P
        provided x not_free_in y, y not_free_in P.

axiom union_axiom ===
   all z:sets
    ex_set x ex y:sets y in z and x in y
        provided x not_free_in z, y not_free_in z, x not_free_in y.

add_operator(525, nonassoc, subset_of).

define X subset_of Y === all z:sets z in X => z in Y.

axiom power_axiom ===
  all x:sets ex_set y y subset_of x
         provided y not_free_in x.

axiom replacement ===
   all z:sets
    (all x:sets ex_one y:sets A) =>
      ex_set y ex x:sets x in z and A
         provided x not_free_in y, x not_free_in z, y not_free_in z,
                  z not_free_in A.

add_operator(700, nonassoc, not_in).

abbreviation X not_in Y === not (X in Y).

axiom infinity ===
    ex x:sets (ex y:sets y in x and (all z:sets z not_in y)) and
          all y:sets y in x  => ex z:sets z in x and
              all w:sets (w in z <=> w in y or w = y)
         provided x not_free_in y, x not_free_in z, x not_free_in w,
                  y not_free_in z,  y not_free_in w,
                  z not_free_in w.

axiom regularity ===
    all x:sets (ex y:sets y in x) =>
            ex y:sets y in x and all z:sets z in y => z not_in x
         provided x not_free_in y, x not_free_in z, y not_free_in z.

%% This axiom makes expicit that a set construction is of type sets
%% if the set exists. Note that we are saying nothing about the
%% type of a set construction if the set does not exist.

axiom ex_set_is_set ===
    (ex_set x P) => (set_of x P):sets.

%% This axiom makes explicit the property of a set that exists.
%% Again we say nothing about what properties the set construction
%% has if the set does not exist.

axiom set_property ===
    A:sets and (ex_set x P) => (A in (set_of x P) <=> [A/x]P)
        provided x not_free_in A.
    

%% End of the axioms 

close_theory.

define empty === set_of x false.

abbreviation not_empty(Y) === not (Y = empty).

define gen_union(X) === set_of x ex y:sets y in X and x in y.

define gen_inter(X) === set_of x x in gen_union(X) and all y:sets (y in X => x in y).

define power(X) === set_of x x subset_of X.

define single(X) === set_of x x = X.

define pair(X,Y) === set_of x (x = X) or (x = Y).

define union(X,Y) === gen_union(pair(X,Y)).

add_operator(525, leftassoc, union).

define inter(X,Y) === gen_inter(pair(X,Y)).

add_operator(520, leftassoc, inter).

define diff(X,Y) === set_of x (x in X and not (x in Y)).

add_operator(525,leftassoc, diff).

define sym_diff(X,Y) === (X diff Y) union (Y diff X).

add_operator(525,leftassoc, sym_diff).

define power_1(X) === power(X) diff single(empty).

define p_subset_of(X,Y) === (X subset_of Y) and not (X = Y).

add_operator(525, nonassoc, p_subset_of).

define ordered_pair(X,Y) === pair(single(X), pair(X,Y)).

%% Using the 'maps to' arrow of Z as an abbreviation for ordered pairs.

abbreviation |-->(X, Y) === ordered_pair(X,Y).
add_operator(500, nonassoc, '|-->').

define cross(X,Y) === 
    set_of z ex x:sets ex y:sets x in X and y in Y and z = ordered_pair(x,y).

add_operator(530, leftassoc, cross).

%% The following definitions are about functions but are not in
%% the func.thy because they are useful for constructing a theory
%% about least and greatest fixed points and so are included here
%% so that such a theory need only inherit this theory rather than
%% the func theory.

define pre_at(F,X) === set_of y (X |--> y) in F.

define at(F, X) === gen_union(pre_at(F,X)). 

add_operator(525, nonassoc, at).


define !!lambda [x:T] Body ===
     set_of y ex x:sets x in T and y = (x |--> Body).

add_operator(805, quant, lambda).



theorem extensionality2 ===
                 X:sets, Y:sets 
    ----------------------------------------
    (all z:sets z in X <=> z in Y) <=> X = Y
        provided z not_free_in X, z not_free_in Y.



theorem ex_subset_of_set ===
        S:sets
 ----------------------
  ex_set x x in S and P
       provided x not_free_in S.

theorem set_of_and_subset_set_of ===
        S:sets
  --------------------
   (set_of x x in S and P) subset_of S
     provided x not_free_in S.

theorem ex_empty === ex_set x false.

theorem empty_type === empty:sets.

theorem empty_is_empty === all x:sets x not_in empty.

theorem empty_is_empty2 === 
         X:sets 
   -------------------
    not (X in empty).

theorem empty_property ===
         X:sets 
  --------------------
  X in empty <=> false.

theorem has_elem_not_empty ===
               X:sets 
  ----------------------------------
  not X = empty <=> ex x:sets x in X
        provided x not_free_in X.

theorem has_elem_not_empty2 ===
    X:sets, Y:sets, X in Y
    ----------------------
       not_empty(Y).

theorem subset_empty_empty === 
             X:sets
   -------------------------------
   X subset_of empty <=> X = empty.

theorem ex_gen_union === 
    X:sets => ex_set x ex y:sets y in X and x in y
        provided x not_free_in X, y not_free_in X, x not_free_in y.

theorem gen_union_type ===
    X:sets
  ------------
  gen_union(X):sets.

theorem gen_union_property ===
             X:sets, Y:sets 
 -------------------------------------------------
 Y in gen_union(X) <=> ex z:sets z in X and Y in z
        provided z not_free_in X, z not_free_in Y.

theorem ex_gen_inter === 
    X:sets => ex_set x x in gen_union(X) and all y:sets y in X => x in y
        provided x not_free_in X, y not_free_in X, x not_free_in y.

theorem gen_inter_type ===
    X:sets
  ------------------------
   gen_inter(X):sets.

theorem gen_inter_property ===
         X:sets, Y:sets 
  -------------------------------------------------
  Y in gen_inter(X) <=> not_empty(X) and all z:sets z in X => Y in z
        provided z not_free_in X, z not_free_in Y.
 
theorem gen_inter_empty ===
     gen_inter(empty) = empty.

theorem ex_diff ===
  X:sets and Y:sets => ex_set x (x in X and not (x in Y))
      provided x not_free_in X, x not_free_in Y.

theorem diff_type ===
  X:sets, Y:sets
 ------------------
  diff(X,Y):sets.

theorem diff_property ===
          X:sets, Y:sets, Z:sets
   ------------------------------------------
   Z in diff(X,Y) <=> Z in X and not (Z in Y).

theorem sym_diff_type ===
  X:sets, Y:sets
 ------------------
  sym_diff(X,Y):sets.

theorem sym_diff_property ===
          X:sets, Y:sets, Z:sets
   ------------------------------------------
   Z in sym_diff(X,Y) <=> 
      Z in X and not (Z in Y) or Z in Y and not (Z in X).



theorem ex_power === 
    X:sets => ex_set x x subset_of X
        provided x not_free_in X.

theorem power_type ===
    X:sets
 -----------
  power(X):sets.

theorem power_property ===
          X:sets, Y:sets
  -------------------------------
  X in power(Y) <=> X subset_of Y.

theorem power_1_type ===
    X:sets
 -----------
  power_1(X):sets.

theorem power_1_property ===
          X:sets, Y:sets
  --------------------------------------------------
  Y in power_1(X) <=> Y subset_of X and not_empty(Y).

theorem power_empty_type ===
    power(empty):sets.

theorem power_empty_property ===
          X:sets
    -----------------------------------
    X in power(empty) <=> X = empty.

theorem open_ex_set_1_iff(x:obvar) ===
    hyp ??? nfi(x) +++ [x:sets] ---> A <=> B
    ---------------------------
   (ex_set x A) <=> (ex_set x B).

theorem ex_single ===
    X:sets => ex_set x x = X
       provided x not_free_in X.

theorem single_type ===
    X:sets
 ------------
 single(X):sets.

theorem single_property ===
    X:sets, Y:sets
  ------------------------
  Y in single(X) <=> Y = X.

theorem single_empty_property ===
    X:sets 
  --------------------------------
  X in single(empty) <=> X = empty.

theorem power_single_empty_property ===
       X:sets
   ----------------------------------------------------
   X in power(single(empty)) <=> (X = empty or X = single(empty)).

theorem empty_not_eq_single_empty ===
    not (empty = single(empty)).

theorem ex_pair ===
    X:sets and Y:sets => ex_set x (x = X) or (x = Y)
        provided x not_free_in X, x not_free_in Y.


theorem set_subset_of_set ===
         X:sets
    -------------------
     X subset_of X.

theorem empty_subset_of_set ===
         X:sets
    -------------------
     empty subset_of X.

theorem subset_subset_imp_eq ===
    X:sets, Y:sets, X subset_of Y, Y subset_of X 
    --------------------------------------------
                  X = Y.

theorem single_not_empty ===
         X:sets
   --------------------
   not_empty(single(X)).

theorem power_not_empty ===
         X:sets
   ---------------------
    not_empty(power(X)).

theorem gen_union_not_empty ===
  X:sets, not_empty(X), not (X = single(empty))
  -------------------------------
    not_empty(gen_union(X)).

theorem not_empty_union_not_empty ===
  X:sets, not_empty(X), Y:sets
 -----------------------------
    not_empty(X union Y).

theorem union_not_empty_not_empty ===
  X:sets, Y:sets, not_empty(Y)
 -----------------------------
    not_empty(X union Y).


theorem pair_not_empty ===
   X:sets, Y:sets
 -----------------
  not_empty(pair(X,Y)).

theorem pair_type ===
    X:sets, Y:sets
  -------------------
    pair(X,Y):sets.

theorem pair_property ===
    X:sets, Y:sets, Z:sets
  ----------------------------------
  Z in pair(X,Y) <=> Z = X or Z = Y.

theorem union_type ===
    X:sets, Y:sets 
  -------------------
   (X union Y):sets.

theorem union_property ===
    X:sets, Y:sets, Z:sets
  --------------------------------
  Z in X union Y <=> (Z in X or Z in Y).

theorem inter_type === 
      X:sets, Y:sets
   --------------------
     (X inter Y):sets.

theorem inter_property ===
    X:sets, Y:sets, Z:sets
  --------------------------------
  Z in X inter Y <=> (Z in X and Z in Y).


theorem eq_ordered_pairs ===
    X:sets, Y:sets, Z:sets, W:sets
  ---------------------------------------------
  ordered_pair(X,Y) = ordered_pair(Z,W) <=> X = Z and Y = W.

theorem ordered_pair_type ===
    X:sets, Y:sets
 --------------------
 ordered_pair(X,Y):sets.

theorem ordered_pair_not_empty ===
    X:sets, Y:sets
 ----------------------
 not_empty(ordered_pair(X,Y)).

theorem ordered_pair_in_power_power ===
    A:sets, B:sets, X:sets, Y:sets
  ---------------------------------------
  X in A and Y in B => ordered_pair(X,Y) in power(power(A union B)).

theorem ex_cross ===
    X:sets and Y:sets => 
      ex_set z ex x:sets ex y:sets x in X and y in Y and z = ordered_pair(x,y)
        provided x not_free_in y, x not_free_in z, y not_free_in z,
                 x not_free_in X, x not_free_in Y,
                 y not_free_in X, y not_free_in Y,
                 z not_free_in X, z not_free_in Y.
    
theorem cross_type ===
    X:sets, Y:sets
 ----------------------
   cross(X,Y):sets.

theorem cross_property ===
   X:sets, Y:sets, Z:sets
 -------------------------------------
 Z in cross(X,Y) 
     <=> 
 ex x:sets ex y:sets x in X and y in Y and Z = ordered_pair(x,y)
        provided x not_free_in y, x not_free_in X, x not_free_in Y, 
                 x not_free_in Z, y not_free_in X, y not_free_in Y, 
                 y not_free_in Z.

theorem cross_not_empty ===
   X:sets, Y:sets, not_empty(X), not_empty(Y)
  -------------------------------------------
      not_empty(X cross Y).

theorem open_subset_of_1_iff ===
        A = B
    ---------------------------
   ID ::: subset_of(A,C) <=> subset_of(B,C)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_subset_of_2_iff ===
        A = B
    ---------------------------
   ID ::: subset_of(C,A) <=> subset_of(C,B)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_p_subset_of_1_iff ===
        A = B
    ---------------------------
   ID ::: p_subset_of(A,C) <=> p_subset_of(B,C)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_p_subset_of_2_iff ===
        A = B
    ---------------------------
   ID ::: p_subset_of(C,A) <=> p_subset_of(C,B)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).


theorem eq_single ===
   X:sets, Y:sets 
 -----------------------------
 single(X) = single(Y) <=> X = Y.

theorem eq_pair ===
   A:sets, B:sets, C:sets, D:sets
 ---------------------------------------------------
 pair(A,B) = pair(C,D) <=> A = C and B = D or A = D and B = C.

theorem single_eq_pair ===
   A:sets, B:sets, C:sets 
 -----------------------------------
 single(A) = pair(B,C) <=> A = B and A = C.

theorem eq_cross ===
   A:sets, B:sets, C:sets, D:sets, 
   not_empty(A), not_empty(B), not_empty(C), not_empty(D)
  -------------------------------------------------------
     (A cross B) = (C cross D) <=> A = C and B = D.

theorem cross_empty_eq_empty ===
   A:sets 
 ----------------------
 A cross empty = empty.

theorem empty_cross_eq_empty ===
   A:sets 
 ----------------------
 empty cross A = empty.

theorem eq_power ===
   A:sets, B:sets
 ------------------------------
 power(A) = power(B) <=> A = B.

theorem in_nonreflexive ===
   X:sets 
 ------------------
  X not_in X.

%% Opening rules for various operators

theorem open_single_1_eq ===
        A = B
    ---------------------------
   ID ::: single(A) = single(B)
      provided context_search(hyp, ID, _, A:sets).

theorem open_gen_union_1_eq ===
        A = B
    ---------------------------
   ID ::: gen_union(A) = gen_union(B)
      provided context_search(hyp, ID, _, A:sets).

theorem open_gen_inter_1_eq ===
        A = B
    ---------------------------
   ID ::: gen_inter(A) = gen_inter(B)
      provided context_search(hyp, ID, _, A:sets).

theorem open_power_1_eq ===
        A = B
    ---------------------------
   ID ::: power(A) = power(B)
      provided context_search(hyp, ID, _, A:sets).

theorem open_power_1_1_eq ===
        A = B
    ---------------------------
   ID ::: power_1(A) = power_1(B)
      provided context_search(hyp, ID, _, A:sets).

theorem open_pair_1_eq ===
        A = B
    ---------------------------
   ID ::: pair(A,C) = pair(B,C)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_pair_2_eq ===
        A = B
    ---------------------------
   ID ::: pair(C,A) = pair(C,B)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_union_1_eq ===
        A = B
    ---------------------------
   ID ::: union(A,C) = union(B,C)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_union_2_eq ===
        A = B
    ---------------------------
   ID ::: union(C,A) = union(C,B)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_inter_1_eq ===
        A = B
    ---------------------------
   ID ::: inter(A,C) = inter(B,C)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_inter_2_eq ===
        A = B
    ---------------------------
   ID ::: inter(C,A) = inter(C,B)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_diff_1_eq ===
        A = B
    ---------------------------
   ID ::: diff(A,C) = diff(B,C)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_diff_2_eq ===
        A = B
    ---------------------------
   ID ::: diff(C,A) = diff(C,B)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_sym_diff_1_eq ===
        A = B
    ---------------------------
   ID ::: sym_diff(A,C) = sym_diff(B,C)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_sym_diff_2_eq ===
        A = B
    ---------------------------
   ID ::: sym_diff(C,A) = sym_diff(C,B)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_ordered_pair_1_eq ===
        A = B
    ---------------------------
   ID ::: ordered_pair(A,C) = ordered_pair(B,C)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_ordered_pair_2_eq ===
        A = B
    ---------------------------
   ID ::: ordered_pair(C,A) = ordered_pair(C,B)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_cross_1_eq ===
        A = B
    ---------------------------
   ID ::: cross(A,C) = cross(B,C)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).

theorem open_cross_2_eq ===
        A = B
    ---------------------------
   ID ::: cross(C,A) = cross(C,B)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, C:sets).



theorem empty_diff_eq_empty ===
    X:sets 
 --------------------
 empty diff X = empty.

theorem diff_empty_eq ===
    X:sets 
 ----------------
 X diff empty = X.

theorem diff_eq_empty ===
    X:sets 
 ----------------
 X diff X = empty.

theorem power_empty_eq_single_empty ===
    power(empty) = single(empty).


theorem empty_subset ===
  X:sets 
 -----------------
 empty subset_of X.

theorem subset_empty_eq_empty ===
  X:sets 
 -------------------------------
 X subset_of empty <=> X = empty.

theorem empty_union_eq ===
  A:sets
 -----------------
 empty union A = A.

theorem union_empty_eq ===
  A:sets
 -----------------
 A union empty = A.

theorem empty_inter_eq_empty ===
  A:sets
 ---------------------
 empty inter A = empty.

theorem inter_empty_eq_empty ===
  A:sets
 ---------------------
 A inter empty = empty.

theorem sym_diff_eq_empty ===
    X:sets
 --------------------
 X sym_diff X = empty.

theorem gen_union_empty_eq_empty ===
    gen_union(empty) = empty.



theorem inter_diff_eq_empty ===
  X:sets, Y:sets
 --------------------------
 X inter (Y diff X) = empty.

theorem in_subset ===
  X:sets, A:sets, B:sets
 ----------------------------------------
 (X in A) and (A subset_of B) => (X in B).

theorem subset_reflexive ===
  X:sets 
 --------------
 X subset_of X.

theorem subset_transitive ===
  X:sets, Y:sets, Z:sets
 ---------------------------------------
 (X subset_of Y) and (Y subset_of Z) => (X subset_of Z).

theorem p_subset_transitive ===
  X:sets, Y:sets, Z:sets
 ---------------------------------------
 (X p_subset_of Y) and (Y p_subset_of Z) => (X p_subset_of Z).


theorem single_is_subset ===
  X:sets, Y:sets
 ----------------------------
 (X in Y) <=> single(X) subset_of Y.


theorem pair_commutes ===
  X:sets, Y:sets
 --------------------------
 pair(X,Y) = pair(Y,X).

theorem subset_of_power === 
  A:sets, B:sets 
 -----------------------------------
 A subset_of B => power(A) subset_of power(B).

theorem in_power ===
  X:sets 
 ---------------
 X in power(X).

theorem in_subset_gen_union ===
  X:sets, Y:sets
 ----------------------------
 X in Y => X subset_of gen_union(Y).

theorem gen_union_single ===
  X:sets 
 ---------------
 X = gen_union(single(X)).

theorem gen_inter_in ===
  X:sets, A:sets, B:sets
 ------------------------------------
 (X in gen_inter(A)) and (B in A) => (X in B).

theorem in_subset_gen_inter ===
  X:sets, Y:sets
 -----------------------------
 X in Y => gen_inter(Y) subset_of X.

theorem gen_inter_single ===
  X:sets
 -----------------------
 X = gen_inter(single(X)).

theorem in_union1 ===
  A:sets, B:sets, X:sets
 ----------------------------
  X in B => X in (A union B).

theorem in_union2 ===
  A:sets, B:sets, X:sets
 ----------------------------
 X in A => X in (A union B).

theorem union_absorption ===
  X:sets 
 ----------------
 (X union X) = X.

theorem subset_union1 ===
  X:sets, Y:sets
 ---------------------
 X subset_of (X union Y).

theorem subset_union2 ===
  X:sets, Y:sets
 ---------------------
Y subset_of (X union Y).

theorem subsets_union ===
  X:sets, Y:sets, Z:sets
 ---------------------------------------
 (X subset_of Y) and (Z subset_of Y) <=> (X union Z) subset_of Y.  

theorem subset_union ===
  X:sets, Y:sets, Z:sets
 ---------------------------------------
  (X subset_of Y) and (X subset_of Z) => (X subset_of (Y union Z)).  

theorem union_inter ===
  X:sets, Y:sets, Z:sets
 ---------------------------------------
 (X subset_of (Y inter Z)) => (X subset_of (Y union Z)).  

theorem union_commutes ===
  A:sets, B:sets
 ------------------------
 (A union B) = (B union A).

theorem union_associates ===
  A:sets, B:sets, C:sets
 -----------------------------------
 ((A union B) union C) = (A union (B union C)).

theorem union_inter_dist ===
  A:sets, B:sets, C:sets
 -----------------------------------
  ((A union B) inter C) = (A inter C) union (B inter C).

theorem gen_union_union_dist ===
  X:sets, Y:sets
 ---------------------------------------
 gen_union(X union Y) = gen_union(X) union gen_union(Y).

theorem gen_inter_union_dist ===
  X:sets, Y:sets, not_empty(X), not_empty(Y)
 ---------------------------------------
 gen_inter(X union Y) = gen_inter(X) inter gen_inter(Y).

theorem inter_absorption ===
  X:sets 
 ------------------
 X inter X = X.

theorem inter_commutes ===
  A:sets, B:sets
 ---------------------
 (A inter B) = (B inter A).

theorem inter_associates ===
  A:sets, B:sets, C:sets
 ---------------------------------------
 ((A inter B) inter C) = (A inter (B inter C)).

theorem inter_union_dist ===
  A:sets, B:sets, C:sets
 ---------------------------------------
 ((A inter B) union C) = (A union C) inter (B union C).

theorem inter_subset1 ===
  X:sets, Y:sets
 -----------------------
 (X inter Y) subset_of X.

theorem inter_subset2 ===
  X:sets, Y:sets
 -----------------------
 (X inter Y) subset_of Y.

theorem subsets_inter ===
  X:sets, Y:sets, Z:sets
 ------------------------------------------------
 (X subset_of Y) and (Z subset_of Y) => ((X inter Z) subset_of Y).  

theorem subset_inter ===
  X:sets, Y:sets, Z:sets
 -------------------------------------------------
 (X subset_of Y) and (X subset_of Z) <=> (X subset_of (Y inter Z)).  

theorem diff_subset ===
  X:sets, Y:sets
 -----------------------
 (X diff Y) subset_of X.

theorem diff_empty_inter ===
  X:sets, Y:sets, Z:sets
 -------------------------------------------------
  (X subset_of Y) and ((X inter Z = empty)) <=> X subset_of (Y diff Z).

theorem diff_partition ===
  X:sets, Y:sets
 -----------------------
  X subset_of Y <=> X union (Y diff X) = Y.

theorem diff_diff ===
  X:sets, Y:sets, Z:sets
 -------------------------------------------------
  X subset_of Y and Y subset_of Z => (Y diff (Z diff X)) = X.

theorem diff_union ===
  X:sets, Y:sets, Z:sets
 -------------------------------------------------
  X diff (Y union Z) = (X diff Y) inter (X diff Z).

theorem diff_inter ===
  X:sets, Y:sets, Z:sets
 -------------------------------------------------
  X diff (Y inter Z) = (X diff Y) union (X diff Z).


theorem ordered_pair_in_cross ===
  X:sets, Y:sets, A:sets, B:sets
 -------------------------------------------------
 ordered_pair(X,Y) in A cross B <=> X in A and Y in B.


theorem sym_diff_commutes ===
  X:sets, Y:sets
 -----------------------
 X sym_diff Y = Y sym_diff X.

theorem sym_diff_associates ===
  X:sets, Y:sets, Z:sets
 -------------------------------------------------
  X sym_diff (Y sym_diff Z) = (X sym_diff Y) sym_diff Z.

theorem sym_diff_subset_union ===
  X:sets, Y:sets
 -----------------------
 (X sym_diff Y) subset_of (X union Y).

theorem sym_diff_is_union ===
  X:sets, Y:sets
 -----------------------
 (X inter Y) = empty <=> (X sym_diff Y) = (X union Y).

theorem ex_set_true ===
  ex x:sets true.


theorem subset_power_gen_union_subset ===
    A:sets, B:sets, A subset_of power(B)
 ---------------------------------------
     gen_union(A) subset_of B.

theorem subset_power_gen_inter_subset ===
    A:sets, B:sets, A subset_of power(B)
 ---------------------------------------
     gen_inter(A) subset_of B.

theorem pair_in_power ===
    A:sets, B:sets, C:sets, A in C, B in C
  ---------------------------------
    pair(A,B) in power(C).


theorem cross_subset_cross1 ===
     A:sets, B:sets, C:sets, A subset_of C
   ----------------------------
    cross(A, B) subset_of cross(C, B).

theorem cross_subset_cross2 ===
     A:sets, B:sets, C:sets, B subset_of C
   ----------------------------
    cross(A, B) subset_of cross(A, C).

theorem disjoint_diff_eq ===
     A:sets, B:sets, A inter B = empty
    ----------------------------------
       A diff B = A.

theorem disjoint_in_not_in1 ===
     A:sets, B:sets, X:sets, A inter B = empty, X in A
     -------------------------------------------------
        X not_in B.
 
theorem disjoint_in_not_in2 ===
     A:sets, B:sets, X:sets, A inter B = empty, X in B
     -------------------------------------------------
        X not_in A.
 
%% Two lemmas that are useful for proving properties about 'at'
theorem pre_at_exists_lemma1 ===
      F:sets, X:sets
  ---------------------
    ex_set y y in F and ex x:sets (X |--> x) = y
      provided y not_free_in F, y not_free_in x, y not_free_in X,
              x not_free_in X.

theorem pre_at_exists_lemma2 ===
     X:sets
 ---------------------------------------
    all [x1:sets] (ex_one [x2:sets] 
      (ex y2:sets x1 = X |--> y2 and x2 = y2)
  or not (ex y2:sets x1 = X |--> y2) and x2 = empty)
     provided x1 not_free_in x2, x1 not_free_in X, x1 not_free_in y2,
              x2 not_free_in X, x2 not_free_in y2, y2 not_free_in X.

theorem pre_at_exists ===
    F:sets, X:sets
  ---------------------
   ex_set y (X |--> y) in F
      provided y not_free_in X, y not_free_in F.

theorem pre_at_type ===
   F:sets, X:sets
  ----------------------
   pre_at(F,X):sets.

theorem at_type ===
   F:sets, X:sets
  ----------------------
    (F at X):sets.

theorem fun_apply ===
    F:sets, X:sets, Y:sets, 
   (X |--> Y) in F, ex_one y:sets (X |--> y) in F
   ------------------------------------------
    at(F, X) = Y
        provided y not_free_in X, y not_free_in F.


theorem ex_lambda ===
   T:sets, all x:sets x in T => B:sets 
  ------------------------------------------
  ex_set y ex x:sets x in T and y = (x |--> B)  
   provided x not_free_in T, x not_free_in y,
            y not_free_in B, y not_free_in T.

theorem lambda_type ===
   T:sets, all x:sets x in T => B:sets 
  ------------------------------------------
    (lambda x:T B):sets
  provided x not_free_in T.

theorem lambda_property ===
   A:sets, T:sets, all x:sets x in T => B:sets 
  ------------------------------------------
   A in (lambda x:T B) <=> ex x:sets x in T and A = (x |--> B)
    provided x not_free_in T, x not_free_in A.

theorem beta_reduction ===
  A:sets, T:sets, A in T, all x:sets x in T => B:sets, local x B
  -------------------------------------------
    (lambda x:T B) at A = [A/x]B
  provided x not_free_in T.

theorem weak_beta_reduction ===
  A:sets, T:sets, A in T, all x:sets x in T => B:sets, 
  all x:sets x in T => all y x = y => B = [y/x]B
  -------------------------------------------
    (lambda x:T B) at A = [A/x]B
  provided x not_free_in T, y not_free_in x, y not_free_in B,
           y not_free_in A, y not_free_in T.


%%
%% Support for Window Inference - loads
%% zfc specific information.
%%

require_global_tactic_file(wi_support).
require_tactic_file(wi_support,zfc).
